"""
usage:
   xdxsmi clock [options] get [<ip>...]
   xdxsmi clock [options] set gfx <mhz>

options:
   -h, --help              print this help message
   -f, --filename=<file>   redirect output to a specified file
   -l, --loop=<interval>   set query interval in seconds
   -i, --index=<idx>       index number of target GPU, starting from 0
   

ip:
   gfx                   GPU frequency in MHz
"""

# from typing import Dict
import sys
from xdxsmi_render import Render
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.smi import XdxSmi
from xdxsmi.api.pyxdxml import *

clock_map = {
    "acpu": "xdx_acpu_freq",
    "cmcu": "xdx_cmcu_freq",
    "gddr": "xdx_gddr_phys_freq",
    "gfx" : "xdx_gfx_freq",
    "ndf" : "xdx_ndf_freq",
}

# get_gfx_radius(lower_bound: None, upper_bound: None)
def get_gfx_radius(lower_bound, upper_bound):
    # default optional arguments
    gfx_radius = ["50","100","200","300","400","500","600","700","800","900","1000"]
    if lower_bound == 50:
        if upper_bound == 400:
            gfx_radius = ["50","100","200","300","400"]
        elif upper_bound == 800:
            gfx_radius = ["50","100","200","300","400","500","600","700","800"]
        elif upper_bound == 1000:
            gfx_radius = ["50","100","200","300","400","500","600","700","800","900","1000"]
        elif upper_bound == 1200:
            gfx_radius = ["50","100","200","300","400","500","600","700","800","900","1000","1050","1100","1150","1200"]
        else:
            gfx_radius = ["50","100","200","300","400","500","600","700","800","900","1000"]
    elif lower_bound == 400:
        if upper_bound == 1000:
            gfx_radius = ["400","500","600","700","800","900","1000"]
        elif upper_bound == 1200:
            gfx_radius = ["400","500","600","700","800","900","1000","1050","1100","1150","1200"]
        elif upper_bound == 1300:
            gfx_radius = ["400","500","600","700","800","900","1000","1050","1100","1150","1200", "1250", "1300"]
        else:
            gfx_radius = ["400","500","600","700","800","900","1000"]
    else:
        gfx_radius = ["400","500","600","700","800","900","1000"]
    return gfx_radius

# (xdx_smi: XdxSmi, args: Dict)
def handle(xdx_smi, args):
    #breakpoint()
    # TODO: shall display info for all xdxgpu by default
    # need to add xdxgpu index filter later as well
    #target_gpus = [xdxml_device_get_handle_by_index(0)]
    #target_gpus = [xdx_smi.handle[0]]

    index_num = -1
    if(args['--index'] != None):
        index_num = int(args['--index'])      
    
    target_gpus = xdx_smi.xdxgpus
    if(index_num >= 0):
        target_gpus = [xdx_smi.xdxgpus[index_num]]

    #Set value for attribute 'gfx'
    if args["set"] and args["gfx"] and args["<mhz>"]:
        try:
            for gpu in target_gpus:
                xdev = gpu.handle
                
                lower_bound, upper_bound = xdxml_device_id_to_gfx_bound(xdev, gpu.device_id_num)
                gfx_list = get_gfx_radius(lower_bound, upper_bound)
                gfx_string = ",".join(map(str, gfx_list))
                
                if args["<mhz>"] in gfx_list:
                    dvfsInfo = xdxml_device_get_dvfs_info(xdev)
                    enable_dvfs = int(getattr(dvfsInfo, "xdx_enable_dvfs"))
                    if(enable_dvfs == 1):
                        print("Warning: Cannot set gfx clock because dvfs is enabled. Please disable dvfs and try again.")
                        sys.exit()

                    newVal = str(int(args["<mhz>"])*1000000)
                    xdxml_device_set_gfx_clock(xdev, newVal)
                else:
                    Render.render_content("ERROR: Invalid clock frequency '{}MHz'".format(args["<mhz>"]), args)
                    Render.render_content("Only the following frequency are supported (unit MHz):", args)
                    Render.render_content("{}".format(gfx_string), args)
                    sys.exit()
        except Exception:
            element_ip = args["<mhz>"]
            sys.exit(("xdxsmi clock set gfx {element_ip} is not a xdxsmi clock command."\
                "See 'xdxsmi clock --help'.".format(element_ip = element_ip)))

            

    #Get values for target attributes
    elif args["get"]:
        target_ip = ["gfx"]
        right_ip=[]
        wrong_ip=[]
        if args["<ip>"]:
            args_ip_lower = [ip.lower() for ip in args["<ip>"]]
#            target_ip = [ip for ip in target_ip if ip.lower() in args_ip_lower] 
            
            for ip in args_ip_lower:
                if ip not in target_ip:
#                    sys.exit("ERROR: unknown '{element_ip}'. Please refer to the help document below.\n{doc}"\
#                .format(element_ip=ip, doc=__doc__))
                    wrong_ip.append(ip)
                else:
                    right_ip.append(ip)
            
        # li: judge the input target_ip is legal
#        if len(target_ip) == 0:
#            element_ip = next(iter(args["<ip>"]))
#            sys.exit("ERROR: unknown '{element_ip}'. Please refer to the help document below.\n{doc}"\
#                .format(element_ip=element_ip, doc=__doc__))
        if len(right_ip)==0:
		if len(wrong_ip)!=0:
                    sys.exit("ERROR: unknown '{element_ip}'. Please refer to the help document below.\n{doc}".format(element_ip=wrong_ip[0], doc=__doc__))
		else:
		    right_ip=target_ip  #default to output all ip info if no input for <ip>
        elif len(wrong_ip)!=0:
            print("Warning: Invalid clock ip '{}'".format(wrong_ip[0]))
            
        data = []
        # i = 0
#        header = ["index", "gpu", "name"] + target_ip
        header = ["index", "gpu", "name"] + right_ip
        data.insert(0, header)
        # instead traverse handle to traverse gpus, cause gpus include handle and user_index
        for gpu in target_gpus:
            xdev = gpu.handle 
            clkInfo = xdxml_device_get_clock(xdev)
            
            row = [
                str(gpu.user_index),
                xdxml_device_get_pci_info(xdev).pci_dbdf.decode("utf8"),
                xdxml_device_get_device_model(xdev).decode('utf-8')]
#            for ip in target_ip:
            for ip in right_ip:
                # TODO: unifiy the xdxml clkinfo unit to Hz or Mhz in uint type
                row.append(str(int(getattr(clkInfo, clock_map[ip])/1e6)) +'MHz' if gpu.is_host else 'N/A')
            data.append(row)
            # i += 1
        Render.render_table(data, args)

    else:
        # no case falls into this situation?
        sys.exit("ERROR: Bad command format.  Please run 'xdxsmi clock -h' for help.\n")

if __name__ == "__main__":
    print(docopt(__doc__))
