"""
usage:
   xdxsmi dvfs [options] get [<ip>...]
   xdxsmi dvfs [options] show [<bound>...]
   xdxsmi dvfs [options] set enable_dvfs <switch>

options:
   -h, --help              print this help message
   -f, --filename=<file>   redirect output to a specified file
   -l, --loop=<interval>   set query interval in seconds
   -i, --index=<idx>       index number of target GPU, starting from 0

ip:
   enable_dvfs           enable to dynamically adjust the frequency of GPU. 0: disable, 1: enable

bound:
   lower_bound           the lower bound of GPU frequency (in MHz) during DVFS
   upper_bound           the upper bound of GPU frequency (in MHz) during DVFS

xdxsmi dvfs get: to query the enable status of DVFS
xdxsmi dvfs show: to see the GPU frequency range adjusted by DVFS
"""

# from typing import Dict
import sys
from xdxsmi_render import Render
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.smi import XdxSmi
from xdxsmi.api.pyxdxml import *

# map from the input ip to the attr name in dvfs structure
dvfs_map = {
    "enable_dvfs" : "xdx_enable_dvfs"
}

dvfs_unit_map = {
    "enable_dvfs" : "",
    "lower_bound" : "MHz",
    "upper_bound" : "MHz"
}

dvfs_gfx_map = {
    "XDX121" : [["500", "1000"], ["500", "1000"]],
    "XDX130" : [["500", "1000"], ["500", "1000"]],
    "XDX151" : [["500", "1000"], ["500", "1000"]],

    "XDX121 Super" : [["500", "1200"], ["500", "1200"]],
    "XDX151 Super" : [["500", "1200"], ["500", "1200"]],

    "XDX151 Ti" : [["500", "1200"], ["500", "1300"]],
    "XDX160 Ti" : [["500", "1200"], ["500", "1300"]], 
}

# (xdx_smi: XdxSmi, args: Dict)
def handle(xdx_smi, args):
    index_num = -1
    if(args['--index'] != None):
        index_num = int(args['--index']) 
    
    target_gpus = xdx_smi.xdxgpus
    if(index_num >= 0):
        target_gpus = [xdx_smi.xdxgpus[index_num]]
       
    if args["set"] and args["enable_dvfs"] and args["<switch>"]:
        try:
            for gpu in target_gpus:
                xdev = gpu.handle
                chip_model = gpu.chip_model
                if chip_model not in dvfs_gfx_map:
                    notify = [["card {} ({}) doesn't support DVFS at this time.".format(gpu.user_index, chip_model)]]
                    Render.render_table(notify, args, upper = False)
                    sys.exit()

                # judge the value of enable_slowdown is valid
                newVal = str(int(args["<switch>"]))
                if(newVal not in ["0","1"]):
                    print ("ERROR: Invalid input enable_dvfs={}".format(args["<switch>"]))
                    print ("Please select one of the following values and reset it: 0, 1")
                    sys.exit()
                xdxml_device_set_dvfs_enable(xdev, newVal)
        except Exception:
            element_ip = args["<switch>"]
            sys.exit('''xdxsmi dvfs set enable_dvfs {element_ip} is not a xdxsmi dvfs command.\
                See 'xdxsmi dvfs --help'.'''.format(element_ip=element_ip))
    #Get values for target attributes
    elif args["get"]:
        target_ip = ["enable_dvfs"]
        right_ip=[]
        wrong_ip=[]
        # filter the same element between target_ip and args["<ip"]
        if args["<ip>"]:
            args_ip_lower = [ip.lower() for ip in args["<ip>"]]
#            target_ip = [ip for ip in target_ip if ip.lower() in args_ip_lower] 
            for ip in args_ip_lower:
                if ip not in target_ip:
                    wrong_ip.append(ip)
                else:
                    right_ip.append(ip)
    
            
        # judge the input target_ip is legal
#        if len(target_ip) == 0:
#            element_ip = next(iter(args["<ip>"]))
#            sys.exit("ERROR: unknown '{element_ip}'. Please refer to the help document below.\n{doc}"\
#                .format(element_ip=element_ip, doc=__doc__))
        if len(right_ip)==0:
	    if len(wrong_ip)!=0:
                sys.exit("ERROR: unknown '{element_ip}'. Please refer to the help document below.\n{doc}".format(element_ip=wrong_ip[0], doc=__doc__))
	    else:
		right_ip=target_ip  #default to output all ip info if no input for <ip>
        elif len(wrong_ip)!=0:
            print("Warning: Invalid dvfs bound '{}'".format(wrong_ip[0]))
            
        data = []
        # i = 0
#        header = ["index", "gpu", "name"] + target_ip
        header = ["index", "gpu", "name"] + right_ip
        data.insert(0, header)
        # instead traverse handle to traverse gpus, cause gpus include handle and user_index
        for gpu in target_gpus:
            xdev = gpu.handle 

            dvfsInfo = xdxml_device_get_dvfs_info(xdev)
            
            row = [
                str(gpu.user_index),
                xdxml_device_get_pci_info(xdev).pci_dbdf.decode("utf8"),
                xdxml_device_get_device_model(xdev).decode('utf-8')]
#            for ip in target_ip:
            for ip in right_ip:
                # unifiy the unit
                row.append(str(int(getattr(dvfsInfo, dvfs_map[ip]))) + dvfs_unit_map[ip] if gpu.is_host else 'N/A')
            data.append(row)
        Render.render_table(data, args)
    elif args["show"]:
        target_ip = ["lower_bound", "upper_bound"]
        right_ip=[]
        wrong_ip=[]
        # filter the same element between target_ip and args["<ip"]
        if args["<bound>"]:
            args_ip_lower = [ip.lower() for ip in args["<bound>"]]
#            target_ip = [ip for ip in target_ip if ip.lower() in args_ip_lower]             
            for ip in args_ip_lower:
                if ip not in target_ip:
                    wrong_ip.append(ip)
                else:
                    right_ip.append(ip)
                    
        # judge the input target_ip is legal
#        if len(target_ip) == 0:
#            element_ip = next(iter(args["<bound>"]))
#            sys.exit("ERROR: unknown '{element_ip}'. Please refer to the help document below.\n{doc}"\
#                .format(element_ip=element_ip, doc=__doc__))

        if len(right_ip)==0:
	    if len(wrong_ip)!=0:
                sys.exit("ERROR: unknown '{element_ip}'. Please refer to the help document below.\n{doc}".format(element_ip=wrong_ip[0], doc=__doc__))
	    else:
		right_ip=target_ip  #default to output all ip info if no input for <ip>
        elif len(wrong_ip)!=0:
            print("Warning: Invalid dvfs bound '{}'".format(wrong_ip[0]))
            

        data = []
#        header = ["index", "gpu", "name"] + target_ip
        header= ["index", "gpu", "name"] + right_ip
        data.insert(0, header)

        for gpu in target_gpus:
            xdev = gpu.handle 
            chip_model = gpu.chip_model
            lower_bound = 0
            upper_bound = 0
            if chip_model in dvfs_gfx_map:
                lower_bound = dvfs_gfx_map[chip_model][1][0] if gpu.gfx_clk == '1300MHz' else dvfs_gfx_map[chip_model][0][0]
                upper_bound = dvfs_gfx_map[chip_model][1][1] if gpu.gfx_clk == '1300MHz' else dvfs_gfx_map[chip_model][0][1]
            else:
                notify = [["card {} ({}) doesn't support DVFS at this time.".format(gpu.user_index, chip_model)]]
                Render.render_table(notify, args, upper = False)
                sys.exit()
            
            row = [
                str(gpu.user_index),
                xdxml_device_get_pci_info(xdev).pci_dbdf.decode("utf8"),
                xdxml_device_get_device_model(xdev).decode('utf-8')]
#            for ip in target_ip:
            for ip in right_ip:
                # unifiy the unit
                row.append(eval(ip) + dvfs_unit_map[ip] if gpu.is_host else 'N/A')
            data.append(row)
        Render.render_table(data, args)

    else:
        # no case falls into this situation?
        sys.exit("ERROR: Bad command format.  Please run 'xdxsmi dvfs -h' for help.\n")

if __name__ == "__main__":
    print(docopt(__doc__))
