"""
usage:
   xdxsmi pmon [options]

options:
    -h, --help              print this help message
    -f, --filename=<file>   redirect query output to a specified file    
    -i, --index=<idx>       show statistics of top-5 processes running on target devices
    -l, --loop=<interval>   set query interval in seconds [default=1sec, max=10sec]
    -c, --count=num         Collect specified number of samples and exit    
    -t, --time              show time in YYYYMMDD HH:MM:SS format
    -s, --select=type       One or more metrics [default: u]
                            Can be any of the following:
                                u - Utilization
                                m - FB and CC Protected memory usage
"""

# from typing import Dict
from datetime import datetime
from xdxsmi_render import Render
from xdxsmi_attrs_ops import setAttrs, getAttrs
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.smi import XdxSmi
from xdxsmi.api.xdxgpu import XdxGpu
from xdxsmi.api.pyxdxml import *


xdxgpu_pmon_head_time           = "    data     time     "
xdxgpu_pmon_head_time_unit      = "YYYYMMDD HH:MM:SS     "

xdxgpu_pmon_head_info           = "GPU     PID     type"
xdxgpu_pmon_head_info_unit      = "Idx       #      C/G"

xdxgpu_pmon_head_util           = "     sm     mem     enc     dec"
xdxgpu_pmon_head_util_unit      = "      %       %       %       %"

xdxgpu_pmon_head_mem            = "          fb          ccpm"
xdxgpu_pmon_head_mem_unit       = "          MB            MB"

xdxgpu_pmon_head_command        = "     command"
xdxgpu_pmon_head_command_unit   = "     name   "

# (xdx_smi: XdxSmi, args: Dict)
def handle(xdx_smi, args):
    index_num = -1
    if(args['--index'] != None):
        index_num = int(args['--index'])
    
    target_gpus = xdx_smi.xdxgpus
    if(index_num >= 0):
        target_gpus = [xdx_smi.xdxgpus[index_num]]

    loop_interval = 1.0
    try:
        if(args['--loop'] != None):
            loop_interval = float(args["--loop"])
            if(loop_interval<=0):
                Render.render_content("Invalid input for --loop: '{}'. A value > 0 is expected. This parameter will be ignored.\n".format(args['--loop']), args)
                loop_interval=1.0
    except Exception as e:
        Render.render_content("Invalid input for --loop: '{}'. A value > 0 is expected. This parameter will be ignored.\n".format(args['--loop']), args)
        loop_interval=1.0

#    if((args['--loop'] != None) and (float(args['--loop']) > 0)):
#        loop_interval = float(args["--loop"])

    if(loop_interval > 10.0):
        loop_interval = 10.0

    Render.render_content('loop interval: {} seconds:\n'.format(loop_interval), args)    
    
    count = -1
    try:
        if(args['--count'] == None):
            count = -1
        elif(args['--count'] != None and args['--count'].isdigit() and int(args['--count']) >= 1 and int(args['--count']) <= 10000000):
            count = int(args['--count'])
        else:
            raise ValueError("The count range is incorrect, the range is from 1 to 10000000")
    except ValueError as e:
        print("ERROR: {}".format(e))
        exit()
        
    select = 'u'
    try:
        if(args[args['--select'] == None or '--select'] == 'u'):
            select = 'u'
        elif(args['--select'] == 'm'):
            select = 'm'
        elif(args['--select'] == 'um' or args['--select'] == 'mu'):
            select = 'um'
        else:
            raise ValueError("The select parameter is incorrect, One or more metrics of 'u','m'")
    except ValueError as e:
        print("ERROR: {}".format(e))
        exit()
        
    show_time = False
    if(args['--time']):
        show_time = True

    # pmon header
    xdxgpu_pmon_head      = "#"
    xdxgpu_pmon_head_unit = "#"

    if(show_time):
        xdxgpu_pmon_head += xdxgpu_pmon_head_time
        xdxgpu_pmon_head_unit += xdxgpu_pmon_head_time_unit

    xdxgpu_pmon_head += xdxgpu_pmon_head_info
    xdxgpu_pmon_head_unit += xdxgpu_pmon_head_info_unit
        
    if(select == 'u'):
        xdxgpu_pmon_head += xdxgpu_pmon_head_util
        xdxgpu_pmon_head_unit += xdxgpu_pmon_head_util_unit
    elif(select == 'm'):
        xdxgpu_pmon_head += xdxgpu_pmon_head_mem
        xdxgpu_pmon_head_unit += xdxgpu_pmon_head_mem_unit
    elif(select == 'um'):
        xdxgpu_pmon_head += xdxgpu_pmon_head_mem
        xdxgpu_pmon_head_unit += xdxgpu_pmon_head_mem_unit
        xdxgpu_pmon_head += xdxgpu_pmon_head_util
        xdxgpu_pmon_head_unit += xdxgpu_pmon_head_util_unit
        
    xdxgpu_pmon_head += xdxgpu_pmon_head_command
    xdxgpu_pmon_head_unit += xdxgpu_pmon_head_command_unit

    Render.render_content(xdxgpu_pmon_head, args)
    Render.render_content(xdxgpu_pmon_head_unit, args)

    # pmon body
    try:
        while(True):
            gpu_id = 0
            for gpu in target_gpus:
                proInfo = xdxml_device_get_process_info(gpu.handle)
                
                for i in range(1, proInfo.pronum+1):
                    
                    attr_name = "process_{}".format(i)
                    pid = getattr(proInfo, attr_name) if getattr(proInfo, attr_name) != None else 'N/A'
                    
                    attr_name = "pro{}NAME".format(i)
                    p_name = getattr(proInfo, attr_name) if getattr(proInfo, attr_name) != None else 'N/A'
                    
                    attr_name = "pro{}LMA".format(i)
                    g_usage = getattr(proInfo, attr_name) if getattr(proInfo, attr_name) != None else 'N/A'
                    if(float(g_usage) == 0):
                        continue

                    # pmon body
                    xdxgpu_pmon_body = " "

                    if(show_time):
                        format_time = datetime.now().strftime("%Y%m%d %H:%M:%S")
                        xdxgpu_pmon_body += "{:<22}".format(format_time)

                    xdxgpu_pmon_body += "{:>3}".format(gpu_id)
                    xdxgpu_pmon_body += "{:>8}".format(pid)
                    xdxgpu_pmon_body += "{:>9}".format("N/A")   #type

                    if(select == 'm'):
                        xdxgpu_pmon_body += "{:>12.3f}".format(float(g_usage)/(1024*1024))  #fb
                        xdxgpu_pmon_body += "{:>14}".format("N/A")  #ccpm
                    elif(select == 'u'):
                        xdxgpu_pmon_body += "{:>7}".format("N/A")   #sm
                        xdxgpu_pmon_body += "{:>8.1f}".format(float(g_usage)/(1024*1024)/gpu.fb_total*100)   #mem
                        xdxgpu_pmon_body += "{:>8}".format("N/A")   #enc
                        xdxgpu_pmon_body += "{:>8}".format("N/A")   #dec
                    elif(select == 'um'):
                        xdxgpu_pmon_body += "{:>12.3f}".format(float(g_usage)/(1024*1024))  #fb
                        xdxgpu_pmon_body += "{:>14}".format("N/A")  #ccpm
                        xdxgpu_pmon_body += "{:>7}".format("N/A")   #sm
                        xdxgpu_pmon_body += "{:>8.1f}".format(float(g_usage)/(1024*1024)/gpu.fb_total*100)   #mem
                        xdxgpu_pmon_body += "{:>8}".format("N/A")   #enc
                        xdxgpu_pmon_body += "{:>8}".format("N/A")   #dec

                    xdxgpu_pmon_body += "     {}".format(p_name)

                    Render.render_content(xdxgpu_pmon_body, args) 
                    
                #add a blank line after each query
                Render.render_content(" ", args)    
                gpu_id += 1
            
            if(count != -1):
                count -= 1
                if(count == 0):
                    exit()

            time.sleep(loop_interval)
    except KeyboardInterrupt:
        pass

if __name__ == "__main__":
    print(docopt(__doc__))
