"""
usage:
   xdxsmi power [options] get [<ip>...]
   xdxsmi power [options] set pwr_warn <mw>
   xdxsmi power [options] set volt <mv>

options:
   -h, --help              print this help message
   -f, --filename=<file>   redirect output to a specified file
   -l, --loop=<interval>   set query interval in seconds
   -i, --index=<idx>       index number of target GPU, starting from 0

ip:
   volt                  voltage (in V) of GPU core
   current               current (in A) of GPU core
   pwr                   current power (in W)of GPU core
   pwr_total             total power (in W) of GPU
   
set volt range:
   CHIP_PANGU_A0 [760mV ~ 880mV]
   CHIP_PANGU_B0 [644mV ~ 1040mV]
   
tips:Modifying GPU voltage may cause system instability and result in system freezing! 
    If this situation occurs, please restart the machine for recovery.
"""

import sys
from xdxsmi_render import Render
from xdxsmi_attrs_ops import setAttrs, getAttrs
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.smi import XdxSmi
from xdxsmi.api.pyxdxml import *


#Dictionary matching abbreviated and real attribute names 
power_map = {
    "gpu":   "xdx_dbdf",
    "name":  "xdx_product_name",
    "volt":  "in0_input",
    "curr":  "curr1_input",
    "pwr":   "power1_average",
    "pwr_max":  "power1_cap_max",
    "pwr_min":  "power1_cap_min",
    "pwr_warn": "power1_cap",
    "pwr_total": "power1_input"
}
power_unit_map = {
    "volt": "V",
    "current": "A",
    "pwr": "W",
    "pwr_total": "W" 
}

# the input function is compatible with python2 and python3
def input_compat(prompt=""):
    try:
        # python 2
        return raw_input(prompt)
    except NameError:
        #python 3
        return input(prompt)

# (xdx_smi: XdxSmi, args: Dict)
def handle(xdx_smi, args):
    ##Collect target gpus
    #target_gpus = [xdx_smi.handle[0]]
    index_num = -1
    if(args['--index'] != None):
        index_num = int(args['--index'])
    
    target_gpus = xdx_smi.xdxgpus
    if(index_num >= 0):
        target_gpus = [xdx_smi.xdxgpus[index_num]]
   
    #Set values of attributes 'pwr_warn'
    if args["set"] and args["pwr_warn"] and args["<mw>"]:
        try:
            newVal = str(int(args["<mw>"]))
            for gpu in target_gpus:
                xdev = gpu.handle
                chip_model = xdxml_device_get_device_model(xdev).decode('utf-8')
                # vendor_name: XDX, BITLAND, GOODTIMES, OTHERS
                vendor_name = xdxml_device_get_vendor(xdev).decode('utf-8')
                if vendor_name == "XDX" and gpu.support_power_info == 1:
                    xdxml_device_set_power_warn(xdev, newVal)
                else:
                    warning = [["Warning: index = {}, chip model = {}, vendor = {}, this card does not support power relation command".format(gpu.user_index, chip_model, vendor_name)]]
                    Render.render_table(warning, args, upper = False)

        except:
            element_ip = next(iter(args["<mw>"]))
            sys.exit('''xdxsmi power set pwr_warn {element_ip} is not a xdxsmi power command.\
                    See 'xdxsmi power --help'.'''.format(element_ip=element_ip))

    #Set values of attributes 'volt'
    elif args["set"] and args["volt"] and args["<mv>"]:
        try:
            for gpu in target_gpus:
                xdev = gpu.handle
                chip_model = xdxml_device_get_device_model(xdev).decode('utf-8')
                # vendor_name: XDX, BITLAND, GOODTIMES, OTHERS
                vendor_name = xdxml_device_get_vendor(xdev).decode('utf-8')

                if vendor_name == "XDX" and gpu.support_power_info == 1:
                    newVal_int = int(args["<mv>"])
                   
                    vlot_min = 760
                    vlot_max = 880
                    if(gpu.asci_type == "CHIP_PANGU_A0"):
                        vlot_min = 760
                        vlot_max = 880
                    elif(gpu.asci_type == "CHIP_PANGU_B0"):
                        vlot_min = 644
                        vlot_max = 1040
                    else:
                        print("Error: The gpu asci type analysis failed.")
                        sys.exit()
                   
                    if newVal_int < vlot_min or newVal_int > vlot_max:
                        print("Warning: You can only set voltage between {0} and {1} mv".format(vlot_min, vlot_max))
                        sys.exit()
                        
                    dvfsInfo = xdxml_device_get_dvfs_info(xdev)
                    enable_dvfs = int(getattr(dvfsInfo, "xdx_enable_dvfs"))
                    if(enable_dvfs == 1):
                        print("Warning: Cannot set power volt because dvfs is enabled. Please disable dvfs and try again.")
                        sys.exit()

                    print("Tips: Modifying GPU voltage may cause system instability and result in system freezing! If this situation occurs, please restart the machine for recovery.")
                    sel = input_compat("please enter <yes> or <no> to continue\n")
                    if sel not in ("yes", "y"):
                        print("Program termination!")
                        sys.exit()
                      
                    newVal = str(newVal_int)
                    xdxml_device_set_power_volt(xdev, newVal)
                else:
                    warning = [["Warning: index = {}, chip model = {}, vendor = {}, this card does not support power relation command".format(gpu.user_index, chip_model, vendor_name)]]
                    Render.render_table(warning, args, upper = False)

        except Exception as e:
            print("error: {}".format(e))
            element_ip = args["<mv>"]
            sys.exit('''xdxsmi power set volt {element_ip} is not a xdxsmi power command."\
                    See 'xdxsmi power --help'.'''.format(element_ip=element_ip))

    #Get values of target attributes   
    elif args["get"]:
        target_ip = ["volt", "current", "pwr", "pwr_total"]

        if args["<ip>"]:
            args_ip_lower = [ip.lower() for ip in args["<ip>"]]
            target_ip = [ip for ip in target_ip if ip.lower() in args_ip_lower]

        if len(target_ip) == 0:
            element_ip = next(iter(args["<ip>"]))
            sys.exit("ERROR: unknown '{element_ip}'. Please refer to the help document below.\n{doc}"\
                    .format(element_ip=element_ip, doc=__doc__))

        data = []
        header = ["index", "gpu", "name"] + target_ip
        data.insert(0, header)
        for gpu in target_gpus:
            xdev = gpu.handle
            chip_model = xdxml_device_get_device_model(xdev).decode('utf-8')
            # vendor_name: XDX, BITLAND, GOODTIMES, OTHERS
            vendor_name = xdxml_device_get_vendor(xdev).decode('utf-8')

            if vendor_name == "XDX" and gpu.support_power_info == 1:
                powerInfo = xdxml_device_get_power_info(xdev)

                # warning to show the info about pwr_total
                if("pwr_total" in target_ip):
                   warning = [["Warning: pwr_total is the simulated total power of the board, which is subject to error and is for reference only."]]
                   Render.render_table(warning, args, upper = False)
            else:
                warning = [["Warning: index = {}, chip model = {}, vendor = {}, this card does not support get power info".format(gpu.user_index, chip_model, vendor_name)]]
                Render.render_table(warning, args, upper = False)

            row = [
                   str(gpu.user_index),
                   xdxml_device_get_pci_info(xdev).pci_dbdf.decode("utf8"),
                   chip_model]
            for ip in target_ip:
                if vendor_name == "XDX" and gpu.support_power_info == 1:
                    val = round(getattr(powerInfo, ip)/1e3, 2)
                    val = 0 if val < 0 or val > 1000 else val

                    # hide soc power data if volt/current/pwer is 0
                    if val == 0:
                        row.append("N/A")
                    else:
                        row.append(str(val) + power_unit_map[ip] if gpu.is_host else 'N/A')
                else:
                   row.append("N/A")
            data.append(row)
        Render.render_table(data, args)
    
   #Stop program if the input command is invalid
    else:
        sys.exit("ERROR: Bad command format. Please run 'xdxsmi power -h' for help.\n")

if __name__ == "__main__":
    print(docopt(__doc__))
