# -*- coding: utf-8 -*-
"""
usage:
   xdxsmi query_gpu get [<ip>...] [options]

ip:
   uuid                   UUID of the GPU
   index                  GPU index number starting from 0
   temperature.gpu        current GPU temperature
   utilization.gpu        GPU utilization
   utilization.memory     GPU memory utilization
   memory.total           total size of GPU memory
   memory.free            free size of GPU memory
   memory.used            used size of GPU memory
   clocks.gr              GDDR frequency

options:
   --format=csv           comma seperated values. value: csv, nocsv. [default: nocsv]
   --noheader=noheader    skip the first line with colum headers. value: 0, 1. [default: 0]
   --nounits=nounits      don't print units for numerical values. value: 0, 1. [default: 0]
   -h, --help             print this help message
   -f, --filename=<file>  redirect output to a specified file
   -l, --loop=<interval>  set query interval in seconds
   -i, --index=<idx>      index number of target GPU, starting from 0     

For example:
   e.g.1 xdxsmi query_gpu get uuid index temperature.gpu utilization.gpu utilization.memory memory.total memory.free memory.used clocks.gr --format=csv --nounits=1 --noheader=1
         query all of the info with comma seperating values, no units, no header
   e.g.2 xdxsmi query_gpu get uuid temperature.gpu
         just query uuid and temperature.gpu with the default settings. (nocsv, show header, show units)
   e.g.3 xdxsmi query_gpu get
         query all of the info with the default settings. (nocsv, show header, show units)
"""

# from typing import Dict
import sys
from xdxsmi_render import Render
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.smi import XdxSmi
from xdxsmi.api.pyxdxml import *

query_unit_map = {
   "uuid":      "",
   "index":     "",
   "temperature.gpu":  "°C",
   "utilization.gpu":  "%",
   "utilization.memory": "%",
   "memory.total":       "MiB",
   "memory.free":        "MiB",
   "memory.used":        "MiB",
   "clocks.gr":          "MHz"
}

# (xdx_smi: XdxSmi, args: Dict)
def handle(xdx_smi, args):
   #Collect target gpus
   
   index_num = -1
   if(args['--index'] != None):
       index_num = int(args['--index'])
   
   target_gpus = xdx_smi.xdxgpus
   if(index_num >= 0):
      target_gpus = [xdx_smi.xdxgpus[index_num]]

   show_units = 1
   try:
      nounits = int(args['--nounits'])
      if(nounits == 1):
         show_units = 0
      elif nounits not in [0, 1]:
         raise ValueError()
   except Exception:
      print("Nounits modifier is not recognized. See 'xdxsmi query_gpu -h'.")
      sys.exit()

   show_header = 1
   try:
      noheader = int(args['--noheader'])
      if(noheader == 1):
         show_header = 0
      elif noheader not in [0, 1]:
         raise ValueError()
   except Exception:
      print("Noheader modifier is not recognized. See 'xdxsmi query_gpu -h'.")
      sys.exit()

   csv = 0
   try:
      format = args['--format']
      if(format == "csv"):
         csv = 1
      elif format not in ["csv", "nocsv"]:
         raise ValueError()
   except Exception:
      print("Format modifier is not recognized. See 'xdxsmi query_gpu -h'.")
      sys.exit()

   if args["get"]:
      target_ip = ["uuid", "index", "temperature.gpu", "utilization.gpu",
                     "utilization.memory", "memory.total", "memory.free",
                     "memory.used", "clocks.gr"]
      
      if args["<ip>"]:
          args_ip_lower = [ip.lower() for ip in args["<ip>"]]
          target_ip = [ip for ip in target_ip if ip.lower() in args_ip_lower]

      # li: judge the input target_ip is legal
      if len(target_ip) == 0:
         element_ip = next(iter(args["<ip>"]))
         sys.exit("ERROR: unknown '{element_ip}'. Please refer to the help document below.\n{doc}"\
                .format(element_ip=element_ip, doc=__doc__))
      
      data = []
      header = [] # build header: target_ip name + units
      for ip in target_ip:
         element = ip
         if ip in query_unit_map and query_unit_map[ip] != "":
            element += ("[" + query_unit_map[ip] + "]")
         header.append(element)

      if show_header == 1:
         data.insert(0, header)

      # get data
      for gpu in target_gpus:
         xdev = gpu.handle
         queryInfo = xdxml_device_query_two_parts(xdev)
         
         query_gpu_dict = {
            "uuid":               "GPU-" + str(gpu.uuid),
            "index":              str(gpu.user_index),
            "temperature.gpu":    str(queryInfo.temp),
            "utilization.gpu":    str(queryInfo.util_gpu),
            "utilization.memory": str(round(gpu.fb_used*100 / gpu.fb_total)),
            "memory.total":       str(gpu.fb_total),
            "memory.free":        str(gpu.fb_free),
            "memory.used":        str(gpu.fb_total - gpu.fb_free),
            "clocks.gr":          str(int(queryInfo.gddr_speed/1e9 /8 * 1000))
         }

         row = []
         for ip in target_ip:
            append_data = query_gpu_dict[ip]
            if show_units == 1 and query_unit_map[ip] != "":
               append_data = append_data + " " + query_unit_map[ip]

            # comma seperated values
            if csv == 1 and ip != target_ip[-1]:
               append_data = append_data + ","
               
            row.append(append_data)
         
         data.append(row)
         # a part of the shown info is from xdxgpu, so we must update dynamic info
         gpu.update_dynamic_info()
      Render.render_table(data, args, upper = False)
   
   else:
      sys.exit("ERROR: Bad command format.  Please run 'xdxsmi query_gpu -h' for help.\n")

if __name__ == "__main__":
   print(docopt(__doc__))