# -*- coding: utf-8 -*-

"""
usage:
   xdxsmi thermal [options] get [<ip>...]
   xdxsmi thermal [options] set temp_slowdown <°C>
   xdxsmi thermal [options] set temp_shutdown <°C>
   xdxsmi thermal [options] set enable_shutdown <switch>

options:
   -h, --help              print this help message
   -f, --filename=<file>   redirect output to a specified file
   -l, --loop=<interval>   set query interval in seconds
   -i, --index=<idx>       index number of target GPU, starting from 0

ip:
   temp_slowdown         The temperature at which the GPU starts to lower its frequency 
   temp_shutdown         The temperature at which the GPU will shut down
   enable_shutdown       enable to feature of GPU shutdown at a certain temperature. 0: disable, 1: enable

Be cautious when using this command to set the thermal behaviors!
e.g. If 'enable_shutdown' is set to 1, the GPU will shut down when its temperature reaches 'temp_shutdown'.
"""

# from typing import Dict
import sys
from xdxsmi_render import Render
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.smi import XdxSmi
from xdxsmi.api.pyxdxml import *

# map from the input ip to the attr name in thm structure
thm_map = {
    "temp_slowdown" : "xdx_temp_slowdown",
    "temp_shutdown" : "xdx_temp_shutdown",
    "enable_slowdown" : "xdx_enable_thm_slowdown",
    "enable_shutdown" : "xdx_enable_thm_shutdown"
}

thm_unit_map = {
    "temp_slowdown" : "°C",
    "temp_shutdown" : "°C",
    "enable_slowdown" : "",
    "enable_shutdown" : ""
}

# (xdx_smi: XdxSmi, args: Dict)
def handle(xdx_smi, args):
    index_num = -1
    if(args['--index'] != None):
        index_num = int(args['--index'])
    
    target_gpus = xdx_smi.xdxgpus
    if(index_num >= 0):
        target_gpus = [xdx_smi.xdxgpus[index_num]]
       
    
    #Set value for attribute 'gfx'
    if args["set"] and args["temp_slowdown"] and args["<°C>"]:
        #Judge whether the value of temp_slowdown is valid
        try:
            newVal = args["<°C>"]
            
            if(newVal.isdigit()==False):
                sys.exit("ERROR: Invalid input temp_slowdown={}".format(args["<°C>"]))

            for gpu in target_gpus:
                xdev = gpu.handle
                temper = xdxmltemperInfo_t()
                xdxml_device_get_temper(xdev, temper, "temp_max")

                thermal = xdxmlthermalInfo_t()
                xdxml_device_get_thermal(xdev, thermal, "temp_shutdown")
                lower = int(temper.temp_max)
                upper = int(thermal.xdx_temp_shutdown)

                if(lower < int(newVal) and int(newVal) < upper):
                    thm_type = "temp_slowdown"
                    xdxml_device_set_thermal_info(xdev, newVal, thm_type)
                else:
                    print ("ERROR: Invalid input temp_slowdown={}".format(args["<°C>"]))
                    print ("Please enter an integer between temp_max {} and temp_shutdown {}".format(lower, upper))
                    sys.exit()
        except Exception:
            sys.exit('''Exception! Please try another way or contact the software provider.''')
    
    elif args["set"] and args["temp_shutdown"] and args["<°C>"]:
        #Judge whether the value of temp_slowdown is valid
        try:
            newVal = args["<°C>"]

            if(newVal.isdigit()==False):
                sys.exit("ERROR: Invalid input temp_shutdown={}".format(args["<°C>"]))
            
            for gpu in target_gpus:
                xdev = gpu.handle
                thermal = xdxmlthermalInfo_t()
                
                xdxml_device_get_thermal(xdev, thermal, "enable_shutdown")

                flag=int(thermal.xdx_enable_thm_shutdown)
                if(flag==0):
                    print("Warning: Temp_shutdown can't be modified,please set enable_shutdown first!")
                    continue
                xdxml_device_get_thermal(xdev, thermal, "temp_slowdown")
                lower = int(thermal.xdx_temp_slowdown)
                upper = 105
                
                if(lower < int(newVal) and int(newVal) < upper):
                    thm_type = "temp_shutdown"
                    xdxml_device_set_thermal_info(xdev, newVal, thm_type)
                else:
                    print ("ERROR: Invalid input temp_shutdown={}".format(args["<°C>"]))
                    print ("Please enter an integer between slowdown temperature {} and protective temperature {}".format(lower, upper))
                    sys.exit()

        except Exception:            
            sys.exit('''Exception! Please try another way or contact the software provider.''')

    elif args["set"] and args["enable_shutdown"] and args["<switch>"]:
        try:
            newVal=args["<switch>"]

            if(newVal.isdigit()==False or newVal not in ["0","1"]):
                print ("ERROR: Invalid input enable_slowdown={}".format(args["<switch>"]))
                print ("Please enter 0 or 1 for <switch>")
                sys.exit()
            for gpu in target_gpus:
                xdev = gpu.handle
                thm_type = "enable_shutdown"
                xdxml_device_set_thermal_info(xdev, newVal, thm_type)
        except Exception:
            sys.exit('''Exception! Please try another way or contact the software provider.''')

    #Get values for target attributes
    elif args["get"]:
        right_ip=[]
        wrong_ip=[]
        target_ip = ["temp_slowdown", "temp_shutdown", "enable_shutdown"]
        # filter the same element between target_ip and args["<ip"]
        
        if args["<ip>"]:
            args_ip_lower = [ip.lower() for ip in args["<ip>"]]
            for ip in args_ip_lower:
                if ip not in target_ip:
                    wrong_ip.append(ip)
                else:
                    right_ip.append(ip)
#            target_ip = [ip for ip in target_ip if ip.lower() in args_ip_lower]            

        # judge the input target_ip is legal
#        if len(target_ip) == 0:
#            element_ip = next(iter(args["<ip>"]))
#            sys.exit("ERROR: unknown '{element_ip}'. Please refer to the help document below.\n{doc}"\
#                .format(element_ip=element_ip, doc=__doc__))
        if len(right_ip)==0:
		if len(wrong_ip)!=0:
                    sys.exit("ERROR: unknown '{element_ip}'. Please refer to the help document below.\n{doc}".format(element_ip=wrong_ip[0], doc=__doc__))
		else:
		    right_ip=target_ip  #default to output all ip info if no input for <ip>
        elif len(wrong_ip)!=0:
            print("Warning: Invalid thermal ip '{}'".format(wrong_ip[0]))
            
        data = []
        # i = 0
#        header = ["index", "gpu", "name"] + target_ip
        header = ["index", "gpu", "name"] + right_ip
        data.insert(0, header)
        # instead traverse handle to traverse gpus, cause gpus include handle and user_index
        for gpu in target_gpus:
            xdev = gpu.handle 
            thmInfo = xdxml_device_get_thermal_info(xdev)
            
            row = [
                str(gpu.user_index),
                xdxml_device_get_pci_info(xdev).pci_dbdf.decode("utf8"),
                xdxml_device_get_device_model(xdev).decode('utf-8')]
#            for ip in target_ip:
            for ip in right_ip:
                # unifiy the unit
                row.append(str(int(getattr(thmInfo, thm_map[ip]))) + thm_unit_map[ip] if gpu.is_host else 'N/A')
            data.append(row)
        if sys.version_info[0] == 2:
            Render.render_table_py2(data, args)
        else:
            Render.render_table(data, args)

    else:
        sys.exit("ERROR: Bad command format.  Please run 'xdxsmi thermal -h' for help.\n")

if __name__ == "__main__":
    print(docopt(__doc__))