/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include "xdxacpu.h"
#include "xdx_device.h"
#include "log2.h"
#include "pvrsrv_error.h"
#include "osfunc.h"
#include "sysinfo.h"
#include "xdx_gfx_convert.h"
#include "rgxdevice.h"

#define PVR_FLOOR_ALIGN(_x, _n)	((_x) & ~((_n)-1U))

u32 DOORBELL_OFFSET(int index)
{
	if (index > 31) {
		printk("ERR: This API only use for db0-31");
		return EPERM;
	}
	return CMCU_DOORBELLS_DOORBELL0_OFFSET + (index << 3);
}
static inline u64 acpu_doorbell_rreg64(IMG_UINT64 reg_base, IMG_UINT32 off)
{
	IMG_UINT64 data;
	rmb();
	data = OSReadHWReg64(reg_base, off);
	mb();
	return data;
}

static inline void acpu_doorbell_wreg64(IMG_UINT64 reg_base, IMG_UINT32 off, IMG_UINT64 value)
{
	wmb();
	OSWriteHWReg64(reg_base, off, value);
	mb();
}

static void DoorBellMemset_callback(void *pvParam)
{
	struct completion *pSync_completion = (struct completion *)pvParam;
	complete(pSync_completion);
}

int acpu_dbcmd_memop_memset_multiple(struct xdx_device *xdev, uint64_t devPAddr, uint32_t size, uint8_t pattern, bool align4k)
{
	uint8_t ret = 0;
	uint32_t ui32MemSize;
	uint32_t ui32MaxPacket =  (align4k) ? MAX_ACPU_MEMSET_SIZE_4K : MAX_ACPU_MEMSET_SIZE_128;

	while (size > 0) {
		ui32MemSize = MIN(ui32MaxPacket, size);
		ret = acpu_dbcmd_memop_memset(xdev, devPAddr, ui32MemSize, pattern, align4k);
		if (ret != 0) {
			return  ret;;
		}

		devPAddr = devPAddr + ui32MemSize;
		size -= ui32MemSize;
	}

	return  0;
}

int acpu_dbcmd_memop_memset(struct xdx_device *xdev, uint64_t devPAddr, uint32_t size, uint8_t pattern, bool align4k)
{
	struct xdx_acpu *acpu = &xdev->acpu;
	struct completion sync_completion;
	unsigned long res;
	acpu_dbcmd_memop_memset_cmd_type acpu_dbcmd_memop_memset_cmd = { .all = 0 };

	uint32_t dbNum = acpu_doorbell_get_timeout(acpu->acpu_db_pool, 1000000);
	if(dbNum == INVALID_DB_INDEX) {
		PVR_DPF((PVR_DBG_ERROR, "%s get doorbell from pool timeout", __func__));
		return -ETIME;
	}

	init_completion(&sync_completion);

	if (align4k) {
		WARN_ON((devPAddr & 0xFFF) != 0);
		/*same with firmware*/
		WARN_ON(size > MAX_ACPU_MEMSET_SIZE_4K);
		acpu_dbcmd_memop_memset_cmd.align4K.cmd = acpu_dbcmd_memop;
		acpu_dbcmd_memop_memset_cmd.align4K.subCmd = acpu_dbcmd_memop_memmet4kaligned;
		acpu_dbcmd_memop_memset_cmd.align4K.num4K = (size >> 12) - 1;
		acpu_dbcmd_memop_memset_cmd.align4K.addr4K = devPAddr >> 12;
		acpu_dbcmd_memop_memset_cmd.align4K.pattern = pattern;
	} else {
		WARN_ON((devPAddr & 0x1F) != 0);
		WARN_ON(size > 0x8000);
		acpu_dbcmd_memop_memset_cmd.align32B.cmd = acpu_dbcmd_memop;
		acpu_dbcmd_memop_memset_cmd.align32B.subCmd = acpu_dbcmd_memop_memset32baligned;
		acpu_dbcmd_memop_memset_cmd.align32B.num4B  = (size >> 2) - 1;
		acpu_dbcmd_memop_memset_cmd.align32B.addr32B = devPAddr >> 5;
		acpu_dbcmd_memop_memset_cmd.align32B.pattern  = pattern;
	}

	/* In order to prevent cpu from being polling all the time, increase the sleep awakening mechanism */
	acpu->acpu_db_pool->callback[dbNum] = DoorBellMemset_callback;
	acpu->acpu_db_pool->callback_param[dbNum] = &sync_completion;

	acpu_doorbell_submit(acpu, dbNum, acpu_dbcmd_memop_memset_cmd.all);

	res = wait_for_completion_timeout(&sync_completion, msecs_to_jiffies(2000));
	if (!res) {
		PVR_DPF((PVR_DBG_ERROR, "%s wait for completion timeout", __func__));
		acpu->acpu_db_pool->callback[dbNum] = NULL;
		acpu->acpu_db_pool->callback_param[dbNum] = NULL;
		return -ETIME;
	}

	return 0;
}

s32 H02CDoorBellPowerGet(struct xdx_device *xdev, u64 value)
{

	u64 reg_base = 0;
	u32 to = 100000;
	s32 ret = 0;
	uint8_t *tmp = (uint8_t *) xdev->rmmio;
	reg_base = (u64) tmp;

	acpu_doorbell_wreg64(reg_base, CMCU_DOORBELLS_H02C_POWER, value);
	while (to--) {
		if (!to) {
			dev_err(xdev->dev, "cntl doorbell timeout\n");
			return EBUSY;
		}
		if (acpu_doorbell_rreg64(reg_base, CMCU_DOORBELLS_H02C_POWER) & 1)
			break;
	}

	ret = acpu_doorbell_rreg64(reg_base, CMCU_DOORBELLS_H02C_POWER);
	dev_dbg(xdev->dev, "h02c power doorbell value =%d", ret >> 1);

	return ret >> 1;
}

int32_t H02CDoorBellPowerSet(struct xdx_device *xdev, u64 value)
{
	power_cmd_type power_cmd;
	u64 reg_base = 0;
	u32 to = 100000;
	uint8_t *tmp = (uint8_t *) xdev->rmmio;

	reg_base = (u64) tmp;

	acpu_doorbell_wreg64(reg_base, CMCU_DOORBELLS_H02C_POWER, value);
	/* FIXME: the feedback form CMCU_DOORBELLS_H02C_INT_STATUS is invalid
	   while(to--){
	   if(!to){
	   dev_err(xdev->dev, "cntl doorbell timeout\n");
	   return EBUSY;
	   }
	   if(cmcu_doorbell_rreg64(reg_base, CMCU_DOORBELLS_H02C_INT_STATUS) & 1)
	   break;
	   }
	 */

	while (to--) {
		if (!to) {
			dev_err(xdev->dev, "cntl doorbell timeout\n");
			return EBUSY;
		}
		if (acpu_doorbell_rreg64(reg_base, CMCU_DOORBELLS_H02C_POWER) &
		    1)
			break;
	}
	power_cmd.all =
	    acpu_doorbell_rreg64(reg_base, CMCU_DOORBELLS_H02C_POWER);

	dev_dbg(xdev->dev, "h02c power doorbell value =%d",
		power_cmd.low >> 1);

	return 0;
}

int32_t H02CDoorBellThermalSet(struct xdx_device *xdev, u64 value)
{
	CMCUThermalCmd_T thermal_cmd;
	u64 reg_base = 0;
	u32 to = 100000;
	uint8_t *tmp = (uint8_t *) xdev->rmmio;

	reg_base = (u64) tmp;

	acpu_doorbell_wreg64(reg_base, CMCU_DOORBELLS_H02C_THERMAL, value);

	while (to--) {
		if (!to) {
			dev_err(xdev->dev, "cntl doorbell timeout\n");
			return EBUSY;
		}
		if (acpu_doorbell_rreg64(reg_base, CMCU_DOORBELLS_H02C_THERMAL) & 1)
			break;
	}
	thermal_cmd.all = acpu_doorbell_rreg64(reg_base, CMCU_DOORBELLS_H02C_THERMAL);

	dev_dbg(xdev->dev, "h02c power doorbell value =%d", thermal_cmd.low31);

	return 0;
}

static void db_task_handler(struct work_struct *data)
{
	acpudb_pool_t *pool =
	    container_of(data, acpudb_pool_t, work);
	if (pool) {
		acpu_doorbell_pool_put(pool);
	}
}

int32_t acpu_doorbell_pool_create(uint32_t max_entries,
				    resource_size_t base, acpudb_pool_t **db_pool)
{
	acpudb_pool_t *pool;
	int32_t error;

	pool = kzalloc(sizeof(acpudb_pool_t), GFP_KERNEL);
	if (pool == NULL){
		pr_err("doorbell alloc fail");
		error = ENOMEM;
		goto err_alloc;
	}
	mutex_init(&pool->pool_mutex);
	pool->wq = create_singlethread_workqueue("xdx_dma_misr");
	INIT_WORK(&(pool->work), db_task_handler);

	WARN_ON(max_entries > 32);
	pool->max_entries = max_entries;
	pool->used_entries = 0;
	pool->free_entries = max_entries;
	pool->db_bitmap =
	    (max_entries > 31) ? ~0x0 : (BIT(max_entries) - 1);
	pool->base = base;
	*db_pool = pool;
	return 0;

err_alloc:
	return error;
}

void acpu_doorbell_pool_destroy(acpudb_pool_t * pool)
{
	if (pool->used_entries != 0) {
		pr_err("%s: Attempt to destroy Acpu doorbell pool"
			 "with %u entries still in use", __func__,
			 pool->used_entries);
		return;
	}
	destroy_workqueue(pool->wq);
	mutex_destroy(&pool->pool_mutex);
	kfree(pool);
}

uint32_t acpu_doorbell_pool_get(acpudb_pool_t * pool)
{
	uint32_t idx;

	WARN_ON(!pool->db_bitmap);
	idx = GetFirstBitsInMaskData32(pool->db_bitmap);
	BIT_UNSET(pool->db_bitmap, idx);
	pool->free_entries--;
	pool->used_entries++;
	PVR_DPF((PVR_DBG_MESSAGE, "%s free=%d busy=%d, BitMap %x, Index %d",
		 __func__, pool->free_entries, pool->used_entries,
		 pool->db_bitmap, idx));

	return idx;
}

void acpu_doorbell_pool_put(acpudb_pool_t *pool)
{
	uint32_t db;
	uint32_t bitmap;

	mutex_lock(&pool->pool_mutex);
	bitmap = acpu_doorbell_rreg64((uint64_t) pool->base, CMCU_DOORBELLS_DOORBELL35_OFFSET);
	bitmap &= pool->db_submit;
	while (bitmap) {
		db = GetFirstBitsInMaskData32(bitmap);
		BIT_UNSET(bitmap, db);

		WARN_ON(BIT_ISSET(pool->db_bitmap, db));
		WARN_ON(!BIT_ISSET(pool->db_submit, db));

		if (acpu_doorbell_rreg64((uint64_t) pool->base, DOORBELL_OFFSET(db)) & 0x1) {
			BIT_SET(pool->db_bitmap, db);
			BIT_UNSET(pool->db_submit, db);
			pool->used_entries--;
			pool->free_entries++;

			if (pool->callback[db] && pool->callback_param[db]) {
				pool->
				    callback[db] (pool->callback_param[db]);
				pool->callback[db] = NULL;
			}

		}
	}

	mutex_unlock(&pool->pool_mutex);
}

int acpu_wait_all_db_done(acpudb_pool_t *pool, int timeout_ms)
{
	int err;
	uint32_t db_submit;

	err = -ETIME;
	LOOP_UNTIL_TIMEOUT(timeout_ms * 1000) {
		mutex_lock(&pool->pool_mutex);
		db_submit = pool->db_submit;
		mutex_unlock(&pool->pool_mutex);

		if (!db_submit)
			return 0;
		else
			usleep_range(10, 20);
	}
	END_LOOP_UNTIL_TIMEOUT();

	return err;
}

int32_t
acpu_waitfor_value(volatile uint64_t __iomem addr,
		   uint32_t val, uint32_t mask)
{
#if defined(NO_HARDWARE)
	(void) addr;
	(void) val;
	(void) mask;
	return 0;
#else
	int32_t err;
	uint32_t actual_val;

	err = -ETIME;	/* Initialiser for following loop */
	LOOP_UNTIL_TIMEOUT(MAX_HW_TIME_US) {
		actual_val = (acpu_doorbell_rreg64(addr, 0) & mask);

		if (actual_val == val) {
			/* Expected value has been found */
			err = 0;
			break;
		} else {
			/* wait for event and retry */
			udelay(1);
		}
	}
	END_LOOP_UNTIL_TIMEOUT();

	return err;

#endif /* NO_HARDWARE */
}

uint32_t acpu_doorbell_get_timeout(acpudb_pool_t * pool,
					  uint32_t timeout)
{
	uint32_t idx = INVALID_DB_INDEX;

	if (timeout == 0) {
		timeout = 1000000;
	}

	LOOP_UNTIL_TIMEOUT(timeout) {
		mutex_lock(&pool->pool_mutex);
		if (pool->db_bitmap) {
			idx = acpu_doorbell_pool_get(pool);
			mutex_unlock(&pool->pool_mutex);
			break;
		}
		mutex_unlock(&pool->pool_mutex);
		udelay(10);
	}
	END_LOOP_UNTIL_TIMEOUT();

	return idx;
}

#if defined(SUPPORT_PANGU_ACPU)
void TinyBufferPoolPuts(PVRSRV_RGXDEV_INFO * dev_info, uint32_t bitmap,
			bool um)
{
	uint32_t idx;
	AcpuMemCopy_T *memory_info =
	    (AcpuMemCopy_T *) dev_info->puiTinyCopyVirAddr;
	mutex_lock(dev_info->hTinyPoolLock);
#if defined(DEBUG)
	if (BITMASK_ANY(dev_info->ui32TinyPoolBitmap, bitmap)) {
		pr_debug("bitmap %x, PoolBitmap %x\n", bitmap,
			 dev_info->ui32TinyPoolBitmap);
		WARN_ON(1);
	}
#endif

	BITMASK_SET(dev_info->ui32TinyPoolBitmap, bitmap);

	while (bitmap) {
		idx = GetFirstBitsInMaskData32(bitmap);
		BIT_UNSET(bitmap, idx);
		if (dev_info->asArrayTinyCopyHostAddr[idx]) {
			if (um) {
				uint32_t ui32Ret;
				ui32Ret =
				    copy_to_user(
					(void *)(dev_info->asArrayTinyCopyHostAddr[idx]),
					&(memory_info[idx].Data[0]),
					memory_info[idx].dataSize);
			} else {
				memcpy(
					(void*)(dev_info->asArrayTinyCopyHostAddr[idx]),
				       &(memory_info[idx].Data[0]),
				       memory_info[idx].dataSize);
			}
			dev_info->asArrayTinyCopyHostAddr[idx] = 0;
		}
	}
	mutex_unlock(dev_info->hTinyPoolLock);
	//PVR_DPF((PVR_DBG_MESSAGE, "%s Bitmap %x, Pool BitMap %x", __func__, bitmap, dev_info->ui32TinyPoolBitmap));
}

uint32_t TinyBufferPoolGet(PVRSRV_RGXDEV_INFO * dev_info, bool tries)
{
	uint32_t idx = INVALID_DB_INDEX;

	do {
		mutex_lock(dev_info->hTinyPoolLock);
		if (dev_info->ui32TinyPoolBitmap) {
			idx =
			    GetFirstBitsInMaskData32(dev_info->ui32TinyPoolBitmap);
			BIT_UNSET(dev_info->ui32TinyPoolBitmap, idx);
			mutex_unlock(dev_info->hTinyPoolLock);
			break;
		}
		mutex_unlock(dev_info->hTinyPoolLock);

		if (!tries) {
			break;
		}

		udelay(10);
	} while (1);

	return idx;
}

static void dump_command(PVRSRV_RGXDEV_INFO * dev_info,
				   uint32_t bitmap)
{
	uint32_t idx;
	AcpuMemCopy_T *memory_info =
	    (AcpuMemCopy_T *) dev_info->puiTinyCopyVirAddr;
	mutex_lock(dev_info->hTinyPoolLock);

	while (bitmap) {
		idx = GetFirstBitsInMaskData32(bitmap);
		BIT_UNSET(bitmap, idx);
		printk
		    ("%s Index cmd %d cmd Type %x, copyType %x dada size %d, addr %llx\n",
		     __func__, idx, memory_info[idx].cmdType,
		     memory_info[idx].copyType,
		     memory_info[idx].dataSize,
		     memory_info[idx].DstAddr);
	}
	mutex_unlock(dev_info->hTinyPoolLock);
}

int32_t Do_ACPUTINYCOPY(PVRSRV_RGXDEV_INFO *psDevInfo, uint32_t bitmap, bool bIsUserBuffer)
{
	uint32_t dbNum;
	acpu_dbcmd_memop_tiny_cmd_type acpu_dbcmd_memop_tiny_cmd;
	PVRSRV_DEVICE_NODE *psDeviceNode = psDevInfo->psDeviceNode;
	struct xdx_device *xdev = gfx_to_xdev(psDeviceNode);
	struct completion sync_completion;
	unsigned long res = 0;

	if (bitmap) {
		WARN_ON((psDevInfo->ui32TinyPoolBitmap & bitmap) != 0);

		dbNum = acpu_doorbell_get_timeout(xdev->acpu.acpu_db_pool, 1000000);

		if (dbNum == INVALID_DB_INDEX) {
			PVR_DPF((PVR_DBG_ERROR, "%s get doorbell from pool timeout", __func__));
			TinyBufferPoolPuts(psDevInfo, bitmap, bIsUserBuffer);
			return -ETIME;
		}

		init_completion(&sync_completion);
		acpu_dbcmd_memop_tiny_cmd.cmd = acpu_dbcmd_memop;
		acpu_dbcmd_memop_tiny_cmd.subCmd = acpu_dbcmd_memop_tiny;
		acpu_dbcmd_memop_tiny_cmd.index = bitmap;
		acpu_dbcmd_memop_tiny_cmd.isFast = IMG_FALSE;

		/* In order to prevent cpu from being polling all the time, increase the sleep awakening mechanism */
		xdev->acpu.acpu_db_pool->callback[dbNum] = DoorBellMemset_callback;
		xdev->acpu.acpu_db_pool->callback_param[dbNum] = &sync_completion;
		acpu_doorbell_submit(&(xdev->acpu), dbNum, acpu_dbcmd_memop_tiny_cmd.all);

		res = wait_for_completion_timeout(&sync_completion, msecs_to_jiffies(2000));
		TinyBufferPoolPuts(psDevInfo, bitmap, bIsUserBuffer);

		if (!res) {
			PVR_DPF((PVR_DBG_ERROR, "%s wait for completion timeout", __func__));
			xdev->acpu.acpu_db_pool->callback[dbNum] = NULL;
			xdev->acpu.acpu_db_pool->callback_param[dbNum] = NULL;
			return -ETIME;
		}
	}

	return 0;
}

uint32_t Do_ACPUTINYCOPYNonBlock(PVRSRV_RGXDEV_INFO * dev_info,
				   uint32_t bitmap, bool um)
{
	uint32_t dbNum;
	acpu_dbcmd_memop_tiny_cmd_type tinycopy_cmd;
	PVRSRV_DEVICE_NODE *dev_node = dev_info->psDeviceNode;
	struct xdx_device *xdev = gfx_to_xdev(dev_node);
	struct xdx_acpu *acpu = xdev_to_acpu(xdev);

	WARN_ON(!bitmap);

	WARN_ON((dev_info->ui32TinyPoolBitmap & bitmap) != 0);

	dbNum =
	    acpu_doorbell_get_timeout(xdev->acpu.acpu_db_pool, 1000000);

	if (dbNum == INVALID_DB_INDEX) {
		dev_err(xdev->dev, "%s get doorbell from pool timeout",
			 __func__);
		TinyBufferPoolPuts(dev_info, bitmap, um);
		return INVALID_DB_INDEX;
	}

	tinycopy_cmd.cmd = acpu_dbcmd_memop;
	tinycopy_cmd.subCmd = acpu_dbcmd_memop_tiny;
	tinycopy_cmd.index = bitmap;
	tinycopy_cmd.isFast = IMG_FALSE;

	acpu_doorbell_submit(acpu, dbNum, tinycopy_cmd.all);
	return dbNum;
}

void Wait_ACPUTINYCOPY(PVRSRV_RGXDEV_INFO * dev_info, uint32_t dbNum,
		       uint32_t bitmap)
{
	PVRSRV_DEVICE_NODE *psDeviceNode = dev_info->psDeviceNode;
	struct xdx_device *xdev = gfx_to_xdev(psDeviceNode);
	struct xdx_acpu *acpu = &xdev->acpu;
	if (acpu_waitfor_value
	    ((uint64_t __iomem) ((uint64_t)acpu->rmmio + DOORBELL_OFFSET(dbNum)), 0x01,
	     0x01) != 0) {
		dump_command(dev_info, bitmap);
		dev_err(xdev->dev, "%s polling Doorbell timeout", __func__);
	}
}

void Sharemem_ACPUMEMSET(PVRSRV_RGXDEV_INFO * dev_info, uint32_t * bitmap,
			 uint64_t dev_phy_addr, uint32_t size, uint32_t pattern)
{
	AcpuMemCopy_T *memory_info = (AcpuMemCopy_T *) dev_info->puiTinyCopyVirAddr;
	uint32_t idx;
	uint32_t aligned_size, aligned_len;
	uint64_t aligned_addr;

	if (size < DMA_MEMFILL_BEST_SIZE || pattern != 0) {
		idx = TinyBufferPoolGet(dev_info, false);
		if (INVALID_DB_INDEX == idx) {
			//TinyBuffer is not available, release have allocated
			Do_ACPUTINYCOPY(dev_info, *bitmap, true);
			*bitmap = 0;
			//try it again
			idx = TinyBufferPoolGet(dev_info, true);
		}

		memory_info[idx].DstAddr = dev_phy_addr;
		memory_info[idx].dataSize = size;
		memory_info[idx].Data[0] = pattern;
		memory_info[idx].cmdType =
		    acpu_dbcmd_memop_subtype_tiny_set;
		memory_info[idx].copyType = 0;
		WARN_ON(0 != MOD_2(memory_info[idx].DstAddr, 4));
		WARN_ON(0 != MOD_2(memory_info[idx].dataSize, 4));
		OSWriteMemoryBarrier(&memory_info[idx]);
		BIT_SET(*bitmap, idx);
		return;
	}

	aligned_addr = PVR_ALIGN(dev_phy_addr, (uint64_t) DMA_MEMFILL_ADDR_ALIGN);
	aligned_size = aligned_addr - dev_phy_addr;

	if (aligned_size) {
		idx = TinyBufferPoolGet(dev_info, false);
		if (INVALID_DB_INDEX == idx) {
			//TinyBuffer is not available, release have allocated
			Do_ACPUTINYCOPY(dev_info, *bitmap, true);
			*bitmap = 0;
			//try it again
			idx = TinyBufferPoolGet(dev_info, true);
		}

		memory_info[idx].DstAddr = dev_phy_addr;
		memory_info[idx].dataSize = aligned_size;
		memory_info[idx].Data[0] = pattern;
		memory_info[idx].cmdType =
		    acpu_dbcmd_memop_subtype_tiny_set;
		memory_info[idx].copyType = 0;
		WARN_ON(0 != MOD_2(memory_info[idx].DstAddr, 4));
		WARN_ON(0 != MOD_2(memory_info[idx].dataSize, 4));
		OSWriteMemoryBarrier(&memory_info[idx]);
		BIT_SET(*bitmap, idx);
		size -= aligned_size;
	}

	if (size) {
		aligned_len =
		    PVR_FLOOR_ALIGN(size, (uint32_t) DMA_MEMFILL_BEST_SIZE);
		aligned_size = size - aligned_len;

		if (aligned_len) {
			idx = TinyBufferPoolGet(dev_info, false);
			if (INVALID_DB_INDEX == idx) {
				//TinyBuffer is not available, release have allocated
				Do_ACPUTINYCOPY(dev_info, *bitmap, true);
				*bitmap = 0;
				//try it again
				idx =
				    TinyBufferPoolGet(dev_info, true);
			}

			memory_info[idx].DstAddr = aligned_addr;
			memory_info[idx].dataSize = aligned_len;
			memory_info[idx].Data[0] = pattern;
			memory_info[idx].cmdType =
			    acpu_dbcmd_memop_subtype_tiny_set;
			memory_info[idx].copyType = 2;
			WARN_ON(0 != MOD_2(memory_info[idx].DstAddr, 32));
			WARN_ON(0 != MOD_2(memory_info[idx].dataSize,
					 DMA_MEMFILL_BEST_SIZE));
			OSWriteMemoryBarrier(&memory_info[idx]);
			BIT_SET(*bitmap, idx);
		}

		if (aligned_size) {
			idx = TinyBufferPoolGet(dev_info, false);
			if (INVALID_DB_INDEX == idx) {
				//TinyBuffer is not available, release have allocated
				Do_ACPUTINYCOPY(dev_info, *bitmap, true);
				*bitmap = 0;
				//try it again
				idx = TinyBufferPoolGet(dev_info, true);
			}

			memory_info[idx].DstAddr = aligned_addr + aligned_len;
			memory_info[idx].dataSize = aligned_size;
			memory_info[idx].Data[0] = pattern;
			memory_info[idx].cmdType =
			    acpu_dbcmd_memop_subtype_tiny_set;
			memory_info[idx].copyType = 0;
			OSWriteMemoryBarrier(&memory_info[idx]);
			BIT_SET(*bitmap, idx);
		}
	}

	return;
}

int32_t ACPUTINYMEMSET(PVRSRV_DEVICE_NODE * dev_node,
			    uint64_t devPAddr, uint32_t size,
			    uint32_t * data)
{
	PVRSRV_RGXDEV_INFO *dev_info = dev_node->pvDevice;
	uint32_t bitmap = 0;
	Sharemem_ACPUMEMSET(dev_info, &bitmap, devPAddr, size, data[0]);
	Do_ACPUTINYCOPY(dev_info, bitmap, false);
	return 0;
}

int acpu_tiny_memset_package(PVRSRV_DEVICE_NODE *psDevNode, uint64_t devPAddr, uint32_t size, uint32_t *data)
{
	PVRSRV_RGXDEV_INFO *psDevInfo = psDevNode->pvDevice;
	struct xdx_device *xdev = gfx_to_xdev(psDevNode);
	uint32_t bitmap = 0;
	uint32_t ui32MemSize;

	if (NULL == psDevInfo->puiTinyCopyVirAddr) {
		acpu_dbcmd_memop_memset_multiple(xdev, devPAddr, size, data[0], (0 == MOD_2(size, 4096)));
		return  0;
	}

	/* The data processed at one time cannot exceed (16MB - 4KB)  */
	while (size > 0) {
		ui32MemSize = MIN(MAX_ACPU_MEMSET_SIZE, size);
		Sharemem_ACPUMEMSET(psDevInfo, &bitmap, devPAddr, ui32MemSize, data[0]);
		devPAddr = devPAddr + ui32MemSize;
		size -= ui32MemSize;
	}

	Do_ACPUTINYCOPY(psDevInfo, bitmap, IMG_FALSE);

	return  0;
}

static void Sharemem_ACPUTINYCOPYD2D(PVRSRV_RGXDEV_INFO * dev_info, uint32_t * bitmap,
		uint64_t devPAddr, uint64_t srcPAddr, uint32_t size)
{
	AcpuMemCopy_T *memory_info = (AcpuMemCopy_T *) dev_info->puiTinyCopyVirAddr;
	uint32_t idx;

	idx = TinyBufferPoolGet(dev_info, false);
	if (INVALID_DB_INDEX == idx) {
		Do_ACPUTINYCOPY(dev_info, *bitmap, true);
		*bitmap = 0;
		idx = TinyBufferPoolGet(dev_info, true);
	}

	memory_info[idx].dataSize = size;
	memory_info[idx].copyType = 2;
	memory_info[idx].cmdType = acpu_dbcmd_memop_subtype_tiny_copy;
	memory_info[idx].MemCopyParam.DstAddr = devPAddr;
	memory_info[idx].MemCopyParam.SrcAddr = srcPAddr;
	OSWriteMemoryBarrier(&memory_info[idx]);
	BIT_SET(*bitmap, idx);

	return;
}

int acpu_tiny_copy_package(PVRSRV_DEVICE_NODE *psDevNode, uint64_t devPAddr, uint64_t srcPAddr, uint32_t size)
{
	PVRSRV_RGXDEV_INFO *psDevInfo = psDevNode->pvDevice;
	uint32_t bitmap = 0;
	uint32_t ui32MemSize;

	/* The data processed at one time cannot exceed 64MB  */
	while (size > 0) {
		ui32MemSize = MIN(MAX_ACPU_MEMCOPY_SIZE, size);
		Sharemem_ACPUTINYCOPYD2D(psDevInfo, &bitmap, devPAddr, srcPAddr, ui32MemSize);
		devPAddr = devPAddr + ui32MemSize;
		srcPAddr = srcPAddr + ui32MemSize;
		size -= ui32MemSize;
	}

	Do_ACPUTINYCOPY(psDevInfo, bitmap, IMG_FALSE);

	return  0;
}
#endif
