/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/time.h>
#include "xdx_ddcci.h"
#include <linux/i2c.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0))
#include <drm/display/drm_dp_helper.h>
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0))
#include <drm/dp/drm_dp_helper.h>
#else
#include <drm/drm_dp_helper.h>
#endif


/* ddc/ci defines*/
#define DDCCI_COMMAND_READ 0x01      /*read ctrl value*/
#define DDCCI_REPLY_READ   0x02      /*read ctrl value reply*/
#define DDCCI_COMMAND_WRITE 0x03     /*write ctrl value*/

#define DDCCI_COMMAND_SAVE 0x0c      /*save current settings*/
#define DDCCI_REPLY_CAPS 0xe3        /*get monitor caps reply*/
#define DDCCI_COMMAND_CAPS 0xf3      /**get monitor caps*/
#define DDCCI_COMMAND_PRESENCE 0xf7  /*ACCESS.bus presence check*/

/*ddc/ci iface tunables*/
#define DELAY      45            /*uS to wait after write*/
#define MAX_BYTES  127              /*max message length*/
#define RETRYS     3

/*magic numbers*/
#define MAGIC_1 0x51                /*first byte to send, host address*/
#define MAGIC_2 0x80                /*second byte to send, ored with length*/
#define MAGIC_XOR 0x50              /*initial xor received frame*/


/*DDCCI_HDMI write len bytes (stored in buf) to ddc/ci at address addr*/
/*return 0 on success, -1 on failure*/
static int xdx_hdmi_ddcci_write(struct hdmi_dev *hdmi, unsigned char addr, unsigned char *buf, unsigned char len)
{
	uint8_t i = 0;
	unsigned char _buf[7];
	unsigned xor = ((unsigned char)addr << 1); /*initial xor value*/

	struct i2c_adapter *adap = &(hdmi->i2c->adap);
	struct i2c_msg *msg = kzalloc(sizeof(struct i2c_msg), GFP_KERNEL);

	_buf[i++] = MAGIC_1;

	/*second magic includes message size*/
	xor ^= (_buf[i++] = MAGIC_2 | len);

	while (len--) {
		/*bytes to send*/
		xor ^= (_buf[i++] = *buf++);
	}

	_buf[i++] = xor;
	/*wait for previous command to complete*/
	mdelay(80);
	msg->addr = addr;
	msg->len = i;
	msg->flags = 0;
	msg->buf = _buf;

	if (hdmi->i2c->adap.algo->master_xfer(adap, msg, 1)) {
		kfree(msg);
		return 1;
	}
	kfree(msg);

	return 0;
}

/*write value to register ctrl of ddc/ci at address addr*/
int xdx_hdmi_ddcci_writectrl(struct hdmi_dev *hdmi, unsigned char addr, unsigned char ctrl, unsigned short value)
{
	unsigned char buf[4];

	buf[0] = DDCCI_COMMAND_WRITE;
	buf[1] = ctrl;
	buf[2] = (value >> 8);
	buf[3] = (value & 255);

	return xdx_hdmi_ddcci_write(hdmi, addr, buf, sizeof(buf));
}

/*read register ctrl raw data of ddc/ci at address addr*/
int xdx_hdmi_ddcci_readctrl(struct hdmi_dev *hdmi, unsigned char addr, unsigned char ctrl, unsigned char *buf, unsigned int len)
{
	unsigned char _buf[2];
	struct i2c_adapter *adap = &(hdmi->i2c->adap);
	struct i2c_msg *msg = kzalloc(sizeof(struct i2c_msg), GFP_KERNEL);

	_buf[0] = DDCCI_COMMAND_READ;
	_buf[1] = ctrl;

	if (xdx_hdmi_ddcci_write(hdmi, addr, _buf, sizeof(_buf)) == 0) {
		return -1;
	}
	mdelay(80);

	msg->addr = addr;
	msg->len = len;
	msg->flags = I2C_M_RD;
	msg->buf = buf;

	if (hdmi->i2c->adap.algo->master_xfer(adap, msg, 1)) {
		kfree(msg);
		return 1;
	}

	kfree(msg);

	return 0;
}


int xdx_hdmi_ddcci_dumpctrl(struct hdmi_dev *hdmi, unsigned char addr, unsigned char ctrl, uint8_t *value, uint8_t *maxinum)
{
	unsigned char buf[13];

	if (xdx_hdmi_ddcci_readctrl(hdmi, addr, ctrl, buf, sizeof(buf)) == 0)
		return -1;
	
	*value = buf[9];
	return 0;
}

/*DDCCI_DP write len bytes (stored in buf) to ddc/ci at address addr*/
static int xdx_dp_ddcci_write(struct dp_dev *dp, unsigned char addr, unsigned char *buf, unsigned char len)
{
	int i = 0;
	unsigned char _buf[8];
	unsigned xor = ((unsigned char)addr << 1); /*initial xor value*/

        struct i2c_adapter *adap = &(dp->aux.ddc);
	struct i2c_msg *msg = kzalloc(sizeof(struct i2c_msg), GFP_KERNEL);

	/*put first magic*/
	xor ^= (_buf[i++] = MAGIC_1);

	/*second magic includes message size*/
	xor ^= (_buf[i++] = MAGIC_2 | len);

	while (len--){
		/*bytes to send*/
		xor ^= (_buf[i++] = *buf++);
	}
	/*finally put checksum*/
	_buf[i++] = xor;

	/*wait for previous command to complete*/
	mdelay(80);

	msg->addr = addr;
	msg->len = i;
	msg->flags = 0;
	msg->buf = _buf;

        if (dp->aux.ddc.algo->master_xfer(adap, msg, 1)) {
		kfree(msg);
		return 1;
	}
	kfree(msg);

	return 0;
}

/*write value to register ctrl of ddc/ci at address addr*/
int xdx_dp_ddcci_writectrl(struct dp_dev *dp, unsigned char addr, unsigned char ctrl, unsigned short value)
{
	unsigned char buf[4];

	buf[0] = DDCCI_COMMAND_WRITE;
	buf[1] = ctrl;
	buf[2] = (value >> 8);
	buf[3] = (value & 255);

	return xdx_dp_ddcci_write(dp, addr, buf, sizeof(buf));
}

/*read register ctrl raw data of ddc/ci at address addr*/
int xdx_dp_ddcci_readctrl(struct dp_dev *dp, unsigned char addr, unsigned char ctrl, unsigned char *buf, unsigned int len)
{
	unsigned char _buf[2];
        struct i2c_adapter *adap = &(dp->aux.ddc);
        struct i2c_msg *msg = kzalloc(sizeof(struct i2c_msg), GFP_KERNEL);

	_buf[0] = DDCCI_COMMAND_READ;
	_buf[1] = ctrl;

	if (xdx_dp_ddcci_write(dp, addr, _buf, sizeof(_buf)) == 0)
		return -1;

	mdelay(80);

	msg->addr = addr;
	msg->len = len;
	msg->flags = I2C_M_RD;
	msg->buf = buf;

	if (dp->aux.ddc.algo->master_xfer(adap, msg, 1)) {
		kfree(msg);
		return 1;
	}

	kfree(msg);

	return 0;
}

int xdx_dp_ddcci_dumpctrl(struct dp_dev *dp, unsigned char addr, unsigned char ctrl, uint8_t *value, uint8_t *maxinum)
{
	unsigned char buf[13];

	if (xdx_dp_ddcci_readctrl(dp, addr, ctrl, buf, sizeof(buf)) == 0)
		return -1;

	*value = buf[9];
	return 0;
}


