/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/component.h>
#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/media-bus-format.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 1, 0)
#include <drm/drm.h>
#include <linux/sched.h>
#endif
#include <drm/drm_framebuffer.h>
#include <drm/drm_atomic.h>
#include <drm/drm_atomic_helper.h>
#include <drm/xdxgpu_drm.h>

#include "dm_dc.h"
#include "dm_crtc.h"
#include "dm_irq.h"
#include "xdx_device.h"
#include "xdx_irq.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0) || (defined(XDXGPU_CENTOS82_SUPPORT))
#include <drm/drm_fourcc.h>
#include <drm/drm_vblank.h>
#endif

#define dmDownAlign(base, alignment) \
( \
	base & ~(alignment -1) \
)

static inline void update_format(u32 format, u64 mod, struct dc_fb *fb)
{
	u8 f = DC_FORMAT_A8R8G8B8;

	switch (format) {
	case DRM_FORMAT_XRGB4444:
	case DRM_FORMAT_RGBX4444:
	case DRM_FORMAT_XBGR4444:
	case DRM_FORMAT_BGRX4444:
		f = DC_FORMAT_X4R4G4B4;
		break;
	case DRM_FORMAT_ARGB4444:
	case DRM_FORMAT_RGBA4444:
	case DRM_FORMAT_ABGR4444:
	case DRM_FORMAT_BGRA4444:
		f = DC_FORMAT_A4R4G4B4;
		break;
	case DRM_FORMAT_XRGB1555:
	case DRM_FORMAT_RGBX5551:
	case DRM_FORMAT_XBGR1555:
	case DRM_FORMAT_BGRX5551:
		f = DC_FORMAT_X1R5G5B5;
		break;
	case DRM_FORMAT_ARGB1555:
	case DRM_FORMAT_RGBA5551:
	case DRM_FORMAT_ABGR1555:
	case DRM_FORMAT_BGRA5551:
		f = DC_FORMAT_A1R5G5B5;
		break;
	case DRM_FORMAT_RGB565:
	case DRM_FORMAT_BGR565:
		f = DC_FORMAT_R5G6B5;
		break;
	case DRM_FORMAT_XRGB8888:
	case DRM_FORMAT_RGBX8888:
	case DRM_FORMAT_XBGR8888:
	case DRM_FORMAT_BGRX8888:
		f = DC_FORMAT_X8R8G8B8;
		break;
	case DRM_FORMAT_ARGB8888:
	case DRM_FORMAT_RGBA8888:
	case DRM_FORMAT_ABGR8888:
	case DRM_FORMAT_BGRA8888:
		f = DC_FORMAT_A8R8G8B8;
		break;
	case DRM_FORMAT_YUYV:
	case DRM_FORMAT_YVYU:
		f = DC_FORMAT_YUY2;
		break;
	case DRM_FORMAT_UYVY:
	case DRM_FORMAT_VYUY:
		f = DC_FORMAT_UYVY;
		break;
	case DRM_FORMAT_YUV420:
	case DRM_FORMAT_YVU420:
		f = DC_FORMAT_YV12;
		break;
	case DRM_FORMAT_NV21:
		f = DC_FORMAT_NV12;
		break;
	case DRM_FORMAT_NV16:
	case DRM_FORMAT_NV61:
		f = DC_FORMAT_NV16;
		break;
	case DRM_FORMAT_P010:
		f = DC_FORMAT_P010;
		break;
	case DRM_FORMAT_ARGB2101010:
	case DRM_FORMAT_RGBA1010102:
	case DRM_FORMAT_ABGR2101010:
	case DRM_FORMAT_BGRA1010102:
		f = DC_FORMAT_A2R10G10B10;
		break;
	case DRM_FORMAT_NV12:
		if (fourcc_mod_xdx_get_type(mod) ==
		    DRM_FORMAT_MOD_XDX_TYPE_CUSTOM_10BIT)
			f = DC_FORMAT_NV12_10BIT;
		else
			f = DC_FORMAT_NV12;
		break;
	case DRM_FORMAT_YUV444:
		if (fourcc_mod_xdx_get_type(mod) ==
		    DRM_FORMAT_MOD_XDX_TYPE_CUSTOM_10BIT)
			f = DC_FORMAT_YUV444_10BIT;
		else
			f = DC_FORMAT_YUV444;
		break;
	default:
		break;
	}

	fb->format = f;
}

static inline void update_swizzle(u32 format, struct dc_fb *fb)
{
	fb->swizzle = DC_SWIZZLE_ARGB;
	fb->uv_swizzle = 0;

	switch (format) {
	case DRM_FORMAT_RGBX4444:
	case DRM_FORMAT_RGBA4444:
	case DRM_FORMAT_RGBX5551:
	case DRM_FORMAT_RGBA5551:
	case DRM_FORMAT_RGBX8888:
	case DRM_FORMAT_RGBA8888:
	case DRM_FORMAT_RGBA1010102:
		fb->swizzle = DC_SWIZZLE_RGBA;
		break;
	case DRM_FORMAT_XBGR4444:
	case DRM_FORMAT_ABGR4444:
	case DRM_FORMAT_XBGR1555:
	case DRM_FORMAT_ABGR1555:
	case DRM_FORMAT_BGR565:
	case DRM_FORMAT_XBGR8888:
	case DRM_FORMAT_ABGR8888:
	case DRM_FORMAT_ABGR2101010:
		fb->swizzle = DC_SWIZZLE_ABGR;
		break;
	case DRM_FORMAT_BGRX4444:
	case DRM_FORMAT_BGRA4444:
	case DRM_FORMAT_BGRX5551:
	case DRM_FORMAT_BGRA5551:
	case DRM_FORMAT_BGRX8888:
	case DRM_FORMAT_BGRA8888:
	case DRM_FORMAT_BGRA1010102:
		fb->swizzle = DC_SWIZZLE_BGRA;
		break;
	case DRM_FORMAT_YVYU:
	case DRM_FORMAT_VYUY:
	case DRM_FORMAT_NV21:
	case DRM_FORMAT_NV61:
		fb->uv_swizzle = 1;
		break;
	default:
		break;
	}
}

static inline void update_watermark(struct drm_property_blob *watermark,
				    struct dc_fb *fb)
{
	struct drm_xdx_watermark *data;
	fb->water_mark = 0;

	if (watermark) {
		data = (struct drm_xdx_watermark *)watermark->data;
		fb->water_mark = data->watermark & 0xFFFFF;
	}
}

static inline u8 to_dc_rotation(unsigned int rotation)
{
	u8 rot;

	switch (rotation & DRM_MODE_REFLECT_MASK) {
	case DRM_MODE_REFLECT_X:
		rot = DC_FLIP_X;
		return rot;
	case DRM_MODE_REFLECT_Y:
		rot = DC_FLIP_Y;
		return rot;
	case DRM_MODE_REFLECT_X | DRM_MODE_REFLECT_Y:
		rot = DC_FLIP_XY;
		return rot;
	default:
		break;
	}

	switch (rotation & DRM_MODE_ROTATE_MASK) {
	case DRM_MODE_ROTATE_0:
		rot = DC_ROT_0;
		break;
	case DRM_MODE_ROTATE_90:
		rot = DC_ROT_90;
		break;
	case DRM_MODE_ROTATE_180:
		rot = DC_ROT_180;
		break;
	case DRM_MODE_ROTATE_270:
		rot = DC_ROT_270;
		break;
	default:
		rot = DC_ROT_0;
		break;
	}

	return rot;
}

static inline u8 to_dc_yuv_color_space(u32 color_space)
{
	u8 cs;

	switch (color_space) {
	case DRM_COLOR_YCBCR_BT601:
		cs = DC_YUV_COLOR_SPACE_601;
		break;
	case DRM_COLOR_YCBCR_BT709:
		cs = DC_YUV_COLOR_SPACE_709;
		break;
	case DRM_COLOR_YCBCR_BT2020:
		cs = DC_YUV_COLOR_SPACE_2020;
		break;
	default:
		cs = DC_YUV_COLOR_SPACE_601;
		break;
	}

	return cs;
}

static inline u8 to_dc_tile_mode(u64 modifier)
{
	return (u8)(modifier & DRM_FORMAT_MOD_XDX_NORM_MODE_MASK);
}

static inline u8 to_dc_display_id(struct dm_dc *dc, struct drm_crtc *crtc)
{
	struct dm_crtc *dm_crtc = to_dm_crtc(crtc);

	return dm_crtc->id;
}

void dm_dc_hw_fini(struct dm_dc *dc)
{
	dc->funcs->enable_interrupt(dc, false);
}

int dm_dc_hw_init(struct dm_dc *dc)
{
	int ret = 0;

	if (dc->funcs->hw_init)
		ret = dc->funcs->hw_init(dc);

	return ret;
}

static u32 calc_factor(u32 src, u32 dest)
{
	u32 factor = 1 << 16;

	if ((src > 1) && (dest > 1))
		factor = ((src - 1) << 16) / (dest - 1);

	return factor;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0) && (!defined(XDXGPU_CENTOS79_SUPPORT))
bool drm_rotation_90_or_270(unsigned int rotation)
{
        return rotation & (BIT(DRM_ROTATE_90) | BIT(DRM_ROTATE_270));
}
#endif

static void update_scale(struct drm_plane_state *state, struct dc_roi *roi,
			 struct dc_scale *scale)
{
	int dst_w = drm_rect_width(&state->dst);
	int dst_h = drm_rect_height(&state->dst);
	int src_w, src_h, temp;

	scale->enable = false;

	if (roi->enable) {
		src_w = roi->width;
		src_h = roi->height;
	} else {
		src_w = drm_rect_width(&state->src) >> 16;
		src_h = drm_rect_height(&state->src) >> 16;
	}

	if (drm_rotation_90_or_270(state->rotation)) {
		temp = src_w;
		src_w = src_h;
		src_h = temp;
	}

	if (src_w != dst_w) {
		scale->scale_factor_x = calc_factor(src_w, dst_w);
		scale->enable = true;
	} else {
		scale->scale_factor_x = 1 << 16;
	}

	if (src_h != dst_h) {
		scale->scale_factor_y = calc_factor(src_h, dst_h);
		scale->enable = true;
	} else {
		scale->scale_factor_y = 1 << 16;
	}
}

static void update_fb(struct dm_plane *plane, struct dc_fb *fb, u8 display_id)
{
	struct drm_plane_state *state = plane->base.state;
	struct dm_plane_state *plane_state = to_dm_plane_state(state);
	struct drm_framebuffer *drm_fb = state->fb;
	struct drm_rect *src = &state->src;
	struct xdx_dm* dm = dm_plane_to_dm(plane);
	struct xdx_device* xdev = dm_to_xdev(dm);
	u32 dma_offset, start_x, start_y;

	start_x = src->x1 >> 16;
	start_y = src->y1 >> 16;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)) || (defined(XDXGPU_CENTOS79_SUPPORT))
	dma_offset = start_y * drm_fb->pitches[0] + start_x * drm_fb->format->cpp[0];
#else
	dma_offset = start_y * drm_fb->pitches[0] + start_x * (drm_fb->bits_per_pixel / 8);
#endif
	fb->y_address = plane->dma_addr[0] + dma_offset;
	fb->y_address = dmDownAlign(fb->y_address, xdev->hw_info.alignment);
	fb->y_stride = drm_fb->pitches[0];
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)) || (defined(XDXGPU_CENTOS79_SUPPORT))
	if (drm_fb->format->format == DRM_FORMAT_YVU420) {
#else
	if (drm_fb->pixel_format == DRM_FORMAT_YVU420) {
#endif
		fb->u_address = plane->dma_addr[2] + dma_offset;
		fb->v_address = plane->dma_addr[1] + dma_offset;
		fb->u_stride = drm_fb->pitches[2];
		fb->v_stride = drm_fb->pitches[1];
	} else {
		fb->u_address = plane->dma_addr[1] + dma_offset;
		fb->v_address = plane->dma_addr[2] + dma_offset;
		fb->u_stride = drm_fb->pitches[1];
		fb->v_stride = drm_fb->pitches[2];
	}
	fb->display_id = display_id;
	fb->width = drm_rect_width(src) >> 16;
	fb->height = drm_rect_height(src) >> 16;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)) || (defined(XDXGPU_CENTOS79_SUPPORT))
	fb->tile_mode = to_dc_tile_mode(drm_fb->modifier);
#else
	fb->tile_mode = to_dc_tile_mode(drm_fb->modifier[0]);
#endif
	fb->rotation = to_dc_rotation(state->rotation);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0)) || (defined(XDXGPU_CENTOS79_SUPPORT))
	fb->yuv_color_space = to_dc_yuv_color_space(state->color_encoding);
#endif
	fb->zpos = state->zpos;
	fb->enable = state->visible;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)) || (defined(XDXGPU_CENTOS79_SUPPORT))
        update_format(drm_fb->format->format, drm_fb->modifier, fb);
        update_swizzle(drm_fb->format->format, fb);
#else
        update_format(drm_fb->pixel_format, drm_fb->modifier[0], fb);
        update_swizzle(drm_fb->pixel_format, fb);
#endif
	update_watermark(plane_state->watermark, fb);

	plane_state->status.tile_mode = fb->tile_mode;
}

#ifdef CONFIG_VERISILICON_DEC
static void update_fbc(struct dm_dc *dc, struct dm_plane *plane, bool *enable)
{
	struct dec_fb dec_fb;
	struct drm_plane_state *state = plane->base.state;
	struct drm_framebuffer *drm_fb = state->fb;
	struct xdx_device *xdev = dm_to_xdev(dc->dm);
	u8 i;

	if (!xdev->flags.dec_support) {
		*enable = false;
		return;
	}

	memset(&dec_fb, 0, sizeof(struct dec_fb));
	dec_fb.fb = drm_fb;

	if (fourcc_mod_xdx_get_type(drm_fb->modifier) !=
	    DRM_FORMAT_MOD_XDX_TYPE_COMPRESSED) {
		*enable = false;
	} else {
		*enable = true;

		for (i = 0; i < DEC_PLANE_MAX; i++) {
			dec_fb.addr[i] = plane->dma_addr[i];
			dec_fb.stride[i] = drm_fb->pitches[i];
		}
	}

	if (dc->funcs->dec_config)
		dc->funcs->dec_config(dc, plane, &dec_fb);
}

static void disable_fbc(struct dm_dc *dc, struct dm_plane *plane)
{
	struct xdx_device *xdev = dm_to_xdev(dc->dm);

	if (!xdev->flags.dec_support)
		return;

	if (dc->funcs->dec_config)
		dc->funcs->dec_config(dc, plane, NULL);
}
#endif

static void update_degamma(struct dm_dc *dc, struct dm_plane *plane,
			   struct dm_plane_state *plane_state)
{
	if (dc->funcs->update_degamma)
		dc->funcs->update_degamma(dc, plane, plane_state->degamma);
	plane_state->degamma_changed = false;
}

static void update_roi(struct dm_dc *dc, struct dm_plane *plane,
		       struct dm_plane_state *plane_state, struct dc_roi *roi)
{
	struct drm_xdx_roi *data;
	struct drm_rect *src = &plane_state->base.src;
	u16 src_w = drm_rect_width(src) >> 16;
	u16 src_h = drm_rect_height(src) >> 16;

	if (plane_state->roi) {
		data = (struct drm_xdx_roi *)plane_state->roi->data;

		if (data->enable) {
			roi->x = data->roi_x;
			roi->y = data->roi_y;
			roi->width = (data->roi_x + data->roi_w > src_w) ?
						   (src_w - data->roi_x) :
						   data->roi_w;
			roi->height = (data->roi_y + data->roi_h > src_h) ?
						    (src_h - data->roi_y) :
						    data->roi_h;
			roi->enable = true;
		} else {
			roi->enable = false;
		}

		if (dc->funcs->update_roi)
			dc->funcs->update_roi(dc, plane, roi);
	} else {
		roi->enable = false;
	}
}

static void update_color_mgmt(struct dm_dc *dc, struct dm_plane *plane,
			      struct dc_fb *fb,
			      struct dm_plane_state *plane_state)
{
	struct drm_xdx_color_mgmt *data;
	struct dc_colorkey colorkey = { 0 };

	if (plane_state->color_mgmt) {
		data = (struct drm_xdx_color_mgmt *)plane_state->color_mgmt->data;

		fb->clear_enable = data->clear_enable;
		fb->clear_value = data->clear_value;

		if (data->colorkey > data->colorkey_high)
			data->colorkey = data->colorkey_high;

		colorkey.colorkey = data->colorkey;
		colorkey.colorkey_high = data->colorkey_high;
		colorkey.transparency = data->transparency;
		if (dc->funcs->update_colorkey)
			dc->funcs->update_colorkey(dc, plane, &colorkey);
	}
}

static void update_plane(struct dm_dc *dc, struct dm_plane *plane)
{
	struct dc_fb fb = { 0 };
	struct dc_scale scale;
	struct dc_position pos;
	#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)) || (defined(XDXGPU_CENTOS82_SUPPORT))
	struct dc_blend blend;
	#endif
	struct dc_roi roi;
	struct drm_plane_state *state = plane->base.state;
	struct dm_plane_state *plane_state = to_dm_plane_state(state);
	struct drm_rect *dest = &state->dst;
	struct dm_crtc *dm_crtc = to_dm_crtc(state->crtc);
	bool dec_enable = false;
	u8 display_id = 0;

#ifdef CONFIG_VERISILICON_DEC
	update_fbc(dc, plane, &dec_enable);
#endif

	display_id = dm_crtc->id;

	update_fb(plane, &fb, display_id);
	fb.dec_enable = dec_enable;

	update_roi(dc, plane, plane_state, &roi);

	update_scale(state, &roi, &scale);

	if (plane_state->degamma_changed)
		update_degamma(dc, plane, plane_state);

	pos.start_x = dest->x1;
	pos.start_y = dest->y1;
	pos.end_x = dest->x2;
	pos.end_y = dest->y2;


	#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)) || (defined(XDXGPU_CENTOS82_SUPPORT))
	blend.alpha = (u8)(state->alpha >> 8);
	blend.blend_mode = (u8)(state->pixel_blend_mode);
	#endif

	update_color_mgmt(dc, plane, &fb, plane_state);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)) || (defined(XDXGPU_CENTOS82_SUPPORT))
	if (dc->funcs->update_plane)
		dc->funcs->update_plane(dc, plane, &fb, &scale, &pos, &blend);
#else
	if (dc->funcs->update_plane)
		dc->funcs->update_plane(dc, plane, &fb, &scale, &pos);
#endif

}

static void update_qos(struct dm_dc *dc, struct dm_plane *plane)
{
	struct drm_plane_state *state = plane->base.state;
	struct dm_plane_state *plane_state = to_dm_plane_state(state);
	struct drm_xdx_watermark *data;
	struct dc_qos qos;

	if (plane_state->watermark) {
		data = (struct drm_xdx_watermark *)plane_state->watermark->data;

		if (data->qos_high) {
			if (data->qos_low > data->qos_high)
				data->qos_low = data->qos_high;

			qos.low_value = data->qos_low & 0x0F;
			qos.high_value = data->qos_high & 0x0F;
			if (dc->funcs->update_qos)
				dc->funcs->update_qos(dc, &qos);
		}
	}
}

static void update_cursor_size(struct drm_plane_state *state,
			       struct dc_cursor *cursor)
{
	u8 size_type;

	switch (state->crtc_w) {
	case 32:
		size_type = DC_CURSOR_SIZE_32X32;
		break;
	case 64:
		size_type = DC_CURSOR_SIZE_64X64;
		break;
	default:
		size_type = DC_CURSOR_SIZE_32X32;
		break;
	}

	cursor->size = size_type;
}

static void update_cursor_plane(struct dm_dc *dc, struct dm_plane *plane)
{
	struct drm_plane_state *state = plane->base.state;
	struct dm_crtc *dm_crtc = to_dm_crtc(state->crtc);
	struct drm_framebuffer *drm_fb = state->fb;
	struct dc_cursor cursor;

	cursor.address = plane->dma_addr[0];

#if (defined(OS_NAME_OPENKYLIN) || defined(OS_NAME_KYLIN) || defined(OS_NAME_UBUNTU)) && LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
	cursor.x = state->crtc_x + drm_fb->width - state->hotspot_x;
	cursor.y = state->crtc_y + drm_fb->height - state->hotspot_y;
#else
	cursor.x = state->crtc_x + drm_fb->width - drm_fb->hot_x;
	cursor.y = state->crtc_y + drm_fb->height - drm_fb->hot_y;
#endif
	cursor.hot_x = drm_fb->width;
	cursor.hot_y = drm_fb->height;
	cursor.display_id = dm_crtc->id;
	update_cursor_size(state, &cursor);
	cursor.enable = true;

	if (dc->funcs->update_cursor)
		dc->funcs->update_cursor(dc, cursor.display_id, &cursor);
}

void dm_dc_plane_update(struct dm_plane *dm_plane)
{
	struct dm_dc *dc = dm_plane->dc;
	enum drm_plane_type type = dm_plane->base.type;

	switch (type) {
	case DRM_PLANE_TYPE_PRIMARY:
	case DRM_PLANE_TYPE_OVERLAY:
		update_plane(dc, dm_plane);
		update_qos(dc, dm_plane);
		break;
	case DRM_PLANE_TYPE_CURSOR:
		update_cursor_plane(dc, dm_plane);
		break;
	default:
		break;
	}
}

void dm_dc_plane_disable(struct dm_plane *dm_plane,
			 struct drm_plane_state *old_state)
{
	struct dm_dc *dc = dm_plane->dc;
	enum drm_plane_type type = dm_plane->base.type;
	struct dc_fb fb = { 0 };
	struct dc_cursor cursor = { 0 };
	struct dm_crtc *dm_crtc = to_dm_crtc(old_state->crtc);

	switch (type) {
	case DRM_PLANE_TYPE_PRIMARY:
	case DRM_PLANE_TYPE_OVERLAY:
		fb.enable = false;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)) || (defined(XDXGPU_CENTOS82_SUPPORT))
		if (dc->funcs->update_plane)
			dc->funcs->update_plane(dc, dm_plane, &fb, NULL, NULL,
						NULL);
#else
		if (dc->funcs->update_plane)
			dc->funcs->update_plane(dc, dm_plane, &fb, NULL, NULL);
#endif

#ifdef CONFIG_VERISILICON_DEC
		disable_fbc(dc, dm_plane);
#endif
		break;
	case DRM_PLANE_TYPE_CURSOR:
		cursor.enable = false;
		cursor.display_id = dm_crtc->id;
		if (dc->funcs->update_cursor)
			dc->funcs->update_cursor(dc, cursor.display_id,
						 &cursor);
		break;
	default:
		break;
	}
}

int dm_dc_universal_init(struct xdx_dm *dm, struct dm_dc *dc,
			 const struct dm_dc_funcs *funcs, u32 id)
{
	int ret;
	if (!dc || !funcs)
		return -EINVAL;

	if (id > MAX_OUTPUT_NUM - 1)
		return -EINVAL;

	dc->id = id;
	dc->funcs = funcs;
	dc->dm = dm;

	dm->dc[id] = dc;
	ret = dm_dc_vsync_irq_init(dc);
	if(ret)
		return ret;
	ret = dm_dc_isr_install(dc);

	return ret;
}

void dm_dc_universal_fini(struct dm_dc *dc)
{
	int ret;

	dm_dc_isr_uninstall(dc);
	ret = dm_dc_vsync_irq_deinit(dc);
	if(ret)
		return;
	dc->funcs->destroy(dc);
}

uint32_t dm_dc_get_crtc_mask(struct dm_dc *dc)
{
	struct xdx_dm *dm = dc->dm;
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct drm_device *ddev = xdev_to_drm(xdev);

	return ((1 << ddev->mode_config.num_crtc) - 1);
}
