/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <drm/drm_atomic.h>
#include <drm/drm_atomic_helper.h>
#include <drm/drm_plane_helper.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0))
#include <drm/drm_fb_dma_helper.h>
#else
#include <drm/drm_fb_cma_helper.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0))
#include <drm/drm_gem_dma_helper.h>
#else
#include <drm/drm_gem_cma_helper.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0))
#include <drm/drm_gem_framebuffer_helper.h>
#endif
#include <linux/fs.h>
#include <drm/xdxgpu_drm.h>
#if (defined(XDXGPU_CENTOS85_SUPPORT))
#include <drm/drm_atomic_uapi.h>
#endif
#include "dm_crtc.h"
#include "dm_plane.h"
#include "dm_fb.h"
#include "xdx_gem.h"
#include "xdx_vmm.h"
#include "xdx_device.h"
#include "dc8200.h"
#include "kernel_compatibility.h"

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0)) && (!defined(XDXGPU_CENTOS79_SUPPORT))
static inline struct drm_rect
drm_plane_state_src(const struct drm_plane_state *state)
{
	struct drm_rect src = {
		.x1 = state->src_x,
		.y1 = state->src_y,
		.x2 = state->src_x + state->src_w,
		.y2 = state->src_y + state->src_h,
	};
	return src;
}

static inline struct drm_rect
drm_plane_state_dest(const struct drm_plane_state *state)
{
	struct drm_rect dest = {
		.x1 = state->crtc_x,
		.y1 = state->crtc_y,
		.x2 = state->crtc_x + state->crtc_w,
		.y2 = state->crtc_y + state->crtc_h,
	};
	return dest;
}
#endif

static void dm_plane_destory(struct drm_plane *plane)
{
	struct dm_plane *dm_plane = to_dm_plane(plane);

	drm_plane_cleanup(plane);

	dm_plane->funcs->destroy(dm_plane);
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0) && (!defined(XDXGPU_CENTOS85_SUPPORT)) && (!defined(XDXGPU_CENTOS82_SUPPORT))
static void xdx_atomic_helper_plane_reset(struct drm_plane *plane,
				     struct drm_plane_state *state)
{
	state->plane = plane;
	state->rotation = DRM_MODE_ROTATE_0;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
	state->alpha = DRM_BLEND_ALPHA_OPAQUE;
#endif
	plane->state = state;
}
#endif
static void dm_plane_reset(struct drm_plane *plane)
{
	struct dm_plane_state *dm_plane_state;
	struct dm_plane *dm_plane = to_dm_plane(plane);

	if (plane->state) {
		dm_plane_state = to_dm_plane_state(plane->state);

		__drm_atomic_helper_plane_destroy_state(plane->state);

		kfree(dm_plane_state);
		plane->state = NULL;
	}

	dm_plane_state = kzalloc(sizeof(*dm_plane_state), GFP_KERNEL);
	if (!dm_plane_state)
		return;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0) || (defined(XDXGPU_CENTOS85_SUPPORT)) || (defined(XDXGPU_CENTOS82_SUPPORT))
	__drm_atomic_helper_plane_reset(plane, &dm_plane_state->base);
#else
	xdx_atomic_helper_plane_reset(plane, &dm_plane_state->base);
#endif

	dm_plane_state->degamma = XDX_DEGAMMA_DISABLE;
	dm_plane_state->degamma_changed = false;
	dm_plane_state->base.zpos = dm_plane->zpos;

	memset(&dm_plane_state->status, 0, sizeof(dm_plane_state->status));
}

static void _dm_plane_duplicate_blob(struct dm_plane_state *state,
				     struct dm_plane_state *ori_state)
{
	state->watermark = ori_state->watermark;
	state->color_mgmt = ori_state->color_mgmt;
	state->roi = ori_state->roi;

	if (state->watermark)
		drm_property_blob_get(state->watermark);
	if (state->color_mgmt)
		drm_property_blob_get(state->color_mgmt);
	if (state->roi)
		drm_property_blob_get(state->roi);
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 14, 0)) && (!defined(XDXGPU_CENTOS79_SUPPORT))
void drm_property_blob_put(struct drm_property_blob *blob)
{
	if (!blob)
		return;

	drm_mode_object_unreference(&blob->base);
}

struct drm_property_blob *drm_property_blob_get(struct drm_property_blob *blob)
{
	drm_mode_object_reference(&blob->base);
	return blob;
}

bool drm_property_replace_blob(struct drm_property_blob **blob,
			       struct drm_property_blob *new_blob)
{
	struct drm_property_blob *old_blob = *blob;

	if (old_blob == new_blob)
		return false;

	drm_property_blob_put(old_blob);
	if (new_blob)
		drm_property_blob_get(new_blob);
	*blob = new_blob;
	return true;
}
#endif

static int _dm_plane_set_property_blob_from_id(struct drm_device *dev,
					       struct drm_property_blob **blob,
					       uint64_t blob_id,
					       size_t expected_size)
{
	struct drm_property_blob *new_blob = NULL;

	if (blob_id) {
		new_blob = drm_property_lookup_blob(dev, blob_id);
		if (new_blob == NULL)
			return -EINVAL;

		if (new_blob->length != expected_size) {
			drm_property_blob_put(new_blob);
			return -EINVAL;
		}
	}

	drm_property_replace_blob(blob, new_blob);
	drm_property_blob_put(new_blob);

	return 0;
}

static struct drm_plane_state *
dm_plane_atomic_duplicate_state(struct drm_plane *plane)
{
	struct dm_plane_state *ori_state;
	struct dm_plane_state *state;

	if (WARN_ON(!plane->state))
		return NULL;

	ori_state = to_dm_plane_state(plane->state);
	state = kzalloc(sizeof(*state), GFP_KERNEL);
	if (!state)
		return NULL;

	__drm_atomic_helper_plane_duplicate_state(plane, &state->base);

	state->degamma = ori_state->degamma;
	state->degamma_changed = ori_state->degamma_changed;

	_dm_plane_duplicate_blob(state, ori_state);
	memcpy(&state->status, &ori_state->status, sizeof(ori_state->status));

	return &state->base;
}

static void dm_plane_atomic_destroy_state(struct drm_plane *plane,
					  struct drm_plane_state *state)
{
	struct dm_plane_state *dm_plane_state = to_dm_plane_state(state);

	__drm_atomic_helper_plane_destroy_state(state);

	drm_property_blob_put(dm_plane_state->watermark);
	drm_property_blob_put(dm_plane_state->color_mgmt);
	drm_property_blob_put(dm_plane_state->roi);
	kfree(dm_plane_state);
}

static int dm_plane_atomic_set_property(struct drm_plane *plane,
					struct drm_plane_state *state,
					struct drm_property *property,
					uint64_t val)
{
	struct drm_device *ddev = plane->dev;
	struct dm_plane *dm_plane = to_dm_plane(plane);
	struct dm_plane_state *dm_plane_state = to_dm_plane_state(state);
	int ret = 0;

	if (property == dm_plane->degamma_mode) {
		if (dm_plane_state->degamma != val) {
			dm_plane_state->degamma = val;
			dm_plane_state->degamma_changed = true;
		} else {
			dm_plane_state->degamma_changed = false;
		}
	} else if (property == dm_plane->watermark_prop) {
		ret = _dm_plane_set_property_blob_from_id(
			ddev, &dm_plane_state->watermark, val,
			sizeof(struct drm_xdx_watermark));
		return ret;
	} else if (property == dm_plane->color_mgmt_prop) {
		ret = _dm_plane_set_property_blob_from_id(
			ddev, &dm_plane_state->color_mgmt, val,
			sizeof(struct drm_xdx_color_mgmt));
		return ret;
	} else if (property == dm_plane->roi_prop) {
		ret = _dm_plane_set_property_blob_from_id(
			ddev, &dm_plane_state->roi, val,
			sizeof(struct drm_xdx_roi));
		return ret;
	} else {
		return -EINVAL;
	}

	return 0;
}

static int dm_plane_atomic_get_property(struct drm_plane *plane,
					const struct drm_plane_state *state,
					struct drm_property *property,
					uint64_t *val)
{
	struct dm_plane *dm_plane = to_dm_plane(plane);
	const struct dm_plane_state *dm_plane_state =
		container_of(state, const struct dm_plane_state, base);

	if (property == dm_plane->degamma_mode)
		*val = dm_plane_state->degamma;
	else if (property == dm_plane->watermark_prop)
		*val = (dm_plane_state->watermark) ?
				     dm_plane_state->watermark->base.id :
				     0;
	else if (property == dm_plane->color_mgmt_prop)
		*val = (dm_plane_state->color_mgmt) ?
				     dm_plane_state->color_mgmt->base.id :
				     0;
	else if (property == dm_plane->roi_prop)
		*val = (dm_plane_state->roi) ? dm_plane_state->roi->base.id : 0;
	else
		return -EINVAL;

	return 0;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0))
static bool dm_plane_format_mod_supported(struct drm_plane *plane,
					  uint32_t format,
					  uint64_t modifier)
{
	//Todo:add other supported modifier later
	return modifier == DRM_FORMAT_MOD_LINEAR;
}
#endif

const struct drm_plane_funcs xdx_dm_plane_funcs = {
	.update_plane = drm_atomic_helper_update_plane,
	.disable_plane = drm_atomic_helper_disable_plane,
	.destroy = dm_plane_destory,
	.reset = dm_plane_reset,
	.atomic_duplicate_state = dm_plane_atomic_duplicate_state,
	.atomic_destroy_state = dm_plane_atomic_destroy_state,
	.atomic_set_property = dm_plane_atomic_set_property,
	.atomic_get_property = dm_plane_atomic_get_property,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0))
	.format_mod_supported = dm_plane_format_mod_supported,
#endif
};

static u8 get_plane_number(struct drm_framebuffer *fb)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)) || (defined(XDXGPU_CENTOS79_SUPPORT))
	const struct drm_format_info *info;

	if (!fb)
		return 0;

	info = drm_format_info(fb->format->format);
	if (!info || info->num_planes > MAX_NUM_PLANES)
		return 0;

	return info->num_planes;
#else
	return drm_format_num_planes(fb->pixel_format);
#endif
}
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 13, 0))
static int xdx_dm_plane_atomic_check(struct drm_plane *plane,
				 struct drm_plane_state *state)
{
	struct dm_plane *dm_plane = to_dm_plane(plane);
	struct drm_framebuffer *fb = state->fb;
	struct drm_crtc *crtc = state->crtc;

	if (!crtc || !fb)
		return 0;

	return dm_plane->funcs->check(dm_plane, state);
}
#else
static int xdx_dm_plane_atomic_check(struct drm_plane *plane,
				 struct drm_atomic_state *state)
{
	struct drm_plane_state *new_plane_state = drm_atomic_get_new_plane_state(state,plane);
	struct dm_plane *dm_plane = to_dm_plane(plane);
	struct drm_framebuffer *fb = new_plane_state->fb;
	struct drm_crtc *crtc = new_plane_state->crtc;

	if (!crtc || !fb)
		return 0;

	return dm_plane->funcs->check(dm_plane, new_plane_state);
}
#endif
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 13, 0))
static void xdx_dm_plane_atomic_update(struct drm_plane *plane,
				  struct drm_plane_state *old_state)
#else
static void xdx_dm_plane_atomic_update(struct drm_plane *plane,
				   struct drm_atomic_state *old_state)
#endif
{
	u8 i, num_planes;
	struct drm_device *ddev = plane->dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);
	struct dm_plane *dm_plane = to_dm_plane(plane);
	struct drm_framebuffer *fb;
	struct dm_dc* dc = dm_plane->dc;
	struct drm_plane_state *state = plane->state;
	struct dm_plane_state *plane_state = to_dm_plane_state(state);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0) && LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)
	struct drm_format_name_buf *name = &plane_state->status.format_name;
#endif
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 13, 0)) && (!defined(XDXGPU_CENTOS79_SUPPORT))
	struct drm_gem_object * obj;
#endif

	if (!state->fb || !state->crtc)
		return;

	fb = state->fb;

	num_planes = get_plane_number(fb);
	for (i = 0; i < num_planes; i++) {
		struct xdx_bo *bo;
		struct xdx_bo_va *va;

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 13, 0)) && (!defined(XDXGPU_CENTOS79_SUPPORT))
		obj = xdx_fb_get_gem_obj(fb, i);
		bo = to_xdx_gem(obj);
#else
		bo = to_xdx_gem(fb->obj[i]);
#endif

		va = xdx_vm_bo_find(&xdev->vmm, &xdev->vmm.dc, bo);
		if (!va) {
			va = xdx_vm_bo_add(&xdev->vmm, &xdev->vmm.dc, bo);
			if (dc->funcs->mmu_flush)
				dc->funcs->mmu_flush(dc);
		}

		WARN_ON(!va);

		dm_plane->dma_addr[i] = va->addr.low + fb->offsets[i];
	}


	plane_state->status.src = drm_plane_state_src(state);
	plane_state->status.dest = drm_plane_state_dest(state);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0) && LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)
	drm_get_format_name(fb->format->format, name);
#endif
	dm_plane->funcs->update(dm_plane);
}
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 13, 0))
static void xdxb_dm_plane_atomic_disable(struct drm_plane *plane,
				    struct drm_plane_state *old_state)
{
	struct dm_plane *dm_plane = to_dm_plane(plane);

	dm_plane->funcs->disable(dm_plane, old_state);
}
static void xdxb_dm_plane_atomic_update(struct drm_plane *plane,
				  struct drm_plane_state *old_state)

{
	u8 i, num_planes;
	struct drm_device *ddev = plane->dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);
	struct dm_plane *dm_plane = to_dm_plane(plane);
	struct drm_framebuffer *fb;
	struct dm_dc* dc = dm_plane->dc;
	struct drm_plane_state *state = plane->state;
	struct dm_plane_state *plane_state = to_dm_plane_state(state);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0)
	struct drm_format_name_buf *name = &plane_state->status.format_name;
#endif
	struct xdx_bo *bo;
	struct xdx_bo_va *va;
#if defined DUMP_DC_BUFFER
	struct file* fp;
	char name_scrn[64];
#endif
	if (!state->fb || !state->crtc)
		return;

	fb = state->fb;

	num_planes = get_plane_number(fb);
	for (i = 0; i < num_planes; i++)
	{
		bo = to_xdx_gem(fb->obj[i]);

		va = xdx_vm_bo_find(&xdev->vmm, &xdev->vmm.dc, bo);
		if (!va) {
			va = xdx_vm_bo_add(&xdev->vmm, &xdev->vmm.dc, bo);
			if (dc->funcs->mmu_flush)
				dc->funcs->mmu_flush(dc);
		}
		WARN_ON(!va);
		dm_plane->dma_addr[i] = va->addr.low + fb->offsets[i];

	}

#if defined DUMP_DC_BUFFER
	if (plane->type == DRM_PLANE_TYPE_PRIMARY)
	{
		xdx_bo_vmap(bo);
		if (bo->vaddr)
		{
			sprintf(name_scrn, "/home/0x%llx.raw", dm_plane->dma_addr[0]);
			dev_info(xdev->dev, "Primary Layer:%d's screen buffer dump into:%s", dm_plane->id == 0 ? 0 :1, name_scrn);
			fp = filp_open(name_scrn, O_CREAT|O_RDWR, 0664);
			if (fp)
			{
				kernel_write(fp, bo->vaddr, fb->width*fb->height*4, 0);
				filp_close(fp, NULL);
			}
		}
	}
#endif
	plane_state->status.src = drm_plane_state_src(state);
	plane_state->status.dest = drm_plane_state_dest(state);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0) && LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)
        drm_get_format_name(fb->format->format, name);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0)
	drm_get_format_name(fb->pixel_format);
#endif
	dm_plane->funcs->update(dm_plane);
}
#endif
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 13, 0))
static void xdx_dm_plane_atomic_disable(struct drm_plane *plane,
				    struct drm_plane_state *old_state)
#else
static void xdx_dm_plane_atomic_disable(struct drm_plane *plane,
				    struct drm_atomic_state *old_state)
#endif
{
	struct drm_plane_state *plane_state;
	struct dm_plane *dm_plane = to_dm_plane(plane);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 13, 0))
	plane_state = old_state;
#else
	plane_state = drm_atomic_get_old_plane_state(old_state, plane);
#endif
	dm_plane->funcs->disable(dm_plane, plane_state);
}

#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 13, 0))
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 13, 0))
static int xdx_dm_plane_atomic_async_check (struct drm_plane *plane,
			  struct drm_plane_state *state)
#else
static int xdx_dm_plane_atomic_async_check (struct drm_plane *plane,
			  struct drm_atomic_state *state)
#endif
{

	if (plane->type != DRM_PLANE_TYPE_CURSOR)
		return -EINVAL;

	return 0;
}
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 13, 0))
static void xdx_dm_plane_atomic_async_update(struct drm_plane *plane,
			  struct drm_plane_state *new_state)
{
	struct drm_plane_state *state = plane->state;
	struct dm_crtc *dm_crtc = to_dm_crtc(state->crtc);
	struct dm_plane *dm_plane = to_dm_plane(plane);
	struct dm_dc* dc = dm_plane->dc;
	struct dc8200_dc* dc8200 = container_of(dc, struct dc8200_dc, base);
	struct drm_plane_state *old_state = NULL;

	mutex_lock(&dc8200->cursor_mtx);
	old_state = drm_atomic_get_old_plane_state(new_state->state, plane);

	new_state->fb->hot_x = 0;
	new_state->fb->hot_y = 0;
	swap(plane->state->fb, new_state->fb);

	plane->state->src_x = new_state->src_x;
	plane->state->src_y = new_state->src_y;
	plane->state->src_w = new_state->src_w;
	plane->state->src_h = new_state->src_h;
	plane->state->crtc_x = new_state->crtc_x;
	plane->state->crtc_y = new_state->crtc_y;
	plane->state->crtc_w = new_state->crtc_w;
	plane->state->crtc_h = new_state->crtc_h;
	if (!plane->state->fb || !plane->state->crtc)
		xdx_dm_plane_atomic_disable(plane, old_state);
	else
		xdx_dm_plane_atomic_update(plane, old_state);

	if(dm_crtc)
		dm_crtc->funcs->cursor_commit(dm_crtc);
	mutex_unlock(&dc8200->cursor_mtx);


}
#else
static void xdx_dm_plane_atomic_async_update(struct drm_plane *plane,
			  struct drm_atomic_state *state_original)
{
	struct drm_plane_state *new_state = drm_atomic_get_new_plane_state(state_original,
										 plane);
	struct drm_plane_state *state = plane->state;
	struct dm_crtc *dm_crtc = to_dm_crtc(state->crtc);
	struct dm_plane *dm_plane = to_dm_plane(plane);
	struct dm_dc* dc = dm_plane->dc;
	struct dc8200_dc* dc8200 = container_of(dc, struct dc8200_dc, base);
	struct drm_plane_state *old_state = NULL;

	mutex_lock(&dc8200->cursor_mtx);
	old_state = drm_atomic_get_old_plane_state(state_original, plane);

#if (defined(OS_NAME_OPENKYLIN) || defined(OS_NAME_KYLIN) || defined(OS_NAME_UBUNTU)) && LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
	new_state->hotspot_x = 0;
	new_state->hotspot_y = 0;
#else
	new_state->fb->hot_x = 0;
	new_state->fb->hot_y = 0;
#endif
	swap(plane->state->fb, new_state->fb);

	plane->state->src_x = new_state->src_x;
	plane->state->src_y = new_state->src_y;
	plane->state->src_w = new_state->src_w;
	plane->state->src_h = new_state->src_h;
	plane->state->crtc_x = new_state->crtc_x;
	plane->state->crtc_y = new_state->crtc_y;
	plane->state->crtc_w = new_state->crtc_w;
	plane->state->crtc_h = new_state->crtc_h;
	if (!plane->state->fb || !plane->state->crtc)
		xdxb_dm_plane_atomic_disable(plane, old_state);
	else
		xdxb_dm_plane_atomic_update(plane, old_state);

	if(dm_crtc)
		dm_crtc->funcs->cursor_commit(dm_crtc);
	mutex_unlock(&dc8200->cursor_mtx);
}
#endif
#endif

#if ((LINUX_VERSION_CODE < KERNEL_VERSION(5, 4, 0)) && (!defined(REDHAT_83)) && (!defined(XDXGPU_CENTOS82_SUPPORT)))

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0))
void xdx_drm_atomic_set_fence_for_plane(struct drm_plane_state *plane_state,
                               struct fence *fence)
{
        if (plane_state->fence) {
                fence_put(fence);
                return;
        }

        plane_state->fence = fence;
}
#endif

static int xdx_resv_prepare_fb(struct drm_plane *plane,
			  struct drm_plane_state *state)
{
	// struct dma_buf *dma_buf;
	// struct drm_device *ddev = plane->dev;
	// struct xdx_device *xdev = drm_to_xdev(ddev);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0))
	struct fence *fence;
#else
	struct dma_fence *fence;
#endif
	struct xdx_bo *bo;
	struct drm_gem_object *obj;

	if (!state->fb)
		return 0;
//drm_plane_state->fb->obj->bo->resv->fence
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 13, 0))
	obj = xdx_fb_get_gem_obj(state->fb, 0);
#else
	obj = drm_gem_fb_get_obj(state->fb, 0);
#endif
//	printk("drm_gem_fb_get_obj addr is %p\n",obj);
	bo = to_xdx_gem(obj);
//	printk("xdx_bo addr is %p\n",bo);
//	printk("bo->resv addr h is %p\n",bo->resv);
	if (bo) {
#if (defined(XDXGPU_CENTOS85_SUPPORT))
		fence = dma_resv_get_excl_rcu(bo->resv);
#else
		fence = reservation_object_get_excl_rcu(bo->resv);
#endif
		//printk("Test xdx_plane_prepare_fb fence=%p state->fence=%p resv:%p\n", fence, state->fence, bo->resv);
		if(fence)
		{
  //    printk("Test fence %llu#%u#%u\n", fence->context, fence->seqno, fence->seqno);
    //  printk("test fence seqno:%u", fence->seqno);
		}
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0))
		xdx_drm_atomic_set_fence_for_plane(state, fence);
#else
		drm_atomic_set_fence_for_plane(state, fence);
#endif
	}

	return 0;
}
#endif
const struct drm_plane_helper_funcs xdx_dm_plane_helper_funcs = {
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0)) || (defined(REDHAT_83)) || (defined(XDXGPU_CENTOS82_SUPPORT)))
	.prepare_fb = drm_gem_plane_helper_prepare_fb,
#else
	.prepare_fb = xdx_resv_prepare_fb,
#endif
	.atomic_check = xdx_dm_plane_atomic_check,
	.atomic_update = xdx_dm_plane_atomic_update,
//#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 13, 0))
	.atomic_disable = xdx_dm_plane_atomic_disable,
//#endif
#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 13, 0))
	.atomic_async_check = xdx_dm_plane_atomic_async_check,
	.atomic_async_update = xdx_dm_plane_atomic_async_update,
#endif
};

int dm_universal_plane_init(struct xdx_dm *dm, struct dm_plane *dm_plane,
			    struct dm_dc *dc, uint32_t possible_crtcs,
			    struct dm_plane_funcs *funcs,
			    const uint32_t *formats, unsigned int format_count,
			    const uint64_t *format_modifiers, enum drm_plane_type type,
			    uint32_t id, u8 zpos, const char* name)
{
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct drm_device *ddev = xdev_to_drm(xdev);
	int ret;

	dm_plane->funcs = funcs;
	dm_plane->id = id;
	dm_plane->zpos = zpos;
	dm_plane->dm = dm;
	dm_plane->dc = dc;
	ret = drm_universal_plane_init(ddev, &dm_plane->base, possible_crtcs,
				       &xdx_dm_plane_funcs, formats,
				       format_count,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)) || (defined(XDXGPU_CENTOS79_SUPPORT))
					format_modifiers,
#endif
					 type, "%s", name);
	if (ret)
		goto err_free_plane;

	drm_plane_helper_add(&dm_plane->base, &xdx_dm_plane_helper_funcs);

	return ret;

err_free_plane:
	dm_plane->funcs->destroy(dm_plane);
	return ret;
}
