/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/component.h>
#include <linux/module.h>
#include <linux/delay.h>

#include <drm/drm_crtc.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_of.h>

#include <drm/drm_atomic_helper.h>
#include <drm/drm_atomic_helper.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0))
#include <drm/display/drm_scdc_helper.h>
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))
#include <drm/drm_scdc_helper.h>
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0)) || (defined(XDXGPU_CENTOS85_SUPPORT))
#include <drm/drm_probe_helper.h>
#endif

#include "xdx_hdmi_drv.h"
#include "xdx_hdmi_reg.h"
#include "doorbell.h"
#include "../../audio/xdxaudio_hal.h"
#include "../pix.h"
#include "drm/xdxgpu_drm.h"
#include "kernel_compatibility.h"

#define DDC_CI_ADDR     0x37
#define DDC_SEGMENT_ADDR    0x30
#define HDMI_EDID_LEN       512
#define SCDC_MIN_SOURCE_VERSION 0x1
#define HDMI14_MAX_TMDSCLK  340000
#define HDMI20_MAX_TMDSCLK_KHZ  600000
static void hdmi_core_regs_dump(struct hdmi_dev *hdmi);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))
static bool xdx_hdmi_support_scdc(struct hdmi_dev *hdmi);
#endif
static void xdx_hdmi_poweroff(struct hdmi_dev *hdmi);
static void xdx_hdmi_poweron(struct hdmi_dev *hdmi);

static const struct drm_display_mode hdmi_default_modes[] = {
    /* 16 - 1920x1080@60Hz 16:9 */
    { DRM_MODE("1920x1080", DRM_MODE_TYPE_DRIVER, 148500, 1920, 2008,
           2052, 2200, 0, 1080, 1084, 1089, 1125, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
    /* 2 - 720x480@60Hz 4:3 */
    { DRM_MODE("720x480", DRM_MODE_TYPE_DRIVER, 27000, 720, 736,
           798, 858, 0, 480, 489, 495, 525, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_NVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_4_3, },
    /* 4 - 1280x720@60Hz 16:9 */
    { DRM_MODE("1280x720", DRM_MODE_TYPE_DRIVER, 74250, 1280, 1390,
           1430, 1650, 0, 720, 725, 730, 750, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
    /* 31 - 1920x1080@50Hz 16:9 */
    { DRM_MODE("1920x1080", DRM_MODE_TYPE_DRIVER, 148500, 1920, 2448,
           2492, 2640, 0, 1080, 1084, 1089, 1125, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
    /* 19 - 1280x720@50Hz 16:9 */
    { DRM_MODE("1280x720", DRM_MODE_TYPE_DRIVER, 74250, 1280, 1720,
           1760, 1980, 0, 720, 725, 730, 750, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
    /* 17 - 720x576@50Hz 4:3 */
    { DRM_MODE("720x576", DRM_MODE_TYPE_DRIVER, 27000, 720, 732,
           796, 864, 0, 576, 581, 586, 625, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_NVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_4_3, },
    /* 2 - 720x480@60Hz 4:3 */
    { DRM_MODE("720x480", DRM_MODE_TYPE_DRIVER, 27000, 720, 736,
           798, 858, 0, 480, 489, 495, 525, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_NVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_4_3, },
};

static inline void xdx_hdmi_h2a_doorbell_w(struct hdmi_dev *hdmi, u64 value)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    xdx_mmio_writel(xdev, SYS_PANGU_REG_HDMIDOORBELL + 8, value & 0xffffffff);

    xdx_mmio_writel(xdev, SYS_PANGU_REG_HDMIDOORBELL + 0xc,
            (value >> 32) & 0xffffffff);
}

static inline u32 xdx_hdmi_h2a_doorbell_r(struct hdmi_dev *hdmi)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    return xdx_mmio_readl(xdev, SYS_PANGU_REG_HDMIDOORBELL + 8);
}

static inline u64 xdx_hdmi_h2a_doorbell_r64(struct hdmi_dev *hdmi)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    u64 lret, hret;

    hret = xdx_mmio_readl(xdev, SYS_PANGU_REG_HDMIDOORBELL + 0xc);
    lret = xdx_mmio_readl(xdev, SYS_PANGU_REG_HDMIDOORBELL + 8);
    return ((0xffffffff & lret) | (0xffffffff & hret) << 32);
}

static inline void xdx_hdmi_h2c_doorbell_w(struct hdmi_dev *hdmi, u64 value)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    xdx_mmio_writel(xdev, SYS_PANGU_REG_HDMIDOORBELL, value & 0xffffffff);

    xdx_mmio_writel(xdev, SYS_PANGU_REG_HDMIDOORBELL + 4,
            (value >> 32) & 0xffffffff);
}

static inline u32 xdx_hdmi_h2c_doorbell_r(struct hdmi_dev *hdmi)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    return xdx_mmio_readl(xdev, SYS_PANGU_REG_HDMIDOORBELL);
}

static inline u32 xdx_hdmi_tile_read(struct hdmi_dev *hdmi, u32 offset)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    return xdx_mmio_readl(xdev,
               xdev->hw_info.hdmi_tile_regs[hdmi->id].base + offset);
}

static inline void xdx_hdmi_tile_write(struct hdmi_dev *hdmi,
                        u32 offset, u32 value)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    xdx_mmio_writel(xdev,
        xdev->hw_info.hdmi_tile_regs[hdmi->id].base + offset, value);
}

static inline void xdx_hdmi_tile_write_mask(struct hdmi_dev *hdmi,
                        u32 offset, u32 mask, u32 value)
{
    u32 tmp = xdx_hdmi_tile_read(hdmi, offset) & ~mask;

    tmp |= (value & mask);
    xdx_hdmi_tile_write(hdmi, offset, tmp);
}

static inline u32 xdx_hdmi_core_read(struct hdmi_dev *hdmi, u32 offset)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    return xdx_mmio_readl(xdev,
               xdev->hw_info.hdmi_core_regs[hdmi->id].base + offset);
}

static inline void xdx_hdmi_core_write(struct hdmi_dev *hdmi,
                        u32 offset, u32 value)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    xdx_mmio_writel(xdev,
        xdev->hw_info.hdmi_core_regs[hdmi->id].base + offset, value);
}

static inline void xdx_hdmi_core_write_mask(struct hdmi_dev *hdmi,
                       u32 offset, u32 mask, u32 value)
{
    u32 tmp = xdx_hdmi_core_read(hdmi, offset) & ~mask;

    tmp |= (value & mask);
    xdx_hdmi_core_write(hdmi, offset, tmp);
}

static inline u16 xdx_hdmi_pcs_read(struct hdmi_dev *hdmi, u32 offset)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    return xdx_mmio_readw(xdev,
               xdev->hw_info.hdmi_pcs_regs[hdmi->id].base + (offset << 2));
}

static inline void xdx_hdmi_pcs_write(struct hdmi_dev *hdmi,
                        u32 offset, u16 value)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    xdx_mmio_writew(xdev,
        xdev->hw_info.hdmi_pcs_regs[hdmi->id].base + (offset << 2), value);
}

static int hdmi_phy_reset(struct hdmi_dev *hdmi)
{
    int i;
    u32 val;
    HDMIDBCmd_T_NOSINK PhyConfig;
    int timeout = 1000;

    PhyConfig.cmd = cDPCmdHDMIPhyReset;
    xdx_hdmi_h2c_doorbell_w(hdmi,(u64)PhyConfig.all);

    for (i = timeout; i; i--) {
        val = xdx_hdmi_h2c_doorbell_r(hdmi);
        if (val & BIT(0))
            break;

        usleep_range(1000, 2000);
    }

    if (!(val & BIT(0))) {
        dev_err(hdmi->dev,"xdx hdmi tx phy powerup fail\n");
        return -ETIMEDOUT;
    }

    return 0;
}

static int hdmi_phy_reset_release(struct hdmi_dev *hdmi)
{
    int i;
    u32 val;
    HDMIDBCmd_T_NOSINK PhyConfig;
    int timeout = 1000;

    PhyConfig.cmd = cDPCmdHDMIPhyResetRelease;
    xdx_hdmi_h2c_doorbell_w(hdmi,(u64)PhyConfig.all);

    for (i = timeout; i; i--) {
    val = xdx_hdmi_h2c_doorbell_r(hdmi);
    if (val & BIT(0))
        break;

    usleep_range(1000, 2000);
    }

    if (!(val & BIT(0))) {
        dev_err(hdmi->dev,"xdx hdmi tx phy powerup fail\n");
        return -ETIMEDOUT;
    }

    for (i = timeout; i; i--) {
        val = xdx_hdmi_tile_read(hdmi, HDMIDP_PHY_IF_RAW);
        if (HDMI_SRAM_INIT_STATUS == (val & HDMI_SRAM_INIT_STATUS))
            break;

        usleep_range(10, 20);
    }
    if (HDMI_SRAM_INIT_STATUS != (val & HDMI_SRAM_INIT_STATUS)) {
        dev_err(hdmi->dev, "sram init done state timeout\n");
        return -ETIMEDOUT;
    }

    return 0;
}

static int hdmi_tx_phy_init(struct hdmi_dev *hdmi)
{
    u32 val;
    int i, ret;
    HDMIDBCmd_T_NOSINK PhyConfig;
    int timeout = 1000;
    u32 PixClock = hdmi->hdmi_mode.clock;
    u32 ColorDeep = hdmi->pVideo.bpc;
    u32 Format = hdmi->display_info.color_formats;
    int  eq_vboost;
    u16 dig_lvl_ovrd_in;

    hdmi_phy_reset(hdmi);
    //ENABLE PLL
    xdx_hdmi_tile_write_mask(hdmi, PHYREF_PLL_EN,
                    PLL0_PHYREF_EN, PLL0_PHYREF_EN);
    for (i = timeout; i; i--) {
        val = xdx_hdmi_tile_read(hdmi, PHYREF_PLL_LOCK);
        if(val & PLL0_PHYREF_CORE_LOCK)
            break;

        usleep_range(1000, 2000);
    }

    if (!(val & PLL0_PHYREF_CORE_LOCK)) {
        dev_err(hdmi->dev,"xdx hdmi wait PHYREF_PLL_LOCK timeout\n");
        return -ETIMEDOUT;
    }

    xdx_hdmi_tile_write_mask(hdmi, HDMI_PHY_CTRL, SRAM_BYPASS, SRAM_BYPASS);

    ret = hdmi_phy_reset_release(hdmi);
    if (ret){
        dev_err(hdmi->dev,"hdmi phy reset release fail\n");
        return -1;
    }

    /* Preferred encoding type is RGB444 */
    if (Format & DRM_COLOR_FORMAT_RGB444)
        Format = RGB;
    else if (Format & DRM_COLOR_FORMAT_YCRCB444)
        Format = YCBCR444;
    else if (Format & DRM_COLOR_FORMAT_YCRCB422)
        Format = YCBCR422;
    else if (Format & DRM_COLOR_FORMAT_YCRCB420)
        Format = YCBCR420;
    else
        Format = RGB;

    if (ColorDeep == 8) {
        switch (Format){
            case RGB:
                xdx_hdmi_tile_write(hdmi, HDMI_CTRL, IPI_8BPC | IPI_RGB);
                break;
            case YCBCR422:
                xdx_hdmi_tile_write(hdmi, HDMI_CTRL, IPI_8BPC | IPI_YCBCR422);
                break;
            case YCBCR444:
                xdx_hdmi_tile_write(hdmi, HDMI_CTRL, IPI_8BPC | IPI_YCBCR444);
                break;
            case YCBCR420:
                xdx_hdmi_tile_write(hdmi, HDMI_CTRL,IPI_8BPC | IPI_YCBCR420);
                break;
            default:
                xdx_hdmi_tile_write(hdmi, HDMI_CTRL, IPI_8BPC | IPI_RGB);
                break;
        }
    }

    if (ColorDeep == 10) {
        switch (Format) {
            case RGB:
                xdx_hdmi_tile_write(hdmi, HDMI_CTRL, IPI_10BPC | IPI_RGB);
                break;
            case YCBCR422:
                xdx_hdmi_tile_write(hdmi, HDMI_CTRL, IPI_10BPC | IPI_YCBCR422);
                break;
            case YCBCR444:
                xdx_hdmi_tile_write(hdmi, HDMI_CTRL, IPI_10BPC | IPI_YCBCR444);
                break;
            case YCBCR420:
                xdx_hdmi_tile_write(hdmi, HDMI_CTRL, IPI_10BPC | IPI_YCBCR420);
                break;
             default:
                xdx_hdmi_tile_write(hdmi, HDMI_CTRL, IPI_10BPC | IPI_RGB);
                break;
        }
    }

    if (Format == YCBCR422) {
        PhyConfig.colorDepth = 8;
    } else {
        PhyConfig.colorDepth = ColorDeep;
    }

    PhyConfig.cmd = cHDMIDBCmdPhyInit;
    PhyConfig.colorFormat = Format;
    PhyConfig.screenMode = PixClock;

    xdx_hdmi_h2a_doorbell_w(hdmi, (u64)PhyConfig.all);

    for (i = timeout; i; i--) {
        val = xdx_hdmi_h2a_doorbell_r(hdmi);
        if (val & BIT(0))
            break;

        usleep_range(1000, 2000);
    }
    if (!(val & BIT(0))) {
        dev_err(hdmi->dev, "HDMI phyinit timeout\n");
        return -ETIMEDOUT;
    }
    //p3 status
    xdx_hdmi_tile_write_mask(hdmi, HDMI_PCS_CTRL,
                    LANE_POWERDOWN, LANE_POWERDOWN);

    for (i = timeout; i; i--) {
        val = xdx_hdmi_tile_read(hdmi, HDMIDP_PHY_IF_RAW);
        if (HDMI_ALL_LANES_PHY_STATUS == (val & HDMI_ALL_LANES_PHY_STATUS))
            break;

        usleep_range(10, 20);
    }
    if (HDMI_ALL_LANES_PHY_STATUS != (val & HDMI_ALL_LANES_PHY_STATUS)) {
        dev_err(hdmi->dev, "hdmi p0 to p3 wait PHY_IF_RAW timeout\n");
        return -ETIMEDOUT;
    }

    xdx_hdmi_tile_write(hdmi, HDMIDP_PHY_IF_CLR, HDMI_ALL_LANES_PHY_INT_CLEAR);

    for (i = timeout; i; i--) {
        val = xdx_hdmi_tile_read(hdmi, HDMIDP_PHY_IF_RAW);
        if (!(val & HDMI_ALL_LANES_PHY_STATUS))
            break;

        xdx_hdmi_tile_write(hdmi, HDMIDP_PHY_IF_CLR,
                        HDMI_ALL_LANES_PHY_INT_CLEAR);

        usleep_range(10, 20);
    }
    if (val & HDMI_ALL_LANES_PHY_STATUS) {
        dev_err(hdmi->dev, "p0->p3 clear irq state timeout\n");
        return -ETIMEDOUT;
    }
    //ALL__HDMI_PCS_CTRL p3->p0
    xdx_hdmi_tile_write_mask(hdmi, HDMI_PCS_CTRL, LANE_POWERDOWN, 0x0);
    for (i = timeout; i; i--) {
        val = xdx_hdmi_tile_read(hdmi, HDMIDP_PHY_IF_RAW);
        if (HDMI_ALL_LANES_PHY_STATUS == (val & HDMI_ALL_LANES_PHY_STATUS))
            break;

        usleep_range(10, 20);
    }
    if (HDMI_ALL_LANES_PHY_STATUS != (val & HDMI_ALL_LANES_PHY_STATUS)) {
        dev_err(hdmi->dev, "hdmip3->p0 wait PHY_IF_RAW timeout\n");
        return -ETIMEDOUT;
    }

    xdx_hdmi_tile_write(hdmi, HDMIDP_PHY_IF_CLR, HDMI_ALL_LANES_PHY_INT_CLEAR);

    for (i = timeout; i; i--) {
        val = xdx_hdmi_tile_read(hdmi, HDMIDP_PHY_IF_RAW);
        if (!(val & HDMI_ALL_LANES_PHY_STATUS))
            break;

        xdx_hdmi_tile_write(hdmi, HDMIDP_PHY_IF_CLR,
                        HDMI_ALL_LANES_PHY_INT_CLEAR);

        usleep_range(10, 20);
    }
    if (val & HDMI_ALL_LANES_PHY_STATUS) {
        dev_err(hdmi->dev, "p3->p0 clear irq state timeout\n");
        return -ETIMEDOUT;
    }

    for (i = timeout; i; i--) {
        val = xdx_hdmi_tile_read(hdmi, HDMIDP_PHY_IF_RAW);
        if (val & HDMI_PHY_READY_POS_STATUS)
            break;

        usleep_range(10, 20);
    }
    if (!(val & HDMI_PHY_READY_POS_STATUS)) {
        dev_err(hdmi->dev, "phy init done wait PHY_IF_RAW timeout\n");
        return -ETIMEDOUT;
    }

    xdx_hdmi_tile_write(hdmi, HDMIDP_PHY_IF_CLR, HDMI_PHY_READY_POS_INT_CLEAR);
    //PHY ready
    for (i = timeout; i; i--) {
        val = xdx_hdmi_tile_read(hdmi, HDMIDP_PHY_IF_RAW);
        if (!(val & HDMI_PHY_READY_POS_STATUS))
            break;

        usleep_range(10, 20);
    }
    if (val & HDMI_PHY_READY_POS_STATUS) {
        dev_err(hdmi->dev, "phy init done wait PHY_IF_RAW timeout\n");
        return -ETIMEDOUT;
    }

    //config HDMI eq value
    if(hdmi->phy.eq_value)
        eq_vboost = hdmi->phy.eq_value;
    else
        eq_vboost = 7;

    if(eq_vboost){
        dig_lvl_ovrd_in = xdx_hdmi_pcs_read(hdmi, 0xf << 2);
        dig_lvl_ovrd_in = (dig_lvl_ovrd_in & (~(7 << 6))) | (eq_vboost << 6) | (1 << 9);
        xdx_hdmi_pcs_write(hdmi, 0xf, dig_lvl_ovrd_in);
    }

    return 0;
}

static void xdx_hdmi_trans_para_fill(struct hdmi_dev *hdmi, struct hdmi_trans_para *para)
{
    u32 PixClock = hdmi->hdmi_mode.clock;
    u32 ColorDeep = hdmi->pVideo.bpc;
    u32 Format = hdmi->display_info.color_formats;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 18, 0))
    if (Format & DRM_COLOR_FORMAT_RGB444)
        Format = RGB;
    else if (Format & DRM_COLOR_FORMAT_YCRCB444)
        Format = YCBCR444;
    else if (Format & DRM_COLOR_FORMAT_YCRCB422)
        Format = YCBCR422;
    else
        Format = YCBCR420;
#else
    if (Format & DRM_COLOR_FORMAT_RGB444)
        Format = RGB;
    else if (Format & DRM_COLOR_FORMAT_YCBCR444)
        Format = YCBCR444;
    else if (Format & DRM_COLOR_FORMAT_YCBCR422)
        Format = YCBCR422;
    else
        Format = YCBCR420;
#endif

    para->pixclock = PixClock;
    para->colorformat = Format;
    para->colordeep = ColorDeep;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0))
    para->picture_aspect_ratio = hdmi->connector.state->picture_aspect_ratio;
#else
	para->picture_aspect_ratio = hdmi->hdmi_mode.picture_aspect_ratio;
#endif
    para->vic_code = hdmi->vic;
    para->dc_idx = hdmi->id;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))
    para->scrambling_support = xdx_hdmi_support_scdc(hdmi);
#endif
    para->eq_vboost = hdmi->phy.eq_value;
}

static int hdmi_tx_phy_powerdown(struct hdmi_dev *hdmi)
{
    int i;
    u32 DBValue;
    HDMIDBCmd_T_NOSINK PhyConfig;
    int timeout = 1000;

    /* For HDMI TX 1.4 PHY, power down by placing PHY in reset */
    PhyConfig.cmd = cDPCmdHDMIPowerdown;
    xdx_hdmi_h2c_doorbell_w(hdmi,(u64)PhyConfig.all);
    for (i = timeout; i; i--) {
        DBValue = xdx_hdmi_h2c_doorbell_r(hdmi);
        if (DBValue & BIT(0))
            break;

        usleep_range(1000, 2000);
    }

    if (!(DBValue & BIT(0))) {
        dev_err(hdmi->dev,"xdx hdmi tx phy poweroff fail\n");
        return -ETIMEDOUT;
    }

    return 0;
}

static int hdmi_tx_phy_powerup(struct hdmi_dev *hdmi)
{
    int i;
    u32 val;
    HDMIDBCmd_T_NOSINK    PhyConfig;
    int timeout = 1000;

    PhyConfig.cmd = cDPCmdHDMIPoweron;
    xdx_hdmi_h2c_doorbell_w(hdmi,(u64)PhyConfig.all);
    for (i = timeout; i; i--) {
        val = xdx_hdmi_h2c_doorbell_r(hdmi);
        if (val & BIT(0))
            break;

        usleep_range(1000, 2000);
    }

    if (!(val & BIT(0))) {
        dev_err(hdmi->dev,"xdx hdmi tx phy powerup fail\n");
        return -ETIMEDOUT;
    }

    return 0;
}

static const struct hdmi_tx_phy_ops xdx_hdmi_phy_ops = {
    .init = hdmi_tx_phy_init,
    .enable = hdmi_tx_phy_powerup,
    .disable = hdmi_tx_phy_powerdown,
};

static void hdmi_config_err_analyze(struct hdmi_dev* hdmi, u32 err_code)
{

       struct xdx_output *output = hdmi_to_output(hdmi);
       struct xdx_dm *dm = output_to_dm(output);
       struct xdx_device *xdev = dm_to_xdev(dm);

       int i;
       const struct err_vector{
           u32 err_bit;
           const char *type;
       }err_vectors[] = {
           {BIT(0), "HDMI_PLL_LOCK_TIME_OUT"},
           {BIT(1), "HDMI_CANNOT_SYNTHESIZE_BITRATE"},
           {BIT(2), "HDMI_PHY_IF_TO_P3_TIME_OUT"},
           {BIT(3), "HDMI_PHY_IRQ_TO_P3_TIME_OUT"},
           {BIT(4), "HDMI_PHY_CTRL_TO_P3_TIME_OUT"},
           {BIT(5), "HDMI_PHY_IF_TO_P0_TIME_OUT"},
           {BIT(6), "HDMI_PHY_IRQ_TO_P0_TIME_OUT"},
           {BIT(7), "HDMI_PHY_IF_RAW_READY_TIME_OUT"},
           {BIT(8), "HDMI_PHY_IF_RAW_DNOE_TIME_OUT"},
           {BIT(9), "HDMI_AVP_CLOCKS_LOCK_FAILED"},
           {BIT(10), "HDMI_VIDEO_CONFIGURE_ERROR"},
           {BIT(11), "HDMI_PACKETS_CONFIGURE_ERROR"},
           {BIT(12), "HDMI_HW_SUPPORT_AVP_ERROR"},
           {BIT(13), "HDMI_INT0_APB_REGBANK_NOTREADY"},
       };

       dev_err(xdev->dev,"HDMI config err: err_code = 0x%0x \n", err_code);
       for (i = 0; i < ARRAY_SIZE(err_vectors); i++){
          if(err_vectors[i].err_bit & err_code)
                  dev_err(xdev->dev,"HDMI CONFIG: ERROR TYPE = %s,   ERROR BIT = BIT(%d) \n", err_vectors[i].type, i);
       }
}

static int hdmi_tx_phy_sink_init(struct hdmi_dev *hdmi)

{
    struct xdx_output * output = hdmi_to_output(hdmi);
    struct hdmi_trans_para para;
    int i;
    //u32 DBValue;
    HDMIDBCmd_T_SINK PhyConfig;
    int ret, timeout = 1000;

    /* For HDMI TX 1.4 PHY, power down by placing PHY in reset */
    if(output->psDoorbellShareMemDesc[hdmi->id] == NULL){
        ret = DPDoorbellShareMemAllocate(output,hdmi->id);
        if (ret) {
            dev_err(hdmi->dev,"hdmi share ram allocate fail\n");
            return ret;
        }
    }

    xdx_hdmi_trans_para_fill(hdmi, &para);
    memcpy_toio(output->psDoorbellVirAddr[hdmi->id], &para, sizeof(struct hdmi_trans_para));

    PhyConfig.all = 0;
    PhyConfig.cmd = cH02ADisplayDBCmdHDMI;
    PhyConfig.subCmd = cH02AHDMISubCmdPhyInit;
    PhyConfig.dword = output->psDoorbellPhyAddr[hdmi->id].uiAddr;
    xdx_hdmi_h2a_doorbell_w(hdmi,(u64)PhyConfig.all);
    for (i = timeout; i; i--) {
        PhyConfig.all = xdx_hdmi_h2a_doorbell_r64(hdmi);
        if (PhyConfig.low32 & BIT(0)){
	    if(PhyConfig.ret & BIT(0))
                hdmi_config_err_analyze(hdmi, PhyConfig.dword);
            break;
	}
        usleep_range(1000, 2000);
    }

    if (!(PhyConfig.low32 & BIT(0))) {
        dev_err(hdmi->dev,"xdx hdmi tx phy sink init fail\n");
        return -ETIMEDOUT;
    }

    //PhyConfig.all = DBValue;

    return PhyConfig.ret;
}

static int hdmi_tx_phy_sink_powerdown(struct hdmi_dev *hdmi)
{
    struct xdx_output * output = hdmi_to_output(hdmi);
    struct hdmi_trans_para para;
    int i, ret;
    u32 DBValue;
    HDMIDBCmd_T_SINK PhyConfig;
    int timeout = 1000;

    /* For HDMI TX 1.4 PHY, power down by placing PHY in reset */
    if(output->psDoorbellShareMemDesc[hdmi->id] == NULL){
        ret = DPDoorbellShareMemAllocate(output,hdmi->id);
        if(ret){
            dev_err(hdmi->dev,"hdmi share ram allocate fail\n");
            return ret;
        }
    }

    xdx_hdmi_trans_para_fill(hdmi, &para);
    memcpy_toio(output->psDoorbellVirAddr[hdmi->id], &para, sizeof(struct hdmi_trans_para));

    PhyConfig.all = 0;
    PhyConfig.cmd = cH02ADisplayDBCmdHDMI;
    PhyConfig.subCmd = cH02AHDMISubCmdPowerDown;
    PhyConfig.dword = output->psDoorbellPhyAddr[hdmi->id].uiAddr;
    xdx_hdmi_h2a_doorbell_w(hdmi,(u64)PhyConfig.all);
    for (i = timeout; i; i--) {
        DBValue = xdx_hdmi_h2a_doorbell_r(hdmi);
        if (DBValue & BIT(0))
            break;

        usleep_range(1000, 2000);
    }

    if (!(DBValue & BIT(0))) {
        dev_err(hdmi->dev,"xdx hdmi tx phy poweroff fail\n");
        return -ETIMEDOUT;
    }

    PhyConfig.all = DBValue;

    return PhyConfig.ret;
}

static const struct hdmi_tx_phy_ops xdx_hdmi_phy_sink_ops = {
    .init = hdmi_tx_phy_sink_init,
    .disable = hdmi_tx_phy_sink_powerdown,
};

static void xdx_hdmi_tx_phy_init(struct hdmi_dev *hdmi)
{
    struct xdx_output * output = hdmi_to_output(hdmi);
    struct xdx_dm* dm = output_to_dm(output);
    struct xdx_device* xdev = dm_to_xdev(dm);

    if (xdev->asic_type == CHIP_PANGU_A0)
        hdmi->phy.ops = &xdx_hdmi_phy_ops;
    else if(xdev->asic_type == CHIP_PANGU_B0)
        hdmi->phy.ops = &xdx_hdmi_phy_sink_ops;
    else
        hdmi->phy.ops = &xdx_hdmi_phy_ops;

    hdmi->phy.name = "xdx_hdmi_tx_phy";
    hdmi->phy.eq_value = 0;
}

static void xdx_hdmi_i2c_hw_init(struct hdmi_dev *hdmi)
{
//  xdx_hdmi_core_write(hdmi, I2CM_CONTROL0, I2CM_BUS_CLEAR_P);
    xdx_hdmi_core_write(hdmi, I2CM_FM_SCL_CONFIG0, 0x01770177);
    xdx_hdmi_core_write(hdmi, I2CM_SM_SCL_CONFIG0, 0x05dc05dc);
    xdx_hdmi_core_write_mask(hdmi, I2CM_INTERFACE_CONTROL0, I2CM_FM_EN, 0);
    xdx_hdmi_core_write_mask(hdmi, SCDC_CONFIG0, SCDC_I2C_FM_EN, 0);
    /* Clear DONE and ERROR interrupts */
    xdx_hdmi_core_write(hdmi, MAINUNIT_1_INT_CLEAR,
                    I2CM_OP_DONE_CLEAR | I2CM_NACK_RCVD_CLEAR);
}

static int xdx_hdmi_i2c_read(struct hdmi_dev *hdmi,
                unsigned char *buf, unsigned int length)
{
    struct xdx_hdmi_i2c *i2c = hdmi->i2c;
    int stat=0, nbytes, i, j;
    unsigned int len;
#ifndef IIC_HAVE_IRQ
    int timeout = 500;
#endif
    if (!i2c->is_regaddr) {
        dev_dbg(hdmi->dev, "set read register address to 0\n");
        i2c->slave_reg = 0x00;
        i2c->is_regaddr = true;
    }
    len = length;
    for (i = 0; i < len;) {
        if ((len - i) >= 16)
            nbytes = 16;
        else
            nbytes = len - i;
#ifdef IIC_HAVE_IRQ
        reinit_completion(&i2c->comp);
#endif

        xdx_hdmi_core_write(hdmi, MAINUNIT_1_INT_CLEAR,
            I2CM_OP_DONE_CLEAR | I2CM_NACK_RCVD_CLEAR | I2CM_ARB_LOST_CLEAR);

        xdx_hdmi_core_write_mask(hdmi, I2CM_INTERFACE_CONTROL0,
                        I2CM_NBYTES, ((u32)(nbytes - 1) << 20) & I2CM_NBYTES);

        xdx_hdmi_core_write_mask(hdmi, I2CM_INTERFACE_CONTROL0,
                        I2CM_ADDR, (i2c->slave_reg + i) << 12);

        xdx_hdmi_core_write_mask(hdmi, I2CM_INTERFACE_CONTROL0,
                        I2CM_WR_MASK, I2CM_FM_READ);
#ifndef IIC_HAVE_IRQ
        for (; timeout; timeout--) {
            stat = xdx_hdmi_core_read(hdmi, MAINUNIT_1_INT_STATUS);

            i2c->stat = stat & (I2CM_OP_DONE_IRQ | I2CM_READ_REQUEST_IRQ |
                        I2CM_NACK_RCVD_IRQ);

            if (i2c->stat) {
                xdx_hdmi_core_write(hdmi, MAINUNIT_1_INT_CLEAR, i2c->stat);
                break;
            }
            usleep_range(10, 100);
        }
#else
        stat = wait_for_completion_timeout(&i2c->comp, HZ / 10);
#endif
        xdx_hdmi_core_write(hdmi, MAINUNIT_1_INT_CLEAR,
            I2CM_OP_DONE_CLEAR | I2CM_NACK_RCVD_CLEAR | I2CM_ARB_LOST_CLEAR);
        if (!stat) {
            dev_warn(hdmi->dev, "i2c read time out!\n");
            xdx_hdmi_core_write(hdmi, I2CM_CONTROL0, 0x01);
            return -EAGAIN;
        }

        /* Check for error condition on the bus */
        if (i2c->stat & I2CM_NACK_RCVD_IRQ) {
            dev_dbg(hdmi->dev, "i2c read nack!\n");
            xdx_hdmi_core_write(hdmi, I2CM_CONTROL0, 0x01);
            return -EIO;
        }

        for (j = 0; j < nbytes; j++) {
            if (!(j % 4))
                stat = xdx_hdmi_core_read(hdmi,
                                I2CM_INTERFACE_RDDATA_0_3 + (4 * (j / 4)));
            *buf = (stat >> (8 * (j % 4))) & 0xff;
            dev_dbg(hdmi->dev, "i2c read done! i2c->stat:%02x 0x%02x\n",
                i2c->stat, *buf);
            buf++;
        }

        xdx_hdmi_core_write_mask(hdmi,
                        I2CM_INTERFACE_CONTROL0, I2CM_WR_MASK, 0);
        i += nbytes;
    }

    i2c->is_segment = false;

    return 0;
}

static int xdx_hdmi_i2c_write(struct hdmi_dev *hdmi,
                    unsigned char *buf, unsigned int length)
{
    struct xdx_hdmi_i2c *i2c = hdmi->i2c;
    int stat, nbytes, i, j;
    unsigned int len;
#ifndef IIC_HAVE_IRQ
    int timeout = 500;
#endif
    if (!i2c->is_regaddr) {
        /* Use the first write byte as register address */
        i2c->slave_reg = buf[0];
        length--;
        buf++;
        i2c->is_regaddr = true;
    }

    len = length;
    for (i = 0; i < len;) {
        if ((len - i) >= 16)
            nbytes = 16;
        else
            nbytes = len - i;
#ifdef IIC_HAVE_IRQ
        reinit_completion(&i2c->comp);
#endif
    /* Clear DONE and ERROR interrupts */
        xdx_hdmi_core_write(hdmi, MAINUNIT_1_INT_CLEAR,
            I2CM_OP_DONE_CLEAR | I2CM_NACK_RCVD_CLEAR | I2CM_ARB_LOST_CLEAR);
        stat = 0x0;
        for (j = 0; j < nbytes; j++) {
            stat |= *buf << (8 * (j % 4));
            if (((j % 4) == 3) || (j == (nbytes - 1))) {
                xdx_hdmi_core_write(hdmi,
                    I2CM_INTERFACE_WRDATA_0_3 + (4 * (j / 4)), stat);
                stat = 0x0;
            }
            dev_dbg(hdmi->dev, "i2c write data[%d] 0x%02x\n", i + j, *buf);
            buf++;
        }
        xdx_hdmi_core_write_mask(hdmi, I2CM_INTERFACE_CONTROL0,
                        I2CM_NBYTES, ((u32)(nbytes - 1) << 20) & I2CM_NBYTES);
        xdx_hdmi_core_write_mask(hdmi, I2CM_INTERFACE_CONTROL0, I2CM_ADDR,
                        (i2c->slave_reg + i) << 12);
        xdx_hdmi_core_write_mask(hdmi, I2CM_INTERFACE_CONTROL0,
                        I2CM_WR_MASK, I2CM_FM_WRITE);

#ifndef IIC_HAVE_IRQ
        for (; timeout; timeout--) {
            stat = xdx_hdmi_core_read(hdmi, MAINUNIT_1_INT_STATUS);

            i2c->stat = stat & (I2CM_OP_DONE_IRQ | I2CM_READ_REQUEST_IRQ |
                        I2CM_NACK_RCVD_IRQ);
            if (i2c->stat) {
                xdx_hdmi_core_write(hdmi, MAINUNIT_1_INT_CLEAR, i2c->stat);
                break;
            }
            usleep_range(10, 100);
        }
#else
        stat = wait_for_completion_timeout(&i2c->comp, HZ / 10);
#endif
        xdx_hdmi_core_write(hdmi, MAINUNIT_1_INT_CLEAR,
            I2CM_OP_DONE_CLEAR | I2CM_NACK_RCVD_CLEAR | I2CM_ARB_LOST_CLEAR);
        if (!stat) {
            dev_dbg(hdmi->dev, "i2c write time out!\n");
            xdx_hdmi_core_write(hdmi, I2CM_CONTROL0, 0x01);
            return -EAGAIN;
        }

        /* Check for error condition on the bus */
        if (i2c->stat & I2CM_NACK_RCVD_IRQ) {
            dev_dbg(hdmi->dev, "i2c write nack!\n");
            xdx_hdmi_core_write(hdmi, I2CM_CONTROL0, 0x01);
            return -EIO;
        }

        xdx_hdmi_core_write_mask(hdmi,
                        I2CM_INTERFACE_CONTROL0, I2CM_WR_MASK, 0);
        i += nbytes;
    }

    return 0;
}

static int hdmi_i2c_master_xfer(struct i2c_adapter *adap,
                struct i2c_msg *msgs, int num)
{
    struct hdmi_dev *hdmi = i2c_get_adapdata(adap);
    struct xdx_hdmi_i2c *i2c = hdmi->i2c;
    u8 addr = msgs[0].addr;
    int i, ret = 0;
    int count = 3;

    dev_dbg(hdmi->dev, "i2c xfer: num: %d, addr: %#x\n", num, addr);

    for (i = 0; i < num; i++) {
        if (msgs[i].len == 0) {
            dev_dbg(hdmi->dev, "unsupported transfer %d/%d, no data\n",
                 i + 1, num);
            return -EOPNOTSUPP;
        }
    }

    mutex_lock(&i2c->lock);
retry:
    /* Set slave device address taken from the first I2C message */
    if (addr == DDC_SEGMENT_ADDR && msgs[0].len == 1)
        addr = DDC_ADDR;
#ifdef IIC_HAVE_IRQ
    /* Unmute DONE and ERROR interrupts */
    xdx_hdmi_core_write_mask(hdmi, MAINUNIT_1_INT_MASK_N,
            I2CM_NACK_RCVD_MASK_N | I2CM_OP_DONE_MASK_N,
            I2CM_NACK_RCVD_MASK_N | I2CM_OP_DONE_MASK_N);
#endif
    xdx_hdmi_core_write_mask(hdmi, I2CM_INTERFACE_CONTROL0,
                    I2CM_SLVADDR, addr << 5);

    /* Set slave device register address on transfer */
    i2c->is_regaddr = false;

    /* Set segment pointer for I2C extended read mode operation */
    i2c->is_segment = false;

    for (i = 0; i < num; i++) {
        dev_dbg(hdmi->dev, "xfer: num: %d/%d, len: %d, flags: %#x\n",
            i + 1, num, msgs[i].len, msgs[i].flags);

        if (msgs[i].addr == DDC_SEGMENT_ADDR && msgs[i].len == 1) {
            i2c->is_segment = true;
            xdx_hdmi_core_write_mask(hdmi, I2CM_INTERFACE_CONTROL1,
                    I2CM_SEG_ADDR, DDC_SEGMENT_ADDR);
            xdx_hdmi_core_write_mask(hdmi, I2CM_INTERFACE_CONTROL1,
                    I2CM_SEG_PTR, *msgs[i].buf);
        } else {
            if (msgs[i].flags & I2C_M_RD)
                ret = xdx_hdmi_i2c_read(hdmi, msgs[i].buf, msgs[i].len);
            else
                ret = xdx_hdmi_i2c_write(hdmi, msgs[i].buf, msgs[i].len);
        }
        if (ret < 0)
            break;
    }

    if (ret) {
        //xdx_hdmi_core_write_mask(hdmi, GLOBAL_SWRESET_REQUEST, I2C_SWINIT_P, I2C_SWINIT_P);
        xdx_hdmi_core_write(hdmi, MAINUNIT_1_INT_CLEAR,
             I2CM_OP_DONE_CLEAR | I2CM_NACK_RCVD_CLEAR | I2CM_ARB_LOST_CLEAR);
        if(count > 0){
            count--;
            goto retry;
        }
    } else {
        ret = num;
    }
#ifdef IIC_HAVE_IRQ
    /* Mute DONE and ERROR interrupts */
    xdx_hdmi_core_write_mask(hdmi, MAINUNIT_1_INT_MASK_N,
          I2CM_OP_DONE_MASK_N | I2CM_NACK_RCVD_MASK_N, 0);
#endif
    mutex_unlock(&i2c->lock);

    return ret;
}

static u32 hdmi_i2c_func(struct i2c_adapter *adapter)
{
    return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}

static const struct i2c_algorithm hdmi_algorithm = {
    .master_xfer    = hdmi_i2c_master_xfer,
    .functionality  = hdmi_i2c_func,
};

static struct i2c_adapter *xdx_hdmi_i2c_adapter(struct hdmi_dev *hdmi)
{
    struct i2c_adapter *adap;
    struct xdx_hdmi_i2c *i2c;
    int ret;

    i2c = kzalloc(sizeof(*i2c), GFP_KERNEL);
    if (!i2c)
        return ERR_PTR(-ENOMEM);

    mutex_init(&i2c->lock);
#ifdef IIC_HAVE_IRQ
    init_completion(&i2c->comp);
#endif
    adap = &i2c->adap;
    adap->class = I2C_CLASS_DDC;
    adap->owner = THIS_MODULE;
    adap->dev.parent = hdmi->dev;
    adap->dev.of_node = hdmi->dev->of_node;
    adap->algo = &hdmi_algorithm;
    strscpy(adap->name, "XDX HDMI DDC", sizeof(adap->name));
    i2c_set_adapdata(adap, hdmi);

    ret = i2c_add_adapter(adap);
    if (ret) {
        dev_warn(hdmi->dev, "Failed to add %s i2c adapter\n", adap->name);
        kfree(i2c);
        return ERR_PTR(ret);
    }

    hdmi->i2c = i2c;

    dev_dbg(hdmi->dev, "registered %s I2C bus success\n", adap->name);

    return adap;
}

static void repo_hpd_event(struct work_struct *p_work)
{
    struct hdmi_dev *hdmi = container_of(p_work, struct hdmi_dev, work.work);
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    struct drm_device *ddev = xdev_to_drm(xdev);
    u32 phy_status;
    u32 timeout = 100;

    phy_status = xdx_hdmi_tile_read(hdmi, HDMI_HPD_INT_STATUS);

    if (!hdmi->hpd_state) {
        while(timeout--) {
            if (phy_status & HPD_STATUS)
                return;
            phy_status = xdx_hdmi_tile_read(hdmi, HDMI_HPD_INT_STATUS);
            msleep(2);
        }
    }

    if (hdmi->hpd_state && (phy_status & HPD_STATUS)) {
        if ((!hdmi->disabled) && (hdmi->hdmi_mode.clock > 340000)){
            /* HDMI2.0 says when unplug , sink will reset the config
             * before,  and should not send scramble data prior to
             * configure the sink ,  so we should power off hdmi
             * tmds and subsequently do a full hdmi setup when system
             * not calling the setup in replug in the same monitor
             */
            mutex_lock(&hdmi->mutex);
            xdx_hdmi_poweroff(hdmi);
            xdx_hdmi_poweron(hdmi);
            mutex_unlock(&hdmi->mutex);
        }
    }
    if(hdmi->hpd_state == false) {
        if(hdmi->edid) {
            kfree(hdmi->edid);
            hdmi->edid = NULL;
        }
    }

    drm_helper_hpd_irq_event(ddev);
}

static void xdx_hdmi_hpd_irq_enable(struct hdmi_dev *hdmi)
{
    xdx_hdmi_tile_write(hdmi, HDMI_HPD_CTRL, 0x1f);
    xdx_hdmi_tile_write(hdmi, HDMI_HPD_CLR, 0x1f);
}

void xdx_hdmi_hpd_irq_disable(struct hdmi_dev *hdmi)
{
    xdx_hdmi_tile_write(hdmi, HDMI_HPD_CTRL, 0x0);
    xdx_hdmi_tile_write(hdmi, HDMI_HPD_CLR, 0x1f);
}

static int hdmi_irq_set(struct xdx_device *xdev, struct xdx_irq_src *src,
                unsigned type, enum xdx_interrupt_state state)
{
    // TODO: enable/disable IRQ
    return -1;
}

static int hdmi_irq_process(struct xdx_device *xdev, struct xdx_irq_src *src,
                struct xdx_irq_entry *entry)
{
    struct xdx_dm *dm = xdev_to_dm(xdev);
    struct xdx_output *output = dm_to_output(dm);
    struct hdmi_dev *hdmi;
#ifdef IIC_HAVE_IRQ
    struct xdx_hdmi_i2c *i2c = hdmi->i2c;
#endif
    u32 stat;

    if (src->id == XDX_IRQ_CLIENTID_HDMI0)
        hdmi = output->hdmi[0];
    else if (src->id == XDX_IRQ_CLIENTID_HDMI1)
        hdmi = output->hdmi[1];
    else
        return -EINVAL;

#ifdef IIC_HAVE_IRQ
    stat = xdx_hdmi_core_read(hdmi, MAINUNIT_1_INT_STATUS);

    i2c->stat = stat & (I2CM_OP_DONE_IRQ | I2CM_READ_REQUEST_IRQ |
                I2CM_NACK_RCVD_IRQ);
    if (i2c->stat) {
        xdx_hdmi_core_write(hdmi, MAINUNIT_1_INT_CLEAR, i2c->stat);
        complete(&i2c->comp);
    }
#endif
    stat = xdx_hdmi_tile_read(hdmi, HDMI_HPD_INT_STATUS);
    if (stat & HPD_HOT_PLUG) {
        hdmi->hpd_state = true;
        mod_delayed_work(hdmi->workqueue, &hdmi->work, msecs_to_jiffies(20));
    } else if (stat & HPD_HOT_UNPLUG) {
        hdmi->hpd_state = false;
        mod_delayed_work(hdmi->workqueue, &hdmi->work, msecs_to_jiffies(20));
    } else {
        dev_dbg(hdmi->dev, "hdmi hpd irq detect none\n");
    }

    if (stat & 0x1f)
        xdx_hdmi_tile_write(hdmi, HDMI_HPD_CLR, 0x1f);

    return 0;
}

static struct xdx_irq_src_funcs hdmi_src_funcs = {
    .set = hdmi_irq_set,
    .process = hdmi_irq_process,
};

static int hdmi_isr_install(struct hdmi_dev *hdmi)
{
    struct xdx_output* output = hdmi_to_output(hdmi);
    struct xdx_dm* dm = output_to_dm(output);
    struct xdx_device* xdev = dm_to_xdev(dm);
    struct xdx_irq_src *irq_src;

    irq_src = kmalloc(sizeof(*irq_src), GFP_KERNEL);
    if (!irq_src) {
        return -ENOMEM;
    }

    if (hdmi->id == 0)
        irq_src->id = XDX_IRQ_CLIENTID_HDMI0;
    else if (hdmi->id == 1)
        irq_src->id = XDX_IRQ_CLIENTID_HDMI1;
    else
        return -EINVAL;

    irq_src->funcs = &hdmi_src_funcs;

    xdx_irq_register_src(xdev, irq_src);

    hdmi->irq_src = irq_src;

    return 0;
}

static int hdmi_isr_uninstall(struct hdmi_dev *hdmi)
{
    struct xdx_output* output = hdmi_to_output(hdmi);
    struct xdx_dm* dm = output_to_dm(output);
    struct xdx_device* xdev = dm_to_xdev(dm);
    int ret;

    ret = xdx_irq_unregister_src(xdev, hdmi->irq_src);
    if (ret) {
        dev_err(hdmi->dev, "failed to unregister irq source\n");
        return ret;
    }

    kfree(hdmi->irq_src);
    hdmi->irq_src = NULL;

    return 0;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))
bool xdx_hdmi_support_scdc(struct hdmi_dev *hdmi)
{
    if (!hdmi->ddc)
        return false;

    if(!hdmi->display_info.hdmi.scdc.supported ||
       !hdmi->display_info.hdmi.scdc.scrambling.supported)
        return false;

    if (!hdmi->display_info.hdmi.scdc.scrambling.low_rates &&
        hdmi->display_info.max_tmds_clock <= 340000)
        return false;

    return true;
}
#endif

static void xdx_hdmi_infoframe_set_checksum(u8 *ptr, int size)
{
    u8 csum = 0;
    int i;

    ptr[3] = 0;
    /* compute checksum */
    for (i = 0; i < size; i++)
        csum += ptr[i];

    ptr[3] = 256 - csum;
}

static int xdx_hdmi_avi_infoframe_check_only(const struct hdmi_avi_infoframe *frame)
{
    if (frame->type != HDMI_INFOFRAME_TYPE_AVI ||
        frame->version != 2 ||
        frame->length != HDMI_AVI_INFOFRAME_SIZE)
        return -EINVAL;

    if (frame->picture_aspect > HDMI_PICTURE_ASPECT_16_9)
        return -EINVAL;

    return 0;
}

static ssize_t xdx_hdmi_avi_infoframe_pack_only(const struct hdmi_avi_infoframe *frame,
                                           void *buffer, size_t size)
{
    u8 *ptr = buffer;
    size_t length;
    int ret;

    ret = xdx_hdmi_avi_infoframe_check_only(frame);
    if (ret)
        return ret;

    length = HDMI_INFOFRAME_HEADER_SIZE + frame->length;

    if (size < length)
        return -ENOSPC;

    memset(buffer, 0, size);

    ptr[0] = frame->type;
    ptr[1] = frame->version;
    ptr[2] = frame->length;
    ptr[3] = 0; /* checksum */

    /* start infoframe payload */
    ptr += HDMI_INFOFRAME_HEADER_SIZE;

    ptr[0] = ((frame->colorspace & 0x3) << 5) | (frame->scan_mode & 0x3);

    /*
     * Data byte 1, bit 4 has to be set if we provide the active format
     * aspect ratio
     */
    if (frame->active_aspect & 0xf)
        ptr[0] |= BIT(4);

    /* Bit 3 and 2 indicate if we transmit horizontal/vertical bar data */
    if (frame->top_bar || frame->bottom_bar)
        ptr[0] |= BIT(3);

    if (frame->left_bar || frame->right_bar)
        ptr[0] |= BIT(2);

    ptr[1] = ((frame->colorimetry & 0x3) << 6) |
             ((frame->picture_aspect & 0x3) << 4) |
              (frame->active_aspect & 0xf);

    ptr[2] = ((frame->extended_colorimetry & 0x7) << 4) |
             ((frame->quantization_range & 0x3) << 2) |
              (frame->nups & 0x3);

    if (frame->itc)
        ptr[2] |= BIT(7);

    ptr[3] = frame->video_code & 0x7f;

    ptr[4] = ((frame->ycc_quantization_range & 0x3) << 6) |
             ((frame->content_type & 0x3) << 4) |
	     (frame->pixel_repeat & 0xf);

    ptr[5] = frame->top_bar & 0xff;
    ptr[6] = (frame->top_bar >> 8) & 0xff;
    ptr[7] = frame->bottom_bar & 0xff;
    ptr[8] = (frame->bottom_bar >> 8) & 0xff;
    ptr[9] = frame->left_bar & 0xff;
    ptr[10] = (frame->left_bar >> 8) & 0xff;
    ptr[11] = frame->right_bar & 0xff;
    ptr[12] = (frame->right_bar >> 8) & 0xff;

    xdx_hdmi_infoframe_set_checksum(buffer, length);

    return length;
}

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 10, 0)) && (!defined(XDXGPU_CENTOS79_SUPPORT))
static enum hdmi_quantization_range
drm_default_rgb_quant_range(const struct drm_display_mode *mode)
{
        /* All CEA modes other than VIC 1 use limited quantization range. */
        return drm_match_cea_mode(mode) > 1 ?
                HDMI_QUANTIZATION_RANGE_LIMITED :
                HDMI_QUANTIZATION_RANGE_FULL;
}

static void drm_hdmi_avi_infoframe_quant_range(struct hdmi_avi_infoframe *frame,
                                   const struct drm_display_mode *mode,
                                   enum hdmi_quantization_range rgb_quant_range,
                                   bool rgb_quant_range_selectable,
                                   bool is_hdmi2_sink)
{
        /*
         * CEA-861:
         * "A Source shall not send a non-zero Q value that does not correspond
         *  to the default RGB Quantization Range for the transmitted Picture
         *  unless the Sink indicates support for the Q bit in a Video
         *  Capabilities Data Block."
         *
         * HDMI 2.0 recommends sending non-zero Q when it does match the
         * default RGB quantization range for the mode, even when QS=0.
         */
        if (rgb_quant_range_selectable ||
            rgb_quant_range == drm_default_rgb_quant_range(mode))
                frame->quantization_range = rgb_quant_range;
        else
                frame->quantization_range = HDMI_QUANTIZATION_RANGE_DEFAULT;

        /*
         * CEA-861-F:
         * "When transmitting any RGB colorimetry, the Source should set the
         *  YQ-field to match the RGB Quantization Range being transmitted
         *  (e.g., when Limited Range RGB, set YQ=0 or when Full Range RGB,
         *  set YQ=1) and the Sink shall ignore the YQ-field."
         *
         * Unfortunate certain sinks (eg. VIZ Model 67/E261VA) get confused
         * by non-zero YQ when receiving RGB. There doesn't seem to be any
         * good way to tell which version of CEA-861 the sink supports, so
         * we limit non-zero YQ to HDMI 2.0 sinks only as HDMI 2.0 is based
         * on on CEA-861-F.
         */
        if (!is_hdmi2_sink ||
            rgb_quant_range == HDMI_QUANTIZATION_RANGE_LIMITED)
                frame->ycc_quantization_range =
                        HDMI_YCC_QUANTIZATION_RANGE_LIMITED;
        else
                frame->ycc_quantization_range =
                        HDMI_YCC_QUANTIZATION_RANGE_FULL;
}
#endif

static void xdx_hdmi_config_avi(struct hdmi_dev *hdmi)
{
    u8 buff[32];
    u32 val;
    int i, j;
    struct hdmi_avi_infoframe frame;
#if (!defined(XDXGPU_CENTOS79_SUPPORT))
    union hdmi_infoframe info;
#endif
    videoParams_t *videoParams = &hdmi->pVideo;
#if ((LINUX_VERSION_CODE < KERNEL_VERSION(5, 1, 0)))  && (!defined(XDXGPU_CENTOS82_SUPPORT)) && (!defined(XDXGPU_CENTOS85_SUPPORT)) || (defined(XDXGPU_CENTOS79_SUPPORT))
    bool rgb_range_selectable = false;
#endif

    hdmi->vic = drm_match_cea_mode(&hdmi->hdmi_mode);
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))) || ((defined(XDXGPU_CENTOS82_SUPPORT)) && (!defined(XDXGPU_CENTOS79_SUPPORT))) || (defined(XDXGPU_CENTOS85_SUPPORT))
    drm_hdmi_avi_infoframe_from_display_mode(&frame,
            &hdmi->connector, &hdmi->hdmi_mode);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)) || (defined(XDXGPU_CENTOS79_SUPPORT))
	drm_hdmi_avi_infoframe_from_display_mode(&frame,
            &hdmi->hdmi_mode, false);
#else
    drm_hdmi_avi_infoframe_from_display_mode(&frame,
            &hdmi->hdmi_mode);
#endif
    //default encoder out bus formate is RGB
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))) || ((defined(XDXGPU_CENTOS82_SUPPORT)) && (!defined(XDXGPU_CENTOS79_SUPPORT))) || (defined(XDXGPU_CENTOS85_SUPPORT))
    drm_hdmi_avi_infoframe_quant_range(&frame, &hdmi->connector,
            &hdmi->hdmi_mode, videoParams->rgb_quant_range);
#else
    rgb_range_selectable = drm_rgb_quant_range_selectable(hdmi->edid);
    drm_hdmi_avi_infoframe_quant_range(&frame, &hdmi->hdmi_mode,
            videoParams->rgb_quant_range, rgb_range_selectable, false);
#endif
    frame.colorspace = videoParams->colorspace;
    frame.colorimetry = videoParams->colorimetry;
    frame.extended_colorimetry = videoParams->extended_colorimetry;
    frame.scan_mode = HDMI_SCAN_MODE_NONE;
    frame.itc = videoParams->itc;
    frame.content_type = videoParams->content_type;
    xdx_hdmi_avi_infoframe_pack_only(&frame, buff, 32);
    /* mode which vic >= 128 must use avi version 3 */
    if (hdmi->vic >= 128) {
        frame.version = 3;
        buff[1] = frame.version;
        buff[4] &= 0x1f;
        buff[4] |= ((frame.colorspace & 0x7) << 5);
        buff[7] = frame.video_code;
        xdx_hdmi_infoframe_set_checksum(buff, 17);
    }

    val = (frame.version << 8) | (frame.length << 16);
    xdx_hdmi_core_write(hdmi, PKT_AVI_CONTENTS0, val);
    dev_dbg(hdmi->dev, "PKT_AVI_CONTENTS0%d:%x\n",PKT_AVI_CONTENTS0, val);
    for (i = 0; i < 4; i++) {
        for (j = 0; j < 4; j++) {
            if (i * 4 + j >= 14)
                break;
            if (!j)
                val = buff[i * 4 + j + 3];
            val |= buff[i * 4 + j + 3] << (8 * j);
        }

        xdx_hdmi_core_write(hdmi, PKT_AVI_CONTENTS1 + i * 4, val);
        dev_dbg(hdmi->dev, "PKT_AVI_CONTENTS%d:%x\n", i * 4, val);
    }
#if (!defined(XDXGPU_CENTOS79_SUPPORT))
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))) || (defined(XDXGPU_CENTOS82_SUPPORT)) || (defined(XDXGPU_CENTOS85_SUPPORT))
    hdmi_infoframe_unpack(&info, buff, sizeof(buff));
#else
    hdmi_infoframe_unpack(&info, buff);
#endif
    hdmi_infoframe_log(KERN_INFO, hdmi->dev, &info);
#endif
}

__attribute__((unused)) static void hdmi_core_regs_dump(struct hdmi_dev *hdmi)
{
    int i;
    const struct reg {
        u32 offset;
        const char *name;
    } regs[] = {
        { LINK_CONFIG0, "LINK_CONFIG0"},
        { VIDEO_INTERFACE_CONFIG2, "VIDEO_INTERFACE_CONFIG2"},
        { PKTSCHED_PKT_EN, "PKTSCHED_PKT_EN"},
        { PKTSCHED_PKT_CONFIG1, "PKTSCHED_PKT_CONFIG1"},
        { PKT_AVI_CONTENTS0, "PKT_AVI_CONTENTS0"},
        { PKT_AVI_CONTENTS1, "PKT_AVI_CONTENTS1"},
        { SCRAMB_CONFIG0, "SCRAMB_CONFIG0"},
    };

    for (i = 0; i < ARRAY_SIZE(regs); i++) {
        dev_dbg(hdmi->dev, "%s: 0x%x\n", regs[i].name, xdx_hdmi_core_read(hdmi, regs[i].offset));
    }
}

static void xdx_hdmi_avp_config(struct hdmi_dev *hdmi)
{
    // Configure TMDS or FRL
    xdx_hdmi_core_write_mask(hdmi, LINK_CONFIG0, OPMODE_FRL, 0);
    // Configure FRL number of lanes
    xdx_hdmi_core_write_mask(hdmi, LINK_CONFIG0, OPMODE_FRL_4LANES, 0);
    //ipi data serialized
    xdx_hdmi_core_write_mask(hdmi, VIDEO_INTERFACE_CONFIG2,
            IPI_DESERIALIZER_EN, IPI_DESERIALIZER_EN);
    //disable avi packet auto send
    xdx_hdmi_core_write_mask(hdmi, PKTSCHED_PKT_EN, PKTSCHED_AVI_TX_EN, 0);
    //avi infoframe packet
    //fc_avi_config(hdmi);
    xdx_hdmi_config_avi(hdmi);
    xdx_hdmi_core_write_mask(hdmi, PKTSCHED_PKT_CONFIG1,
            PKTSCHED_AVI_FIELDRATE, PKTSCHED_AVI_FIELDRATE);
    //avi packets auto send
    xdx_hdmi_core_write_mask(hdmi, PKTSCHED_PKT_EN,
            PKTSCHED_AVI_TX_EN | PKTSCHED_GCP_TX_EN |
            PKTSCHED_SPDI_TX_EN | PKTSCHED_AUDI_TX_EN,
            PKTSCHED_AVI_TX_EN | PKTSCHED_GCP_TX_EN |
            PKTSCHED_SPDI_TX_EN | PKTSCHED_AUDI_TX_EN);
}

static int xdx_hdmi_mode_setup(struct hdmi_dev *hdmi)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    int ret;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))
    u8 bytes = 0;
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 9, 0)) && (!defined(XDXGPU_CENTOS82_SUPPORT)) && (!defined(XDXGPU_CENTOS85_SUPPORT))
    dev_dbg(hdmi->dev, "%s: Pixel clock: %d, Mode: %dx%d@%d\n", __func__,
        hdmi->hdmi_mode.clock, hdmi->hdmi_mode.hdisplay,
        hdmi->hdmi_mode.vdisplay, hdmi->hdmi_mode.vrefresh);
#else
    dev_dbg(hdmi->dev, "%s: Pixel clock: %d, Mode: %dx%d\n", __func__,
        hdmi->hdmi_mode.clock, hdmi->hdmi_mode.hdisplay,
        hdmi->hdmi_mode.vdisplay);
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))
    if (hdmi->hdmi_mode.clock > 340000) {
        drm_scdc_readb(hdmi->ddc, SCDC_SINK_VERSION, &bytes);
        drm_scdc_writeb(hdmi->ddc, SCDC_SOURCE_VERSION,
                min_t(u8, bytes, SCDC_MIN_SOURCE_VERSION));
#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 4, 0))
        drm_scdc_set_high_tmds_clock_ratio(hdmi->ddc, 1);
        drm_scdc_set_scrambling(hdmi->ddc, 1);
#else
        drm_scdc_set_high_tmds_clock_ratio(&hdmi->connector, 1);
        drm_scdc_set_scrambling(&hdmi->connector, 1);
#endif
    } else {
        if (xdx_hdmi_support_scdc(hdmi)) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 4, 0))
            drm_scdc_set_high_tmds_clock_ratio(hdmi->ddc, 0);
            drm_scdc_set_scrambling(hdmi->ddc, 0);
#else
            drm_scdc_set_high_tmds_clock_ratio(&hdmi->connector, 0);
            drm_scdc_set_scrambling(&hdmi->connector, 0);
#endif
        }
    }
#endif

    ret = hdmi->phy.ops->init(hdmi);
    if (ret) {
        hdmi->phy.enabled = false;
        return ret;
    } else {
        hdmi->phy.enabled = true;
    }

    msleep(5);

    if (!hdmi->sink_is_hdmi) {
        xdx_hdmi_core_write_mask(hdmi, LINK_CONFIG0, OPMODE_DVI, OPMODE_DVI);
        dev_dbg(hdmi->dev, "%s DVI mode\n", __func__);
        return 0;
    } else {
        xdx_hdmi_core_write_mask(hdmi, LINK_CONFIG0, OPMODE_DVI, 0);
    }

    if (xdev->asic_type == CHIP_PANGU_B0) {
        hdmi_core_regs_dump(hdmi);
        return 0;
    }

    xdx_hdmi_core_write_mask(hdmi, HDCP14_CONFIG0,
            HDCP14_BYPASS, HDCP14_BYPASS);
    xdx_hdmi_core_write_mask(hdmi, HDCP2LOGIC_CONFIG0,
            HDCP2_BYPASS, HDCP2_BYPASS);

    if (hdmi->hdmi_mode.clock > 340000) {
        xdx_hdmi_core_write(hdmi, SCRAMB_CONFIG0, 1);
    } else {
        xdx_hdmi_core_write(hdmi, SCRAMB_CONFIG0, 0);
    }

    xdx_hdmi_avp_config(hdmi);
    //hdmi_core_regs_dump(hdmi);

    return 0;
}

static void xdx_hdmi_poweron(struct hdmi_dev *hdmi)
{
#ifdef SUPPORT_AUDIO
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
#endif
    int ret;

    ret = xdx_hdmi_mode_setup(hdmi);
    if (ret)
        return;

    hdmi->disabled = false;

#ifdef SUPPORT_AUDIO
    if (hdmi->sink_has_audio && xdev->xdx_hdmi_audio_hotplug_ptr) {
        if (hdmi->id == 0) {
            xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_HDMI,
                XDX_AUDIO_MODULE_CONNECT);
        }
        else if (hdmi->id == 1) {
            xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_HDMI1,
                XDX_AUDIO_MODULE_CONNECT);
        } else {}
    }
#endif
}

static void xdx_hdmi_poweroff(struct hdmi_dev *hdmi)
{
#ifdef SUPPORT_AUDIO
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
#endif

    if (hdmi->phy.enabled) {
        hdmi->phy.ops->disable(hdmi);
        hdmi->phy.enabled = false;
    }
    hdmi->disabled = true;
    dev_dbg(hdmi->dev, "%s hdmi->sink_has_audio:%d\n", __func__, hdmi->sink_has_audio);
#ifdef SUPPORT_AUDIO
    if (hdmi->sink_has_audio && xdev->xdx_hdmi_audio_hotplug_ptr) {
        if (hdmi->id == 0) {
            xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_HDMI,
                XDX_AUDIO_MODULE_DISCONNECT);
        }
        else if (hdmi->id == 1) {
            xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_HDMI1,
                XDX_AUDIO_MODULE_DISCONNECT);
        } else {}
    }
#endif
}

static void xdx_hdmi_encoder_dpms(struct drm_encoder *encoder, int mode)
{
    struct hdmi_dev *hdmi = encoder_to_hdmi(encoder);

    mutex_lock(&hdmi->mutex);

    if (mode != DRM_MODE_DPMS_ON) {
        xdx_hdmi_poweroff(hdmi);
    }
    else {
        xdx_hdmi_poweron(hdmi);
    }

    mutex_unlock(&hdmi->mutex);
}

static enum drm_connector_status
xdx_hdmi_helper_connector_detect(struct drm_connector *connector, bool force)
{
    struct hdmi_dev *hdmi =
        container_of(connector, struct hdmi_dev, connector);
    enum drm_connector_status status;
    u32 phy_status;

    phy_status = xdx_hdmi_tile_read(hdmi, HDMI_HPD_INT_STATUS);
    if (phy_status & HPD_STATUS)
        status = connector_status_connected;
    else
        status = connector_status_disconnected;

    return status;
}

static void xdx_hdmi_helper_connector_force(struct drm_connector *connector)
{
    struct hdmi_dev *hdmi =
        container_of(connector, struct hdmi_dev, connector);

    mutex_lock(&hdmi->mutex);
    hdmi->force = connector->force;
    mutex_unlock(&hdmi->mutex);
}

static void xdx_hdmi_helper_connector_destroy(struct drm_connector *connector)
{
    drm_connector_unregister(connector);
    drm_connector_cleanup(connector);
}

static void set_display_info(struct drm_connector *connector, struct drm_display_mode *mode)
{
    struct drm_display_info *info = &connector->display_info;

    info->bpc = 8;
    info->color_formats = DRM_COLOR_FORMAT_RGB444;
}

static int xdx_hdmi_connector_get_modes(struct drm_connector *connector)
{
    struct hdmi_dev *hdmi =
        container_of(connector, struct hdmi_dev, connector);
    struct drm_device *dev = connector->dev;
    struct xdx_device *xdev = drm_to_xdev(dev);
    struct drm_display_mode *mode;
    struct edid *edid = NULL;
    int count = 0;
    u16 i;

    if (xdev->flags.fake_mode) {
        count = drm_add_modes_noedid(connector, 800, 600);
        drm_set_preferred_mode(connector, 800, 600);
        hdmi->display_info = connector->display_info;

        return count;
    }

    if (!hdmi->ddc)
        return 0;

    if (!hdmi->edid) {
        edid = drm_get_edid(connector, hdmi->ddc);
    }
    else {
        edid = hdmi->edid;
    }
    
    if (edid) {
        hdmi->edid = edid;
        drm_connector_update_edid_property(connector, edid);
        count = drm_add_edid_modes(connector, edid);
        hdmi->sink_is_hdmi = drm_detect_hdmi_monitor(edid);
        hdmi->sink_has_audio = drm_detect_monitor_audio(edid);
    } else {
        hdmi->sink_is_hdmi = true;
        hdmi->sink_has_audio = true;
        for (i = 0; i < ARRAY_SIZE(hdmi_default_modes); i++) {
            const struct drm_display_mode *ptr =
                &hdmi_default_modes[i];

            mode = drm_mode_duplicate(connector->dev, ptr);
            if (mode) {
                if (!i)
                    mode->type = DRM_MODE_TYPE_PREFERRED;
                drm_mode_probed_add(connector, mode);
                count++;
            }
        }
        set_display_info(connector,mode);
    }

    hdmi->display_info = connector->display_info;

    dev_dbg(dev->dev, "bpc %d format %d\n",
        hdmi->display_info.bpc, hdmi->display_info.color_formats);

    if(!hdmi->sink_is_hdmi) {
        dev_info(dev->dev, "sink is not HDMI, probably be DVI !\n");
        if(hdmi->display_info.bpc == 0)
            hdmi->display_info.bpc = 8;
        if(hdmi->display_info.color_formats == 0)
            hdmi->display_info.color_formats = DRM_COLOR_FORMAT_RGB444;
    } else {
        if(hdmi->display_info.bpc > 10)
            hdmi->display_info.bpc = 8;
        if((hdmi->display_info.bpc == 0) || (hdmi->display_info.color_formats == 0)) {
             dev_err(dev->dev, "bpc or color_formats is null !!\n");
             return 0;
        }
    }

    if (edid)
        dev_dbg(dev->dev, "\nHDMI:Version: %d\nRevision: %d\nExtensions: %d\n# of modes: %d\n",
            edid->version, edid->revision, edid->extensions, count);

    return count;
}

static int xdx_hdmi_mode_valid(struct drm_connector *connector,
                   struct drm_display_mode *mode)
{
    struct drm_device *dev = connector->dev;

    if(connector->connector_type == DRM_MODE_CONNECTOR_DVII)
    {
        // dvi-i
        if(mode->clock > 165000) {
            return MODE_ERROR;
        }
    }
    if (!mode->hdisplay || !mode->vdisplay || !drm_mode_vrefresh(mode) )
    {
        dev_dbg(dev->dev,"[DRM]par:mode:%dx%d vrefresh:%d pix_clock:%d un-supported\n",
            mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode), mode->clock);
        return MODE_ERROR;
    }else{
//        dev_dbg(dev->dev,"[DRM]par:mode:%dx%d vrefresh:%d pix_clock:%d supported\n",
//            mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode), mode->clock);
        return MODE_OK;
    }

    return MODE_OK;
}

static struct drm_encoder *
xdx_hdmi_connector_best_encoder(struct drm_connector *connector)
{
    struct hdmi_dev *hdmi =
        container_of(connector, struct hdmi_dev, connector);

    return &hdmi->encoder;
}

static void xdx_hdmi_encoder_destroy(struct drm_encoder *encoder)
{
    drm_encoder_cleanup(encoder);
}

static void xdx_hdmi_encoder_mode_set(struct drm_encoder *encoder,
                      struct drm_display_mode *mode,
                      struct drm_display_mode *adjusted_mode)
{
    struct hdmi_dev *hdmi = encoder_to_hdmi(encoder);
    struct drm_device *ddev = encoder->dev;
    struct drm_crtc* crtc;
    struct dm_crtc_state* dm_state;
    struct xdx_device *xdev = drm_to_xdev(ddev);
    //u32 PixClock = mode->clock;
    u32 ColorDeep = hdmi->display_info.bpc;
    u32 Format = hdmi->display_info.color_formats;

    drm_for_each_crtc (crtc, ddev)
    {
        if (drm_crtc_index(crtc) == drm_encoder_index(encoder))
            break;
    }
    dm_state = to_dm_crtc_state(crtc->state);
    dm_state->output_fmt = Format;

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 18, 0)
    Format &= (DRM_COLOR_FORMAT_RGB444 | DRM_COLOR_FORMAT_YCRCB444);
#else
    Format &= (DRM_COLOR_FORMAT_RGB444 | DRM_COLOR_FORMAT_YCBCR444);
#endif
    //if format is RGB or YcrCb444, check clock limit
    //if(PixClock > 480000 && ColorDeep > 8 && Format){
    if(ColorDeep > 8){ //temporarily disable 10bpc for compatibility
        ColorDeep = 8;
        dev_warn(xdev->dev,"change to 8 bpc\n");
    }
    hdmi->pVideo.bpc = ColorDeep;
    dm_state->bpc = ColorDeep;


    mutex_lock(&hdmi->mutex);
    memcpy(&hdmi->hdmi_mode, mode, sizeof(hdmi->hdmi_mode));
    mutex_unlock(&hdmi->mutex);
}

static void xdx_hdmi_encoder_commit(struct drm_encoder *encoder)
{
    const struct drm_encoder_helper_funcs * helper = encoder->helper_private;

    helper->dpms(encoder, DRM_MODE_DPMS_ON);
}

static bool xdx_hdmi_encoder_mode_fixup(struct drm_encoder *encoder,
                    const struct drm_display_mode *mode,
                    struct drm_display_mode *adjusted_mode)
{
    return true;
}

static void xdx_hdmi_encoder_prepare(struct drm_encoder *encoder)
{
    const struct drm_encoder_helper_funcs * helper = encoder->helper_private;

    helper->dpms(encoder, DRM_MODE_DPMS_OFF);
}

static const struct drm_encoder_funcs xdx_hdmi_encoder_funcs = {
    .destroy = xdx_hdmi_encoder_destroy,
};

static ssize_t xdx_debugfs_connector_read(struct file *f, char __user *buf,
					size_t size, loff_t *pos)
{
    struct drm_connector *connector = file_inode(f)->i_private;
    u32 len, buffer_size = 4096;
    char *str;
    size_t ret;

    str = kzalloc(buffer_size, GFP_KERNEL);
    if(str == NULL)
	    return -ENOMEM;
    len = scnprintf(str, buffer_size,  "connector:%s\n",connector->name);
    ret = simple_read_from_buffer(buf, size ,pos, str, len);
    kfree(str);
    return ret;
}

static ssize_t xdx_debugfs_connector_write(struct file *f, const char __user *buf,
					 size_t size, loff_t *pos)
{
    struct drm_connector *connector = file_inode(f)->i_private;
    struct hdmi_dev *hdmi =
        container_of(connector, struct hdmi_dev, connector);
    int retval;
    char my_buffer[512];
    int temp = 0;

    //retval = copy_from_user(my_buffer,buf,size);
    retval = simple_write_to_buffer(my_buffer,size,pos,buf,4);
    if(retval < 0)
        return retval;
    sscanf(my_buffer, "%d", &temp);
    hdmi->phy.eq_value = temp;

    return retval;
}

static const struct file_operations xdx_connector_debugfs_fops = {
	.read = xdx_debugfs_connector_read,
	.write = xdx_debugfs_connector_write,
};


static int xdx_connector_late_register(struct drm_connector *connector)
{
	debugfs_create_file("eqdebug", 0444, connector->debugfs_entry, connector,
	            &xdx_connector_debugfs_fops);
	return 0;
}
static const struct drm_encoder_helper_funcs xdx_hdmi_encoder_helper_funcs = {
    .dpms = xdx_hdmi_encoder_dpms,
    .prepare = xdx_hdmi_encoder_prepare,
    .commit = xdx_hdmi_encoder_commit,
    .mode_set = xdx_hdmi_encoder_mode_set,
    .mode_fixup = xdx_hdmi_encoder_mode_fixup,
//  .atomic_check = xdx_hdmi_encoder_atomic_check,
};

static const struct drm_connector_funcs xdx_hdmi_connector_funcs = {
    .dpms = drm_helper_connector_dpms,
    .reset = drm_atomic_helper_connector_reset,
    .detect = xdx_hdmi_helper_connector_detect,
    .force = xdx_hdmi_helper_connector_force,
    .fill_modes = drm_helper_probe_single_connector_modes,
    .destroy = xdx_hdmi_helper_connector_destroy,
    //.set_property = xdx_hdmi_helper_connector_set_property,
    .atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
    .atomic_destroy_state = drm_atomic_helper_connector_destroy_state,
    .late_register = xdx_connector_late_register,
};

static const struct drm_connector_helper_funcs xdx_hdmi_connector_helper_funcs = {
    .get_modes = xdx_hdmi_connector_get_modes,
    .mode_valid = xdx_hdmi_mode_valid,
    /*
     * For atomic, don't set atomic_best_encoder or best_encoder. This will
     * cause the DRM core to fallback to drm_atomic_helper_best_encoder().
     * This is fine as we only have a single connector and encoder.
     */
    .best_encoder = xdx_hdmi_connector_best_encoder,
};

static int xdx_hdmi_register(struct hdmi_dev *hdmi, u32 crtc_mask)
{
    struct xdx_output *output = hdmi_to_output(hdmi);
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    struct drm_device *ddev = xdev_to_drm(xdev);
    int ret;

    hdmi->connector.polled = DRM_CONNECTOR_POLL_HPD;
    hdmi->connector.interlace_allowed = 0;
    hdmi->connector.doublescan_allowed = 0;

    if((xdev->fw_st_info.vgaDetectFlag == HDMI_VGA_DVI) && (hdmi->id == 1)) {
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0)))
        drm_connector_init_with_ddc(ddev, &hdmi->connector,
                &xdx_hdmi_connector_funcs, DRM_MODE_CONNECTOR_DVII, hdmi->ddc);
#else
        drm_connector_init(ddev, &hdmi->connector,
                &xdx_hdmi_connector_funcs, DRM_MODE_CONNECTOR_DVII);
#endif
    } else {
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0)))
        drm_connector_init_with_ddc(ddev, &hdmi->connector,
                &xdx_hdmi_connector_funcs, DRM_MODE_CONNECTOR_HDMIA, hdmi->ddc);
#else
        drm_connector_init(ddev, &hdmi->connector,
                &xdx_hdmi_connector_funcs, DRM_MODE_CONNECTOR_HDMIA);
#endif
    }
    drm_connector_helper_add(&hdmi->connector,
                 &xdx_hdmi_connector_helper_funcs);

    if (!hdmi->id) {
        hdmi->encoder.possible_clones = 1 << 0;
        hdmi->encoder.possible_crtcs = BIT(0);
    } else {
        hdmi->encoder.possible_clones = 1 << 2;
        hdmi->encoder.possible_crtcs = BIT(2);
    }

    ret = drm_encoder_init(ddev, &hdmi->encoder,
                   &xdx_hdmi_encoder_funcs, DRM_MODE_ENCODER_TMDS,
                   NULL);

    if (ret)
        return ret;

    drm_encoder_helper_add(&hdmi->encoder, &xdx_hdmi_encoder_helper_funcs);

    ret = drm_connector_attach_encoder(&hdmi->connector, &hdmi->encoder);
    if (ret) {
        dev_err(hdmi->dev,"attach [ENCODER:%d:%s] to [CONNECTOR:%d:%s] err\n",
            hdmi->encoder.base.id, hdmi->encoder.name,
            hdmi->connector.base.id, hdmi->connector.name);
        return ret;
    }

    return 0;
}

static void video_parms_init(struct hdmi_dev *hdmi)
{
    hdmi->pVideo.colorspace = HDMI_COLORSPACE_RGB;
    hdmi->pVideo.rgb_quant_range = HDMI_QUANTIZATION_RANGE_FULL;
    //hdmi->pVideo.mColorimetry = ITU601 ;
    hdmi->pVideo.content_type = HDMI_CONTENT_TYPE_CINEMA;
    hdmi->pVideo.itc = true;
    hdmi->pVideo.colorimetry = HDMI_COLORIMETRY_ITU_709;
    hdmi->pVideo.extended_colorimetry = HDMI_EXTENDED_COLORIMETRY_XV_YCC_709;
}

int xdx_hdmi_init(struct xdx_dm *dm, u32 id)
{
    struct xdx_device *xdev = dm_to_xdev(dm);
    struct xdx_output *output = dm_to_output(dm);
    struct hdmi_dev *hdmi = NULL;
    int ret;

    hdmi = kzalloc(sizeof(*hdmi), GFP_KERNEL);
    if (!hdmi) {
        return -ENOMEM;
    }

    output->psDoorbellShareMemDesc[id] =NULL;
    output->psDoorbellVirAddr[id] = NULL;
    hdmi->id = id;
    output->hdmi[id] = hdmi;
    hdmi->output = output;
    hdmi->dev = xdev->dev;
    hdmi->core_write = xdx_hdmi_core_write;
    hdmi->core_read = xdx_hdmi_core_read;
    hdmi->disabled = true;
    hdmi->hpd_state = true;
    hdmi->force = DRM_FORCE_UNSPECIFIED;
    hdmi->phy.enabled = false;
    mutex_init(&hdmi->mutex);

    video_parms_init(hdmi);

    hdmi->workqueue = create_workqueue("hdmi_hpd_queue");
    INIT_DELAYED_WORK(&hdmi->work, repo_hpd_event);

    hdmi->ddc = xdx_hdmi_i2c_adapter(hdmi);
    if (IS_ERR(hdmi->ddc)) {
        ret = PTR_ERR(hdmi->ddc);
        hdmi->ddc = NULL;
        goto err_alloc;
    }

    if (hdmi->i2c)
        xdx_hdmi_i2c_hw_init(hdmi);

    ret = xdx_hdmi_register(hdmi, dm_dc_get_crtc_mask(dm->dc[id]));
    if (ret)
        goto err_put_adapter;

    ret = hdmi_isr_install(hdmi);
    if (ret)
        goto err_cleanup_hdmi;

    memcpy(&hdmi->hdmi_mode, &hdmi_default_modes[0], sizeof(hdmi->hdmi_mode));

    xdx_hdmi_tx_phy_init(hdmi);
    xdx_hdmi_hpd_irq_enable(hdmi);

    return 0;
err_cleanup_hdmi:
    hdmi->connector.funcs->destroy(&hdmi->connector);
    hdmi->encoder.funcs->destroy(&hdmi->encoder);
err_put_adapter:
    if (hdmi->i2c)
        i2c_del_adapter(&hdmi->i2c->adap);
    else
        i2c_put_adapter(hdmi->ddc);
err_alloc:
    if (hdmi->workqueue)
        destroy_workqueue(hdmi->workqueue);
    kfree(hdmi);

    return ret;
}

int xdx_hdmi_hw_init(struct xdx_dm *dm, u32 id)
{
    struct xdx_output *output = dm_to_output(dm);
    struct hdmi_dev *hdmi = output->hdmi[id];

    if (hdmi->i2c){
        dev_info(hdmi->dev,"resume i2c init\n");
        xdx_hdmi_i2c_hw_init(hdmi);
    }
    xdx_hdmi_hpd_irq_enable(hdmi);

    return 0;
}

int xdx_hdmi_deinit(struct xdx_dm *dm, u32 id)
{
    struct xdx_output *output = dm_to_output(dm);
    struct hdmi_dev *hdmi = output->hdmi[id];

    xdx_hdmi_hpd_irq_disable(hdmi);

    hdmi_isr_uninstall(hdmi);

    if (hdmi->workqueue)
        destroy_workqueue(hdmi->workqueue);

    if (hdmi->i2c) {
        i2c_del_adapter(&hdmi->i2c->adap);
        kfree(hdmi->i2c);
    } else {
        i2c_put_adapter(hdmi->ddc);
    }

    if(hdmi->edid)
        kfree(hdmi->edid);

    kfree(hdmi);

    output->hdmi[id] = NULL;

    return 0;
}
