/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#ifndef VIRT_DMA_H
#define VIRT_DMA_H

#include <linux/dmaengine.h>
#include <linux/interrupt.h>

#include "dmaengine.h"

struct xdx_virt_dma_desc {
	struct dma_async_tx_descriptor tx;
	struct dmaengine_result tx_result;
	/* protected by vc.lock */
	struct list_head node;
};

struct xdx_virt_dma_chan {
	struct dma_chan	chan;
	struct tasklet_struct task;
	void (*desc_free)(struct xdx_virt_dma_desc *);

	spinlock_t lock;

	/* protected by vc.lock */
	struct list_head desc_allocated;
	struct list_head desc_submitted;
	struct list_head desc_issued;
	struct list_head desc_issued_pending;
	struct list_head desc_completed;

	struct xdx_virt_dma_desc *cyclic;
	struct xdx_virt_dma_desc *vd_terminated;
};

static inline struct xdx_virt_dma_chan *xdx_to_virt_chan(struct dma_chan *chan)
{
	return container_of(chan, struct xdx_virt_dma_chan, chan);
}

void xdx_vchan_dma_desc_free_list(struct xdx_virt_dma_chan *vc, struct list_head *head);
void xdx_vchan_init(struct xdx_virt_dma_chan *vc, struct dma_device *dmadev);

struct xdx_virt_dma_desc *xdx_vchan_find_desc(struct xdx_virt_dma_chan *, dma_cookie_t);
extern dma_cookie_t xdx_vchan_tx_submit(struct dma_async_tx_descriptor *);
extern int xdx_vchan_tx_desc_free(struct dma_async_tx_descriptor *);

/**
 * vchan_tx_prep - prepare a descriptor
 * @vc: virtual channel allocating this descriptor
 * @vd: virtual descriptor to prepare
 * @tx_flags: flags argument passed in to prepare function
 */
static inline struct dma_async_tx_descriptor *xdx_vchan_tx_prep(struct xdx_virt_dma_chan *vc,
	struct xdx_virt_dma_desc *vd, unsigned long tx_flags)
{
	unsigned long flags;

	dma_async_tx_descriptor_init(&vd->tx, &vc->chan);
	vd->tx.flags = tx_flags;
	vd->tx.tx_submit = xdx_vchan_tx_submit;
#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 10, 0)
	vd->tx.desc_free = xdx_vchan_tx_desc_free;
#endif

	vd->tx_result.result = DMA_TRANS_NOERROR;
	vd->tx_result.residue = 0;

	spin_lock_irqsave(&vc->lock, flags);
	list_add_tail(&vd->node, &vc->desc_allocated);
	spin_unlock_irqrestore(&vc->lock, flags);

	return &vd->tx;
}

/**
 * vchan_issue_pending - move submitted descriptors to issued list
 * @vc: virtual channel to update
 *
 * vc.lock must be held by caller
 */
static inline bool xdx_vchan_issue_pending(struct xdx_virt_dma_chan *vc)
{
	list_splice_tail_init(&vc->desc_submitted, &vc->desc_issued);
	return !list_empty(&vc->desc_issued);
}

/**
 * vchan_cookie_complete - report completion of a descriptor
 * @vd: virtual descriptor to update
 *
 * vc.lock must be held by caller
 */
static inline void xdx_vchan_cookie_complete(struct xdx_virt_dma_desc *vd)
{
	struct xdx_virt_dma_chan *vc = xdx_to_virt_chan(vd->tx.chan);
	dma_cookie_t cookie;

	cookie = vd->tx.cookie;
	list_del(&vd->node);
	dma_cookie_complete(&vd->tx);
	dev_vdbg(vc->chan.device->dev, "xdx_vchan_cookie_complete txd %p[%x]: marked complete\n",
		 vd, cookie);
	list_add_tail(&vd->node, &vc->desc_completed);

	tasklet_schedule(&vc->task);
}

/**
 * vchan_vdesc_fini - Free or reuse a descriptor
 * @vd: virtual descriptor to free/reuse
 */
static inline void xdx_vchan_vdesc_fini(struct xdx_virt_dma_desc *vd)
{
	struct xdx_virt_dma_chan *vc = xdx_to_virt_chan(vd->tx.chan);
#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 10, 0)
	if (dmaengine_desc_test_reuse(&vd->tx))
		list_add(&vd->node, &vc->desc_allocated);
	else
#endif
		vc->desc_free(vd);
}

/**
 * vchan_cyclic_callback - report the completion of a period
 * @vd: virtual descriptor
 */
static inline void xdx_vchan_cyclic_callback(struct xdx_virt_dma_desc *vd)
{
	struct xdx_virt_dma_chan *vc = xdx_to_virt_chan(vd->tx.chan);

	vc->cyclic = vd;
	tasklet_schedule(&vc->task);
}

/**
 * vchan_terminate_vdesc - Disable pending cyclic callback
 * @vd: virtual descriptor to be terminated
 *
 * vc.lock must be held by caller
 */
static inline void xdx_vchan_terminate_vdesc(struct xdx_virt_dma_desc *vd)
{
	struct xdx_virt_dma_chan *vc = xdx_to_virt_chan(vd->tx.chan);

	/* free up stuck descriptor */
	if (vc->vd_terminated)
		xdx_vchan_vdesc_fini(vc->vd_terminated);

	vc->vd_terminated = vd;
	if (vc->cyclic == vd)
		vc->cyclic = NULL;
}

/**
 * vchan_next_desc - peek at the next descriptor to be processed
 * @vc: virtual channel to obtain descriptor from
 *
 * vc.lock must be held by caller
 */
static inline struct xdx_virt_dma_desc *xdx_vchan_next_desc(struct xdx_virt_dma_chan *vc)
{
	return list_first_entry_or_null(&vc->desc_issued,
					struct xdx_virt_dma_desc, node);
}

/**
 * vchan_get_all_descriptors - obtain all submitted and issued descriptors
 * @vc: virtual channel to get descriptors from
 * @head: list of descriptors found
 *
 * vc.lock must be held by caller
 *
 * Removes all submitted and issued descriptors from internal lists, and
 * provides a list of all descriptors found
 */
static inline void xdx_vchan_get_all_descriptors(struct xdx_virt_dma_chan *vc,
	struct list_head *head)
{
	list_splice_tail_init(&vc->desc_allocated, head);
	list_splice_tail_init(&vc->desc_submitted, head);
	list_splice_tail_init(&vc->desc_issued, head);
	list_splice_tail_init(&vc->desc_issued_pending, head);
	list_splice_tail_init(&vc->desc_completed, head);
}

static inline void xdx_vchan_free_chan_resources(struct xdx_virt_dma_chan *vc)
{
#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 10, 0)
	struct xdx_virt_dma_desc *vd;
#endif
	unsigned long flags;
	LIST_HEAD(head);

	spin_lock_irqsave(&vc->lock, flags);
	xdx_vchan_get_all_descriptors(vc, &head);
#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 10, 0)
	list_for_each_entry(vd, &head, node)
		dmaengine_desc_clear_reuse(&vd->tx);
#endif
	spin_unlock_irqrestore(&vc->lock, flags);

	xdx_vchan_dma_desc_free_list(vc, &head);
}

/**
 * vchan_synchronize() - synchronize callback execution to the current context
 * @vc: virtual channel to synchronize
 *
 * Makes sure that all scheduled or active callbacks have finished running. For
 * proper operation the caller has to ensure that no new callbacks are scheduled
 * after the invocation of this function started.
 * Free up the terminated cyclic descriptor to prevent memory leakage.
 */
static inline void xdx_vchan_synchronize(struct xdx_virt_dma_chan *vc)
{
	unsigned long flags;

	tasklet_kill(&vc->task);

	spin_lock_irqsave(&vc->lock, flags);
	if (vc->vd_terminated) {
		xdx_vchan_vdesc_fini(vc->vd_terminated);
		vc->vd_terminated = NULL;
	}
	spin_unlock_irqrestore(&vc->lock, flags);
}

#endif
