/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#include <linux/device.h>
#include <linux/dmaengine.h>
#include <linux/module.h>
#include <linux/spinlock.h>

#include "virt-dma.h"
#include "xdx_dma.h"
#include "xdx_dma_drv.h"

static struct xdx_virt_dma_desc *xdx_to_virt_desc(struct dma_async_tx_descriptor *tx)
{
	return container_of(tx, struct xdx_virt_dma_desc, tx);
}

dma_cookie_t xdx_vchan_tx_submit(struct dma_async_tx_descriptor *tx)
{
	struct xdx_virt_dma_chan *vc = xdx_to_virt_chan(tx->chan);
	struct xdx_virt_dma_desc *vd = xdx_to_virt_desc(tx);
	unsigned long flags;
	dma_cookie_t cookie;
	int ret;
//	struct xdx_vchan *vchan;
	ret = xdx_dma_package_desc(tx);
	if (ret < 0)
		return (dma_cookie_t)ret;
	spin_lock_irqsave(&vc->lock, flags);
	cookie = dma_cookie_assign(tx);

	list_move_tail(&vd->node, &vc->desc_submitted);
	spin_unlock_irqrestore(&vc->lock, flags);

	//dev_dbg(vc->chan.device->dev, "vchan %p: txd %p[%x]: submitted\n",
	//	vc, vd, cookie);

//	vchan = to_xdx_vchan(tx->chan);
	//vchan->desc_cnt++;
	return cookie;
}
EXPORT_SYMBOL_GPL(xdx_vchan_tx_submit);

/**
 * vchan_tx_desc_free - free a reusable descriptor
 * @tx: the transfer
 *
 * This function frees a previously allocated reusable descriptor. The only
 * other way is to clear the DMA_CTRL_REUSE flag and submit one last time the
 * transfer.
 *
 * Returns 0 upon success
 */
int xdx_vchan_tx_desc_free(struct dma_async_tx_descriptor *tx)
{
	struct xdx_virt_dma_chan *vc = xdx_to_virt_chan(tx->chan);
	struct xdx_virt_dma_desc *vd = xdx_to_virt_desc(tx);
	unsigned long flags;

	spin_lock_irqsave(&vc->lock, flags);
	list_del(&vd->node);
	spin_unlock_irqrestore(&vc->lock, flags);

	//dev_dbg(vc->chan.device->dev, "vchan %p: txd %p[%x]: freeing\n",
	//	vc, vd, vd->tx.cookie);
	vc->desc_free(vd);
	return 0;
}
EXPORT_SYMBOL_GPL(xdx_vchan_tx_desc_free);

struct xdx_virt_dma_desc *xdx_vchan_find_desc(struct xdx_virt_dma_chan *vc,
	dma_cookie_t cookie)
{
	struct xdx_virt_dma_desc *vd;

	list_for_each_entry(vd, &vc->desc_issued, node)
		if (vd->tx.cookie == cookie)
			return vd;

	return NULL;
}
EXPORT_SYMBOL_GPL(xdx_vchan_find_desc);

/*
 * This tasklet handles the completion of a DMA descriptor by
 * calling its callback and freeing it.
 */
static void xdx_vchan_complete(unsigned long arg)
{
	struct xdx_virt_dma_chan *vc = (struct xdx_virt_dma_chan *)arg;
	struct xdx_virt_dma_desc *vd, *_vd;
	struct dmaengine_desc_callback cb;
	struct xdx_dma_dev *sdc;

	LIST_HEAD(head);
	spin_lock_irq(&vc->lock);
	list_splice_tail_init(&vc->desc_completed, &head);

	spin_unlock_irq(&vc->lock);

	list_for_each_entry_safe(vd, _vd, &head, node) {
		dmaengine_desc_get_callback(&vd->tx, &cb);

		if (vd->tx_result.result != DMA_TRANS_NOERROR) {
			sdc = to_xdx_dma_dev(vc->chan.device);
			sdc->ops->print_reg(sdc);
		}
		list_del(&vd->node);
		dmaengine_desc_callback_invoke(&cb, &vd->tx_result);
		xdx_vchan_vdesc_fini(vd);
	}
}

void xdx_vchan_dma_desc_free_list(struct xdx_virt_dma_chan *vc, struct list_head *head)
{
	struct xdx_virt_dma_desc *vd, *_vd;

	list_for_each_entry_safe(vd, _vd, head, node) {
#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 10, 0)
		if (dmaengine_desc_test_reuse(&vd->tx)) {
			list_move_tail(&vd->node, &vc->desc_allocated);
			XDX_DMA_DBG(NULL, "txd %p: reuse freeing\n", vd);
		} else
#endif
		{
			XDX_DMA_DBG(NULL, "txd %p: freeing\n", vd);
			list_del(&vd->node);
			vc->desc_free(vd);
		}
	}
	XDX_DMA_DBG(vc->chan.device->dev, "txd freeing Done\n");
}
EXPORT_SYMBOL_GPL(xdx_vchan_dma_desc_free_list);

void xdx_vchan_init(struct xdx_virt_dma_chan *vc, struct dma_device *dmadev)
{
	dma_cookie_init(&vc->chan);

	spin_lock_init(&vc->lock);
	INIT_LIST_HEAD(&vc->desc_allocated);
	INIT_LIST_HEAD(&vc->desc_submitted);
	INIT_LIST_HEAD(&vc->desc_issued);
	INIT_LIST_HEAD(&vc->desc_issued_pending);
	INIT_LIST_HEAD(&vc->desc_completed);

	tasklet_init(&vc->task, xdx_vchan_complete, (unsigned long)vc);

	vc->chan.device = dmadev;
	list_add_tail(&vc->chan.device_node, &dmadev->channels);
}
EXPORT_SYMBOL_GPL(xdx_vchan_init);


MODULE_LICENSE("GPL");

