/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

/*xdxgpu headers */
#include "xdx_gfx.h"
#include "xdx_drm.h"

/* pvr headers */
#include "pvrversion.h"
//#include "services_kernel_client.h"
#include "syscommon.h"
#include "rgx_fwif_shared.h"
#include "pvr_debug.h"
#include "rgxdevice.h"
#include "power.h"
#include "pvrsrv.h"

/* bridge headers */
#ifdef SUPPORT_NATIVE_FENCE_SYNC
#include "pvr_sync.h"
#endif
#include "ospvr_gputrace.h"
#include "pvr_bridge_k.h"
#include "pvr_fence.h"
#include "env_connection.h"
#include "pvr_sync_ioctl_drm.h"
#include "private_data.h"
#include "pvr_device.h"
#include "rgxfwutils.h"
#include "dm_irq.h"

#ifdef XDXGPU_ENABLE_MTS_KICK_IN_VSYNC_IRQ
static void dc_irq_process_mts_kick(void *user_data)
{
	PVRSRV_DEVICE_NODE *psDeviceNode =(PVRSRV_DEVICE_NODE *)user_data;
	PVRSRV_RGXDEV_INFO *psDevInfo = psDeviceNode->pvDevice;

	RGXMTSScheduleWrite(psDevInfo,(RGXFWIF_DM_GP & ~RGX_CR_MTS_SCHEDULE_DM_CLRMSK));
}
#endif

CONNECTION_DATA *LinuxServicesConnectionFromFile(struct file *pFile)
{
	if (pFile)
	{
		struct drm_file *psDRMFile = pFile->private_data;
        struct xdx_drm_fpriv *fpriv = psDRMFile->driver_priv;
		PVRSRV_CONNECTION_PRIV *psConnectionPriv = fpriv->pvr_fpriv;

		return (CONNECTION_DATA*)psConnectionPriv->pvConnectionData;
	}

	return NULL;
}

CONNECTION_DATA *LinuxSyncConnectionFromFile(struct file *pFile)
{
	if (pFile)
	{
		struct drm_file *psDRMFile = pFile->private_data;
        struct xdx_drm_fpriv *fpriv = psDRMFile->driver_priv;
		PVRSRV_CONNECTION_PRIV *psConnectionPriv = fpriv->pvr_fpriv;

#if (PVRSRV_DEVICE_INIT_MODE == PVRSRV_LINUX_DEV_INIT_ON_CONNECT)
		return (CONNECTION_DATA*)psConnectionPriv->pvConnectionData;
#else
		return (CONNECTION_DATA*)psConnectionPriv->pvSyncConnectionData;
#endif
	}

	return NULL;
}

/**************************************************************************/ /*!
@Function     PVRSRVDeviceInit
@Description  Common device related initialisation.
@Input        psDeviceNode  The device node for which initialisation should be
                            performed
@Return       int           0 on success and a Linux error code otherwise
*/ /***************************************************************************/
static int PVRSRVDeviceInit(PVRSRV_DEVICE_NODE *psDeviceNode)
{
#if defined(SUPPORT_NATIVE_FENCE_SYNC)
	{
		PVRSRV_ERROR eError = pvr_sync_device_init(psDeviceNode->psDevConfig->pvOSDevice);
		if (eError != PVRSRV_OK)
		{
			PVR_DPF((PVR_DBG_ERROR, "%s: unable to create sync (%d)",
					 __func__, eError));
			return -EBUSY;
		}
	}
#endif

#if defined(SUPPORT_RGX)
	{
		int error = PVRGpuTraceInitDevice(psDeviceNode);
		if (error != 0)
		{
			PVR_DPF((PVR_DBG_WARNING,
				 "%s: failed to initialise PVR GPU Tracing on device%d (%d)",
				 __func__, psDeviceNode->sDevId.i32OsDeviceID, error));
		}
	}
#endif

	return 0;
}

/**************************************************************************/ /*!
@Function     PVRSRVDeviceDeinit
@Description  Common device related de-initialisation.
@Input        psDeviceNode  The device node for which de-initialisation should
                            be performed
@Return       void
*/ /***************************************************************************/
static void PVRSRVDeviceDeinit(PVRSRV_DEVICE_NODE *psDeviceNode)
{
#if defined(SUPPORT_RGX)
	PVRGpuTraceDeInitDevice(psDeviceNode);
#endif

#if defined(SUPPORT_NATIVE_FENCE_SYNC)
	pvr_sync_device_deinit(psDeviceNode->psDevConfig->pvOSDevice);
#endif
	pvr_fence_cleanup();
}

/**************************************************************************/ /*!
@Function     PVRSRVDeviceShutdown
@Description  Common device shutdown.
@Input        xdev  The device node representing the device that should
                     be shutdown
@Return       void
*/ /***************************************************************************/

static void PVRSRVDeviceShutdown(struct xdx_device *xdev)
{
	PVRSRV_DEVICE_NODE *psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;

	/* Since this is a shutdown request ignore the returned error and try to
	 * to power off the device. This is done because there is no way of
	 * signalling the OS that this call failed. */
	(void) LinuxBridgeBlockClientsAccess(xdev, IMG_TRUE);

	/* Passing PVRSRV_POWER_FLAGS_NONE as there are no special actions required
	 * from the shutdown call beside the regular device power off. */
	(void) PVRSRVSetDeviceSystemPowerState(psDeviceNode,
	                                       PVRSRV_SYS_POWER_STATE_OFF,
	                                       PVRSRV_POWER_FLAGS_NONE);
}

/**************************************************************************/ /*!
@Function     PVRSRVDeviceSuspend
@Description  Common device suspend.
@Input        xdev  The device node representing the device that should
                            be suspended
@Return       int           0 on success and a Linux error code otherwise
*/ /***************************************************************************/
static int PVRSRVDeviceSuspend(struct xdx_device *xdev)
{
	PVRSRV_DEVICE_NODE *psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;
	PVRSRV_ERROR eError;

	/* LinuxBridgeBlockClientsAccess prevents processes from using the driver
	 * while it's suspended (this is needed for Android)
	 * XDXGPU UPDATE: per OS distribution venders' request
	 * sometimes xdxgpu.ko would be packed into initrd
	 * after loading hibernation image when exiting hibernation
	 * devcie pm_freeze callback will be called
	 * without stopping kernel and user threads 
	 * so LinuxBridgeBlockClientsAccess(xdev, IMG_FALSE) would be blocked
	 * make it non-blocking as Android is not on support list
	 */
	eError = LinuxBridgeBlockClientsAccess(xdev, IMG_TRUE);
	PVR_LOG_RETURN_IF_FALSE_VA(eError == PVRSRV_OK, -EFAULT,
	                           "LinuxBridgeBlockClientsAccess() failed with error %u",
	                           eError);

#if defined(SUPPORT_AUTOVZ)
	/* To allow the driver to power down the GPU under AutoVz, the firmware must
	 * declared as offline, otherwise all power requests will be ignored. */
	psDeviceNode->bAutoVzFwIsUp = IMG_FALSE;
#endif

	if (PVRSRVSetDeviceSystemPowerState(psDeviceNode,
										PVRSRV_SYS_POWER_STATE_OFF,
										PVRSRV_POWER_FLAGS_SUSPEND_REQ) != PVRSRV_OK)
	{
		/* Ignore return error as we're already returning an error here. */
		(void) LinuxBridgeUnblockClientsAccess(xdev);
		return -EINVAL;
	}

	return 0;
}

/**************************************************************************/ /*!
@Function     PVRSRVDeviceResume
@Description  Common device resume.
@Input        psDeviceNode  The device node representing the device that should
                            be resumed
@Return       int           0 on success and a Linux error code otherwise
*/ /***************************************************************************/
static int PVRSRVDeviceResume(struct xdx_device *xdev)
{
	PVRSRV_DEVICE_NODE *psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;

	if (PVRSRVSetDeviceSystemPowerState(psDeviceNode,
										PVRSRV_SYS_POWER_STATE_ON,
										PVRSRV_POWER_FLAGS_RESUME_REQ) != PVRSRV_OK)
	{
		return -EINVAL;
	}

	/* Ignore return error. We should proceed even if this fails. */
	(void) LinuxBridgeUnblockClientsAccess(xdev);

	/*
	 * Reprocess the device queues in case commands were blocked during
	 * suspend.
	 */
	if (psDeviceNode->eDevState == PVRSRV_DEVICE_STATE_ACTIVE)
	{
		PVRSRVCheckStatus(NULL);
	}

	return 0;
}


static int PVRSRVDeviceServicesOpen(PVRSRV_DEVICE_NODE *psDeviceNode,
                             struct xdx_drm_fpriv *fpriv)
{
	static DEFINE_MUTEX(sDeviceInitMutex);
	PVRSRV_DATA *psPVRSRVData = PVRSRVGetPVRSRVData();
	ENV_CONNECTION_PRIVATE_DATA sPrivData;
	PVRSRV_CONNECTION_PRIV *psConnectionPriv;
	PVRSRV_ERROR eError;
	int iErr = 0;

	if (!psPVRSRVData)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: No device data", __func__));
		iErr = -ENODEV;
		goto out;
	}

	mutex_lock(&sDeviceInitMutex);
	/*
	 * If the first attempt already set the state to bad,
	 * there is no point in going the second time, so get out
	 */
	if (psDeviceNode->eDevState == PVRSRV_DEVICE_STATE_BAD)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: Driver already in bad state. Device open failed.",
				 __func__));
		iErr = -ENODEV;
		mutex_unlock(&sDeviceInitMutex);
		goto out;
	}

	if (fpriv->pvr_fpriv == NULL)
	{
		/* Allocate psConnectionPriv (stores private data and release pfn under driver_priv) */
		psConnectionPriv = kzalloc(sizeof(*psConnectionPriv), GFP_KERNEL);
		if (!psConnectionPriv)
		{
			PVR_DPF((PVR_DBG_ERROR, "%s: No memory to allocate driver_priv data", __func__));
			iErr = -ENOMEM;
			mutex_unlock(&sDeviceInitMutex);
			goto fail_alloc_connection_priv;
		}
	}
	else
	{
		psConnectionPriv = (PVRSRV_CONNECTION_PRIV*)fpriv->pvr_fpriv;
	}

	if (psDeviceNode->eDevState == PVRSRV_DEVICE_STATE_INIT)
	{
		eError = PVRSRVCommonDeviceInitialise(psDeviceNode);
		if (eError != PVRSRV_OK)
		{
			PVR_DPF((PVR_DBG_ERROR, "%s: Failed to initialise device (%s)",
					 __func__, PVRSRVGetErrorString(eError)));
			iErr = -ENODEV;
			mutex_unlock(&sDeviceInitMutex);
			goto fail_device_init;
		}

#if defined(SUPPORT_RGX)
		PVRGpuTraceInitIfEnabled(psDeviceNode);
#endif
	}
	mutex_unlock(&sDeviceInitMutex);

	sPrivData.psDevNode = psDeviceNode;

	/*
	 * Here we pass the file pointer which will passed through to our
	 * OSConnectionPrivateDataInit function where we can save it so
	 * we can back reference the file structure from its connection
	 */
	eError = PVRSRVCommonConnectionConnect(&psConnectionPriv->pvConnectionData,
	                                       (void *)&sPrivData);
	if (eError != PVRSRV_OK)
	{
		iErr = -ENOMEM;
		goto fail_connect;
	}

#if (PVRSRV_DEVICE_INIT_MODE == PVRSRV_LINUX_DEV_INIT_ON_CONNECT)
	psConnectionPriv->pfDeviceRelease = PVRSRVCommonConnectionDisconnect;
#endif
	fpriv->pvr_fpriv = (void*)psConnectionPriv;
	goto out;

fail_connect:
fail_device_init:
	kfree(psConnectionPriv);
fail_alloc_connection_priv:
out:
	return iErr;
}

#if defined(SUPPORT_NATIVE_FENCE_SYNC) && !defined(USE_PVRSYNC_DEVNODE)
static void PVRSRVSyncConnectionRelease(void *pvDataPtr)
{
	CONNECTION_DATA *psConnection = pvDataPtr;

	pvr_sync_close(psConnection);
	OSConnectionPrivateDataDeInit(psConnection->hOsPrivateData);
	kfree(psConnection);
}
#endif

static int PVRSRVDeviceSyncOpen(PVRSRV_DEVICE_NODE *psDeviceNode,
                                struct xdx_drm_fpriv *fpriv)
{
	PVRSRV_DATA *psPVRSRVData = PVRSRVGetPVRSRVData();
	CONNECTION_DATA *psConnection = NULL;
	ENV_CONNECTION_PRIVATE_DATA sPrivData;
	PVRSRV_CONNECTION_PRIV *psConnectionPriv;
	PVRSRV_ERROR eError;
	int iErr = 0;

	if (!psPVRSRVData)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: No device data", __func__));
		iErr = -ENODEV;
		goto out;
	}

	if (fpriv->pvr_fpriv == NULL)
	{
		/* Allocate psConnectionPriv (stores private data and release pfn under driver_priv) */
		psConnectionPriv = kzalloc(sizeof(*psConnectionPriv), GFP_KERNEL);
		if (!psConnectionPriv)
		{
			PVR_DPF((PVR_DBG_ERROR, "%s: No memory to allocate driver_priv data", __func__));
			iErr = -ENOMEM;
			goto out;
		}
	}
	else
	{
		psConnectionPriv = (PVRSRV_CONNECTION_PRIV*)fpriv->pvr_fpriv;
	}

	/* Allocate connection data area, no stats since process not registered yet */
	psConnection = kzalloc(sizeof(*psConnection), GFP_KERNEL);
	if (!psConnection)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: No memory to allocate connection data", __func__));
		iErr = -ENOMEM;
		goto fail_alloc_connection;
	}

#if (PVRSRV_DEVICE_INIT_MODE == PVRSRV_LINUX_DEV_INIT_ON_CONNECT)
	psConnectionPriv->pvConnectionData = (void*)psConnection;
#else
	psConnectionPriv->pvSyncConnectionData = (void*)psConnection;
#endif

	sPrivData.psDevNode = psDeviceNode;

	/* Call environment specific connection data init function */
	eError = OSConnectionPrivateDataInit(&psConnection->hOsPrivateData, &sPrivData);
	if (eError != PVRSRV_OK)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: OSConnectionPrivateDataInit() failed (%s)",
		        __func__, PVRSRVGetErrorString(eError)));
		goto fail_private_data_init;
	}

#if defined(SUPPORT_NATIVE_FENCE_SYNC) && !defined(USE_PVRSYNC_DEVNODE)
#if (PVRSRV_DEVICE_INIT_MODE == PVRSRV_LINUX_DEV_INIT_ON_CONNECT)
	iErr = pvr_sync_open(psConnectionPriv->pvConnectionData, fpriv->dfile);
#else
	iErr = pvr_sync_open(psConnectionPriv->pvSyncConnectionData, fpriv->dfile);
#endif
	if (iErr)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: pvr_sync_open() failed(%d)",
				__func__, iErr));
		goto fail_pvr_sync_open;
	}
#endif

#if defined(SUPPORT_NATIVE_FENCE_SYNC) && !defined(USE_PVRSYNC_DEVNODE)
#if (PVRSRV_DEVICE_INIT_MODE == PVRSRV_LINUX_DEV_INIT_ON_CONNECT)
	psConnectionPriv->pfDeviceRelease = PVRSRVSyncConnectionRelease;
#endif
#endif
	fpriv->pvr_fpriv = psConnectionPriv;
	goto out;

#if defined(SUPPORT_NATIVE_FENCE_SYNC) && !defined(USE_PVRSYNC_DEVNODE)
fail_pvr_sync_open:
	OSConnectionPrivateDataDeInit(psConnection->hOsPrivateData);
#endif
fail_private_data_init:
	kfree(psConnection);
fail_alloc_connection:
	kfree(psConnectionPriv);
out:
	return iErr;
}

int xdx_gfx_drm_open(struct xdx_device *xdev, struct xdx_drm_fpriv *fpriv)
{
	int ret = 0;

#if (PVRSRV_DEVICE_INIT_MODE != PVRSRV_LINUX_DEV_INIT_ON_CONNECT)
	ret = PVRSRVDeviceServicesOpen(xdev->gfx.pvrdev, fpriv);
#endif
	return ret;
}

/**************************************************************************/ /*!
@Function     PVRSRVDeviceRelease
@Description  Common device release.
@Input        psDeviceNode  The device node for the device that the given file
                            represents
@Input        psDRMFile     The DRM file data that's being released
@Return       void
*/ /***************************************************************************/
static void PVRSRVDeviceRelease(PVRSRV_DEVICE_NODE *psDeviceNode,
                         struct xdx_drm_fpriv *fpriv)
{
	PVR_UNREFERENCED_PARAMETER(psDeviceNode);

	if (fpriv->pvr_fpriv)
	{
		PVRSRV_CONNECTION_PRIV *psConnectionPriv = (PVRSRV_CONNECTION_PRIV*)fpriv->pvr_fpriv;

		if (psConnectionPriv->pvConnectionData || psConnectionPriv->pvSyncConnectionData)
		{
#if (PVRSRV_DEVICE_INIT_MODE == PVRSRV_LINUX_DEV_INIT_ON_CONNECT)
			if (psConnectionPriv->pfDeviceRelease)
			{
				psConnectionPriv->pfDeviceRelease(psConnectionPriv->pvConnectionData);
			}
#else
			if (psConnectionPriv->pvConnectionData)
				PVRSRVCommonConnectionDisconnect(psConnectionPriv->pvConnectionData);

#if defined(SUPPORT_NATIVE_FENCE_SYNC) && !defined(USE_PVRSYNC_DEVNODE)
			if (psConnectionPriv->pvSyncConnectionData)
				PVRSRVSyncConnectionRelease(psConnectionPriv->pvSyncConnectionData);
#endif
#endif
		}

		kfree(fpriv->pvr_fpriv);
		fpriv->pvr_fpriv = NULL;
	}
}

void xdx_gfx_drm_release(struct xdx_device *xdev, struct xdx_drm_fpriv *fpriv)
{
    PVRSRVDeviceRelease(xdev->gfx.pvrdev, fpriv);
}

int drm_pvr_srvkm_init(struct drm_device *dev, void *arg, struct drm_file *psDRMFile)
{
	struct drm_pvr_srvkm_init_data *data = arg;
	struct xdx_device *xdev = drm_to_xdev(dev);
	struct xdx_drm_fpriv *fpriv = psDRMFile->driver_priv;
	int iErr = 0;

	switch (data->init_module)
	{
		case PVR_SRVKM_SYNC_INIT:
		{
			iErr = PVRSRVDeviceSyncOpen(xdev->gfx.pvrdev, fpriv);
			break;
		}
		case PVR_SRVKM_SERVICES_INIT:
		{
			iErr = PVRSRVDeviceServicesOpen(xdev->gfx.pvrdev, fpriv);
			break;
		}
		default:
		{
			PVR_DPF((PVR_DBG_ERROR, "%s: invalid init_module (%d)",
			        __func__, data->init_module));
			iErr = -EINVAL;
		}
	}

	return iErr;
}

static int volcanic_early_init(void *handle)
{
	return 0;
}

static int volcanic_late_init(void *handle)
{
	return 0;
}

static int volcanic_sw_init(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
	struct drm_device *ddev = xdev_to_drm(xdev);
	int ret;
	enum PVRSRV_ERROR_TAG srv_err;
	PVRSRV_DEVICE_NODE *psDeviceNode = NULL;
	int deviceId;

	if (ddev->render)
		deviceId = ddev->render->index;
	else /* when render node is NULL, fallback to primary node */
		deviceId = ddev->primary->index;

	srv_err = PVRSRVCommonDeviceCreate(ddev->dev, deviceId, &psDeviceNode);
	if (srv_err != PVRSRV_OK) {
		dev_err(xdev->dev,
			"failed to create device node for gfx (%d)\n", srv_err);
		if (srv_err == PVRSRV_ERROR_PROBE_DEFER)
			ret = -EPROBE_DEFER;
		else
			ret = -ENODEV;
		goto err_end;
	}

	xdev->gfx.pvrdev = psDeviceNode;
	psDeviceNode->uiAsicType = xdev->asic_type;

	ret = PVRSRVDeviceInit(psDeviceNode);
	if (ret) {
		dev_err(xdev->dev, "gfx initialisation failed (err=%d)\n", ret);
		goto err_device_destroy;
	}

	return 0;

err_device_destroy:
	PVRSRVCommonDeviceDestroy(psDeviceNode);

err_end:
	return ret;
}

static int volcanic_sw_fini(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
	PVRSRV_DEVICE_NODE *psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;

	PVRSRVDeviceDeinit(psDeviceNode);
	PVRSRVCommonDeviceDestroy(psDeviceNode);

	return 0;
}

static int volcanic_early_fini(void *handle)
{
	return 0;
}

static int volcanic_hw_init(void *handle)
{
	int ret = 0;
#ifdef XDXGPU_ENABLE_MTS_KICK_IN_VSYNC_IRQ
	int i = 0;
#endif
	struct xdx_device *xdev = (struct xdx_device *)handle;

#if (PVRSRV_DEVICE_INIT_MODE == PVRSRV_LINUX_DEV_INIT_ON_PROBE)
	struct drm_device *ddev = xdev_to_drm(xdev);
	enum PVRSRV_ERROR_TAG srv_err;
	PVRSRV_DEVICE_NODE *psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;

	srv_err = PVRSRVCommonDeviceInitialise(psDeviceNode);
	if (srv_err != PVRSRV_OK) {
		ret = -ENODEV;
		dev_err(xdev->dev,
			"gfx common initialisation failed (err=%d)\n", ret);
	}
#endif
#ifdef XDXGPU_ENABLE_MTS_KICK_IN_VSYNC_IRQ
	for(i = 0;i < xdev->dm.dc_num; i++)
	{
		ret = dm_register_dc_vsync_irq_callback(xdev->dm.dc[i],dc_irq_process_mts_kick,xdev->gfx.pvrdev);
		if(ret)
		{
			dev_err(xdev->dev,
				"dm dc vsync irq register failed (err=%d)\n", ret);
			return ret;
		}
	}
#endif
	PVR_UNREFERENCED_PARAMETER(xdev);
	return ret;
}

static int volcanic_hw_fini(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
#ifdef XDXGPU_ENABLE_MTS_KICK_IN_VSYNC_IRQ
	int ret = 0;
	int i = 0;
	for(i = 0;i < xdev->dm.dc_num; i++)
	{
		ret = dm_unregister_dc_vsync_irq_callback(xdev->dm.dc[i],dc_irq_process_mts_kick);
		if(ret)
		{
			dev_err(xdev->dev,
				"dm dc vsync irq unregister failed (err=%d)\n", ret);
			return ret;
		}
	}
#endif
	PVRSRVDeviceShutdown(xdev);

	return 0;
}

static int volcanic_suspend(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
        int ret = 0;

	if (atomic_read(&xdev->in_s3) || 
            atomic_read(&xdev->in_s4)) {
	        ret = PVRSRVDeviceSuspend(xdev);
	} else {
	        PVRSRVDeviceShutdown(xdev);
	}
        xdxgpu_irq_set(xdev, xdev->gfx.irq_src, 0, XDX_IRQ_STATE_DISABLE);

        return ret;
}

static int volcanic_resume(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;

	if (xdev->asic_type == CHIP_PANGU_B0) {
                if (atomic_read(&xdev->in_s4)) {
                        /* 
                         * FIXME:
                         *  On PANGU B0, META will hang during S4 thaw 
                         *  Workaround the problem by skip thaw temporarily
                         *
                         *  Thaw is mainly used for write S4 image to disk
                         *  No major function loss after workaround
                         */
                        return 0;
                }
        }

	xdxgpu_irq_set(xdev, xdev->gfx.irq_src, 0, XDX_IRQ_STATE_ENABLE);
	return PVRSRVDeviceResume(xdev);
}

static bool volcanic_is_idle(void *handle)
{
	return true;
}

static int volcanic_wait_for_idle(void *handle)
{
	return 0;
}

const struct xdx_ip_block_version volcanic_block_version = {
	.name = "gfx",
	.type = XDX_IP_BLOCK_TYPE_GFX,
	.major = 1,
	.minor = 0,
	.rev = 0,
	.funcs = {
		.early_init = volcanic_early_init,
		.late_init = volcanic_late_init,
		.sw_init = volcanic_sw_init,
		.sw_fini = volcanic_sw_fini,
		.early_fini = volcanic_early_fini,
		.hw_init = volcanic_hw_init,
		.hw_fini = volcanic_hw_fini,
		.suspend = volcanic_suspend,
		.resume = volcanic_resume,
		.is_idle = volcanic_is_idle,
		.wait_for_idle = volcanic_wait_for_idle,
	}
};
