/*************************************************************************/ /*!
@File
@Title          PowerVR devfreq device implementation
@Copyright      Copyright (c) Imagination Technologies Ltd. All Rights Reserved
@Description    Linux module setup
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

#if !defined(NO_HARDWARE)

#include <linux/devfreq.h>
#if defined(SUPPORT_LINUX_DVFS) && defined(CONFIG_DEVFREQ_THERMAL)
#include "pm/xdx_thm.h"
#endif /* SUPPORT_LINUX_DVFS && CONFIG_DEVFREQ_THERMAL */
#include <linux/version.h>
#include <linux/device.h>
#include <drm/drm.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0)) || (defined(REDHAT_83))
#include <drm/drm_device.h>
#include <drm/drm_file.h>
#else
#include <drm/drmP.h>
#endif

#include "power.h"
#include "pvrsrv.h"
#include "pvrsrv_device.h"

#include "rgxdevice.h"
#include "rgxinit.h"
#include "sofunc_rgx.h"

#include "syscommon.h"

#include "pvr_dvfs_device.h"

#include "kernel_compatibility.h"

#ifdef XDXGPU
#include "pm/xdx_pm.h"

#include "xdx_device.h"
#include "xdx_gfx_convert.h"

static inline int xdx_devfreq_update_stats(struct devfreq *df)
{
	if (!df->profile->get_dev_status)
		return -EINVAL;

	return df->profile->get_dev_status(df->dev.parent, &df->last_status);
}

static int xdx_devfreq_get_target_freq(struct devfreq *df,
					unsigned long *freq)
{
	int err;
	struct devfreq_dev_status *stat;
	unsigned long long a, b;
	unsigned int dfso_upthreshold = 90;
	unsigned int dfso_downdifferential = 5;
	struct devfreq_simple_ondemand_data *data = df->data;

	err = xdx_devfreq_update_stats(df);
	if (err)
		return err;

	stat = &df->last_status;

	if (data) {
		if (data->upthreshold)
			dfso_upthreshold = data->upthreshold;
		if (data->downdifferential)
			dfso_downdifferential = data->downdifferential;
	}
	if (dfso_upthreshold > 100 ||
	    dfso_upthreshold < dfso_downdifferential)
		return -EINVAL;

	/* Assume MAX if it is going to be divided by zero */
	if (stat->total_time == 0) {
		*freq = ULONG_MAX;
		return 0;
	}

	/* Prevent overflow */
	if (stat->busy_time >= (1 << 24) || stat->total_time >= (1 << 24)) {
		stat->busy_time >>= 7;
		stat->total_time >>= 7;
	}

	/* Set to Max level if it's busy enough */
	if (stat->busy_time * 100 >
	    stat->total_time * dfso_upthreshold) {
		*freq = ULONG_MAX;
		return 0;
	}

	/* Set MAX if we do not know the initial frequency */
	if (stat->current_frequency == 0) {
		*freq = ULONG_MAX;
		return 0;
	}

	/* Keep the current frequency */
	if (stat->busy_time * 100 >
	    stat->total_time * (dfso_upthreshold - dfso_downdifferential)) {
		*freq = stat->current_frequency;
		return 0;
	}

	/* Set the desired frequency based on the load */
	a = stat->busy_time;
	a *= stat->current_frequency;
	b = div_u64(a, stat->total_time);
	b *= 100;
	b = div_u64(b, (dfso_upthreshold - dfso_downdifferential / 2));
	*freq = (unsigned long) b;

	return 0;
}

static int xdx_devfreq_event_handler(struct devfreq *devfreq,
				unsigned int event, void *data)
{
	switch (event) {
	case DEVFREQ_GOV_START:
		devfreq_monitor_start(devfreq);
		break;

	case DEVFREQ_GOV_STOP:
		devfreq_monitor_stop(devfreq);
		break;

	case DEVFREQ_GOV_UPDATE_INTERVAL:
	#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 7, 0))
		devfreq_update_interval(devfreq, (unsigned int *)data);
	#else
		devfreq_interval_update(devfreq, (unsigned int *)data);
	#endif
		break;

	case DEVFREQ_GOV_SUSPEND:
		devfreq_monitor_suspend(devfreq);
		break;

	case DEVFREQ_GOV_RESUME:
		devfreq_monitor_resume(devfreq);
		break;

	default:
		break;
	}

	return 0;
}

static struct devfreq_governor xdx_devfreq_dvfs = {
	.name = "xdx_ondemand",
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0))
    .attrs = BIT(0) | BIT(1),
#endif
	.get_target_freq = xdx_devfreq_get_target_freq,
	.event_handler = xdx_devfreq_event_handler,
};

PVRSRV_ERROR XdxDVFSGovInit(void)
{
	return devfreq_add_governor(&xdx_devfreq_dvfs);
}

PVRSRV_ERROR XdxDVFSGovDeinit(void)
{
	return devfreq_remove_governor(&xdx_devfreq_dvfs);
}

#endif /* XDXGPU */

static int _device_get_devid(struct device *dev)
{
	struct drm_device *ddev = dev_get_drvdata(dev);
	int deviceId;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 12, 0))
	/*
	 * Older kernels do not have render drm_minor member in drm_device,
	 * so we fallback to primary node for device identification
	 */
	deviceId = ddev->primary->index;
#else
	if (ddev->render)
		deviceId = ddev->render->index;
	else /* when render node is NULL, fallback to primary node */
		deviceId = ddev->primary->index;
#endif

	return deviceId;
}

#ifdef XDXGPU
void xdx_dvfs_enable(struct device *dev)
{
	int deviceId = _device_get_devid(dev);
	PVRSRV_DEVICE_NODE *psDeviceNode = PVRSRVGetDeviceInstanceByOSId(deviceId);
	IMG_DVFS_DEVICE *psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;

	if (!psDVFSDevice->bUserEnabled) {
		if (PVRSRV_OK != InitDVFSDevice(psDeviceNode)) {
			dev_err(dev, "xdx InitDVFSDevice failed\n");
			return;
		}

		psDVFSDevice->bUserEnabled = IMG_TRUE;
		if (PVRSRV_OK != RegisterDVFSDevice(psDeviceNode)) {
			DeinitDVFSDevice(psDeviceNode);
			psDVFSDevice->bUserEnabled = IMG_FALSE;
			dev_err(dev, "xdx RegisterDVFSDevice failed\n");
			return;
		}

		dev_info(dev, "%s : dvfs enabled\n", __func__);
	}
}

void xdx_dvfs_disable(struct device *dev)
{
	int err;
	u32 default_volt;
	u32 default_freq;
	u32 cur_freq;
	struct xdx_device *xdev;
	int deviceId = _device_get_devid(dev);
	IMG_DVFS_DEVICE_CFG	*psDVFSDeviceCfg = NULL;
	PVRSRV_DEVICE_NODE *psDeviceNode = PVRSRVGetDeviceInstanceByOSId(deviceId);
	IMG_DVFS_DEVICE *psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	psDVFSDeviceCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;

	if (!psDVFSDevice->bUserEnabled) {
		dev_err(psDeviceNode->psDevConfig->pvOSDevice, "%s : already disabled\n", __func__);
		return;
	}

	UnregisterDVFSDevice(psDeviceNode);
	DeinitDVFSDevice(psDeviceNode);

	xdev = gfx_to_xdev(psDeviceNode);
	default_volt = (xdev->fw_st_info.gpu_core0_volt) * 1000;
	default_freq = xdev->fw_st_info.default_gpu_freq;
	cur_freq = get_gpu_freq(xdev);

	if (cur_freq > default_freq){
		dev_err(psDeviceNode->psDevConfig->pvOSDevice, "%s : current frequency is higher than default frequency!\n", __func__);
	}

	if (PVRSRV_OK != PVRSRVDevicePreClockSpeedChange(psDeviceNode,
													 psDVFSDeviceCfg->bIdleReq,
													 NULL))
	{
		dev_err(dev, "PVRSRVDevicePreClockSpeedChange failed\n");
		return;
	}

	/* Increasing frequency, change voltage first */
	err = xdx_dvfs_early_set_volt(psDeviceNode, default_volt);
	if (err != 0) {
		dev_err(psDeviceNode->psDevConfig->pvOSDevice, "%s : failed to set volt!\n", __func__);
	}

	err = xdx_dvfs_early_set_freq(psDeviceNode, default_freq);
	if (err != 0) {
		dev_err(psDeviceNode->psDevConfig->pvOSDevice, "%s : failed to set freq!\n", __func__);
	}

	PVRSRVDevicePostClockSpeedChange(psDeviceNode, psDVFSDeviceCfg->bIdleReq, NULL);

	psDVFSDevice->bUserEnabled = IMG_FALSE;
	dev_info(psDeviceNode->psDevConfig->pvOSDevice, "%s : dvfs disabled\n", __func__);
}

bool xdx_dvfs_enabled(struct device *dev)
{
	int deviceId = _device_get_devid(dev);
	PVRSRV_DEVICE_NODE *psDeviceNode = PVRSRVGetDeviceInstanceByOSId(deviceId);
	IMG_DVFS_DEVICE *psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;

	return psDVFSDevice->bUserEnabled;
}
#endif /* XDXGPU */

static IMG_INT32 devfreq_target(struct device *dev, unsigned long *requested_freq, IMG_UINT32 flags)
{
	int err;
	int deviceId = _device_get_devid(dev);
	PVRSRV_DEVICE_NODE *psDeviceNode = PVRSRVGetDeviceInstanceByOSId(deviceId);
	RGX_DATA		*psRGXData = NULL;
	IMG_DVFS_DEVICE		*psDVFSDevice = NULL;
	IMG_DVFS_DEVICE_CFG	*psDVFSDeviceCfg = NULL;
	RGX_TIMING_INFORMATION	*psRGXTimingInfo = NULL;
	IMG_UINT32		ui32Freq, ui32CurFreq, ui32Volt;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 13, 0))
	struct opp *opp;
#else
	struct dev_pm_opp *opp;
#endif
	/* Check the device is registered */
	if (!psDeviceNode)
	{
		return -ENODEV;
	}

	psRGXData = (RGX_DATA*) psDeviceNode->psDevConfig->hDevData;
	psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	psDVFSDeviceCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;

	/* Check the RGX device is initialised */
	if (!psRGXData)
	{
		return -ENODATA;
	}

/* FIXME: Shall involve temperature protection feature here */
	psRGXTimingInfo = psRGXData->psRGXTimingInfo;
	if (!(psDVFSDevice->bEnabled && psDVFSDevice->bUserEnabled))
	{
		*requested_freq = psRGXTimingInfo->ui32CoreClockSpeed;
		return 0;
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0))
	rcu_read_lock();
#endif

	opp = devfreq_recommended_opp(dev, requested_freq, flags);
	if (IS_ERR(opp)) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0))
		rcu_read_unlock();
#endif
		PVR_DPF((PVR_DBG_ERROR, "Invalid OPP"));
		return PTR_ERR(opp);
	}

	ui32Freq = dev_pm_opp_get_freq(opp);
	ui32Volt = dev_pm_opp_get_voltage(opp);

	if (psDVFSDeviceCfg->bSetVgid0Volt && ui32Volt > psDVFSDeviceCfg->ui32Vgid0Volt) {
		ui32Volt = psDVFSDeviceCfg->ui32Vgid0Volt;
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0))
	rcu_read_unlock();
#else
	dev_pm_opp_put(opp);
#endif

	ui32CurFreq = psRGXTimingInfo->ui32CoreClockSpeed;

	if (ui32CurFreq == ui32Freq)
	{
		return 0;
	}

	if (PVRSRV_OK != PVRSRVDevicePreClockSpeedChange(psDeviceNode,
													 psDVFSDeviceCfg->bIdleReq,
													 NULL))
	{
		dev_err(dev, "PVRSRVDevicePreClockSpeedChange failed\n");
		return -EPERM;
	}

	/* Increasing frequency, change voltage first */
	if (ui32Freq > ui32CurFreq)
	{
		err = psDVFSDeviceCfg->pfnSetVoltage(psDeviceNode, ui32Volt);
		if (err != 0) {
			dev_err(psDeviceNode->psDevConfig->pvOSDevice, "%s : failed to set volt!\n", __func__);
			goto clock_speed_change_error;
		}
	}

	err = psDVFSDeviceCfg->pfnSetFrequency(psDeviceNode, ui32Freq);
	if (err != 0) {
		dev_err(psDeviceNode->psDevConfig->pvOSDevice, "%s : failed to set freq!\n", __func__);
		goto clock_speed_change_error;
	}

	/* Decreasing frequency, change frequency first */
	if (ui32Freq < ui32CurFreq)
	{
		err = psDVFSDeviceCfg->pfnSetVoltage(psDeviceNode, ui32Volt);
		if (err != 0) {
			dev_err(psDeviceNode->psDevConfig->pvOSDevice, "%s : failed to set volt!\n", __func__);
			goto clock_speed_change_error;
		}
	}

clock_speed_change_error:
	PVRSRVDevicePostClockSpeedChange(psDeviceNode, psDVFSDeviceCfg->bIdleReq,
									 NULL);

	return err;
}

static int devfreq_get_dev_status(struct device *dev, struct devfreq_dev_status *stat)
{
	int                      deviceId = _device_get_devid(dev);
	PVRSRV_DEVICE_NODE      *psDeviceNode = PVRSRVGetDeviceInstanceByOSId(deviceId);
	PVRSRV_RGXDEV_INFO      *psDevInfo = NULL;
	IMG_DVFS_DEVICE         *psDVFSDevice = NULL;
	RGX_DATA                *psRGXData = NULL;
	RGX_TIMING_INFORMATION  *psRGXTimingInfo = NULL;
	RGXFWIF_GPU_UTIL_STATS   sGpuUtilStats;
	PVRSRV_ERROR             eError;

	/* Check the device is registered */
	if (!psDeviceNode)
	{
		return -ENODEV;
	}

	psDevInfo = psDeviceNode->pvDevice;
	psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	psRGXData = (RGX_DATA*) psDeviceNode->psDevConfig->hDevData;

	/* Check the RGX device is initialised */
	if (!psDevInfo || !psRGXData)
	{
		return -ENODATA;
	}

	psRGXTimingInfo = psRGXData->psRGXTimingInfo;
	stat->current_frequency = psRGXTimingInfo->ui32CoreClockSpeed;

	if (psDevInfo->pfnGetGpuUtilStats == NULL)
	{
		/* Not yet ready. So set times to something sensible. */
		stat->busy_time = 0;
		stat->total_time = 0;
		return 0;
	}

	eError = psDevInfo->pfnGetGpuUtilStats(psDeviceNode,
						psDVFSDevice->hGpuUtilUserDVFS,
						&sGpuUtilStats);

	if (eError != PVRSRV_OK)
	{
		return -EAGAIN;
	}

	stat->busy_time = sGpuUtilStats.ui64GpuStatActive;
	stat->total_time = sGpuUtilStats.ui64GpuStatCumulative;

	return 0;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0))
static IMG_INT32 devfreq_cur_freq(struct device *dev, unsigned long *freq)
{
	int deviceId = _device_get_devid(dev);
	PVRSRV_DEVICE_NODE *psDeviceNode = PVRSRVGetDeviceInstanceByOSId(deviceId);
	RGX_DATA *psRGXData = NULL;

	/* Check the device is registered */
	if (!psDeviceNode)
	{
		return -ENODEV;
	}

	psRGXData = (RGX_DATA*) psDeviceNode->psDevConfig->hDevData;

	/* Check the RGX device is initialised */
	if (!psRGXData)
	{
		return -ENODATA;
	}

	*freq = psRGXData->psRGXTimingInfo->ui32CoreClockSpeed;

	return 0;
}
#endif

static struct devfreq_dev_profile img_devfreq_dev_profile =
{
	.target             = devfreq_target,
	.get_dev_status     = devfreq_get_dev_status,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0))
	.get_cur_freq       = devfreq_cur_freq,
#endif
};

static int FillOPPTable(struct device *dev, PVRSRV_DEVICE_NODE *psDeviceNode)
{
	const IMG_OPP *iopp;
	int i, err = 0;
	IMG_DVFS_DEVICE_CFG *psDVFSDeviceCfg = NULL;

	/* Check the device exists */
	if (!dev || !psDeviceNode)
	{
		return -ENODEV;
	}

	psDVFSDeviceCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;

	for (i = 0, iopp = psDVFSDeviceCfg->pasOPPTable;
	     i < psDVFSDeviceCfg->ui32OPPTableSize;
	     i++, iopp++)
	{
		err = dev_pm_opp_add(dev, iopp->ui32Freq, iopp->ui32Volt);
		if (err) {
			dev_err(dev, "Could not add OPP entry, %d\n", err);
			return err;
		}
	}

	return 0;
}

static void ClearOPPTable(struct device *dev, PVRSRV_DEVICE_NODE *psDeviceNode)
{
#if (defined(CHROMIUMOS_KERNEL) && (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 18, 0))) || \
	(LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0))
	const IMG_OPP *iopp;
	int i;
	IMG_DVFS_DEVICE_CFG *psDVFSDeviceCfg = NULL;

	/* Check the device exists */
	if (!dev || !psDeviceNode)
	{
		return;
	}

	psDVFSDeviceCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;

	for (i = 0, iopp = psDVFSDeviceCfg->pasOPPTable;
	     i < psDVFSDeviceCfg->ui32OPPTableSize;
	     i++, iopp++)
	{
		dev_pm_opp_remove(dev, iopp->ui32Freq);
	}
#endif
}

static int GetOPPValues(struct device *dev,
                        unsigned long *min_freq,
                        unsigned long *min_volt,
                        unsigned long *max_freq)
{
	int deviceId = _device_get_devid(dev);
	PVRSRV_DEVICE_NODE *psDeviceNode = PVRSRVGetDeviceInstanceByOSId(deviceId);
	IMG_DVFS_DEVICE_CFG *psDVFSDeviceCfg = NULL;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 13, 0))
	struct opp *opp;
#else
	struct dev_pm_opp *opp;
#endif
	int count, i, err = 0;
	unsigned long freq;
	unsigned long volt;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)) && \
	(!defined(CHROMIUMOS_KERNEL) || (LINUX_VERSION_CODE < KERNEL_VERSION(3, 18, 0)))
	unsigned int *freq_table;
#else
	unsigned long *freq_table;
#endif

	count = dev_pm_opp_get_opp_count(dev);
	if (count < 0)
	{
		dev_err(dev, "Could not fetch OPP count, %d\n", count);
		return count;
	}

	dev_info(dev, "Found %d OPP points.\n", count);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0))
	freq_table = devm_kcalloc(dev, count, sizeof(*freq_table), GFP_ATOMIC);
#else
	freq_table = kcalloc(count, sizeof(*freq_table), GFP_ATOMIC);
#endif
	if (! freq_table)
	{
		return -ENOMEM;
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0))
	/* Start RCU read-side critical section to map frequency to OPP */
	rcu_read_lock();
#endif

	/* Iterate over OPP table; Iteration 0 finds "opp w/ freq >= 0 Hz".	 */
	freq = 0;
	opp = dev_pm_opp_find_freq_ceil(dev, &freq);
	if (IS_ERR(opp))
	{
		err = PTR_ERR(opp);
		dev_err(dev, "Couldn't find lowest frequency, %d\n", err);
		goto exit;
	}

	*min_volt = dev_pm_opp_get_voltage(opp);
	*max_freq = *min_freq = freq_table[0] = freq;
	dev_info(dev, "opp[%d/%d]: (%lu Hz, %lu uV)\n", 1, count, freq, *min_volt);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0))
	dev_pm_opp_put(opp);
#endif

	/* Iteration i > 0 finds "opp w/ freq >= (opp[i-1].freq + 1)". */
	for (i = 1; i < count; i++)
	{
		freq++;
		opp = dev_pm_opp_find_freq_ceil(dev, &freq);
		if (IS_ERR(opp))
		{
			err = PTR_ERR(opp);
			dev_err(dev, "Couldn't find %dth frequency, %d\n", i, err);
			goto exit;
		}

		freq_table[i] = freq;
		*max_freq = freq;
		if (i != count - 1) {
			volt = dev_pm_opp_get_voltage(opp);
			dev_info(dev,
				 "opp[%d/%d]: (%lu Hz, %lu uV)\n",
				  i + 1,
				  count,
				  freq,
				  volt);
		} else {
			psDVFSDeviceCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;
			if (psDVFSDeviceCfg->bSetVgid0Volt)
				volt = psDVFSDeviceCfg->ui32Vgid0Volt;
			else
				volt = dev_pm_opp_get_voltage(opp);
			dev_info(dev,
				"opp[%d/%d]: (%lu Hz, %lu uV) (Vgid0: %u uV, Vgid1: %lu uV)\n",
				  i + 1,
				  count,
				  freq,
				  volt,
				  psDVFSDeviceCfg->ui32Vgid0Volt,
				  dev_pm_opp_get_voltage(opp));
		}
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0))
		dev_pm_opp_put(opp);
#endif
	}

exit:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0))
	rcu_read_unlock();
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0))
	if (!err)
	{
		img_devfreq_dev_profile.freq_table = freq_table;
		img_devfreq_dev_profile.max_state = count;
	}
	else
#endif
	{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0))
		devm_kfree(dev, freq_table);
#else
		kfree(freq_table);
#endif
	}

	return err;
}

#define TO_IMG_ERR(err) ((err == -EPROBE_DEFER) ? PVRSRV_ERROR_PROBE_DEFER : PVRSRV_ERROR_INIT_FAILURE)

#ifdef XDXGPU
int xdx_dvfs_init(PPVRSRV_DEVICE_NODE psDeviceNode)
{
	IMG_DVFS_DEVICE_CFG    *psDVFSDeviceCfg = NULL;
	struct xdx_device      *xdev;
	u16                     psdevice_id;

	psDVFSDeviceCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;
	xdev = gfx_to_xdev(psDeviceNode);
	psdevice_id = xdev->fw_st_info.device_id;

	/* by default: set bUserEnabled to FALSE */
	psDeviceNode->psDevConfig->sDVFS.sDVFSDevice.bUserEnabled = IMG_FALSE;

	switch(psdevice_id) {
	case 0x1160:
	case 0x11E0:
	case 0x1340:
	case 0x1170:
	case 0x1350:
	case 0x1360:
	case 0x13C0:
		/* we have DVFS support on these cards.
		 * Must keep the above PCIe-IDs same with func xdx_match_sku_opptable provided.
		 */
		psDVFSDeviceCfg->bSupportDVFS = IMG_TRUE;
		dev_info(xdev->dev, "DVFS supported, device ID: 0x%x", psdevice_id);
		break;
	default:
		psDVFSDeviceCfg->bSupportDVFS = IMG_FALSE;
		dev_info(xdev->dev, "DVFS not supported, device ID: 0x%x", psdevice_id);
		break;
	}

	return PVRSRV_OK;
}

void xdx_dvfs_deinit(PPVRSRV_DEVICE_NODE psDeviceNode)
{
	struct xdx_device      *xdev;
	IMG_DVFS_DEVICE_CFG    *psDVFSDeviceCfg = NULL;
	psDVFSDeviceCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;

	if (IMG_FALSE == psDVFSDeviceCfg->bSupportDVFS)
		return;

	if (IMG_FALSE == psDeviceNode->psDevConfig->sDVFS.sDVFSDevice.bUserEnabled)
		return;

	xdev = gfx_to_xdev(psDeviceNode);

	xdx_dvfs_disable(xdev->dev);
}

static void xdx_opptable_val(IMG_OPP *pasOPPTable, const IMG_OPP* xdx_opp_group, int tsize) {
	int i;
	for (i = 0; i < tsize; i++) {
		pasOPPTable[i].ui32Volt = xdx_opp_group[i].ui32Volt;
		pasOPPTable[i].ui32Freq = xdx_opp_group[i].ui32Freq;
	}
}

static int xdx_match_sku_opptable(PPVRSRV_DEVICE_NODE psDeviceNode){
	IMG_DVFS_DEVICE        *psDVFSDevice = NULL;
	IMG_DVFS_DEVICE_CFG    *psDVFSDeviceCfg = NULL;
	IMG_OPP                *pasOPPTable;
	struct xdx_device      *xdev;
	u16                    psdevice_id;
	u32 				   max_gpu_freq;
	u32                    default_volt;

	psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	psDVFSDeviceCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;
	xdev = gfx_to_xdev(psDeviceNode);
	psdevice_id = xdev->fw_st_info.device_id;
	max_gpu_freq = xdev->fw_st_info.default_gpu_freq;
	default_volt = (xdev->fw_st_info.gpu_core0_volt) * 1000;

	pasOPPTable = OSAllocZMem(sizeof(IMG_OPP) * MAX_OPP_TABLE_SIZE);
	if (!pasOPPTable) {
		PVR_DPF(( PVR_DBG_ERROR, "%s: Failed to allocate pasOPPTable for DVFS",
			__func__));
		return -ENOMEM;
	}

	psDVFSDeviceCfg->pasOPPTable = pasOPPTable;
	switch(psdevice_id) {
	case 0x1160:
	case 0x11E0:
	case 0x1340:
		psDVFSDeviceCfg->ui32OPPTableSize = 4;
		xdx_opptable_val(pasOPPTable, xdx_opp_group0,
				    psDVFSDeviceCfg->ui32OPPTableSize);
		if (max_gpu_freq != xdx_opp_group0[psDVFSDeviceCfg->ui32OPPTableSize-1].ui32Freq)
			dev_warn(xdev->dev, "DEV 0x%x, Its Init Freq %u HZ Not Match OPPtable\n",
				psdevice_id, max_gpu_freq);

		psDVFSDeviceCfg->ui32Vgid0Volt = xdx_opp_group0_vgid0_volt;
		if (default_volt != xdx_opp_group0[psDVFSDeviceCfg->ui32OPPTableSize-1].ui32Volt) {
			if (default_volt == xdx_opp_group0_vgid0_volt) {
				psDVFSDeviceCfg->bSetVgid0Volt = IMG_TRUE;
			} else {
				dev_err(xdev->dev, "The max volt in the opptable mismatch with FW default volt: %u\n", default_volt);
			}
		}
		break;
	case 0x1170:
	case 0x1350:
		psDVFSDeviceCfg->ui32OPPTableSize = 5;
		xdx_opptable_val(pasOPPTable, xdx_opp_group1,
				    psDVFSDeviceCfg->ui32OPPTableSize);
		if (max_gpu_freq != xdx_opp_group1[psDVFSDeviceCfg->ui32OPPTableSize-1].ui32Freq)
			dev_warn(xdev->dev, "DEV 0x%x, Its Init Freq %u HZ Not Match OPPtable\n",
				psdevice_id, max_gpu_freq);

		psDVFSDeviceCfg->ui32Vgid0Volt = xdx_opp_group1_vgid0_volt;
		if (default_volt != xdx_opp_group1[psDVFSDeviceCfg->ui32OPPTableSize-1].ui32Volt) {
			if (default_volt == xdx_opp_group1_vgid0_volt) {
				psDVFSDeviceCfg->bSetVgid0Volt = IMG_TRUE;
			} else {
				dev_err(xdev->dev, "The max volt in the opptable mismatch with FW default volt: %u\n", default_volt);
			}
		}
		break;
	case 0x1360:
	case 0x13C0:
		if (max_gpu_freq == 1300000000) {
			psDVFSDeviceCfg->ui32OPPTableSize = 6;
			xdx_opptable_val(pasOPPTable, xdx_opp_group2,
						psDVFSDeviceCfg->ui32OPPTableSize);
			psDVFSDeviceCfg->ui32Vgid0Volt = xdx_opp_group2_vgid0_volt;
			if (default_volt != xdx_opp_group2[psDVFSDeviceCfg->ui32OPPTableSize-1].ui32Volt) {
				if (default_volt == xdx_opp_group2_vgid0_volt)
					psDVFSDeviceCfg->bSetVgid0Volt = IMG_TRUE;
				else
					dev_err(xdev->dev, "The max volt in the opptable mismatch with FW default volt: %u\n", default_volt);
			}
		} else {
			psDVFSDeviceCfg->ui32OPPTableSize = 5;
			xdx_opptable_val(pasOPPTable, xdx_opp_group2_extra,
						psDVFSDeviceCfg->ui32OPPTableSize);
		}
		if (max_gpu_freq != xdx_opp_group2[psDVFSDeviceCfg->ui32OPPTableSize-1].ui32Freq &&
			max_gpu_freq != xdx_opp_group2_extra[psDVFSDeviceCfg->ui32OPPTableSize-1].ui32Freq)
			dev_warn(xdev->dev, "DEV 0x%x, Its Init Freq %u HZ Not Match OPPtable\n",
				psdevice_id, max_gpu_freq);
		break;
	default:
		dev_err(xdev->dev, "Fatal, can not be here! DVFS is not supported with deviceID: %d", psdevice_id);
		return -EINVAL;
	}

	return 0;
}
#endif /* XDXGPU */

PVRSRV_ERROR InitDVFSDevice(PPVRSRV_DEVICE_NODE psDeviceNode)
{
	IMG_DVFS_DEVICE        *psDVFSDevice = NULL;
	IMG_DVFS_DEVICE_CFG    *psDVFSDeviceCfg = NULL;
	struct device          *psDev;
	struct xdx_device      *xdev;
	PVRSRV_ERROR            eError;
	int                     err;
	u32                     init_freq;
	u32                     init_volt;
	u32                     max_freq;
	u32                     max_volt;

	PVRSRV_VZ_RET_IF_MODE(GUEST, PVRSRV_OK);

#if !defined(CONFIG_PM_OPP)
	return PVRSRV_ERROR_NOT_SUPPORTED;
#endif

	if (!psDeviceNode)
	{
		return PVRSRV_ERROR_INVALID_PARAMS;
	}

	if (psDeviceNode->psDevConfig->sDVFS.sDVFSDevice.bInitPending)
	{
		PVR_DPF((PVR_DBG_ERROR,
				 "DVFS initialise pending for device node %p",
				 psDeviceNode));
		return PVRSRV_ERROR_INIT_FAILURE;
	}

	psDev = psDeviceNode->psDevConfig->pvOSDevice;
	psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	psDVFSDeviceCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;
	psDeviceNode->psDevConfig->sDVFS.sDVFSDevice.bInitPending = IMG_TRUE;
#ifdef XDXGPU
	psDVFSDeviceCfg->bSetVgid0Volt = IMG_FALSE;
	psDVFSDeviceCfg->ui32Vgid0Volt = 0;
	psDVFSDeviceCfg->bIdleReq = IMG_TRUE;

	psDVFSDeviceCfg->pfnSetFrequency = xdx_dvfs_early_set_freq;
	psDVFSDeviceCfg->pfnSetVoltage = xdx_dvfs_early_set_volt;
	psDVFSDeviceCfg->ui32PollMs = 100;
	psDeviceNode->psDevConfig->sDVFS.sDVFSGovernorCfg.ui32UpThreshold = 90;
	psDeviceNode->psDevConfig->sDVFS.sDVFSGovernorCfg.ui32DownDifferential = 5;

	err = xdx_match_sku_opptable(psDeviceNode);
	if (err != PVRSRV_OK)
	{
		PVR_DPF((PVR_DBG_ERROR, "Failed to match opptable, %d", err));
		eError = TO_IMG_ERR(err);
		goto err_exit;
	}
#endif /* XDXGPU */

#if defined(SUPPORT_SOC_TIMER)
	if (! psDeviceNode->psDevConfig->pfnSoCTimerRead)
	{
		PVR_DPF((PVR_DBG_ERROR, "System layer SoC timer callback not implemented"));
		/* return PVRSRV_ERROR_NOT_IMPLEMENTED; */
	}
#endif

	eError = SORgxGpuUtilStatsRegister(&psDVFSDevice->hGpuUtilUserDVFS);
	if (eError != PVRSRV_OK)
	{
		PVR_DPF((PVR_DBG_ERROR, "Failed to register to the GPU utilisation stats, %d", eError));
		return eError;
	}

#if defined(CONFIG_OF)
	err = dev_pm_opp_of_add_table(psDev);
	if (err)
	{
		/*
		 * If there are no device tree or system layer provided operating points
		 * then return an error
		 */
		if (err != -ENODEV || !psDVFSDeviceCfg->pasOPPTable)
		{
			PVR_DPF((PVR_DBG_ERROR, "Failed to init opp table from devicetree, %d", err));
			eError = TO_IMG_ERR(err);
			goto err_exit;
		}
	}
#endif

	if (psDVFSDeviceCfg->pasOPPTable)
	{
		err = FillOPPTable(psDev, psDeviceNode);
		if (err)
		{
			PVR_DPF((PVR_DBG_ERROR, "Failed to fill OPP table with data, %d", err));
			eError = TO_IMG_ERR(err);
			goto err_exit;
		}
	}

	PVR_TRACE(("PVR DVFS init pending: dev = %p, PVR device = %p",
			   psDev, psDeviceNode));

	xdev = gfx_to_xdev(psDeviceNode);
	init_freq = get_gpu_freq(xdev);
	init_volt = get_gpu_volt(xdev) * 1000;
	max_freq = psDVFSDeviceCfg->pasOPPTable[psDVFSDeviceCfg->ui32OPPTableSize-1].ui32Freq;
	max_volt = psDVFSDeviceCfg->pasOPPTable[psDVFSDeviceCfg->ui32OPPTableSize-1].ui32Volt;

	if (init_freq != max_freq)
	{
		dev_warn(xdev->dev, "%s : initial freq is not equal to DVFS max freq, initial freq: %d, max freq: %d\n",
						__func__, init_freq, max_freq);
	}

	if (init_volt != max_volt && !psDVFSDeviceCfg->bSetVgid0Volt)
	{
		dev_warn(xdev->dev, "%s : initial volt is not equal to DVFS max volt, initial volt: %d, max volt: %d\n",
						__func__, init_volt, max_volt);
	}

	return PVRSRV_OK;

err_exit:
	DeinitDVFSDevice(psDeviceNode);
	return eError;
}

PVRSRV_ERROR RegisterDVFSDevice(PPVRSRV_DEVICE_NODE psDeviceNode)
{
	IMG_DVFS_DEVICE        *psDVFSDevice = NULL;
	IMG_DVFS_DEVICE_CFG    *psDVFSDeviceCfg = NULL;
	IMG_DVFS_GOVERNOR_CFG  *psDVFSGovernorCfg = NULL;
	RGX_TIMING_INFORMATION *psRGXTimingInfo = NULL;
	struct device          *psDev;
	unsigned long           min_freq = 0, max_freq = 0, min_volt = 0;
	PVRSRV_ERROR            eError;
	int                     err;

	if (!psDeviceNode)
	{
		return PVRSRV_ERROR_INVALID_PARAMS;
	}

	if (!psDeviceNode->psDevConfig->sDVFS.sDVFSDevice.bInitPending)
	{
		PVR_DPF((PVR_DBG_ERROR,
				 "DVFS initialise not yet pending for device node %p",
				 psDeviceNode));
		return PVRSRV_ERROR_INIT_FAILURE;
	}

	psDev = psDeviceNode->psDevConfig->pvOSDevice;
	psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	psDVFSDeviceCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;
	psDVFSGovernorCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSGovernorCfg;
	psRGXTimingInfo = ((RGX_DATA *)psDeviceNode->psDevConfig->hDevData)->psRGXTimingInfo;
	psDeviceNode->psDevConfig->sDVFS.sDVFSDevice.bInitPending = IMG_FALSE;
	psDeviceNode->psDevConfig->sDVFS.sDVFSDevice.bReady = IMG_TRUE;

	err = GetOPPValues(psDev, &min_freq, &min_volt, &max_freq);
	if (err)
	{
		PVR_DPF((PVR_DBG_ERROR, "Failed to read OPP points, %d", err));
		eError = TO_IMG_ERR(err);
		goto err_exit;
	}

	img_devfreq_dev_profile.initial_freq = max_freq;
	img_devfreq_dev_profile.polling_ms = psDVFSDeviceCfg->ui32PollMs;

	psDVFSDevice->data.upthreshold = psDVFSGovernorCfg->ui32UpThreshold;
	psDVFSDevice->data.downdifferential = psDVFSGovernorCfg->ui32DownDifferential;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 16, 0))
	psDVFSDevice->psDevFreq = devm_devfreq_add_device(psDev,
													  &img_devfreq_dev_profile,
													  "xdx_ondemand",
													  &psDVFSDevice->data);
#else
	psDVFSDevice->psDevFreq = devfreq_add_device(psDev,
												 &img_devfreq_dev_profile,
												 "xdx_ondemand",
												 &psDVFSDevice->data);
#endif

	if (IS_ERR(psDVFSDevice->psDevFreq))
	{
		PVR_DPF((PVR_DBG_ERROR,
				 "Failed to add as devfreq device %p, %ld",
				 psDVFSDevice->psDevFreq,
				 PTR_ERR(psDVFSDevice->psDevFreq)));
		eError = TO_IMG_ERR(PTR_ERR(psDVFSDevice->psDevFreq));
		goto err_exit;
	}

#if defined(CHROMIUMOS_KERNEL) && (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 0)) && (LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0))
	psDVFSDevice->psDevFreq->policy.user.min_freq = min_freq;
	psDVFSDevice->psDevFreq->policy.user.max_freq = max_freq;
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0))
	psDVFSDevice->psDevFreq->scaling_min_freq = min_freq;
	psDVFSDevice->psDevFreq->scaling_max_freq = max_freq;
#else
	psDVFSDevice->psDevFreq->min_freq = min_freq;
	psDVFSDevice->psDevFreq->max_freq = max_freq;
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
	psDVFSDevice->psDevFreq->suspend_freq = max_freq;
#endif

	err = devfreq_register_opp_notifier(psDev, psDVFSDevice->psDevFreq);
	if (err)
	{
		PVR_DPF((PVR_DBG_ERROR, "Failed to register opp notifier, %d", err));
		eError = TO_IMG_ERR(err);
		goto err_exit;
	}

#if defined(SUPPORT_LINUX_DVFS) && defined(CONFIG_DEVFREQ_THERMAL)
	err = RegisterThermalZoneDevice(psDev, psDVFSDevice, psDVFSDeviceCfg->psPowerOps);
	if (err)
	{
		eError = TO_IMG_ERR(err);
		goto err_exit;
	}
#endif /* SUPPORT_LINUX_DVFS && CONFIG_DEVFREQ_THERMAL */

	PVR_TRACE(("PVR DVFS activated: %lu-%lu Hz, Period: %ums",
			   min_freq,
			   max_freq,
			   psDVFSDeviceCfg->ui32PollMs));

	return PVRSRV_OK;

err_exit:
	UnregisterDVFSDevice(psDeviceNode);
	return eError;
}

void UnregisterDVFSDevice(PPVRSRV_DEVICE_NODE psDeviceNode)
{
	IMG_DVFS_DEVICE *psDVFSDevice = NULL;
	struct device *psDev = NULL;
	IMG_INT32 i32Error;

	/* Check the device exists */
	if (!psDeviceNode)
	{
		return;
	}

	PVRSRV_VZ_RETN_IF_MODE(GUEST);

	psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	psDev = psDeviceNode->psDevConfig->pvOSDevice;

	if (! psDVFSDevice)
	{
		return;
	}

#if defined(SUPPORT_LINUX_DVFS) && defined(CONFIG_DEVFREQ_THERMAL)
	UnregisterThermalZoneDevice(psDev, psDVFSDevice);
#endif /* SUPPORT_LINUX_DVFS && CONFIG_DEVFREQ_THERMAL */

	if (psDVFSDevice->psDevFreq)
	{
		i32Error = devfreq_unregister_opp_notifier(psDev, psDVFSDevice->psDevFreq);
		if (i32Error < 0)
		{
			PVR_DPF((PVR_DBG_ERROR, "Failed to unregister OPP notifier"));
		}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 16, 0))
		devfreq_remove_device(psDVFSDevice->psDevFreq);
#else
		devm_devfreq_remove_device(psDev, psDVFSDevice->psDevFreq);
#endif

		psDVFSDevice->psDevFreq = NULL;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0) && \
     LINUX_VERSION_CODE < KERNEL_VERSION(3, 13, 0))
	kfree(img_devfreq_dev_profile.freq_table);
#endif

	psDVFSDevice->bInitPending = IMG_FALSE;
	psDVFSDevice->bReady = IMG_FALSE;
}

void DeinitDVFSDevice(PPVRSRV_DEVICE_NODE psDeviceNode)
{
	IMG_DVFS_DEVICE *psDVFSDevice = NULL;
	struct device *psDev = NULL;
	IMG_DVFS_DEVICE_CFG *psDVFSDeviceCfg = NULL;
	IMG_OPP *pasOPPTable = NULL;

	/* Check the device exists */
	if (!psDeviceNode)
	{
		return;
	}

	PVRSRV_VZ_RETN_IF_MODE(GUEST);

	psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	psDev = psDeviceNode->psDevConfig->pvOSDevice;

	/* Remove OPP entries for this device */
	ClearOPPTable(psDev, psDeviceNode);

#if defined(CONFIG_OF)
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)) || \
	(defined(CHROMIUMOS_KERNEL) && (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 18, 0)))
	dev_pm_opp_of_remove_table(psDev);
#endif
#endif

	psDVFSDeviceCfg = &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;
	pasOPPTable = psDVFSDeviceCfg->pasOPPTable;

	/* Free mem allocated for OPP table */
	if (pasOPPTable) {
		OSFreeMem(pasOPPTable);
	}

	SORgxGpuUtilStatsUnregister(psDVFSDevice->hGpuUtilUserDVFS);
	psDVFSDevice->hGpuUtilUserDVFS = NULL;
	psDVFSDevice->bInitPending = IMG_FALSE;
	psDVFSDevice->bReady = IMG_FALSE;
}

PVRSRV_ERROR SuspendDVFS(PPVRSRV_DEVICE_NODE psDeviceNode)
{
	IMG_DVFS_DEVICE	*psDVFSDevice = NULL;
	RGX_DATA *psRGXData = NULL;

	/* Check the device is registered */
	if (!psDeviceNode)
	{
		return PVRSRV_ERROR_INVALID_DEVICE;
	}

	if (psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg.bSupportDVFS) {
		psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;

		/*
		* FIXME: Shall take an atomic lock here to prevent inconsistency
		* of bUserEnabled between suspend and resume function
		*/
		if (psDVFSDevice->bUserEnabled) {
			psRGXData = (RGX_DATA*) psDeviceNode->psDevConfig->hDevData;
			/* Check the RGX device is initialised */
			if (!psRGXData)
			{
				return -ENODATA;
			}
		}
		psDVFSDevice->bEnabled = IMG_FALSE;
	}

	return PVRSRV_OK;
}

PVRSRV_ERROR ResumeDVFS(PPVRSRV_DEVICE_NODE psDeviceNode)
{
	IMG_DVFS_DEVICE	*psDVFSDevice = NULL;

	/* Check the device is registered */
	if (!psDeviceNode)
	{
		return PVRSRV_ERROR_INVALID_DEVICE;
	}

	if (psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg.bSupportDVFS) {
		psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
		/* Not supported in GuestOS drivers */
		psDVFSDevice->bEnabled = !PVRSRV_VZ_MODE_IS(GUEST);

	}

	return PVRSRV_OK;
}

int xdx_dvfs_early_set_freq(PPVRSRV_DEVICE_NODE psDeviceNode, IMG_UINT32 ui32Freq)
{
	u32 cur_freq;
#if defined(SUPPORT_PANGU_ACPU)
	int err;
	CMCUPowerCmd_T power_cmd = { .cmd = cH02CPowerDBCmdDVFS,
				     .subcmd = cH02CPowerDVFSSubCmdSetFreq,
				     .channel = 1,
				     .data = ui32Freq };
#endif
	struct xdx_device *xdev = gfx_to_xdev(psDeviceNode);
	const RGX_DATA *psRGXData = (RGX_DATA*)psDeviceNode->psDevConfig->hDevData;
	PVR_UNREFERENCED_PARAMETER(psRGXData);

#if defined(SUPPORT_PANGU_ACPU)
	err = H02CDoorBellPowerSet(psDeviceNode->psDevConfig->hSysData, power_cmd.all);
	if (err != 0) {
		dev_err(psDeviceNode->psDevConfig->pvOSDevice, "%s : DVFS failed to set freq!\n", __func__);
		return err;
	}
#endif

	((struct xdx_device*)psDeviceNode->psDevConfig->hSysData)->pm.current_gfxclk = ui32Freq;
	dev_dbg(psDeviceNode->psDevConfig->pvOSDevice, "DVFS set freq to %d \n", ui32Freq);

	cur_freq = get_gpu_freq(xdev);
	if(cur_freq != ui32Freq){
		dev_err(psDeviceNode->psDevConfig->pvOSDevice, "%s : adjusted frequency is not equal to current frequency!\n", __func__);
	}
	psRGXData->psRGXTimingInfo->ui32CoreClockSpeed = cur_freq;

	return 0;
}

int xdx_dvfs_early_set_volt(PPVRSRV_DEVICE_NODE psDeviceNode, IMG_UINT32 ui32Volt)
{
#if defined(SUPPORT_PANGU_ACPU)
	int err;
	CMCUPowerCmd_T power_cmd = { .cmd = cH02CPowerDBCmdDVFS,
				     .subcmd = cH02CPowerDVFSSubCmdSetVolt,
				     .channel = 1,
				     .data = ui32Volt / 1000 };

	err = H02CDoorBellPowerSet(psDeviceNode->psDevConfig->hSysData, power_cmd.all);
	if (err != 0) {
		dev_err(psDeviceNode->psDevConfig->pvOSDevice, "%s : DVFS failed to set volt!\n", __func__);
		return err;
	}
#endif

	((struct xdx_device*)psDeviceNode->psDevConfig->hSysData)->pm.current_gfxvol = ui32Volt;
	dev_dbg(psDeviceNode->psDevConfig->pvOSDevice, "DVFS set volt to %d \n", ui32Volt);

	mdelay(1);

	return 0;
}

void xdx_dvfs_set_opp_availability_above_floor(PVRSRV_DEVICE_NODE *psDeviceNode, bool enable_opp) {
	struct xdx_device 	*xdev 			= gfx_to_xdev(psDeviceNode);
	IMG_DVFS_DEVICE_CFG	*psDVFSDeviceCfg= &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;
	const IMG_OPP 		*iopp			= psDVFSDeviceCfg->pasOPPTable;
	const int			opp_size 		= psDVFSDeviceCfg->ui32OPPTableSize;
	int i;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 13, 0))
	struct opp *opp;
#else
	struct dev_pm_opp *opp;
#endif

	// Set OPPs' availability except the first
	iopp++;
	for (i = 1; i < opp_size; i++, iopp++) {
		opp = dev_pm_opp_find_freq_exact(xdev->dev, iopp->ui32Freq, !enable_opp);

		if (IS_ERR(opp)) {
			dev_err(xdev->dev, "Failed to find OPP for frequency %u, ptr_err %ld\n", iopp->ui32Freq, PTR_ERR(opp));
			if (PTR_ERR(opp) == -ERANGE)
				continue;
			return;
		}
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0))
		dev_pm_opp_put(opp);
#endif

		if (enable_opp && dev_pm_opp_enable(xdev->dev, iopp->ui32Freq))
			dev_err(xdev->dev, "Failed to enable %u HZ OPP\n", iopp->ui32Freq);
		if (!enable_opp && dev_pm_opp_disable(xdev->dev, iopp->ui32Freq))
			dev_err(xdev->dev, "Failed to disable %u HZ OPP\n", iopp->ui32Freq);
	}
}

#endif /* !NO_HARDWARE */
