/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include "pvrversion.h"
#include "services_kernel_client.h"
#include "syscommon.h"
#include "rgx_fwif_shared.h"
#include "pvr_debug.h"
#include "pvrsrv_memallocflags.h"

#include "pvr_device.h"

#include "xdx_device.h"
#include "xdx_gem.h"

static void pvr_heap_vram_cpu_to_dev(IMG_HANDLE hPrivData,
				     IMG_UINT32 ui32NumOfAddr,
				     IMG_DEV_PHYADDR *psDevPAddr,
				     IMG_CPU_PHYADDR *psCpuPAddr)
{
	struct xdx_device *xdev = hPrivData;
	uint32_t i;

	for (i = 0; i < ui32NumOfAddr; i++)
		psDevPAddr[i].uiAddr = psCpuPAddr[i].uiAddr - xdev->aper_base +
				       xdev->vram_start;
}

static void pvr_heap_vram_dev_to_cpu(IMG_HANDLE hPrivData,
				     IMG_UINT32 ui32NumOfAddr,
				     IMG_CPU_PHYADDR *psCpuPAddr,
				     IMG_DEV_PHYADDR *psDevPAddr)
{
	struct xdx_device *xdev = hPrivData;
	uint32_t i;

	for (i = 0; i < ui32NumOfAddr; i++)
		psCpuPAddr[i].uiAddr = psDevPAddr[i].uiAddr -
				       xdev->vram_start + xdev->aper_base;
}

static PHYS_HEAP_FUNCTIONS pvr_heap_pl_vram_funcs = {
	.pfnCpuPAddrToDevPAddr = pvr_heap_vram_cpu_to_dev,
	.pfnDevPAddrToCpuPAddr = pvr_heap_vram_dev_to_cpu,
};

static void pvr_heap_gtt_cpu_to_dev(IMG_HANDLE hPrivData,
				    IMG_UINT32 uiNumOfAddr,
				    IMG_DEV_PHYADDR *psDevPAddr,
				    IMG_CPU_PHYADDR *psCpuPAddr)
{
	struct xdx_device *xdev = hPrivData;
	uint32_t i;

	/* Optimise common case */
	psDevPAddr[0].uiAddr = psCpuPAddr[0].uiAddr + xdev->gtt_start;
	if (uiNumOfAddr > 1) {
		for (i = 1; i < uiNumOfAddr; ++i)
			psDevPAddr[i].uiAddr =
				psCpuPAddr[i].uiAddr + xdev->gtt_start;
	}
}

static void pvr_heap_gtt_dev_to_cpu(IMG_HANDLE hPrivData,
				    IMG_UINT32 uiNumOfAddr,
				    IMG_CPU_PHYADDR *psCpuPAddr,
				    IMG_DEV_PHYADDR *psDevPAddr)
{
	struct xdx_device *xdev = hPrivData;
	uint32_t i;

	/* Optimise common case */
	psCpuPAddr[0].uiAddr =
		(phys_addr_t)(psDevPAddr[0].uiAddr - xdev->gtt_start);
	if (uiNumOfAddr > 1) {
		for (i = 1; i < uiNumOfAddr; ++i)
			psCpuPAddr[i].uiAddr = (phys_addr_t)(
				psDevPAddr[i].uiAddr - xdev->gtt_start);
	}
}

static PHYS_HEAP_FUNCTIONS pvr_heap_pl_gtt_funcs = {
	.pfnCpuPAddrToDevPAddr = pvr_heap_gtt_cpu_to_dev,
	.pfnDevPAddrToCpuPAddr = pvr_heap_gtt_dev_to_cpu,
};

#if defined(SUPPORT_PANGU_ACPU)
static void PanguAcpuCpuPAddrToDevPAddr(IMG_HANDLE hPrivData,
					  IMG_UINT32 ui32NumOfAddr,
					  IMG_DEV_PHYADDR *psDevPAddr,
					  IMG_CPU_PHYADDR *psCpuPAddr)
{
	struct xdx_device *xdev = hPrivData;

	/* Optimise common case */
	psDevPAddr[0].uiAddr = psCpuPAddr[0].uiAddr -
		xdev->shareram_start +	xdev->shareram_base;

	if (ui32NumOfAddr > 1) {
		IMG_UINT32 ui32Idx;

		for (ui32Idx = 1; ui32Idx < ui32NumOfAddr; ++ui32Idx) {
			psDevPAddr[ui32Idx].uiAddr = psCpuPAddr[ui32Idx].uiAddr -
			xdev->shareram_start +	xdev->shareram_base;
		}
	}
}

static void PanguAcpuDevPAddrToCpuPAddr(IMG_HANDLE hPrivData,
					IMG_UINT32 ui32NumOfAddr,
					IMG_CPU_PHYADDR *psCpuPAddr,
					IMG_DEV_PHYADDR *psDevPAddr)
{
	struct xdx_device *xdev = hPrivData;

	/* Optimise common case */
	psCpuPAddr[0].uiAddr = psDevPAddr[0].uiAddr -
		xdev->shareram_start +	xdev->shareram_base;

	if (ui32NumOfAddr > 1) {
		IMG_UINT32 ui32Idx;

		for (ui32Idx = 1; ui32Idx < ui32NumOfAddr; ++ui32Idx) {
			psCpuPAddr[ui32Idx].uiAddr = psDevPAddr[ui32Idx].uiAddr -
				xdev->shareram_start + xdev->shareram_base;
		}
	}
}

static PHYS_HEAP_FUNCTIONS gsAcpuPhysHeapFuncs = {
	/* pfnCpuPAddrToDevPAddr */
	PanguAcpuCpuPAddrToDevPAddr,
	/* pfnDevPAddrToCpuPAddr */
	PanguAcpuDevPAddrToCpuPAddr,
};
#endif

void pvr_init_phys_heaps(struct xdx_device *xdev,
			     PHYS_HEAP_CONFIG *phys_heaps, uint32_t *pcount)
{
	uint32_t count = 0;

	if (!phys_heaps)
		return;

	/* CPU visable vram heap */
	phys_heaps[count].pszPDumpMemspaceName = "GPU_LOCAL";
	phys_heaps[count].eType = PHYS_HEAP_TYPE_LMA;
	phys_heaps[count].psMemFuncs = &pvr_heap_pl_vram_funcs;
	phys_heaps[count].hPrivData = xdev;
	phys_heaps[count].ui32UsageFlags = PHYS_HEAP_USAGE_GPU_LOCAL;
	phys_heaps[count].sStartAddr.uiAddr = xdev->aper_base + PANGU_VRAM_RESERVE_SIZE;
	phys_heaps[count].sCardBase.uiAddr = xdev->vram_start + PANGU_VRAM_RESERVE_SIZE ;
	if (xdev->real_vram_size > xdev->aper_size)
		phys_heaps[count].uiSize = xdev->visible_vram_size - MMU_HEAP_SIZE - PANGU_VRAM_RESERVE_SIZE;
	else
		phys_heaps[count].uiSize = xdev->visible_vram_size -
			MMU_HEAP_SIZE - xdev->dc_heap_size - PANGU_VRAM_RESERVE_SIZE;

	dev_info(xdev->dev, "visable heap: sStartAddr 0x%016llX, sCardBase 0x%016llX, uiSize %lldM\n",
		phys_heaps[count].sStartAddr.uiAddr,
		phys_heaps[count].sCardBase.uiAddr,
		phys_heaps[count].uiSize >> 20
	);
	count++;

	if (xdev->real_vram_size > xdev->aper_size) {
		/* CPU invisable vram heap */
		phys_heaps[count].pszPDumpMemspaceName = "GPU_PRIVATE";
		phys_heaps[count].eType = PHYS_HEAP_TYPE_LMA;
		phys_heaps[count].psMemFuncs = NULL;
		phys_heaps[count].hPrivData = xdev;
		phys_heaps[count].ui32UsageFlags = PHYS_HEAP_USAGE_GPU_PRIVATE;
		phys_heaps[count].sCardBase.uiAddr = xdev->vram_start + xdev->aper_size;
		phys_heaps[count].uiSize = xdev->real_vram_size -
			xdev->aper_size - xdev->dc_heap_size;
		phys_heaps[count].sStartAddr.uiAddr = 0;
		dev_info(xdev->dev, "invisable heap: sStartAddr 0x%016llX, sCardBase 0x%016llX, uiSize %lldM\n",
			phys_heaps[count].sStartAddr.uiAddr,
			phys_heaps[count].sCardBase.uiAddr,
			phys_heaps[count].uiSize >> 20
		);
		count++;
	}

	/* vram heap only for MMU */
	phys_heaps[count].pszPDumpMemspaceName = "MMU_GPU_LOCAL";
	phys_heaps[count].eType = PHYS_HEAP_TYPE_LMA;
	phys_heaps[count].psMemFuncs = &pvr_heap_pl_vram_funcs;
	phys_heaps[count].hPrivData = xdev;
	phys_heaps[count].ui32UsageFlags = PHYS_HEAP_USAGE_MMU;
	/*
	 * when resize bar is enabled, a 32M space next to the reserved_vram_size
	 * is reserved for MMU table; the high address 32M on the bar2 is reserved
	 * while it is disabled.
	 */
	if (xdev->aper_size == xdev->vram_size) {
		phys_heaps[count].sStartAddr.uiAddr = xdev->aper_base +
			xdev->aper_size - xdev->hw_info.reserved_vram_size -
			MMU_HEAP_SIZE;
		phys_heaps[count].sCardBase.uiAddr = xdev->vram_start +
			xdev->aper_size - xdev->hw_info.reserved_vram_size -
			MMU_HEAP_SIZE;
	} else {
		phys_heaps[count].sStartAddr.uiAddr = xdev->aper_base +
			xdev->aper_size - MMU_HEAP_SIZE;
		phys_heaps[count].sCardBase.uiAddr = xdev->vram_start +
			xdev->aper_size - MMU_HEAP_SIZE;
	}
	phys_heaps[count].uiSize = MMU_HEAP_SIZE;

	dev_info(xdev->dev, "MMU heap: sStartAddr 0x%016llX, sCardBase 0x%016llX, uiSize %lldM\n",
		phys_heaps[count].sStartAddr.uiAddr,
		phys_heaps[count].sCardBase.uiAddr,
		phys_heaps[count].uiSize >> 20
	);
	count++;

	/* vram heap only for DC */
	phys_heaps[count].pszPDumpMemspaceName = "DC_GPU_LOCAL";
	phys_heaps[count].eType = PHYS_HEAP_TYPE_LMA;
	phys_heaps[count].psMemFuncs = &pvr_heap_pl_vram_funcs;
	phys_heaps[count].hPrivData = xdev;
	phys_heaps[count].ui32UsageFlags = PHYS_HEAP_USAGE_DC;
	if (xdev->aper_size == xdev->vram_size) {
		phys_heaps[count].sStartAddr.uiAddr = xdev->aper_base +
			xdev->aper_size - xdev->hw_info.reserved_vram_size -
			MMU_HEAP_SIZE - xdev->dc_heap_size;
		phys_heaps[count].sCardBase.uiAddr = xdev->vram_start +
			xdev->aper_size - xdev->hw_info.reserved_vram_size -
			MMU_HEAP_SIZE - xdev->dc_heap_size;
	} else {
		/* invisble DC heap when resizebar is disabled */
		phys_heaps[count].sStartAddr.uiAddr = 0;
		phys_heaps[count].sCardBase.uiAddr = xdev->vram_start +
			xdev->vram_size - xdev->hw_info.reserved_vram_size -
			xdev->dc_heap_size;
	}
	phys_heaps[count].uiSize = xdev->dc_heap_size;

	dev_info(xdev->dev, "DC heap: sStartAddr 0x%016llX, sCardBase 0x%016llX, uiSize %lldM\n",
		phys_heaps[count].sStartAddr.uiAddr,
		phys_heaps[count].sCardBase.uiAddr,
		phys_heaps[count].uiSize >> 20
	);
	count++;

	/* GPU visable system heap */
	phys_heaps[count].pszPDumpMemspaceName = "SYSMEM";
	phys_heaps[count].eType = PHYS_HEAP_TYPE_UMA;
	phys_heaps[count].psMemFuncs = &pvr_heap_pl_gtt_funcs;
	phys_heaps[count].hPrivData = xdev;
	phys_heaps[count].ui32UsageFlags = PHYS_HEAP_USAGE_CPU_LOCAL;
	phys_heaps[count].sCardBase.uiAddr = xdev->gtt_start;
	phys_heaps[count].uiSize = 0;
	phys_heaps[count].sStartAddr.uiAddr = 0;
	dev_info(xdev->dev, "system heap: sStartAddr 0x%016llX, sCardBase 0x%016llX, uiSize %lldM\n",
		phys_heaps[count].sStartAddr.uiAddr,
		phys_heaps[count].sCardBase.uiAddr,
		phys_heaps[count].uiSize >> 20
	);
	count++;

#ifdef CONFIG_FW_MAIN_ON_SYSMEM
	phys_heaps[count].pszPDumpMemspaceName = "SYSMEM";
	phys_heaps[count].eType = PHYS_HEAP_TYPE_UMA;
	phys_heaps[count].psMemFuncs = &pvr_heap_pl_gtt_funcs;
	phys_heaps[count].hPrivData = xdev;
	phys_heaps[count].ui32UsageFlags = PHYS_HEAP_USAGE_FW_MAIN;
	phys_heaps[count].sCardBase.uiAddr = xdev->gtt_start;
	phys_heaps[count].uiSize = 0;
	phys_heaps[count].sStartAddr.uiAddr = 0;
	dev_info(xdev->dev, "FW Main heap: sStartAddr 0x%016llX, sCardBase 0x%016llX, uiSize %dM\n",
		phys_heaps[count].sStartAddr.uiAddr,
		phys_heaps[count].sCardBase.uiAddr,
		phys_heaps[count].uiSize >> 20
	);
	count++;
#endif

#if defined(SUPPORT_PANGU_ACPU)
	phys_heaps[count].eType = PHYS_HEAP_TYPE_LMA;
	phys_heaps[count].pszPDumpMemspaceName = "LMA_ACPU_SRAM";
	phys_heaps[count].psMemFuncs = &gsAcpuPhysHeapFuncs;
	phys_heaps[count].hPrivData = xdev;
	phys_heaps[count].ui32UsageFlags = PHYS_HEAP_USAGE_FW_ACPU;

	if(xdev->shramhas_offset)
	{
		phys_heaps[count].sCardBase.uiAddr = xdev->shareram_start + PANGU_SSRAM_DEV_PHYSICAL_OFFSET;
		phys_heaps[count].sStartAddr.uiAddr = xdev->shareram_base + PANGU_SSRAM_DEV_PHYSICAL_OFFSET;
		phys_heaps[count].uiSize = xdev->shareram_useful_size - PANGU_SSRAM_DEV_PHYSICAL_OFFSET;
	}
	else
	{
		phys_heaps[count].sCardBase.uiAddr = xdev->shareram_start;
		phys_heaps[count].sStartAddr.uiAddr = xdev->shareram_base;
		phys_heaps[count].uiSize = xdev->shareram_useful_size;
	}


	dev_info(xdev->dev, "shareram heap: sStartAddr 0x%016llX, sCardBase 0x%016llX, uiSize %lldK\n",
		phys_heaps[count].sStartAddr.uiAddr,
		phys_heaps[count].sCardBase.uiAddr,
		phys_heaps[count].uiSize >> 10
	);
	count++;
#endif

	*pcount = count;
}
