/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include "xdx_pm.h"
#include "xdx_device.h"
#include "xdx_irq.h"
#include "xdx_cmcu.h"
#include "osfunc.h"
#include "xdx_shared.h"
#include <linux/delay.h>

const char* acpu_err_msg[][4] = {
	{"GddrReadCrcCount", "GddrWriteCrcCount"},
	{"UnexpectedInterrupt", "Assert"},
	{"ErrorException", "UnexceptedException", "GddrReadRetryCount", "GddrWriteRetryCount"}};

const char* err_msg[][13] = {
	{"Warning"},
	{"UnexpectedInterrupt", "Assert"},
	{"NMIException",      "HardFaultException",    "MMFaultException",
	 "BusFaultException", "UsageFaultException",   "SecureFaultException",
	 "SVCallException",   "DebugMonitorException", "PendSVException",
	 "SysTickException",  "WatchDog",              "SOCTempOverProtect",
	 "GDDRTempOverProtect"}};

static bool is_done_irq(uint64_t data) {
	/**
	 * Fixme: after setting done 1 in a db, it triggers another irq from the db and thus 
	 * it traps a irq cycle. Temporaily, apply the below code to judge and stop it
	 */
	return data & 1;
}

static int xdx_cmcu_irq_set(struct xdx_device *xdev,
			   struct xdx_irq_src *src,
			   unsigned type,
			   enum xdx_interrupt_state state)
{
	// TODO: enable/disable IRQ
	return -1;
}

static u64 c2h0_doorbell_rreg64(void *reg_base, u32 off)
{
	u64 data;
	wmb();
	data = OSReadHWReg64(reg_base, off);
	mb();
	return data;
}

static void c2h0_doorbell_wreg64(void  *reg_base, u32 off, u64 value)
{
	wmb();
	OSWriteHWReg64(reg_base, off, value);
	mb();
}

#ifndef BIT
#define BIT(b) (1UL << (b))
#endif

//TODO: later, all below print functions will be replaced by Xmacro
static void print_msg(struct xdx_device *xdev, const char* chip, ReportDBCmd_T rpt, uint8_t errors_size[]) {
	uint8_t log_level = rpt.cmd;
	uint8_t err_type = rpt.type;
	uint32_t err_info = rpt.data;

	uint8_t warn_size = errors_size[0];
	uint8_t err_size = errors_size[1];
	uint8_t fatal_size = errors_size[2];

	switch(log_level) {
	case Warning:
			dev_warn(xdev->dev, "FW DB WARN MSG: %s %s, Value 0x%x\n", chip, \
			err_type < warn_size ? (strcmp(chip, "ACPU") ? err_msg[log_level][err_type] : acpu_err_msg[log_level][err_type]) : "Unknown Warning", err_info);
		break;
	case Error:
			dev_err(xdev->dev, "FW DB ERROR MSG: %s %s, Value 0x%x\n", chip, \
			err_type < err_size ? (strcmp(chip, "ACPU") ? err_msg[log_level][err_type] : acpu_err_msg[log_level][err_type]) : "Unknown Error", err_info);
		break;
	case FatalError:
			dev_err(xdev->dev, "FW DB FATAL MSG: %s %s, Value 0x%x\n", chip, \
			err_type < fatal_size ? (strcmp(chip, "ACPU") ? err_msg[log_level][err_type] : acpu_err_msg[log_level][err_type]) : "Unknown Fatal", err_info);
		break;
	default:
		dev_warn(xdev->dev, "FW DB WARN MSG: %s Unknown Type Error, Value 0x%x\n", chip, err_info);
		break;
	}
}

static void cmcu_err_rpt(struct work_struct *para) {
	struct cmcu_data *data = container_of(para, struct cmcu_data, work);
	void __iomem * reg_base =  (void __iomem *)data->xdev->cmcu.cmcu_rmmio;

	ReportDBCmd_T cmcu_rpt = (ReportDBCmd_T) c2h0_doorbell_rreg64(reg_base, CMCU_DOORBELLS_DOORBELL51_OFFSET);
	uint8_t log_level = cmcu_rpt.cmd;
	uint8_t err_type = cmcu_rpt.type;

	uint8_t warn_size = 1;
	uint8_t err_size = cCMCUErrorMax;
	uint8_t fatal_size = cCMCUFatalErrorMax;

	uint8_t errors_size[3] = {warn_size, err_size, fatal_size};

	if(is_done_irq(cmcu_rpt.all))
		return;
	
	print_msg(data->xdev, "CMCU", cmcu_rpt, errors_size);

	switch(log_level) {
	case cCMCUWarning:
		//Operation
		break;
		
	case cCMCUError:
		switch(err_type) {
		case cCMCUUnexpectedInterrupt:
			//Operation
			break;
		case cCMCUAssert:
			//Operation
			break;
		default:
			//Operation
			break;
		}
		break;

	/* TODO: Operation Plus Reset FW */
	case cCMCUFatalError:
		switch(err_type) {
		case cCMCUNMIException:
			//Operation
			break;
		case cCMCUHardFaultException:
			//Operation
			break;
		case cCMCUMMFaultException:
			//Operation
			break;
		case cCMCUBusFaultException:
			//Operation
			break;
		case cCMCUUsageFaultException:
			//Operation
			break;
		case cCMCUSecureFaultException:
			//Operation
			break;
		case cCMCUSVCallException:
			//Operation
			break;
		case cCMCUDebugMonitorException:
			//Operation
			break;
		case cCMCUPendSVException:
			//Operation
			break;
		case cCMCUSysTickException:
			//Operation
			break;
		case cCMCUWatchDog:
			//Operation
			break;
		case cCMCUSocTempOverProtect:
			//Operation
			break;
		case cCMCUGddrTempOverProtect:
			//Operation
			break;
		default:
			//Operation
			break;
		}
		break;

	default:
		//Operation
		break;
	}

	c2h0_doorbell_wreg64(reg_base, CMCU_DOORBELLS_DOORBELL51_OFFSET, cmcu_rpt.all | 1);
}

static void acpu_err_rpt(struct work_struct *para) {
	struct acpu_data *data = container_of(para, struct acpu_data, work);
	void __iomem * reg_base =  (void __iomem *)data->xdev->cmcu.cmcu_rmmio;

	ReportDBCmd_T acpu_rpt = (ReportDBCmd_T) c2h0_doorbell_rreg64(reg_base, CMCU_DOORBELLS_DOORBELL52_OFFSET);
	uint8_t log_level = acpu_rpt.cmd;
	uint8_t err_type = acpu_rpt.type;

	uint8_t warn_size = cACPUWarningMax;
	uint8_t err_size = cACPUErrorMax;
	uint8_t fatal_size = cACPUFatalErrorMax;

	uint8_t errors_size[3] = {warn_size, err_size, fatal_size};

	if(is_done_irq(acpu_rpt.all))
		return;
	
	print_msg(data->xdev, "ACPU", acpu_rpt, errors_size);

	switch(log_level) {
	case cACPUWarning:
		switch(err_type) {
		case eGddrReadCrcCount:
			//Operation
			break;
		case eGddrWriteCrcCount:
			//Operation
			break;
		default:
			//Operation
			break;
		}
		break;
		
	case cACPUError:
		switch(err_type) {
		case cACPUUnexpectedInterrupt:
			//Operation
			break;
		case cACPUAssert:
			//Operation
			break;
		default:
			//Operation
			break;
		}
		break;
		
	/* TODO: Reset FW */
	case cACPUFatalError:
		switch(err_type) {
		case cACPUErrorException:
			//Operation
			break;
		case cACPUUnexceptedException:
			//Operation
			break;
		case eGddrReadRetryCount:
			//Operation
			break;
		case eGddrWriteRetryCount:
			//Operation
			break;
		default:
			//Operation
			break;
		}
		break;

	default:
		//Operation
		break;
	}

	c2h0_doorbell_wreg64(reg_base, CMCU_DOORBELLS_DOORBELL52_OFFSET, acpu_rpt.all | 1);
}

static void vdmcu0_err_rpt(struct work_struct *para) {
	struct vdmcu0_data *data = container_of(para, struct vdmcu0_data, work);
	void __iomem * reg_base =  (void __iomem *)data->xdev->cmcu.cmcu_rmmio;

	ReportDBCmd_T vdmcu0_rpt = (ReportDBCmd_T) c2h0_doorbell_rreg64(reg_base, CMCU_DOORBELLS_DOORBELL53_OFFSET);
	uint8_t log_level = vdmcu0_rpt.cmd;
	uint8_t err_type = vdmcu0_rpt.type;

	uint8_t warn_size = 1;
	uint8_t err_size = cVDMCU0ErrorMax;
	uint8_t fatal_size = cVDMCU0FatalErrorMax;

	uint8_t errors_size[3] = {warn_size, err_size, fatal_size};

	if(is_done_irq(vdmcu0_rpt.all))
		return;
	
	print_msg(data->xdev, "VDMCU0", vdmcu0_rpt, errors_size);

	switch(log_level) {
	case cVDMCU0Warning:
		//Operation
		break;
		
	case cVDMCU0Error:
		switch(err_type) {
		case cVDMCU0UnexpectedInterrupt:
			//Operation
			break;
		default:
			//Operation
			break;
		}
		break;

	/* TODO: Operation plus Reset FW */
	case cVDMCU0FatalError:
		switch(err_type) {
		case cVDMCU0NMIException:
			//Operation
			break;
		case cVDMCU0HardFaultException:
			//Operation
			break;
		case cVDMCU0MMFaultException:
			//Operation
			break;
		case cVDMCU0BusFaultException:
			//Operation
			break;
		case cVDMCU0UsageFaultException:
			//Operation
			break;
		case cVDMCU0SecureFaultException:
			//Operation
			break;
		case cVDMCU0SVCallException:
			//Operation
			break;
		case cVDMCU0DebugMonitorException:
			//Operation
			break;
		case cVDMCU0PendSVException:
			//Operation
			break;
		case cVDMCU0SysTickException:
			//Operation
			break;
		default:
			//Operation
			break;
		}
		break;

	default:
		//Operation
		break;
	}

	c2h0_doorbell_wreg64(reg_base, CMCU_DOORBELLS_DOORBELL53_OFFSET, vdmcu0_rpt.all | 1);
}

/* after test it, apply the same method on cmcu, acpu, vdmcu0 */
static void vdmcu1_err_rpt(struct work_struct *para) {
	struct vdmcu1_data *data = container_of(para, struct vdmcu1_data, work);
	void __iomem * reg_base =  (void __iomem *)data->xdev->cmcu.cmcu_rmmio;

	ReportDBCmd_T vdmcu1_rpt = (ReportDBCmd_T) c2h0_doorbell_rreg64(reg_base, CMCU_DOORBELLS_DOORBELL54_OFFSET);
	uint8_t log_level = vdmcu1_rpt.cmd;
	uint8_t err_type = vdmcu1_rpt.type;

	uint8_t warn_size = 1;
	uint8_t err_size = cVDMCU1ErrorMax;
	uint8_t fatal_size = cVDMCU1FatalErrorMax;

	uint8_t errors_size[3] = {warn_size, err_size, fatal_size};

	if(is_done_irq(vdmcu1_rpt.all))
		return;
	
	print_msg(data->xdev, "VDMCU1", vdmcu1_rpt, errors_size);

	switch(log_level) {
	case cVDMCU1Warning:
		//Operation
		break;
		
	case cVDMCU1Error:
		switch(err_type) {
		case cVDMCU1UnexpectedInterrupt:
			//Operation
			break;
		default:
			//Operation
			break;
		}
		break;
		
	/* TODO: Operation Plus Reset FW */
	case cVDMCU1FatalError:
		switch(err_type){
		case cVDMCU1NMIException:
			//Operation
			break;
		case cVDMCU1HardFaultException:
			//Operation
			break;
		case cVDMCU1MMFaultException:
			//Operation
			break;
		case cVDMCU1BusFaultException:
			//Operation
			break;
		case cVDMCU1UsageFaultException:
			//Operation
			break;
		case cVDMCU1SecureFaultException:
			//Operation
			break;
		case cVDMCU1SVCallException:
			//Operation
			break;
		case cVDMCU1DebugMonitorException:
			//Operation
			break;
		case cVDMCU1PendSVException:
			//Operation
			break;
		case cVDMCU1SysTickException:
			//Operation
			break;
		default:
			//Operation
			break;
		}
		break;

	default:
		//Operation
		break;
	}

	c2h0_doorbell_wreg64(reg_base, CMCU_DOORBELLS_DOORBELL54_OFFSET, vdmcu1_rpt.all | 1);
}


static int xdx_cmcu_irq_process(struct xdx_device *xdev,
			       struct xdx_irq_src *src,
			       struct xdx_irq_entry *entry)
{
	void __iomem * reg_base;
	uint64_t int_status;

	reg_base = (void __iomem *)xdev->cmcu.cmcu_rmmio;
	int_status = c2h0_doorbell_rreg64(reg_base, CMCU_DOORBELLS_C2H0_STATS_REGS_OFFSET);

	if(int_status != 0) {
		c2h0_doorbell_wreg64(reg_base, CMCU_DOORBELLS_C2H0_CLR_REGS_OFFSET, int_status);
	}
	
	/**
	 * TODO:
	 * 1. write a counter to count each type of errors for each cpu
	 * 2. set processing priority, fatal > err > warning
	 * 3. c2h0_doorbell_wreg64 changes to c2h0_doorbell_wreg32
	 * 4. set independent wq for each work item
	 * 5. set ratelimit on error handling to avoid poor performance
	*/

	if(int_status & BIT(51)) {
		int_status = int_status & (~BIT(51));
		schedule_work(&xdev->cmcu.cmcu.work);
	}

	if(int_status & BIT(52)) {
		int_status = int_status & (~BIT(52));
		schedule_work(&xdev->cmcu.acpu.work);
	}

	if(int_status & BIT(53)) {
		int_status = int_status & (~BIT(53));
		schedule_work(&xdev->cmcu.vdmcu0.work);
	}

	if(int_status & BIT(54)) {
		int_status = int_status & (~BIT(54));
		schedule_work(&xdev->cmcu.vdmcu1.work);
	}

	if(int_status) {
		dev_warn(xdev->dev, "Warning: Unexpected IRQs %0llx", int_status);
	}

	return 0;
}

static struct xdx_irq_src_funcs xdx_cmcu_src_funcs = {
	.set = xdx_cmcu_irq_set,
	.process = xdx_cmcu_irq_process,
};

#define PANGU_REG_DOORBELL_SIZE 0x600
int xdx_cmcu_init(struct xdx_device *xdev)
{
	int ret;
	struct xdx_irq_src *irq_src;
	irq_src = kmalloc(sizeof(*irq_src), GFP_KERNEL);
	if (!irq_src)
		return -ENOMEM;

	irq_src->id = XDX_IRQ_CLIENTID_CMCU;
	irq_src->funcs = &xdx_cmcu_src_funcs;
	xdx_irq_register_src(xdev, irq_src);

	xdev->cmcu.cmcu_rmmio_base =xdev->rmmio_base + PANGU_REG_DOORBELLS_BASE;
	xdev->cmcu.cmcu_rmmio_size = PANGU_REG_DOORBELL_SIZE;
	xdev->cmcu.cmcu_rmmio = ioremap(xdev->cmcu.cmcu_rmmio_base, xdev->cmcu.cmcu_rmmio_size); 
	if (xdev->cmcu.cmcu_rmmio == NULL) {
		dev_err(xdev->dev, "cmcu failed to map mmio registers\n");
		ret = -ENODEV;
		goto err1; 
	}

	xdev->cmcu.irq_src = irq_src;

	xdev->cmcu.cmcu.xdev = xdev;
	xdev->cmcu.acpu.xdev = xdev;
	xdev->cmcu.vdmcu0.xdev = xdev;
	xdev->cmcu.vdmcu1.xdev = xdev; 

	INIT_WORK(&xdev->cmcu.cmcu.work, cmcu_err_rpt);
	INIT_WORK(&xdev->cmcu.acpu.work, acpu_err_rpt);
	INIT_WORK(&xdev->cmcu.vdmcu0.work, vdmcu0_err_rpt);
	INIT_WORK(&xdev->cmcu.vdmcu1.work, vdmcu1_err_rpt);
		
	return 0;

err1:
	xdx_irq_unregister_src(xdev, irq_src);
	kfree(irq_src);
	return ret;
}

int xdx_cmcu_fini(struct xdx_device *xdev)
{
	int ret;

	flush_work(&xdev->cmcu.cmcu.work);
	flush_work(&xdev->cmcu.acpu.work);
	flush_work(&xdev->cmcu.vdmcu0.work);
	flush_work(&xdev->cmcu.vdmcu1.work);

	iounmap(xdev->cmcu.cmcu_rmmio);
	
	ret = xdx_irq_unregister_src(xdev, xdev->cmcu.irq_src);
	if (ret) {
		dev_err(xdev->dev, "failed to unregister irq source\n");
	}

	kfree(xdev->cmcu.irq_src);
	xdev->cmcu.irq_src = NULL;

	return 0;
}
