/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include "xdx_fw.h"
#include "xdx_device.h"
#include "pangu.h"
#include "fwshare/FirmwareInfo.h"
#include "fwshare/ParameterPage.h"
#include "xdxaudio_hal.h"

const static u32 fw_bootup_gpu_freq_list[] = {
	800000000,
	700000000,
	600000000,
	400000000,
	900000000,
	1000000000,
	1050000000,
	1100000000,
	1150000000,
	1200000000,
	1250000000,
	1300000000,
	1350000000,
	1400000000,
	1500000000,
	500000000,
	300000000,
	200000000,
	100000000,
	50000000
};

s32 get_gpu_temp(struct xdx_device *xdev)
{
	s16 gpu_temp = *((s32*)(xdev->rmmio + PANGU_PUBLIC_FW_INFO_BASE + offsetof(PublicFirmwareInfo_T, curTemp)));

	if (gpu_temp < -50 || gpu_temp > 250) {
		dev_warn(xdev->dev, "GPU Temp %d Not in Range[-50, 250]\n", gpu_temp);
		return 0;
	}
		
	return gpu_temp;
}

s32 get_gpu_maxHistoryTemp(struct xdx_device *xdev)
{
	s16 gpu_maxHistoryTemp = *((s32*)(xdev->rmmio + PANGU_PUBLIC_FW_INFO_BASE + offsetof(PublicFirmwareInfo_T, maxHistoryTemp)));
	return gpu_maxHistoryTemp;
}

s32 get_gddr_temp(struct xdx_device *xdev)
{
    return *((s32*)(xdev->rmmio + PANGU_PUBLIC_FW_INFO_BASE + offsetof(PublicFirmwareInfo_T, gddrTemp)));
}

u32 get_gpu_freq(struct xdx_device *xdev)
{
	u64 gpu_info = *((u64*)(xdev->rmmio + PANGU_PUBLIC_FW_INFO_BASE + offsetof(PublicFirmwareInfo_T, gpuInfo)));
	u8 gpu_clock = (gpu_info >> 4) & 0xFF;
	return xdx_gpu_clock_freq_list[gpu_clock];
}

u32 get_fan_rpm(struct xdx_device *xdev)
{
    return *((u32*)(xdev->rmmio + PANGU_PUBLIC_FW_INFO_BASE + offsetof(PublicFirmwareInfo_T, curRPM)));
}

u32 get_fan_pwm(struct xdx_device *xdev)
{
    return *((u32*)(xdev->rmmio + PANGU_PUBLIC_FW_INFO_BASE + offsetof(PublicFirmwareInfo_T, curDuty)));
}

u32 get_gpu_volt(struct xdx_device *xdev)
{
	return *((u32*)(xdev->rmmio + PANGU_PUBLIC_FW_INFO_BASE + offsetof(PublicFirmwareInfo_T, curVolt)));
}

u32 get_gpu_current(struct xdx_device *xdev)
{
	return *((u32*)(xdev->rmmio + PANGU_PUBLIC_FW_INFO_BASE + offsetof(PublicFirmwareInfo_T, curCurr)));
}

u32 get_gpu_power(struct xdx_device *xdev)
{
	return *((u32*)(xdev->rmmio + PANGU_PUBLIC_FW_INFO_BASE + offsetof(PublicFirmwareInfo_T, curPower)));
}

GDDRStat_T get_gddr_crc(struct xdx_device *xdev)
{
	u32 offset = offsetof(PublicFirmwareInfo_T, gddrstate);
	
	return *((GDDRStat_T*)(xdev->rmmio + PANGU_PUBLIC_FW_INFO_BASE + offset));
}

u32 get_gddr_type(struct xdx_device *xdev)
{
	u32 offset = PARAMETER_PAGE_OFFSET + offsetof(ParameterPage_T, ddrParameter) + offsetof(DDRParameter_T, gddrType);
	if (xdev->pdev->device == 0x15A4) {  // X1900
#ifdef SUPPORT_PANGU_ACPU
		return *((u32*)(xdev->share_ram +  offset));
#else
		return -1;
#endif
	} else {
#ifdef SUPPORT_AUDIO
		return *((u32*)(xdev->xdx_hdmi_audio_hotplug_ptr->iobase +  offset));
#else
		return -1;
#endif
	}
}

int xdx_fwinfo_init(struct xdx_device *xdev)
{
	dev_info(xdev->dev, "xdxgpu public firware info init\n");
	xdev->fwinfo.fwinfo_base = xdev->rmmio_base + PANGU_PUBLIC_FW_INFO_BASE;
	xdev->fwinfo.fwinfo_size = PANGU_PUBLIC_FW_INFO_SIZE;
	xdev->fwinfo.fwinfo = ioremap(xdev->fwinfo.fwinfo_base, xdev->fwinfo.fwinfo_size);
	if (xdev->fwinfo.fwinfo == NULL) {
		dev_err(xdev->dev, "xdev failed to map xdxfw info`\n");
		return -ENODEV;
	}

    return 0;
}

int cache_fw_static_info(struct xdx_device *xdev)
{
	struct xdx_fw_static_info *info;
	u32 i = 0;
	PublicFirmwareInfo_T *fwinfo = kmalloc(sizeof(PublicFirmwareInfo_T), GFP_KERNEL);
	memcpy_fromio(fwinfo,
		      xdev->rmmio + PANGU_PUBLIC_FW_INFO_BASE,
		      sizeof(PublicFirmwareInfo_T));

	info = &xdev->fw_st_info;

	info->default_gpu_freq = xdx_gpu_clock_freq_list[fwinfo->gpuInfo.gpuFreq];

	/* Acquire GDDR Info */
	info->gddr_size = fwinfo->gddrInfo.gddrTotalSizeGB;
	info->gddr_speed = xdx_gddr_clock_speed_list[fwinfo->gddrInfo.gddrSpeed];
	info->gddr_temp = fwinfo->gddrTemp;
	info->gddr_vendor = fwinfo->gddrInfo.vendorID;

	/* Acquire GPU Info */
	info->gpu_freq = fwinfo->gpuInfo.gpuFreq;
	info->gpu_core0_volt = fwinfo->gpuInfo.gCore0Volt;
	info->gpu_core1_volt = fwinfo->gpuInfo.gCore0Volt;

	/* Acquire FW Info */
	info->fw_info_ver = fwinfo->fwInfoVersion;
	info->fw_ver = fwinfo->fwVersion;
	for(i=0; i<64; ++i) {
		info->fw_ver_str[i] = (char)((fwinfo->fwVersionString[i/4] >> 8*(i%4)) & 0xFF);
	}

	/* Acquire Vbios Info */
	info->vbios_ver = fwinfo->vbiosVersion;
	for(i=0; i<64; ++i) {
		info->vbios_ver_str[i] = (char)((fwinfo->vbiosVersionString[i/4] >> 8*(i%4)) & 0xFF);
	}

	/* Acquire Board Info */
	info->vendor = fwinfo->boardInfo;
	info->chip_id = fwinfo->chipID[1];
	for(i=0; i<32; ++i) {
		info->serial_number[i] = (char)((fwinfo->snNumber[i/4] >> 8*(i%4)) & 0xFF);
	}

	/* Acquire RPM Info */
	info->cur_rpm = fwinfo->curRPM;
	info->min_rpm = fwinfo->minRPM;
	info->max_rpm = fwinfo->maxRPM;

	/* Acquire Temp Info */
	info->gpu_temp = (s16)fwinfo->curTemp;
	info->max_temp_history = (s16)fwinfo->maxHistoryTemp;
	info->temp_min = fwinfo->tempTH1;
	info->temp_max = fwinfo->tempTH2;
	info->temp_prot = fwinfo->tempProtect;

	/* Acquire Duty Info */
	info->curr_duty = fwinfo->curDuty;
	info->min_duty = fwinfo->minDuty;
	info->max_duty = fwinfo->maxDuty;

	/* Acquire Fan Ctrl Mode */
	info->fan_ctrl_mode = fwinfo->fanControlMode;

	/* Acquire support fan info */
	info->support_fan_info = fwinfo->supportFanInfo;

	/* Acquire support power info */
	info->support_power_info = fwinfo->supportPowerInfo;

	/* Acquire support temp info */
	info->support_temp_info = fwinfo->supportTempInfo;

	/* Acquire support power info */
	info->support_power_info = fwinfo->supportPowerInfo;

	/* Acquire card type info */
	info->card_type = fwinfo->cardType;

	if((xdev->asic_type == CHIP_PANGU_B0) && (fwinfo->fwInfoVersion >= 0x10004)) {
		/* Acquire vga flag info */
		info->vgaDetectFlag = fwinfo->vgaDetectFlag;
	} else {
		/* for A0 or fwversion < 10004 ,no indication available 
		 for dp-vga, set it to default HDMI_DP */
		info->vgaDetectFlag = HDMI_DP;
	}

	kfree(fwinfo);

	return 0;
}
