/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/suspend.h>
#include <linux/vmstat.h>
#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 13, 0))
#include <linux/umh.h>
#endif

#include "xdx_pm.h"
#include "xdx_device.h"
#include "acpu/xdxacpu.h"
#include "xdx_cmcu.h"

#include "../gfx/services/server/include/device.h"
#include "../gfx/services/server/devices/volcanic/rgxdevice.h"

/*
 * In PCI thaw noirq stage during S4, it is observed there are PCIe MWr directly after device switch to D0
 *
 * This will cause host hang on PANGU, because firmware will NOT wakeup and GDDR is still NOT accessible
 * And in PCI freeze stage before that, pci_disable_device will NOT disable PCI mem region so above problem WILL happen
 *
 * Workaround the problem by explicitly disable PCI mem region when switch to PS2
 *
 * NOTE: Root cause is found: MWr is caused by fbcon not handled properly during freeze
 * Enable the workaround if similar problem found later
 *
 */
#undef PM_WR_DISABLE_PCI_MEM_EXPLICITLY

/*
 * On small memory machine (<=8GB), S4 might fail when memory usage is high (>=4GB)
 * Drop page cache and slab object only won't help. We need to swapout more mem onto disk
 *
 * S4 have knob on tuning image size. When choose to use small image size, kernel will
 * release more mem including swapout mem onto disk, thus also release mem pressure
 *
 * Enable hiber tune on small memory machine to release mem pressure
 *
 */
#define PM_HIBER_TUNE_ENABLE

static inline int pci_ps_set_check(struct xdx_device *xdev,int status)
{
	switch (status) {
	case 0:
		break;
	case -EIO:
		dev_warn(xdev->dev, "device doesn't support PCI PM\n");
		break;
	case -EINVAL:
		dev_warn(xdev->dev,
			 "can't enter requested pci power state\n");
		break;
	default:
		dev_warn(xdev->dev, "pci_set_power_state failed %d\n", status);
		break;
	}
	return status;
}

static int xdx_set_ps(struct xdx_device *xdev, pci_power_t state)
{
	int r;
	dev_dbg(xdev->dev,
		 "target D-state: %d\n",
		 state);
	r = pci_set_power_state(xdev->pdev, state);
	if (pci_ps_set_check(xdev, r))
		return r;
	r = pci_set_power_state(xdev_to_audio_pdev(xdev), state);
	dev_dbg(xdev->dev,
		 "gfx function D-state: %d, audio function D-state: %d\n",
		 xdev->pdev->current_state,
		 xdev_to_audio_pdev(xdev)->current_state);
	return pci_ps_set_check(xdev, r);
}

static int curr_ps_check(struct xdx_device *xdev,
			 enum xdx_power_state expected_ps)
{
	/* TODO: skip the check before related h02c doorbell cmd is ready */
	return 0;
}

static int xdx_ps0_to_ps1(struct xdx_device *xdev)
{
	int r;
	struct pci_dev *audio_pdev = xdev_to_audio_pdev(xdev);

	r = curr_ps_check(xdev, XDX_PS0);
	if (r)
		return r;

	r = pci_save_state(xdev->pdev);
	r = pci_save_state(audio_pdev);
	if (r) {
		dev_warn(xdev->dev, "pci_save_state failed\n");
		return r;
	}

	pci_disable_device(xdev->pdev);
	if (pci_is_enabled(audio_pdev))
		pci_disable_device(audio_pdev);

	r = xdx_set_ps(xdev, PCI_D1);
	if (r)
		return r;

	xdev->power_state = XDX_PS1;
	return 0;

}

static int xdx_ps0_to_ps2(struct xdx_device *xdev)
{
	int r;
	struct pci_dev *pdev = xdev->pdev;
	struct pci_dev *audio_pdev = xdev_to_audio_pdev(xdev);
#ifdef PM_WR_DISABLE_PCI_MEM_EXPLICITLY
        u16 cmd;
#endif

	r = curr_ps_check(xdev, XDX_PS0);
	if (r)
		return r;

	r = pci_save_state(pdev);
	r = pci_save_state(xdev_to_audio_pdev(xdev));
	if (r) {
		dev_warn(xdev->dev, "pci_save_state failed\n");
		return r;
	}
	pci_disable_device(pdev);
	if (pci_is_enabled(audio_pdev))
		pci_disable_device(audio_pdev);

#ifdef PM_WR_DISABLE_PCI_MEM_EXPLICITLY
        pci_read_config_word(pdev, PCI_COMMAND, &cmd);
        if (cmd & (PCI_COMMAND_IO | PCI_COMMAND_MEMORY)) {
                dev_dbg(xdev->dev, "clear Func0 io and mem\n");
                cmd &= ~PCI_COMMAND_IO;
                cmd &= ~PCI_COMMAND_MEMORY;
                pci_write_config_word(pdev, PCI_COMMAND, cmd);
        }
        pci_read_config_word(xdev_to_audio_pdev(xdev), PCI_COMMAND, &cmd);
        if (cmd & (PCI_COMMAND_IO | PCI_COMMAND_MEMORY)) {
                dev_dbg(xdev->dev, "clear Func1 io and mem\n");
                cmd &= ~PCI_COMMAND_IO;
                cmd &= ~PCI_COMMAND_MEMORY;
                pci_write_config_word(xdev_to_audio_pdev(xdev), PCI_COMMAND, cmd);
        }
#endif

	r = xdx_set_ps(xdev, PCI_D3hot);

	if (r)
		return r;

	xdev->power_state = XDX_PS2;
	return 0;
}

static int xdx_set_ps3(struct xdx_device *xdev)
{
	int r = xdx_set_ps(xdev, PCI_D3cold);

	if (r)
		return r;

	xdev->power_state = XDX_PS3;
	return 0;
}

static int xdx_ps2_to_ps0(struct xdx_device *xdev)
{
	int r;
	struct pci_dev *pdev = NULL;
	r = curr_ps_check(xdev, XDX_PS2);
	if (r)
		return r;
	pdev = xdev->pdev;
	r = xdx_set_ps(xdev, PCI_D0);
	if (r)
		return r;
	pci_restore_state(pdev);
	pci_restore_state(xdev_to_audio_pdev(xdev));
	r = pci_enable_device(pdev);
	r = pci_enable_device(xdev_to_audio_pdev(xdev));
	pci_set_master(pdev);
	xdev->power_state = XDX_PS0;
	return 0;
}

static int xdx_ps1_to_ps0(struct xdx_device *xdev)
{
	int r;
	struct pci_dev *audio_pdev = xdev_to_audio_pdev(xdev);

	r = curr_ps_check(xdev, XDX_PS1);
	if (r)
		return r;

	r = xdx_set_ps(xdev, PCI_D0);
	if (r)
		return r;

	pci_restore_state(xdev->pdev);
	pci_restore_state(audio_pdev);
	r = pci_enable_device(xdev->pdev);
	r = pci_enable_device(audio_pdev);

	xdev->power_state = XDX_PS0;
	return 0;
}

int xdx_switch_power_state(struct xdx_device *xdev, enum xdx_power_state target)
{
	int r = 0;

	dev_dbg(xdev->dev, "switch xdxgpu power state from PS%d to PS%d\n",
		 xdev->power_state, target);

	if (unlikely(target == XDX_PS3))
		return xdx_set_ps3(xdev);

	switch (xdev->power_state) {
	case XDX_PS0:
		switch (target) {
		case XDX_PS1:
			r = xdx_ps0_to_ps1(xdev);
			break;
		case XDX_PS2:
			r = xdx_ps0_to_ps2(xdev);
			break;
		default:
			goto invalid_ps_switch;
		}
		break;
	case XDX_PS1:
		switch (target) {
		case XDX_PS0:
			r = xdx_ps1_to_ps0(xdev);
			break;
		case XDX_PS2:
			r = xdx_ps1_to_ps0(xdev);
			if (!r)
				r = xdx_ps0_to_ps2(xdev);
			break;
		default:
			goto invalid_ps_switch;
		}
		break;
	case XDX_PS2:
		switch (target) {
		case XDX_PS0:
			r = xdx_ps2_to_ps0(xdev);
			break;
		case XDX_PS1:
			r = xdx_ps2_to_ps0(xdev);
			if (!r)
				r = xdx_ps0_to_ps1(xdev);
			break;
		default:
			goto invalid_ps_switch;
		}
		break;
	/* TODO: cold reset to transfer from PS3 to PS0 again */
	default:
invalid_ps_switch:
		DRM_WARN("Doesn't support switch from PS%d to PS%d\n",
			 xdev->power_state, target);
		r = -EINVAL;
	}
	return curr_ps_check(xdev, target);
}

void xdx_dvfs_set_freq(struct xdx_device *xdev, u32 gfx_clk)
{
	PVRSRV_DEVICE_NODE *psDeviceNode;
	IMG_HANDLE devData = ((PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev)->psDevConfig->hDevData;

	CMCUPowerCmd_T power_cmd;
	power_cmd.cmd = cH02CPowerDBCmdDVFS;
	power_cmd.subcmd = cH02CPowerDVFSSubCmdSetFreq;
	power_cmd.channel = 1;
	power_cmd.data = gfx_clk;
	psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;

	if (PVRSRV_OK != PVRSRVDevicePreClockSpeedChange(psDeviceNode,
													 IMG_TRUE,
													 NULL))
	{
		PVR_DPF((PVR_DBG_ERROR, "PVRSRVDevicePreClockSpeedChange failed"));
		return;
	}

#if defined(SUPPORT_PANGU_ACPU)
	H02CDoorBellPowerSet(xdev, power_cmd.all);
#endif

	xdev->pm.current_gfxclk = gfx_clk;
	((RGX_DATA*) devData)->psRGXTimingInfo->ui32CoreClockSpeed = gfx_clk;

	PVRSRVDevicePostClockSpeedChange(psDeviceNode, IMG_TRUE, NULL);
}

enum xdx_power_state xdx_power_state_get(struct xdx_device *xdev)
{
	/* TODO: send cmd to FW through h02c doorbell */
	return XDX_PS0;
}

long xdx_power_get(struct xdx_device *xdev, uint8_t subcmd,
		   uint8_t type, uint8_t channel)
{
	power_cmd_type power_cmd;

	power_cmd.thermal.cmd = cCmcuHostDBCmdDDKPower;
	power_cmd.thermal.subCmd = subcmd;
	power_cmd.thermal.type = type;
	power_cmd.thermal.channel = channel;
#if defined(SUPPORT_PANGU_ACPU)
	return H02CDoorBellPowerGet(xdev, power_cmd.all);
#endif
	return 0;
}

void xdx_power_set(struct xdx_device *xdev, uint8_t cmd,
		   uint8_t subcmd, uint8_t channel, long val)
{
	CMCUPowerCmd_T power_cmd;
	PVRSRV_DEVICE_NODE *psDeviceNode;

	power_cmd.cmd = cmd;
	power_cmd.subcmd = subcmd;
	power_cmd.channel = channel;
	power_cmd.data = val;

	if (cmd == cH02CPowerDBCmdDVFS && subcmd == cH02CPowerDVFSSubCmdSetVolt) {
		psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;
		if (PVRSRV_OK != PVRSRVDevicePreClockSpeedChange(psDeviceNode,
														IMG_TRUE,
														NULL))
		{
			PVR_DPF((PVR_DBG_ERROR, "PVRSRVDevicePreClockSpeedChange failed"));
			return;
		}

#if defined(SUPPORT_PANGU_ACPU)
		H02CDoorBellPowerSet(xdev, power_cmd.all);
#endif

		PVRSRVDevicePostClockSpeedChange(psDeviceNode, IMG_TRUE, NULL);
	} else {
#if defined(SUPPORT_PANGU_ACPU)
		H02CDoorBellPowerSet(xdev, power_cmd.all);
#endif
	}
}

void xdx_thermal_set(struct xdx_device *xdev, uint8_t cmd,
		   uint8_t subcmd, uint8_t channel, long val)
{
	CMCUThermalCmd_T thermal_cmd;

	thermal_cmd.cmd = cmd;
	thermal_cmd.subcmd = subcmd;
	thermal_cmd.channel = channel;
	thermal_cmd.data = val;

#if defined(SUPPORT_PANGU_ACPU)
	H02CDoorBellThermalSet(xdev, thermal_cmd.all);
#endif
}

#ifdef PM_HIBER_TUNE_ENABLE
/* Small memory machine threshold */
static uint64_t xdx_pm_hiber_tune_threshold = (8ULL << 30); /* <=8GB ram */

/* S4 image percentage of ram */
#define XDX_PM_IMG_PERCENT_VERY_LOWMEM   0      /* very low mem pressure but slow to restore */
#define XDX_PM_IMG_PERCENT_LOWMEM        22     /* low mem pressure */
#define XDX_PM_IMG_PERCENT_DEFAULT       40     /* sys default with high mem pressure */
static uint8_t xdx_pm_hiber_tune_percent = XDX_PM_IMG_PERCENT_LOWMEM;

/* S4 image tuning */
static void xdx_pm_hiber_tune(struct xdx_device *xdev)
{
        static bool xdx_pm_hiber_tune_init = false;
        struct sysinfo si = {0};
        uint64_t totalram = 0;
        char sizearg[256] = {0};
        char *argv[] = {
                "/bin/sh",
                "-c",
                sizearg,
                NULL
        };
        char *envp[] = {
                "HOME=/",
                "TERM=linux",
                "PATH=/usr/sbin",
                 NULL
        };
        int ret = 0;

        if (xdx_pm_hiber_tune_init == true)
                return;
        xdx_pm_hiber_tune_init = true;

        si_meminfo(&si);
        totalram = si.totalram << PAGE_SHIFT;
        if (totalram > xdx_pm_hiber_tune_threshold)
                return;

        /* TODO: take consider used size and use default percent when mem pressure is low  */
        dev_info(xdev->dev, "pm hiber tune: ram=%lu(page) percent=%u\n",
                 si.totalram, xdx_pm_hiber_tune_percent);
        sprintf(sizearg, "echo %llu > /sys/power/image_size",
                totalram / 100 * xdx_pm_hiber_tune_percent);
        ret = call_usermodehelper("/bin/sh", argv, envp, UMH_WAIT_PROC);
        if (ret < 0) {
                dev_warn(xdev->dev, "pm hiber tune fail: call user mode ret=%d!!\n", ret);
                BUG_ON(0);
        }

        return;
}
#else
static void xdx_pm_hiber_tune(struct xdx_device *xdev)
{
        return;
}
#endif

static void xdx_pm_drop_cache(struct xdx_device *xdev)
{
        char *argv[] = {
                "/usr/sbin/sysctl",
                "-w",
                "vm.drop_caches=3",
                NULL
        };
        char *envp[] = {
                "HOME=/",
                "TERM=linux",
                "PATH=/usr/sbin",
                 NULL
        };
        int ret = 0;

        dev_dbg(xdev->dev, "pm drop cache\n");
#if (defined(XDXGPU_CENTOS79_SUPPORT))
        dev_dbg(xdev->dev, "pm drop cache begin pages=%ld\n", global_page_state(NR_FILE_PAGES));
#else
        dev_dbg(xdev->dev, "pm drop cache begin pages=%ld\n", global_node_page_state(NR_FILE_PAGES));
#endif
        ret = call_usermodehelper("/usr/sbin/sysctl", argv, envp, UMH_WAIT_PROC);
        if (ret < 0) {
                dev_warn(xdev->dev, "pm drop cache fail: call user mode ret=%d!!\n", ret);
                BUG_ON(0);
        }
#if (defined(XDXGPU_CENTOS79_SUPPORT))
        dev_dbg(xdev->dev, "pm drop cache begin pages=%ld\n", global_page_state(NR_FILE_PAGES));
#else
        dev_dbg(xdev->dev, "pm drop cache end pages=%ld\n", global_node_page_state(NR_FILE_PAGES));
#endif

        return;
}
static int xdx_pm_notifier(struct notifier_block *nb, unsigned long val,
			   void *ptr)
{
	struct xdx_pm *xpm =
		container_of(nb, struct xdx_pm, notifier);
        struct xdx_device *xdev = xpm->dev;

	switch (val) {
	case PM_HIBERNATION_PREPARE:
                dev_info(xdev->dev, "pm notifier: val=%ld\n", val);
                xdx_pm_hiber_tune(xdev);

                /* TODO: May be unncessary after set size. Remove after test */
                xdx_pm_drop_cache(xdev);
                break;

        case PM_SUSPEND_PREPARE:
                dev_info(xdev->dev, "pm notifier: val=%ld\n", val);
                /* TODO: May be unncessary. Remove after test */
                xdx_pm_drop_cache(xdev);
		break;

	default:
		break;
	}

	return 0;
}

int xdx_pm_init(struct xdx_device *xdev)
{
	int ret;

	dev_info(xdev->dev, "pm init\n");

        /* base */
        xdev->pm.dev = xdev;

        /* thm */
	ret = xdx_thm_init(xdev);
	if(ret != 0) {
		dev_err(xdev->dev, "%s : failed to init thm!!!\n", __func__);
		goto err1;
	}

        /* cmcu */
	ret = xdx_cmcu_init(xdev);
	if(ret != 0) {
		dev_err(xdev->dev, "%s : failed to init cmcu!!!\n", __func__);
		goto err2;
	}

        /* notifier */
        xdev->pm.notifier.notifier_call = xdx_pm_notifier;
        register_pm_notifier(&xdev->pm.notifier);

	return 0;

err2:
	xdx_thm_fini(xdev);
err1:
	return ret;
}

int xdx_pm_fini(struct xdx_device *xdev) {
	int ret;

        /* notifier */
        unregister_pm_notifier(&xdev->pm.notifier);

	/**
	 * TODO:
	 * If xdx_cmcu_fini() or xdx_thm_fini() failed, then a strategy should
	 * be applied to deal with it, like init thm again or other methods
	 */
        /* cmcu */
	ret = xdx_cmcu_fini(xdev);
	if(ret != 0) {
		dev_err(xdev->dev, "%s : failed to fini cmcu!!!\n", __func__);
		return ret;
	}

        /* thm */
	ret = xdx_thm_fini(xdev);
	if(ret != 0) {
		dev_err(xdev->dev, "%s : failed to fini thm!!!\n", __func__);
		return ret;
	}

	return 0;
}

void xdx_runtime_suspend(struct xdx_device *xdev)
{
	int r;

	r =  xdx_switch_power_state(xdev, XDX_PS1);

	if (r) {
		dev_err(xdev->dev,
			"failed to switch xgpu power state to PS%u\n",
			XDX_PS1);
	}
}

void xdx_runtime_resume(struct xdx_device *xdev)
{
	int r;

	r =  xdx_switch_power_state(xdev, XDX_PS0);

	if (r) {
		dev_err(xdev->dev,
			"failed to switch xgpu power state to PS%u\n",
			XDX_PS0);
	}
}

/*
 * to support runtime low power mode with display on
 * one gddr phy (phy3) is still active during low power mode
 * on 2 phy GPU, it's 4K interleave on phy23
 * on 4 phy GPU, it's 4K interleave on phy0123
 * input is the dev_phy_addr of a gddr page
 * then return value is its offset to the next valid addr on phy3
 */
uint64_t rpm_valid_gddr_offset(struct xdx_device *xdev,
			       uint64_t dev_phy_addr)
{
	return PAGE_SIZE * (xdev->fw_st_info.gddr_count - 1) -
	       dev_phy_addr % (PAGE_SIZE * xdev->fw_st_info.gddr_count);
}
