/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#include <linux/kernel.h>
#include "osfunc.h"
#include "smmu_platform.h"
#include "smmu_common.h"
#include "xdx_smmu_drv.h"
#include "xdx_device.h"

struct smmu_level_para_s {
	uint32_t lvl_starts[SMMU_MAX_LEVEL];
	uint32_t lvl_ends[SMMU_MAX_LEVEL];
	enum smmu_level_e lvls[SMMU_MAX_LEVEL];
};
static const uint64_t gsSmmuBadDevPhyAddr = MMU_BAD_PHYS_ADDR;

static inline const struct smmu_pxe_config_s *xdx_smmu_get_pxe_cfg(
	struct smmu_context_s *smmu_ctx, enum smmu_level_e lvl
)
{
	if (lvl <= SMMU_LEVEL_0 || lvl >= SMMU_LEVEL_LAST)
		return NULL;
	return smmu_ctx->page_cfg->pxe_cfgs[lvl - 1];
}

static const struct smmu_pxe_config_s *xdx_smmu_get_pte_cfg(
		struct smmu_context_s *smmu_ctx)
{
	return xdx_smmu_get_pxe_cfg(smmu_ctx, SMMU_LEVEL_1);
}

static const struct smmu_page_config_s *xdx_smmu_get_page_cfg(
		struct smmu_context_s *smmu_ctx)
{
	return smmu_ctx->page_cfg;
}

static uint32_t xdx_smmu_page_size(struct smmu_context_s *smmu_ctx)
{
	return smmu_ctx->page_cfg->page_size;
}

static uint32_t xdx_smmu_get_pxe_idx(
		struct smmu_context_s *smmu_ctx, int32_t level,
		uint64_t dev_vaddr, bool round_up)
{
	uint64_t dev_vaddr_temp;
	uint32_t idx;
	const struct smmu_page_config_s *page_cfg = xdx_smmu_get_page_cfg(smmu_ctx);
	const struct smmu_pxe_config_s *pxe_cfg;

	dev_vaddr_temp = dev_vaddr;
	dev_vaddr_temp -= page_cfg->offset_in_page;
	pxe_cfg = xdx_smmu_get_pxe_cfg(smmu_ctx, level);
	if (!pxe_cfg)
		return ~0;
	if (round_up)
		dev_vaddr_temp--;
	idx = (uint32_t) ((dev_vaddr_temp & pxe_cfg->uiIndexMask) >> pxe_cfg->uiIndexShift);
	if (round_up)
		idx++;

	return idx;
}

static int xdx_smmu_alloc_memdesc(struct smmu_context_s *smmu_ctx,
		const struct smmu_pxe_config_s *pxe_cfg, struct smmu_mem_desc_s *psMemDesc)
{
	int ret;
	size_t uiBytes;
	size_t uiAlign;

	uiBytes = pxe_cfg->uiNumEntries * pxe_cfg->uiBytesPerEntry;
	uiAlign = 1 << pxe_cfg->uiAddrLog2Align;
	uiBytes = ALIGN(uiBytes, uiAlign);

	ret = xdx_smmu_alloc_phys_mem(smmu_ctx, psMemDesc, uiBytes, uiAlign);
	if (ret != 0) {
		SMMU_ERR(xdx_smmu_dev(smmu_ctx), "alloc phys failed. %d",  ret);
		return ret;
	}
	OSDeviceMemSetIO(psMemDesc->pvCpuVAddr, 0, uiBytes);
	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
		"alloc phys pvCpuVAddr. %pK + %lx", psMemDesc->pvCpuVAddr, uiBytes);
	ret = xdx_smmu_phypage_clean(smmu_ctx, psMemDesc, pxe_cfg,
			psMemDesc->uiOffset, psMemDesc->uiSize);
	if (ret < 0)
		goto e0;
	return 0;
e0:
	xdx_smmu_free_phys_mem(smmu_ctx, psMemDesc);
	return ret;
}

static void xdx_smmu_memdesc_free(
		struct smmu_context_s *smmu_ctx,
		struct smmu_mem_desc_s *psMemDesc)
{
#if defined(SMMU_CLEARMEM_ON_FREE)
	OSDeviceMemSetIO(psMemDesc->pvCpuVAddr, 0, psMemDesc->uiSize);
#endif
	xdx_smmu_free_phys_mem(smmu_ctx, psMemDesc);
}

static INLINE int xdx_smmu_setup_pte(
		struct smmu_context_s *smmu_ctx,
		struct smmu_lvlx_info_s *level_entry,
		uint32_t uiIndex, const uint64_t pdev_phy_addr)
{
	struct smmu_mem_desc_s *psMemDesc = &level_entry->sMemDesc;
	uint64_t ui64PxE64;
	uint64_t uiAddr = pdev_phy_addr;
	union TTEntry_u *pui64Px = (union TTEntry_u *)psMemDesc->pvCpuVAddr;
	union TTEntry_u TTentry;

	/* Set the entry */
	TTentry.all = UINT64_C(0);
	TTentry.all = uiAddr;
	TTentry.ptPage.valid = TT_VALID;
	TTentry.ptPage.type = TT_TYPE_PAGE;
	TTentry.ptPage.AttrIndx = NORMAL_MEMORY_Outer_Inner_WriteBack_ReadWriteAllocate_INDEX;
	TTentry.ptPage.SH = SH_OUTER_SHAREABLE;
	TTentry.ptPage.AF = TT_ENTRY_ACCESSED;
	TTentry.ptPage.XN = TT_EXECUTE_NEVER;
	TTentry.ptPage.AP = 1;
	ui64PxE64 = TTentry.all;
	xdx_smmu_copy_pxe_entry(&pui64Px[uiIndex], &TTentry, sizeof(TTentry));
	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
		"pui64Px[%d] = 0x%llx uiAddr = 0x%llx level_entry = 0x%llx ui32RefCount = 0x%x",
		uiIndex, (uint64_t)&pui64Px[uiIndex],
		uiAddr, (uint64_t)level_entry, level_entry->ui32RefCount);
	return 0;
}

static int xdx_smmu_setup_pxe(struct smmu_context_s *smmu_ctx,
		struct smmu_lvlx_info_s *level_entry,
		uint32_t uiIndex, enum smmu_level_e smmu_lvl,
		const uint64_t pdev_phy_addr)
{
	struct smmu_mem_desc_s *psMemDesc = &level_entry->sMemDesc;

	if (pdev_phy_addr == MMU_BAD_PHYS_ADDR)
		level_entry->ui32RefCount--;
	else
		level_entry->ui32RefCount++;

	switch (smmu_lvl) {
	case SMMU_LEVEL_4:
	case SMMU_LEVEL_3:
	case SMMU_LEVEL_2:
	{
		union TTEntry_u *pui64Px = (union TTEntry_u *)psMemDesc->pvCpuVAddr;
		union TTEntry_u TTentry;

		TTentry.all = UINT64_C(0);
		TTentry.tbl.valid = TT_VALID;
		TTentry.tbl.type = TT_TYPE_TABLE;
		TTentry.tbl.nextTblAddr = pdev_phy_addr >> NEXT_TT_ADDRESS_SHIFT;
		xdx_smmu_copy_pxe_entry(&pui64Px[uiIndex], &TTentry, sizeof(TTentry));
		break;
	}
	case SMMU_LEVEL_1:
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static struct smmu_lvlx_info_s *smmu_alloc_lvl_info(
	struct smmu_context_s *smmu_ctx, const struct smmu_pxe_config_s *pxe_cfg)
{
	uint32_t desc_size;
	struct smmu_lvlx_info_s *level_info;
	int ret;

	desc_size = sizeof(struct smmu_lvlx_info_s) +
				(sizeof(struct smmu_lvlx_info_s *) * pxe_cfg->uiNumEntries);
	level_info = kzalloc(desc_size, GFP_KERNEL);
	if (!level_info)
		return NULL;
	level_info->ui32RefCount = 0;
	level_info->ui32NumOfEntries = pxe_cfg->uiNumEntries;
	ret = xdx_smmu_alloc_memdesc(smmu_ctx, pxe_cfg, &level_info->sMemDesc);
	if (ret) {
		kfree(level_info);
		return NULL;
	}
	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
			"Alloc lvl info:%x %pK", pxe_cfg->uiNumEntries, level_info);
	return level_info;
}

static void smmu_free_lvl_info(
	struct smmu_context_s *smmu_ctx, struct smmu_lvlx_info_s *level_info)
{
	SMMU_DBG(xdx_smmu_dev(smmu_ctx), "Free lvl info:%pK", level_info);
	xdx_smmu_memdesc_free(smmu_ctx, &level_info->sMemDesc);
	kfree(level_info);
}

static bool xdx_smmu_free_level(struct smmu_context_s *smmu_ctx,
		struct smmu_lvlx_info_s *level_entry, struct smmu_level_para_s *lvl_para,
		uint32_t lvl_idx, uint32_t start_idx, uint32_t end_idx,
		bool is_first, bool is_last)
{
	uint32_t i;
	int ret;

	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
			"current lvl:%d :%x -> %x ", lvl_para->lvls[lvl_idx],
			lvl_para->lvl_starts[lvl_idx], lvl_para->lvl_ends[lvl_idx]);
	for (i = start_idx; (i < end_idx) && (level_entry != NULL); i++) {
		if (lvl_para->lvls[lvl_idx] != SMMU_LEVEL_1) {
			const struct smmu_pxe_config_s *next_pxe_cfg;
			struct smmu_lvlx_info_s *psNextLevel;
			uint32_t next_start_idx;
			uint32_t next_end_idx;
			bool is_first_next;
			bool is_end_next;

			next_pxe_cfg = xdx_smmu_get_pxe_cfg(smmu_ctx, lvl_para->lvls[lvl_idx + 1]);
			if (!level_entry->next_level_info[i]) {
				SMMU_DBG(xdx_smmu_dev(smmu_ctx),
					"lvl:%d idx:%x not alloc", lvl_para->lvls[lvl_idx], i);
				continue;
			}
			psNextLevel = level_entry->next_level_info[i];
			if (is_last && (i == (end_idx - 1))) {
				next_end_idx = lvl_para->lvl_ends[lvl_idx + 1];
				is_end_next = 1;
			} else {
				next_end_idx = next_pxe_cfg->uiNumEntries;
				is_end_next = 0;
			}
			if (next_end_idx == 0)
				return 0;
			if (is_first && (i == start_idx)) {
				next_start_idx = lvl_para->lvl_starts[lvl_idx + 1];
				is_first_next = 1;
			} else {
				next_start_idx = 0;
				is_first_next = 0;
			}
			ret = xdx_smmu_free_level(smmu_ctx, psNextLevel, lvl_para, lvl_idx + 1,
					next_start_idx, next_end_idx, is_first_next, is_end_next);
			if (ret) {
				if (!psNextLevel->ui32RefCount) {
					ret = xdx_smmu_setup_pxe(smmu_ctx, level_entry,
						i, lvl_para->lvls[lvl_idx], MMU_BAD_PHYS_ADDR);
					SMMU_DBG(xdx_smmu_dev(smmu_ctx),
						"Free Nextlvl:%d idx:%x  psNextLevel:%pK",
						lvl_para->lvls[lvl_idx], i, psNextLevel);
					smmu_free_lvl_info(smmu_ctx, psNextLevel);
					level_entry->next_level_info[i] = NULL;
				}
			}
		} else {
			level_entry->ui32RefCount--;
		}
	}

	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
			"end: level = %d, refcount = %d level_entry = 0x%llx",
			lvl_para->lvls[lvl_idx],
			level_entry->ui32RefCount, (uint64_t)level_entry);
	/* Level one flushing is done when we actually write the table entries */
	if ((lvl_para->lvls[lvl_idx] != SMMU_LEVEL_1) && (level_entry != NULL)) {
		const struct smmu_pxe_config_s *pxe_cfg;

		pxe_cfg = xdx_smmu_get_pxe_cfg(smmu_ctx, lvl_para->lvls[lvl_idx + 1]);
		xdx_smmu_phypage_clean(smmu_ctx,
				&level_entry->sMemDesc, pxe_cfg, start_idx, (end_idx - start_idx));
	}
	return 1;
}

static int xdx_smmu_alloc_level(struct smmu_context_s *smmu_ctx,
	struct smmu_lvlx_info_s *level_entry, struct smmu_level_para_s *lvl_para,
	struct smmu_level_para_s *lvl_para_done, uint32_t current_lvl,
	uint32_t start_idx, uint32_t end_idx,
	bool is_first, bool is_last)
{
	int ret = 0;
	uint32_t i;
	uint32_t next_start_idx;
	uint32_t next_end_idx;
	const struct smmu_pxe_config_s *pxe_cfg;

	lvl_para_done->lvl_starts[current_lvl] = lvl_para->lvl_starts[current_lvl];
	lvl_para_done->lvl_ends[current_lvl] = 0;
	lvl_para_done->lvls[current_lvl] = lvl_para->lvls[current_lvl];

	SMMU_DBG(xdx_smmu_dev(smmu_ctx), "current lvl level_entry:%pK :%d :%x -> %x ",
			level_entry, lvl_para->lvls[current_lvl],
			lvl_para->lvl_starts[current_lvl], lvl_para->lvl_ends[current_lvl]);
	for (i = start_idx; i < end_idx; i++) {
		if (lvl_para->lvls[current_lvl] != SMMU_LEVEL_1) {
			const struct smmu_pxe_config_s *next_pxe_cfg;
			bool is_first_next;
			bool is_end_next;
			struct smmu_lvlx_info_s *psNextLevel;

			next_pxe_cfg = xdx_smmu_get_pxe_cfg(smmu_ctx,
					lvl_para->lvls[current_lvl + 1]);
			if (!level_entry->next_level_info[i]) {
				psNextLevel = smmu_alloc_lvl_info(smmu_ctx, next_pxe_cfg);
				if (!psNextLevel)
					goto errout;
				level_entry->next_level_info[i] = psNextLevel;
				/* Wire up the entry */
				ret = xdx_smmu_setup_pxe(smmu_ctx, level_entry, i,
						lvl_para->lvls[current_lvl],
						psNextLevel->sMemDesc.dev_phy_addr.uiAddr);
			} else {
//				level_entry->ui32RefCount++;
				psNextLevel = level_entry->next_level_info[i];
				ret = 0;
			}
			lvl_para_done->lvl_ends[current_lvl] = i + 1;
			if (ret)
				goto errout;

			/* If we're crossing a Px then the start index changes */
			if (is_first && (i == start_idx)) {
				next_start_idx = lvl_para->lvl_starts[current_lvl + 1];
				is_first_next = 1;
			} else {
				next_start_idx = 0;
				is_first_next = 0;
			}

			/* If we're crossing a Px then the end index changes */
			if (is_last && (i == (end_idx - 1))) {
				next_end_idx = lvl_para->lvl_ends[current_lvl + 1];
				is_end_next = 1;
			} else {
				next_end_idx = next_pxe_cfg->uiNumEntries;
				is_end_next = 0;
			}
			/* Recurse into the next level */
			ret = xdx_smmu_alloc_level(smmu_ctx, psNextLevel, lvl_para, lvl_para_done,
					current_lvl + 1, next_start_idx,
					next_end_idx, is_first_next, is_end_next);
			if (ret)
				goto errout;
		} else {
			level_entry->ui32RefCount++;
			lvl_para_done->lvl_ends[current_lvl] = i + 1;
		}
		if (level_entry->ui32RefCount > level_entry->ui32NumOfEntries) {
			SMMU_ERR(xdx_smmu_dev(smmu_ctx),
					"level_entry ui32RefCount:%x > ui32NumOfEntries :%x",
					level_entry->ui32RefCount, level_entry->ui32NumOfEntries);
		}
	}

	/* Level one flushing is done when we actually write the table entries */
	if (lvl_para->lvls[current_lvl] != SMMU_LEVEL_1) {
		pxe_cfg = xdx_smmu_get_pxe_cfg(smmu_ctx, lvl_para->lvls[current_lvl]);
		ret = xdx_smmu_phypage_clean(smmu_ctx, &level_entry->sMemDesc, pxe_cfg,
				start_idx, end_idx - start_idx);
		if (ret != 0)
			goto errout;
	}
	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
			"alloc level end:level=%d,refcount=%d level_entry=0x%llx",
			lvl_para->lvls[current_lvl], level_entry->ui32RefCount,
			(uint64_t)level_entry);
	return 0;
errout:
	if (current_lvl == 0 && lvl_para_done->lvl_ends[current_lvl] != 0)
		xdx_smmu_free_level(smmu_ctx, level_entry, lvl_para_done, current_lvl,
			lvl_para_done->lvl_starts[current_lvl],
			lvl_para_done->lvl_ends[current_lvl], 1, 1);
	return ret;
}

static void _xdx_smmu_get_level_data(
		struct smmu_context_s *smmu_ctx,
		uint64_t dev_vaddrStart,
		uint64_t dev_vaddrEnd, struct smmu_level_para_s *lvl_para
)
{
	const struct smmu_page_config_s *page_cfg = xdx_smmu_get_page_cfg(smmu_ctx);
	uint32_t i = 0;
	enum smmu_level_e smmu_level;
	const struct smmu_pxe_config_s *pxe_cfg;

	for (smmu_level = page_cfg->top_level, i = 0;
					smmu_level > SMMU_LEVEL_0; smmu_level--,  i++) {
		pxe_cfg = xdx_smmu_get_pxe_cfg(smmu_ctx, smmu_level);
		if (pxe_cfg->uiIndexMask != 0) {
			lvl_para->lvl_starts[i] =
				xdx_smmu_get_pxe_idx(smmu_ctx, smmu_level, dev_vaddrStart, 0);
			lvl_para->lvl_ends[i] =
				xdx_smmu_get_pxe_idx(smmu_ctx, smmu_level, dev_vaddrEnd, 1);
			SMMU_DBG(xdx_smmu_dev(smmu_ctx),
				"level:%d lvl_starts:%x, lvl_ends:%x", smmu_level,
				lvl_para->lvl_starts[i], lvl_para->lvl_ends[i]);
			lvl_para->lvls[i] = smmu_level;
		}
	}
}

static int xdx_smmu_alloc_page_tabs(struct smmu_context_s *smmu_ctx,
		uint64_t dev_vaddrStart, uint64_t dev_vaddrEnd, uint32_t uiLog2DataPageSize)
{
	int ret;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	struct smmu_level_para_s lvl_para = {0};
	struct smmu_level_para_s lvl_done_para = {0};
#else
	struct smmu_level_para_s lvl_para;
	struct smmu_level_para_s lvl_done_para;

	memset(&lvl_para, 0, sizeof(lvl_para));
	memset(&lvl_done_para, 0, sizeof(lvl_done_para));
#endif

	_xdx_smmu_get_level_data(smmu_ctx, dev_vaddrStart, dev_vaddrEnd, &lvl_para);
	ret = xdx_smmu_alloc_level(smmu_ctx, smmu_ctx->base_lvl_info, &lvl_para,
				&lvl_done_para, 0,
				lvl_para.lvl_starts[0], lvl_para.lvl_ends[0], 1, 1);
	return ret;
}

static void xdx_smmu_free_page_tabs(struct smmu_context_s *smmu_ctx,
		uint64_t dev_vaddrStart, uint64_t dev_vaddrEnd, uint32_t uiLog2DataPageSize)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	struct smmu_level_para_s lvl_para = {0};
#else
	struct smmu_level_para_s lvl_para;
	memset(&lvl_para, 0, sizeof(lvl_para));
#endif

	_xdx_smmu_get_level_data(smmu_ctx, dev_vaddrStart, dev_vaddrEnd, &lvl_para);
	xdx_smmu_free_level(smmu_ctx, smmu_ctx->base_lvl_info, &lvl_para, 0,
			lvl_para.lvl_starts[0], lvl_para.lvl_ends[0], 1, 1);
}

static INLINE int xdx_smmu_get_pt_info(
		struct smmu_context_s *smmu_ctx, uint64_t dev_vaddr,
		struct smmu_lvlx_info_s **plevel_entry, uint32_t *pte_index)
{
	struct smmu_lvlx_info_s *lvl_info = NULL;
	enum smmu_level_e smmu_lvl;
	uint32_t pxe_idx;
	const struct smmu_page_config_s *page_cfg = xdx_smmu_get_page_cfg(smmu_ctx);

	for (smmu_lvl = page_cfg->top_level ; smmu_lvl > SMMU_LEVEL_0; smmu_lvl--) {
		pxe_idx = xdx_smmu_get_pxe_idx(smmu_ctx, smmu_lvl, dev_vaddr, 0);
		if (smmu_lvl != SMMU_LEVEL_1) {
			if (lvl_info != NULL)
				lvl_info = lvl_info->next_level_info[pxe_idx];
			else
				lvl_info = smmu_ctx->base_lvl_info->next_level_info[pxe_idx];
		} else {
/* find PTE index into page table */
			*pte_index = pxe_idx;
			if (lvl_info == NULL)
				lvl_info = smmu_ctx->base_lvl_info;
		}
	}
	*plevel_entry = lvl_info;
	return 0;
}

static int xdx_smmu_ctx_creat_top_lvl_info(struct smmu_context_s *smmu_ctx)
{
	struct smmu_lvlx_info_s *base_info;
	uint32_t size;
	const struct smmu_pxe_config_s *top_pxe_cfg;
	const struct smmu_page_config_s *page_cfg = xdx_smmu_get_page_cfg(smmu_ctx);

	top_pxe_cfg = xdx_smmu_get_pxe_cfg(smmu_ctx, page_cfg->top_level);
	size = sizeof(struct smmu_lvlx_info_s) +
			top_pxe_cfg->uiNumEntries * sizeof(struct smmu_lvlx_info_s *);
	base_info =  smmu_alloc_lvl_info(smmu_ctx, top_pxe_cfg);
	if (base_info == NULL)
		return -ENOMEM;
	SMMU_DBG(xdx_smmu_dev(smmu_ctx), "Top Level creat. %pK", base_info);
	smmu_ctx->base_lvl_info = base_info;
	return 0;
}

static int xdx_smmu_ctx_destroy_top_lvl_info(struct smmu_context_s *smmu_ctx)
{
	smmu_free_lvl_info(smmu_ctx, smmu_ctx->base_lvl_info);
	smmu_ctx->base_lvl_info = NULL;
	SMMU_DBG(xdx_smmu_dev(smmu_ctx), "Top Level destroy.");
	return 0;
}

int xdx_smmu_ctx_creat(struct xdx_device *xdev,
			const struct smmu_grp_para_s *grp_para,
			const struct smmu_page_config_s *pcfg)
{
	struct smmu_context_s *smmu_ctx;
	int ret = 0;
	struct xdx_smmu_s *xsmmu = xdev_to_smmu_dev(xdev);

	if (!xsmmu) {
		dev_info(xdev->dev, "smmu context init failed (XSmmu not inited.)");
		return -ENOMEM;
	}
	if (!grp_para|| !grp_para->gsid_tab || !grp_para->gsid_cnt || !pcfg ) {
		dev_info(xdev->dev, "smmu context init failed (Invalid context parameters.)");
		return -EINVAL;
	}
	smmu_ctx = devm_kcalloc(xdev->dev, 1, sizeof(struct smmu_context_s), GFP_KERNEL);
	if (!smmu_ctx)
		return -ENOMEM;

	INIT_LIST_HEAD(&smmu_ctx->list);
	INIT_LIST_HEAD(&smmu_ctx->maped_head);
	mutex_init(&smmu_ctx->lock);
	smmu_ctx->page_cfg = pcfg;
	smmu_ctx->group = xdx_smmu_creat_group(xsmmu,
			grp_para->gsid_cnt, grp_para->gsid_tab, grp_para->flags);
	if (!smmu_ctx->group) {
		ret = -EIO;
		SMMU_ERR(xdx_smmu_dev(smmu_ctx), "XDX SMMU group creat failed.");
		goto e0;
	}
	smmu_ctx->xsmmu = xsmmu;
	SMMU_DBG(xdx_smmu_dev(smmu_ctx), "XDX SMMU group creat Success.");
	/* Record Device specific attributes in the context for subsequent use */
	ret = xdx_smmu_ctx_env_init(xdev, smmu_ctx,
			grp_para->vaddr_rang_start, grp_para->vaddr_rang_size);
	if (ret < 0)
		goto e1;
	ret = xdx_smmu_ctx_creat_top_lvl_info(smmu_ctx);
	if (ret < 0)
		goto e2;
	xdx_smmu_group_config_pagetab(smmu_ctx->group,
			smmu_ctx->base_lvl_info->sMemDesc.dev_phy_addr.uiAddr);
	xdev_add_smmu_ctx(xdev, smmu_ctx);
	return 0;
e2:
	xdx_smmu_ctx_env_destroy(smmu_ctx);
e1:
	if (smmu_ctx->group)
		xdx_smmu_destroy_group(smmu_ctx->group);
e0:
	smmu_ctx->group = NULL;
	devm_kfree(xdev->dev, smmu_ctx);
	return ret;
}

void xdx_smmu_ctx_destroy(struct smmu_context_s *smmu_ctx)
{
	struct xdx_device *xdev = smmu_ctx->xsmmu->xdx_dev;

	mutex_lock(&smmu_ctx->lock);
	list_del(&smmu_ctx->list);
	xdx_smmu_ctx_unmap_all(smmu_ctx);
	xdx_smmu_ctx_destroy_top_lvl_info(smmu_ctx);
	xdx_smmu_ctx_env_destroy(smmu_ctx->ctx_env_data);

	if (smmu_ctx->group)
		xdx_smmu_destroy_group(smmu_ctx->group);
	smmu_ctx->group = NULL;
	mutex_unlock(&smmu_ctx->lock);
	devm_kfree(xdev->dev, smmu_ctx);
}

int xdx_smmu_alloc(struct smmu_context_s *smmu_ctx,
		uint64_t uSize,
		uint64_t *puActualSize,
		uint32_t uiProtflags,
		uint64_t uDevVAddrAlignment,
		uint64_t pdev_vaddr,
		uint32_t uiLog2PageSize)
{
	int ret;
	uint64_t dev_vaddrEnd;
	const struct smmu_page_config_s *page_cfg = xdx_smmu_get_page_cfg(smmu_ctx);

	/* size and alignment must be datapage granular */
	dev_vaddrEnd = uSize + pdev_vaddr;
	if (((pdev_vaddr & page_cfg->page_mask) != 0)
			|| ((dev_vaddrEnd & page_cfg->page_mask) != 0))
		return -EINVAL;

	mutex_lock(&smmu_ctx->lock);
	ret = xdx_smmu_alloc_page_tabs(smmu_ctx, pdev_vaddr, dev_vaddrEnd, uiLog2PageSize);
	mutex_unlock(&smmu_ctx->lock);
	if (ret != 0) {
		SMMU_ERR(xdx_smmu_dev(smmu_ctx), "xdx_smmu_alloc_page_tabs failed.(%d)", ret);
		return ret;
	}
	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
		"pdev_vaddr:%llx + %llx ALLOC success.", pdev_vaddr, uSize);
	return 0;
}

void xdx_smmu_free(struct smmu_context_s *smmu_ctx,
		uint64_t dev_vaddr, uint64_t uiSize, uint32_t uiLog2DataPageSize)
{
	uint64_t dev_vaddrEnd;

/* ensure the address range to free is inside the heap */
	dev_vaddrEnd = dev_vaddr + uiSize;
/* The Cleanup lock has to be taken before the MMUContext hLock to
 * prevent deadlock scenarios. It is necessary only for parts of
 * _SetupCleanup_FreeMMUMapping though.
 */
	mutex_lock(&smmu_ctx->lock);
	xdx_smmu_free_page_tabs(smmu_ctx, dev_vaddr, dev_vaddrEnd, uiLog2DataPageSize);
	xdx_smmu_cleanup_smmu_env(smmu_ctx->xsmmu);
	mutex_unlock(&smmu_ctx->lock);
}

uint32_t smmu_unmap_pages_fast(struct smmu_context_s *smmu_ctx,
		uint64_t dev_vaddr_base, uint32_t pages_cnt)
{
	int ret;
	uint32_t page_size = xdx_smmu_page_size(smmu_ctx);
	struct smmu_lvlx_info_s *level_entry = NULL;
	uint32_t err_cnt = 0;
	uint32_t pte_idx = 0;
	uint32_t map_idx = 0;
	uint64_t dev_vaddr = dev_vaddr_base;
	uint32_t pte_start = 0;
	const struct smmu_pxe_config_s *pte_cfg = xdx_smmu_get_pte_cfg(smmu_ctx);

	mutex_lock(&smmu_ctx->lock);
	for (map_idx = 0; map_idx < pages_cnt; map_idx++) {
		if (!level_entry) {
			ret = xdx_smmu_get_pt_info(smmu_ctx, dev_vaddr, &level_entry, &pte_idx);
			if (ret) {
				level_entry = NULL;
				err_cnt++;
				goto next_page;
			}
			pte_start = pte_idx;
		}
		ret = xdx_smmu_setup_pte(smmu_ctx, level_entry, pte_idx, gsSmmuBadDevPhyAddr);
		if (ret)
			err_cnt++;

		pte_idx++;
		if (pte_idx == pte_cfg->uiNumEntries) {
			ret = xdx_smmu_phypage_clean(smmu_ctx,
					&level_entry->sMemDesc, pte_cfg, pte_start, pte_idx);
			if (ret)
				err_cnt++;

			level_entry = NULL;
		}
next_page:
		dev_vaddr += page_size;
	}
	if (level_entry != NULL) {
		ret = xdx_smmu_phypage_clean(smmu_ctx,
				&level_entry->sMemDesc, pte_cfg, pte_start, pte_idx);
		if (ret)
			err_cnt++;
	}
	mutex_unlock(&smmu_ctx->lock);
	return err_cnt;
}

int smmu_map_pages_linear(
		struct smmu_context_s *smmu_ctx, uint64_t dev_vaddr_base,
		uint64_t dev_phy_addr, uint32_t pages_cnt)
{
	int ret;
	uint32_t page_size = xdx_smmu_page_size(smmu_ctx);
	struct smmu_lvlx_info_s *level_entry = NULL;
	uint32_t pte_idx = 0;
	uint32_t map_idx = 0;
	uint32_t pte_start = 0;
	uint64_t dev_vaddr = dev_vaddr_base;
	uint64_t dev_paddr = dev_phy_addr;
	const struct smmu_pxe_config_s *pte_cfg = xdx_smmu_get_pte_cfg(smmu_ctx);

	mutex_lock(&smmu_ctx->lock);
	for (map_idx = 0; map_idx < pages_cnt; map_idx++) {
		if (!level_entry) {
			ret = xdx_smmu_get_pt_info(smmu_ctx, dev_vaddr, &level_entry, &pte_idx);
			if (ret)
				goto e0;
			pte_start = pte_idx;
		}
		ret = xdx_smmu_setup_pte(smmu_ctx, level_entry, pte_idx, dev_paddr);
		if (ret)
			goto e0;

		pte_idx++;
		dev_vaddr += page_size;
		dev_paddr += page_size;
		if (pte_idx == pte_cfg->uiNumEntries) {
			ret = xdx_smmu_phypage_clean(smmu_ctx,
					&level_entry->sMemDesc, pte_cfg, pte_start, pte_idx);
			if (ret)
				goto e1;

			level_entry = NULL;
		}
	}
	if (level_entry != NULL) {
		ret = xdx_smmu_phypage_clean(smmu_ctx,
				&level_entry->sMemDesc, pte_cfg, pte_start, pte_idx);
		if (ret)
			goto e1;

		level_entry = NULL;
	}
	mutex_unlock(&smmu_ctx->lock);
	return 0;
e1:
	map_idx++;
e0:
	smmu_unmap_pages_fast(smmu_ctx, dev_vaddr_base, map_idx);
	mutex_unlock(&smmu_ctx->lock);
	return ret;
}

int smmu_map_pages_fast(struct smmu_context_s *smmu_ctx,
		uint64_t dev_vaddr_base, uint32_t page_offset,
		uint64_t dev_phy_addr[], uint32_t pages_cnt)
{
	int ret;
	uint32_t page_size = xdx_smmu_page_size(smmu_ctx);
	struct smmu_lvlx_info_s *level_entry = NULL;
	struct smmu_lvlx_info_s *pre_entry = NULL;
	uint32_t pte_idx = 0;
	uint32_t map_idx = 0;
	uint32_t pte_end = 0, pte_start = 0;
	uint64_t dev_vaddr = dev_vaddr_base;
	uint64_t dev_paddr;
	const struct smmu_pxe_config_s *pte_cfg = xdx_smmu_get_pte_cfg(smmu_ctx);

	mutex_lock(&smmu_ctx->lock);
	for (map_idx = 0; map_idx < pages_cnt; map_idx++) {
		dev_paddr = dev_phy_addr[map_idx];
		xdx_smmu_get_pt_info(smmu_ctx, dev_vaddr, &level_entry, &pte_idx);
		if (pre_entry == level_entry) {
			if (pte_idx > pte_end)
				pte_end = pte_idx;
			else if (pte_idx < pte_start)
				pte_start = pte_idx;
		} else {
			/* Flush if we moved to another level_entry, i.e. page table */
			if (pre_entry != NULL) {
				ret = xdx_smmu_phypage_clean(smmu_ctx,
						&pre_entry->sMemDesc, pte_cfg,
						pte_start, (pte_end + 1 - pte_start));
				if (ret)
					goto e0;

			}
			pte_start = pte_idx;
			pte_end = pte_start;
		}
		ret = xdx_smmu_setup_pte(smmu_ctx, level_entry, pte_idx, dev_paddr);
		if (ret)
			goto e0;

		dev_vaddr += page_size;
	}
	/* Flush the last level we touched */
	if (level_entry != NULL) {
		ret = xdx_smmu_phypage_clean(smmu_ctx, &level_entry->sMemDesc,
						pte_cfg, pte_start, (pte_end+1 - pte_start));
		if (ret)
			goto e0;
	}
	mutex_unlock(&smmu_ctx->lock);
	return 0;
e0:
	mutex_unlock(&smmu_ctx->lock);
	smmu_unmap_pages_fast(smmu_ctx, dev_vaddr_base, map_idx);
	return ret;
}

void smmu_unmap_pages(struct smmu_context_s *smmu_ctx, uint64_t dev_vaddr_base,
	uint32_t map_indices[], uint32_t pages_cnt)
{
	int ret;
	uint32_t page_size = xdx_smmu_page_size(smmu_ctx);
	struct smmu_lvlx_info_s *level_entry = NULL;
	struct smmu_lvlx_info_s *pre_entry = NULL;
	uint32_t err_cnt = 0;
	uint32_t pte_idx = 0;
	uint32_t map_idx = 0;
	uint64_t dev_vaddr = dev_vaddr_base;
	uint32_t pte_start = 0;
	uint32_t pte_end = 0;
	const struct smmu_pxe_config_s *pte_cfg = xdx_smmu_get_pte_cfg(smmu_ctx);

	mutex_lock(&smmu_ctx->lock);
	for (map_idx = 0; map_idx < pages_cnt; map_idx++) {
		if (map_indices != NULL)
			dev_vaddr = dev_vaddr_base + (map_indices[map_idx] * page_size);

		xdx_smmu_get_pt_info(smmu_ctx, dev_vaddr, &level_entry, &pte_idx);
		if (pre_entry == level_entry) {
			if (pte_idx > pte_end)
				pte_end = pte_idx;
			else if (pte_idx < pte_start)
				pte_start = pte_idx;
		} else {
			if (pre_entry != NULL) {
				ret = xdx_smmu_phypage_clean(smmu_ctx,
						&pre_entry->sMemDesc,
						pte_cfg, pte_start,
						(pte_end + 1 - pte_start));
			}
			pte_start = pte_idx;
			pte_end = pte_start;
			pre_entry = level_entry;
		}
		ret = xdx_smmu_setup_pte(smmu_ctx, level_entry, pte_idx, gsSmmuBadDevPhyAddr);
		if (ret)
			err_cnt++;
		dev_vaddr += page_size;
	}

	if (level_entry != NULL) {
		ret = xdx_smmu_phypage_clean(smmu_ctx, &level_entry->sMemDesc,
				pte_cfg, pte_start, (pte_end + 1 - pte_start));
		if (ret)
			err_cnt++;
	}
	mutex_unlock(&smmu_ctx->lock);
}

int smmu_map_pages(struct smmu_context_s *smmu_ctx,
		uint64_t dev_vaddr_base, uint32_t page_offset,
		uint64_t dev_phy_addr[], bool valid[],
		uint32_t map_indices[], uint32_t pages_cnt)
{
	int ret;
	uint32_t page_size = xdx_smmu_page_size(smmu_ctx);
	struct smmu_lvlx_info_s *level_entry = NULL;
	struct smmu_lvlx_info_s *pre_entry = NULL;
	uint32_t pte_idx = 0;
	uint32_t map_idx = 0;
	uint32_t maped_cnt = 0;
	uint64_t page_off = 0;
	uint32_t pte_end = 0, pte_start = 0;
	uint64_t dev_vaddr = dev_vaddr_base;
	uint64_t dev_paddr;
	bool dummy_map_flag = 0;
	const struct smmu_pxe_config_s *pte_cfg = xdx_smmu_get_pte_cfg(smmu_ctx);

	mutex_lock(&smmu_ctx->lock);
	for (map_idx = 0; map_idx < pages_cnt; map_idx++) {
		if (map_indices != NULL) {
			page_off = map_indices[map_idx];
			dev_vaddr = dev_vaddr_base + (page_off * page_size);
		}
		dev_paddr = dev_phy_addr[map_idx];

		if (valid[map_idx] || dummy_map_flag) {
			if (!valid[map_idx])
				dev_paddr = 0;

			xdx_smmu_get_pt_info(smmu_ctx, dev_vaddr, &level_entry, &pte_idx);
			if (pre_entry == level_entry) {
				if (pte_idx > pte_end)
					pte_end = pte_idx;
				else if (pte_idx < pte_start)
					pte_start = pte_idx;
			} else {
				if (pre_entry != NULL) {
					ret = xdx_smmu_phypage_clean(smmu_ctx, &pre_entry->sMemDesc,
						pte_cfg, pte_start, (pte_end + 1 - pte_start));
					if (ret)
						goto e0;
				}
				pte_start = pte_idx;
				pte_end = pte_start;
				pre_entry = level_entry;
			}
			ret = xdx_smmu_setup_pte(smmu_ctx, level_entry, pte_idx, dev_paddr);
			if (ret)
				goto e0;

			if (valid[map_idx])
				maped_cnt++;
		}
		dev_vaddr += page_size;
	}
	if (level_entry != NULL) {
		ret = xdx_smmu_phypage_clean(smmu_ctx,
				&level_entry->sMemDesc, pte_cfg,
				pte_start, (pte_end+1 - pte_start));
		if (ret)
			goto e0;
	}
	mutex_unlock(&smmu_ctx->lock);
	return 0;
e0:
	smmu_unmap_pages(smmu_ctx, dev_vaddr_base, map_indices, map_idx);
	mutex_unlock(&smmu_ctx->lock);
	return ret;
}

struct smmu_context_s *xdx_smmu_get_ctx_sid(struct xdx_device *xdev, uint32_t gsid)
{
	struct xdx_smmu *smmu = xdev_to_smmu(xdev);
	struct smmu_context_s *_ctx;
	struct smmu_group_s *grp;

	grp = xdx_smmu_find_group_gsid(xdev_to_smmu_dev(xdev), gsid);
	if (!grp)
		return NULL;
	list_for_each_entry(_ctx, &smmu->ctxs_head, list) {
		if (xdx_smmu_ctx_group(_ctx) == grp)
			return _ctx;
	}
	return NULL;
}

