/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#include <drm/drm_atomic.h>
#include <drm/drm_atomic_helper.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0)) || (defined(XDXGPU_CENTOS85_SUPPORT)) || ((defined(XDXGPU_CENTOS82_SUPPORT)) && (!defined(XDXGPU_CENTOS79_SUPPORT)))
#include <drm/drm_probe_helper.h>
#else
#include <drm/drm_crtc_helper.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0))
#include <drm/drm_vblank.h>
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0))
#include <drm/drm_irq.h>
#else
#include <drm/drmP.h>
#endif
#include "xdx_vkms.h"
#include "xdx_gem.h"
#include "xdx_device.h"

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))
static enum hrtimer_restart vkms_vblank_simulate(struct hrtimer *timer)
{
	struct vkms_output *output =
		container_of(timer, struct vkms_output, vblank_hrtimer);
	struct vkms_device *vkmsdev = vkms_output_to_vkmsdev(output);
	struct xdx_device *xdev = vkmsdev_to_xdev(vkmsdev);
	struct drm_crtc *crtc = &output->crtc;
	struct vkms_crtc_state *state;
	u64 ret_overrun;
	bool ret;

	spin_lock(&output->lock);

	ret_overrun =
		hrtimer_forward_now(&output->vblank_hrtimer, output->period_ns);
	if (ret_overrun != 1)
		dev_warn(xdev->dev, "%s: vblank timer overrun\n", __func__);

	ret = drm_crtc_handle_vblank(crtc);
	if (!ret)
		dev_err(xdev->dev, "vkms failure on handling vblank");

	state = output->composer_state;
	if (state && output->composer_enabled) {
		u64 frame = drm_crtc_accurate_vblank_count(crtc);

		/* update frame_start only if a queued xdx_vkms_composer_worker()
		 * has read the data
		 */
		spin_lock(&output->composer_lock);
		if (!state->crc_pending)
			state->frame_start = frame;
		else
			DRM_DEBUG_DRIVER(
				"crc worker falling behind, frame_start: %llu, frame_end: %llu\n",
				state->frame_start, frame);
		state->frame_end = frame;
		state->crc_pending = true;
		spin_unlock(&output->composer_lock);

		ret = queue_work(output->composer_workq, &state->composer_work);
		if (!ret)
			DRM_DEBUG_DRIVER("Composer worker already queued\n");
	}

	spin_unlock(&output->lock);

	return HRTIMER_RESTART;
}

static int vkms_enable_vblank(struct drm_crtc *crtc)
{
	struct drm_device *dev = crtc->dev;
	unsigned int pipe = drm_crtc_index(crtc);
	struct drm_vblank_crtc *vblank = &dev->vblank[pipe];
	struct vkms_output *out = drm_crtc_to_vkms_output(crtc);

	drm_calc_timestamping_constants(crtc, &crtc->mode);

	hrtimer_init(&out->vblank_hrtimer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
	out->vblank_hrtimer.function = vkms_vblank_simulate;
	out->period_ns = ktime_set(0, vblank->framedur_ns);
	hrtimer_start(&out->vblank_hrtimer, out->period_ns, HRTIMER_MODE_REL);

	return 0;
}

static void vkms_disable_vblank(struct drm_crtc *crtc)
{
	struct vkms_output *out = drm_crtc_to_vkms_output(crtc);

	hrtimer_cancel(&out->vblank_hrtimer);
}
#endif

bool xdx_vkms_get_vblank_timestamp(struct drm_device *dev, unsigned int pipe,
				   int *max_error, ktime_t *vblank_time,
				   bool in_vblank_irq)
{
	struct xdx_device *xdev = drm_to_xdev(dev);
	struct vkms_device *vkmsdev = xdev_to_vkmsdev(xdev);
	struct vkms_output *output = &vkmsdev->output;
	struct drm_vblank_crtc *vblank = &dev->vblank[pipe];

	if (!READ_ONCE(vblank->enabled)) {
		*vblank_time = ktime_get();
		return true;
	}

	*vblank_time = READ_ONCE(output->vblank_hrtimer.node.expires);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
	if (*vblank_time == vblank->time)
#elif (defined(XDXGPU_CENTOS79_SUPPORT))
	if ((*vblank_time).tv64 == vblank->time.tv64)
#else
	if (vblank_time->tv64 == timeval_to_ns(&vblank->time))
#endif
	{
		dev_warn(xdev->dev, "vblank is not updated: 0x%lld\n",
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
			vblank->time);
#elif (defined(XDXGPU_CENTOS79_SUPPORT))
			vblank->time.tv64);
#else
			timeval_to_ns(&vblank->time));
#endif
		return true;
	}

	/*
	 * To prevent races we roll the hrtimer forward before we do any
	 * interrupt processing - this is how real hw works (the interrupt is
	 * only generated after all the vblank registers are updated) and what
	 * the vblank core expects. Therefore we need to always correct the
	 * timestampe by one frame.
	 */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
	*vblank_time -= output->period_ns;
#else
	vblank_time->tv64 -= output->period_ns.tv64;
#endif

	return true;
}

static struct drm_crtc_state *
vkms_atomic_crtc_duplicate_state(struct drm_crtc *crtc)
{
	struct vkms_crtc_state *vkms_state;

	if (WARN_ON(!crtc->state))
		return NULL;

	vkms_state = kzalloc(sizeof(*vkms_state), GFP_KERNEL);
	if (!vkms_state)
		return NULL;

	__drm_atomic_helper_crtc_duplicate_state(crtc, &vkms_state->base);

	INIT_WORK(&vkms_state->composer_work, xdx_vkms_composer_worker);

	return &vkms_state->base;
}

static void vkms_atomic_crtc_destroy_state(struct drm_crtc *crtc,
					   struct drm_crtc_state *state)
{
	struct vkms_crtc_state *vkms_state = to_vkms_crtc_state(state);

	__drm_atomic_helper_crtc_destroy_state(state);

	WARN_ON(work_pending(&vkms_state->composer_work));
	kfree(vkms_state->active_planes);
	kfree(vkms_state);
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0) && (!defined(XDXGPU_CENTOS85_SUPPORT))
static void xdx_vkms_atomic_helper_crtc_reset(struct drm_crtc *crtc,
			       struct drm_crtc_state *crtc_state)
{
	if (crtc_state)
		crtc_state->crtc = crtc;

	crtc->state = crtc_state;
}
#endif
static void vkms_atomic_crtc_reset(struct drm_crtc *crtc)
{
	struct vkms_crtc_state *vkms_state =
		kzalloc(sizeof(*vkms_state), GFP_KERNEL);

	if (crtc->state)
		vkms_atomic_crtc_destroy_state(crtc, crtc->state);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0) || (defined(XDXGPU_CENTOS85_SUPPORT))
	__drm_atomic_helper_crtc_reset(crtc, &vkms_state->base);
#else
	xdx_vkms_atomic_helper_crtc_reset(crtc, &vkms_state->base);
#endif
	if (vkms_state)
		INIT_WORK(&vkms_state->composer_work, xdx_vkms_composer_worker);
}

static const struct drm_crtc_funcs vkms_crtc_funcs = {
	.set_config = drm_atomic_helper_set_config,
	.destroy = drm_crtc_cleanup,
	.page_flip = drm_atomic_helper_page_flip,
	.reset = vkms_atomic_crtc_reset,
	.atomic_duplicate_state = vkms_atomic_crtc_duplicate_state,
	.atomic_destroy_state = vkms_atomic_crtc_destroy_state,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))
	.enable_vblank = vkms_enable_vblank,
	.disable_vblank = vkms_disable_vblank,
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)) || (defined(XDXGPU_CENTOS85_SUPPORT))
	.get_crc_sources = xdx_vkms_get_crc_sources,
	.set_crc_source = xdx_vkms_set_crc_source,
	.verify_crc_source = xdx_vkms_verify_crc_source,
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	.set_crc_source = xdx_vkms_set_crc_source,
#endif	
};
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)) || (defined(XDXGPU_CENTOS85_SUPPORT))
static int vkms_crtc_atomic_check(struct drm_crtc *crtc,
				  struct drm_atomic_state *state)
{
	struct drm_crtc_state *old_crtc_state = drm_atomic_get_new_crtc_state(state, crtc);
		
	struct vkms_crtc_state *vkms_state = to_vkms_crtc_state(old_crtc_state);
	struct drm_plane *plane;
	struct drm_plane_state *plane_state;
	int i = 0, ret;

    printk("drm_atomic_get_existing_crtc_state succeed.\n");
	if (vkms_state->active_planes)
		return 0;
    ret = drm_atomic_add_affected_planes(old_crtc_state->state, crtc);
	if (ret < 0)
		return ret;
	drm_for_each_plane_mask (plane, crtc->dev, old_crtc_state->plane_mask) 
	{
		plane_state = drm_atomic_get_existing_plane_state(old_crtc_state->state,
								  plane);
		WARN_ON(!plane_state);
		if (!plane_state->visible)
			continue;
		i++;
	}
	vkms_state->active_planes = kcalloc(i, sizeof(plane), GFP_KERNEL);
	if (!vkms_state->active_planes)
		return -ENOMEM;
	vkms_state->num_active_planes = i;

	i = 0;
	drm_for_each_plane_mask (plane, crtc->dev, old_crtc_state->plane_mask) 
	{
		plane_state = drm_atomic_get_existing_plane_state(old_crtc_state->state,
								  plane);

		if (!plane_state->visible)
			continue;

		vkms_state->active_planes[i++] =
			to_vkms_plane_state(plane_state);
	}

	return 0;
}
#else
static int vkms_crtc_atomic_check(struct drm_crtc *crtc,
				  struct drm_crtc_state *state)
{
	struct vkms_crtc_state *vkms_state = to_vkms_crtc_state(state);
	struct drm_plane *plane;
	struct drm_plane_state *plane_state;
	int i = 0, ret;

	if (vkms_state->active_planes)
		return 0;

	ret = drm_atomic_add_affected_planes(state->state, crtc);
	if (ret < 0)
		return ret;

	drm_for_each_plane_mask (plane, crtc->dev, state->plane_mask) {
		plane_state = drm_atomic_get_existing_plane_state(state->state,
								  plane);
		WARN_ON(!plane_state);

		if (!plane_state->visible)
			continue;

		i++;
	}

	vkms_state->active_planes = kcalloc(i, sizeof(plane), GFP_KERNEL);
	if (!vkms_state->active_planes)
		return -ENOMEM;
	vkms_state->num_active_planes = i;

	i = 0;
	drm_for_each_plane_mask (plane, crtc->dev, state->plane_mask) {
		plane_state = drm_atomic_get_existing_plane_state(state->state,
								  plane);

		if (!plane_state->visible)
			continue;

		vkms_state->active_planes[i++] =
			to_vkms_plane_state(plane_state);
	}

	return 0;
}
#endif
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 11, 0)) && (!defined(XDXGPU_CENTOS85_SUPPORT))
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0))
static void vkms_crtc_atomic_enable(struct drm_crtc *crtc,
				    struct drm_crtc_state *old_state)
{
	drm_crtc_vblank_on(crtc);
}
#endif
#else
static void vkms_crtc_atomic_enable(struct drm_crtc *crtc,
				    struct drm_atomic_state *old_state)
{
	drm_crtc_vblank_on(crtc);
}
#endif
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 11, 0)) && (!defined(XDXGPU_CENTOS85_SUPPORT))
static void vkms_crtc_atomic_disable(struct drm_crtc *crtc,
				     struct drm_crtc_state *old_state)
{
	drm_crtc_vblank_off(crtc);
}
#else
static void vkms_crtc_atomic_disable(struct drm_crtc *crtc,
				     struct drm_atomic_state *old_state)
{
	drm_crtc_vblank_off(crtc);
}
#endif
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 11, 0)) && (!defined(XDXGPU_CENTOS85_SUPPORT))
static void vkms_crtc_atomic_begin(struct drm_crtc *crtc,
				   struct drm_crtc_state *old_crtc_state)
#else
static void vkms_crtc_atomic_begin(struct drm_crtc *crtc,
struct drm_atomic_state *old_crtc_state)
#endif
{
	struct vkms_output *vkms_output = drm_crtc_to_vkms_output(crtc);

	/* This lock is held across the atomic commit to block vblank timer
	 * from scheduling xdx_vkms_composer_worker until the composer is updated
	 */
	spin_lock_irq(&vkms_output->lock);
}
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 11, 0)) && (!defined(XDXGPU_CENTOS85_SUPPORT))
static void vkms_crtc_atomic_flush(struct drm_crtc *crtc,
				   struct drm_crtc_state *old_crtc_state)
#else
static void vkms_crtc_atomic_flush(struct drm_crtc *crtc,
				   struct drm_atomic_state *old_crtc_state)
#endif
{
	struct vkms_output *vkms_output = drm_crtc_to_vkms_output(crtc);

	if (crtc->state->event) {
		spin_lock(&crtc->dev->event_lock);

		if (drm_crtc_vblank_get(crtc) != 0)
			drm_crtc_send_vblank_event(crtc, crtc->state->event);
		else
			drm_crtc_arm_vblank_event(crtc, crtc->state->event);

		spin_unlock(&crtc->dev->event_lock);

		crtc->state->event = NULL;
	}

	vkms_output->composer_state = to_vkms_crtc_state(crtc->state);

	spin_unlock_irq(&vkms_output->lock);
}

static const struct drm_crtc_helper_funcs vkms_crtc_helper_funcs = {
	.atomic_check = vkms_crtc_atomic_check,
	.atomic_begin = vkms_crtc_atomic_begin,
	.atomic_flush = vkms_crtc_atomic_flush,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)
	.atomic_enable = vkms_crtc_atomic_enable,
#endif
	.atomic_disable = vkms_crtc_atomic_disable,
};

int xdx_vkms_crtc_init(struct drm_device *dev, struct drm_crtc *crtc,
		       struct drm_plane *primary, struct drm_plane *cursor)
{
	struct vkms_output *vkms_out = drm_crtc_to_vkms_output(crtc);
	int ret;

	ret = drm_crtc_init_with_planes(dev, crtc, primary, cursor,
					&vkms_crtc_funcs, NULL);
	if (ret) {
		DRM_ERROR("Failed to init CRTC\n");
		return ret;
	}

	drm_crtc_helper_add(crtc, &vkms_crtc_helper_funcs);

	spin_lock_init(&vkms_out->lock);
	spin_lock_init(&vkms_out->composer_lock);

	vkms_out->composer_workq = alloc_ordered_workqueue("vkms_composer", 0);
	if (!vkms_out->composer_workq)
		return -ENOMEM;

	return ret;
}
