/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <drm/drm_atomic_helper.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0)) || (defined(REDHAT_83)) || (defined(XDXGPU_CENTOS85_SUPPORT)) || ((defined(XDXGPU_CENTOS82_SUPPORT))&&(!defined(XDXGPU_CENTOS79_SUPPORT)))
#include <drm/drm_probe_helper.h>
#else
#include <drm/drm_crtc_helper.h>
#endif

#include "xdx_vkms.h"
#include "xdx_gem.h"
#include "xdx_device.h"
#include "kernel_compatibility.h"

static void vkms_connector_destroy(struct drm_connector *connector)
{
	drm_connector_cleanup(connector);
}

static const struct drm_connector_funcs vkms_connector_funcs = {
	.fill_modes = drm_helper_probe_single_connector_modes,
	.destroy = vkms_connector_destroy,
	.reset = drm_atomic_helper_connector_reset,
	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
	.atomic_destroy_state = drm_atomic_helper_connector_destroy_state,
};

static void vkms_encoder_destroy(struct drm_encoder *encoder)
{
	struct vkms_output *output = drm_encoder_to_vkms_output(encoder);
	vkms_output_to_vkmsdev(output);

	drm_encoder_cleanup(encoder);
}

static const struct drm_encoder_funcs vkms_encoder_funcs = {
	.destroy = vkms_encoder_destroy,
};

static void vkms_mode_set(struct drm_encoder *encoder,
			  struct drm_display_mode *mode,
			  struct drm_display_mode *adjusted_mode)
{
	struct vkms_output *output = drm_encoder_to_vkms_output(encoder);
	vkms_output_to_vkmsdev(output);

	// TODO:
}

static void vkms_encoder_disable(struct drm_encoder *encoder)
{
	struct vkms_output *output = drm_encoder_to_vkms_output(encoder);
	vkms_output_to_vkmsdev(output);

	// TODO:
}

static void vkms_encoder_enable(struct drm_encoder *encoder)
{
	struct vkms_output *output = drm_encoder_to_vkms_output(encoder);
	vkms_output_to_vkmsdev(output);

	// TODO:
}

static const struct drm_encoder_helper_funcs vkms_encoder_helper_funcs = {
	.mode_set = vkms_mode_set,
	.enable = vkms_encoder_enable,
	.disable = vkms_encoder_disable,
};

static int vkms_conn_get_modes(struct drm_connector *connector)
{
	int count;

	count = drm_add_modes_noedid(connector, XRES_MAX, YRES_MAX);
	drm_set_preferred_mode(connector, XRES_DEF, YRES_DEF);

	return count;
}

static const struct drm_connector_helper_funcs vkms_conn_helper_funcs = {
	.get_modes = vkms_conn_get_modes,
};

int xdx_vkms_output_init(struct xdx_device *xdev, int index)
{
	struct vkms_device *vkmsdev = xdev_to_vkmsdev(xdev);
	struct vkms_output *output = &vkmsdev->output;
	struct drm_device *dev = xdev_to_drm(xdev);
	struct drm_connector *connector = &output->connector;
	struct drm_encoder *encoder = &output->encoder;
	struct drm_crtc *crtc = &output->crtc;
	struct drm_plane *primary, *cursor = NULL;
	int ret;

	primary = xdx_vkms_plane_init(xdev, DRM_PLANE_TYPE_PRIMARY, index);
	if (IS_ERR(primary))
		return PTR_ERR(primary);

	if (enable_cursor) {
		cursor =
			xdx_vkms_plane_init(xdev, DRM_PLANE_TYPE_CURSOR, index);
		if (IS_ERR(cursor)) {
			dev_err(xdev->dev, "failed to init plane\n");
			ret = PTR_ERR(cursor);
			goto err_cursor;
		}
	}

	ret = xdx_vkms_crtc_init(dev, crtc, primary, cursor);
	if (ret) {
		dev_err(xdev->dev, "failed to init crtc\n");
		goto err_crtc;
	}

	ret = drm_connector_init(dev, connector, &vkms_connector_funcs,
				 DRM_MODE_CONNECTOR_VIRTUAL);
	if (ret) {
		dev_err(xdev->dev, "failed to init connector\n");
		goto err_connector;
	}
	drm_connector_helper_add(connector, &vkms_conn_helper_funcs);

	ret = drm_encoder_init(dev, encoder, &vkms_encoder_funcs,
			       DRM_MODE_ENCODER_VIRTUAL, NULL);
	if (ret) {
		dev_err(xdev->dev, "failed to init encoder\n");
		goto err_encoder;
	}
	drm_encoder_helper_add(encoder, &vkms_encoder_helper_funcs);

	encoder->possible_crtcs = 1;

	ret = drm_connector_attach_encoder(connector, encoder);
	if (ret) {
		dev_err(xdev->dev, "Failed to attach connector to encoder\n");
		goto err_attach;
	}

	drm_mode_config_reset(dev);

	return 0;

err_attach:
	drm_encoder_cleanup(encoder);

err_encoder:
	drm_connector_cleanup(connector);

err_connector:
	drm_crtc_cleanup(crtc);

err_crtc:
	if (enable_cursor)
		drm_plane_cleanup(cursor);

err_cursor:
	drm_plane_cleanup(primary);

	return ret;
}
