/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#include <linux/version.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/debugfs.h>
#include <linux/types.h>
#include <linux/device.h>
#include <linux/pagemap.h>
#include <linux/sched.h>
#include <linux/stddef.h>
#include <linux/platform_device.h>
#include <linux/mod_devicetable.h>
#include <linux/dma-buf.h>
#include "hantro_mmu.h"
#include "vpu_xmem.h"
#include "hantro_fpga.h"
#include "dc8200_mmu.h"
#include "xdx_gem.h"
#include "kernel_compatibility.h"

#ifndef NULL
#ifdef __cplusplus
#define NULL 0
#else
#define NULL ((void *)0)
#endif
#endif
#define PD_MODE

#define MMU_REG_OFFSET 0
#define MMU_REG_HW_ID (MMU_REG_OFFSET + 6 * 4)
#define MMU_REG_FLUSH (MMU_REG_OFFSET + 97 * 4)
#define MMU_REG_PAGE_TABLE_ID (MMU_REG_OFFSET + 107 * 4)
#define MMU_REG_CONTROL (MMU_REG_OFFSET + 226 * 4)
#define MMU_REG_ADDRESS (MMU_REG_OFFSET + 227 * 4)
#define MMU_REG_ADDRESS_MSB (MMU_REG_OFFSET + 228 * 4)
#define MMU_REG_PDENTRY0 (MMU_REG_OFFSET + 237 * 4)


/***** New MMU Defination ********************************************/
#define VPU_MMU_MTLB_SHIFT 22
#define MMU_STLB_4K_SHIFT 12
#define MMU_STLB_64K_SHIFT 16

#define MMU_MTLB_BITS (32 - MMU_MTLB_SHIFT)
#define MMU_PAGE_4K_BITS MMU_STLB_4K_SHIFT
#define MMU_STLB_4K_BITS (32 - MMU_MTLB_BITS - MMU_PAGE_4K_BITS)
#define MMU_PAGE_64K_BITS MMU_STLB_64K_SHIFT
#define MMU_STLB_64K_BITS (32 - MMU_MTLB_BITS - MMU_PAGE_64K_BITS)

#define VPU_MMU_MTLB_ENTRY_NUM BIT(MMU_MTLB_BITS)
#define MMU_MTLB_SIZE (MMU_MTLB_ENTRY_NUM << 2)
#define VPU_MMU_STLB_4K_ENTRY_NUM BIT(MMU_STLB_4K_BITS)
#define MMU_STLB_4K_SIZE (MMU_STLB_4K_ENTRY_NUM << 2)
#define VPU_MMU_PAGE_4K_SIZE BIT(MMU_STLB_4K_SHIFT)
#define MMU_STLB_64K_ENTRY_NUM BIT(MMU_STLB_64K_BITS)
#define MMU_STLB_64K_SIZE (MMU_STLB_64K_ENTRY_NUM << 2)
#define MMU_PAGE_64K_SIZE BIT(MMU_STLB_64K_SHIFT)

#define MMU_MTLB_MASK (~((1U << MMU_MTLB_SHIFT) - 1))
#define MMU_STLB_4K_MASK ((~0U << MMU_STLB_4K_SHIFT) ^ MMU_MTLB_MASK)
#define MMU_PAGE_4K_MASK (MMU_PAGE_4K_SIZE - 1)
#define MMU_STLB_64K_MASK ((~((1U << MMU_STLB_64K_SHIFT) - 1)) ^ MMU_MTLB_MASK)
#define MMU_PAGE_64K_MASK (MMU_PAGE_64K_SIZE - 1)

/* Page offset definitions. */
#define MMU_OFFSET_4K_BITS (32 - MMU_MTLB_BITS - MMU_STLB_4K_BITS)
#define MMU_OFFSET_4K_MASK ((1U << MMU_OFFSET_4K_BITS) - 1)
#define MMU_OFFSET_16K_BITS (32 - MMU_MTLB_BITS - MMU_STLB_16K_BITS)
#define MMU_OFFSET_16K_MASK ((1U << MMU_OFFSET_16K_BITS) - 1)

#define MMU_MTLB_ENTRY_HINTS_BITS 6
#define MMU_MTLB_ENTRY_STLB_MASK (~((1U << MMU_MTLB_ENTRY_HINTS_BITS) - 1))

#define MMU_MTLB_PRESENT 0x00000001
#define MMU_MTLB_EXCEPTION 0x00000002
#define MMU_MTLB_4K_PAGE 0x00000000

#define MMU_STLB_PRESENT 0x00000001
#define MMU_STLB_EXCEPTION 0x00000002
#define MMU_STLB_4K_PAGE 0x00000000

#define MMU_FALSE 0
#define MMU_TRUE 1

#define MMU_ERR_OS_FAIL (0xffff)
#define MMU_EFAULT MMU_ERR_OS_FAIL
#define MMU_ENOTTY MMU_ERR_OS_FAIL

#define MMU_INFINITE ((u32)~0U)

#define MAX_NOPAGED_SIZE 0x20000
#define MMU_SUPPRESS_OOM_MESSAGE 1

#define MTLB_PCIE_START_ADDRESS 0x00100000
#define PAGE_PCIE_START_ADDRESS 0x00200000 /* page_table_entry start address */
#define STLB_PCIE_START_ADDRESS 0x00300000
#define PAGE_TABLE_ENTRY_SIZE 64

#if MMU_SUPPRESS_OOM_MESSAGE
#define MMU_NOWARN __GFP_NOWARN
#else
#define MMU_NOWARN 0
#endif

#define MMU_IS_ERROR(status) ((status) < 0)
#define MMU_NO_ERROR(status) ((status) >= 0)
#define MMU_IS_SUCCESS(status) ((status) == MMU_STATUS_OK)

#undef MMUDEBUG
#ifdef HANTROMMU_DEBUG
#ifdef __KERNEL__
#define MMUDEBUG(fmt, args...) pr_info("hantrommu: " fmt, ##args)
#else
#define MMUDEBUG(fmt, args...) fprintf(stderr, fmt, ##args)
#endif
#else
#define MMUDEBUG(fmt, args...)
#endif

#define MMU_ON_ERROR(func)                                                     \
	do {                                                                   \
		status = func;                                                 \
		if (MMU_IS_ERROR(status)) {                                    \
			goto onerror;                                          \
		}                                                              \
	} while (MMU_FALSE)

#define WritePageEntry(page_entry, entry_value)                                \
	(*(unsigned int *)(page_entry) = (unsigned int)(entry_value))


/* simple map mode: generate mmu address which is same as input bus address*/
static unsigned int simple_map = 0;
/* this shift should be an integral multiple of mmu page size(4096).
 *It can generate a mmu address shift in simple map mode
 */
static unsigned int map_shift = 0;

/* module_param(name, type, perm) */
module_param(simple_map, uint, 0);
module_param(map_shift, uint, 0);

#define MMU_TYPE_ENC "_ENC"
#define MMU_TYPE_DEC "_DEC"
#ifndef USE_DTB_PROBE
static struct mmu_core_cfg mmu_core_array[] = {
#ifdef HAS_VC8000D
	{ 0x1a01400, 228 * 4, 0, HANTRO_CORE_DEC },
#endif
#ifdef HAS_VC8000E
	{ 0x1b02000, 228 * 4, 0, HANTRO_CORE_ENC },
#endif
};
#endif

enum MMURegion {
	MMU_REGION_IN,
	MMU_REGION_OUT,
	MMU_REGION_PRIVATE,
	MMU_REGION_PUB,

	MMU_REGION_COUNT
};

struct MMUNode {
	void *buf_virtual_address;
	unsigned int buf_bus_address; /* used in kernel map mode */
	int mtlb_start;
	int stlb_start;
	int mtlb_end;
	int stlb_end;
	unsigned int page_count;
	int process_id;
	struct file *filp;

	struct MMUNode *next;
	struct MMUNode *prev;
};

struct MMUDDRRegion {
	unsigned long long physical_address;
	unsigned long long virtual_address;
	unsigned int page_count;

	void *node_mutex;
	struct MMUNode *simple_map_head;
	struct MMUNode *simple_map_tail;
	struct MMUNode *free_map_head;
	struct MMUNode *map_head;
	struct MMUNode *free_map_tail;
	struct MMUNode *map_tail;
};

struct MMU {
	void *page_table_mutex;
	void *hw_mutex;
	/* Master TLB information. */
	unsigned int mtlb_size;
	unsigned long long mtlb_physical;
	void *mtlb_virtual;
	unsigned int mtlb_entries;

	int enabled;
	unsigned int stlb_size;
	unsigned long long stlb_physical;
	void *stlb_virtual;
	struct MMUDDRRegion region[MMU_REGION_COUNT];
	unsigned int page_table_array_size;
	unsigned long long page_table_array_physical;
	void *page_table_array;
};

unsigned int mmu_enable = MMU_FALSE;
static unsigned int pcie;
static unsigned long gBaseDDRHw = (unsigned long)-1; /* PCI base register address (memalloc) */
static unsigned int region_in_mmu_start = REGION_IN_MMU_START;
static unsigned int region_in_mmu_end = REGION_IN_MMU_END;
static unsigned int region_out_mmu_start = REGION_OUT_MMU_START;
static unsigned int region_out_mmu_end = REGION_OUT_MMU_END;
static unsigned int region_private_mmu_start = REGION_PRIVATE_MMU_START;
static unsigned int region_private_mmu_end = REGION_PRIVATE_MMU_END;

static enum MMUStatus MMUEnableStep0(struct xdx_device *xdev,
				     unsigned long BaseDDRAddr,
				     struct mmu_t *mmucore);
static enum MMUStatus MMURelease(void *filp, struct MMU *mmu);

static enum MMUStatus ZeroMemory(void *memory, unsigned int bytes)
{
	memset(memory, 0, bytes);

	return MMU_STATUS_OK;
}

static enum MMUStatus AllocateMemory(unsigned int bytes, void **memory)
{
	void *pointer;
	enum MMUStatus status;

	if (bytes > MAX_NOPAGED_SIZE) {
		pointer = vmalloc(bytes);
		XDX_DEV_DEBUG(" *****VMALLOC size*****%d\n", bytes);
	} else {
		pointer = kmalloc(bytes, GFP_KERNEL | MMU_NOWARN);
		XDX_DEV_DEBUG(" *****KMALLOC size*****%d\n", bytes);
	}

	if (!pointer) {
		/* Out of memory. */
		status = MMU_STATUS_OUT_OF_MEMORY;
		goto onerror;
	}

	/* Return pointer to the memory allocation. */
	*memory = pointer;

	return MMU_STATUS_OK;

onerror:
	/* Return the status. */
	return status;
}

static enum MMUStatus FreeMemory(void *memory)
{
	/* Free the memory from the OS pool. */
	if (is_vmalloc_addr(memory)) {
		XDX_DEV_DEBUG(" *****VFREE*****%p\n", memory);
		vfree(memory);
	} else {
		XDX_DEV_DEBUG(" *****KFREE*****%p\n", memory);
		kfree(memory);
	}
	return MMU_STATUS_OK;
}

static enum MMUStatus SMDeleteNode(struct MMUNode **pp)
{
	(*pp)->prev->next = (*pp)->next;
	(*pp)->next->prev = (*pp)->prev;

	XDX_DEV_DEBUG(" *****DeleteNode size*****%d\n", (*pp)->page_count);
	FreeMemory(*pp);

	return MMU_STATUS_OK;
}

static enum MMUStatus DeleteNode(struct MMUNode **pp)
{
	(*pp)->prev->next = (*pp)->next;
	(*pp)->next->prev = (*pp)->prev;

	XDX_DEV_DEBUG(" *****%s size*****%d\n", __func__, (*pp)->page_count);
	FreeMemory(*pp);

	return MMU_STATUS_OK;
}

static enum MMUStatus MergeNode(struct MMUNode *h, struct MMUNode **pp)
{
	struct MMUNode *tmp0 = h->next;
	struct MMUNode *tmp1 = h->next;

	while (tmp0) {
		/* 1th step: find front contiguous memory node */
		if (tmp0->mtlb_end == (*pp)->mtlb_start &&
		    tmp0->stlb_end == (*pp)->stlb_start) {
			tmp0->mtlb_end = (*pp)->mtlb_end;
			tmp0->stlb_end = (*pp)->stlb_end;
			tmp0->page_count += (*pp)->page_count;
			DeleteNode(pp);
			XDX_DEV_DEBUG(
				" *****first merge to front. node size*****%d\n",
				tmp0->page_count);
			/* after merge to front contiguous memory node,
			 *find if there is behind contiguous memory node.
			 */
			while (tmp1) {
				/* merge */
				if (tmp1->mtlb_start == tmp0->mtlb_end &&
				    tmp1->stlb_start == tmp0->stlb_end) {
					tmp1->mtlb_start = tmp0->mtlb_start;
					tmp1->stlb_start = tmp0->stlb_start;
					tmp1->page_count += tmp0->page_count;
					XDX_DEV_DEBUG(
						" *****second merge to behind. node size*****%d\n",
						tmp1->page_count);
					DeleteNode(&tmp0);
					return MMU_STATUS_OK;
				}
				tmp1 = tmp1->next;
			}
			return MMU_STATUS_OK;
			/* 1th step: find behind contiguous memory node */
		} else if (tmp0->mtlb_start == (*pp)->mtlb_end &&
			   tmp0->stlb_start == (*pp)->stlb_end) {
			tmp0->mtlb_start = (*pp)->mtlb_start;
			tmp0->stlb_start = (*pp)->stlb_start;
			tmp0->page_count += (*pp)->page_count;
			DeleteNode(pp);
			XDX_DEV_DEBUG(
				" *****first merge to behind. node size*****%d\n",
				tmp0->page_count);
			/* after merge to behind contiguous memory node,
			 *find if there is front contiguous memory node
			 */
			while (tmp1) {
				/* merge */
				if (tmp1->mtlb_end == tmp0->mtlb_start &&
				    tmp1->stlb_end == tmp0->stlb_start) {
					tmp1->mtlb_end = tmp0->mtlb_end;
					tmp1->stlb_end = tmp0->stlb_end;
					tmp1->page_count += tmp0->page_count;
					XDX_DEV_DEBUG(
						" *****second merge to front. node size*****%d\n",
						tmp1->page_count);
					DeleteNode(&tmp0);
					return MMU_STATUS_OK;
				}
				tmp1 = tmp1->next;
			}
			return MMU_STATUS_OK;
		}
		tmp0 = tmp0->next;
	}
	return MMU_STATUS_FALSE;
}

#if 0
/* Insert a node to map list */
static enum MMUStatus SMInsertNode(enum MMURegion e, struct MMUNode **pp,
				   struct MMU *mmu)
{
	struct MMUNode *h;

	h = mmu->region[e].simple_map_head;

	h->next->prev = *pp;
	(*pp)->next = h->next;
	(*pp)->prev = h;
	h->next = *pp;
	XDX_DEV_DEBUG(" *****insert bm node*****%d\n", (*pp)->page_count);

	return MMU_STATUS_OK;
}
#endif

static enum MMUStatus InsertNode(enum MMURegion e, struct MMUNode **pp,
				 struct MMU *mmu, unsigned int free)
{
	enum MMUStatus status;
	struct MMUNode *h, *b;

	if (free) {
		h = mmu->region[e].free_map_head;
		b = mmu->region[e].map_head;
		status = MergeNode(h, pp);
		XDX_DEV_DEBUG(" *****insert free*****%d\n", (*pp)->page_count);
		if (MMU_IS_ERROR(status)) {
			/* remove from map*/
			if ((*pp)->prev && (*pp)->next) {
				(*pp)->prev->next = (*pp)->next;
				(*pp)->next->prev = (*pp)->prev;
			}
			/* insert to free map */
			h->next->prev = *pp;
			(*pp)->next = h->next;
			(*pp)->prev = h;
			h->next = *pp;
		}
	} else {
		h = mmu->region[e].map_head;

		h->next->prev = *pp;
		(*pp)->next = h->next;
		(*pp)->prev = h;
		h->next = *pp;
		XDX_DEV_DEBUG(" *****insert unfree*****%d\n", (*pp)->page_count);
	}

	return MMU_STATUS_OK;
}

#if 0
/* Create a Node */
static enum MMUStatus SMCreateNode(enum MMURegion e, struct MMUNode **node,
				   struct MMU *mmu, unsigned int page_count)
{
	struct MMUNode *p, **new;

	p = kmalloc(sizeof(struct MMUNode), GFP_KERNEL | MMU_NOWARN);
	new = &p;

	(*new)->mtlb_start = -1;
	(*new)->stlb_start = -1;
	(*new)->mtlb_end = -1;
	(*new)->stlb_end = -1;
	(*new)->process_id = 0;
	(*new)->filp = NULL;
	(*new)->page_count = 0;
	(*new)->prev = NULL;
	(*new)->next = NULL;
	/* Insert a uncomplete Node, it will be initialized later */
	SMInsertNode(e, new, mmu);

	/* return a new node for map buffer */
	*node = *new;
	return MMU_STATUS_OK;
}
#endif

/* Create initial Nodes */
static enum MMUStatus SMCreateNodes(struct MMU *mmu)
{
	struct MMUNode *simple_map_head;
	struct MMUNode *simple_map_tail;
	int i;

	/* Init each region map node */
	for (i = 0; i < MMU_REGION_COUNT; i++) {
		simple_map_head = kmalloc(sizeof(struct MMUNode),
					  GFP_KERNEL | MMU_NOWARN);
		simple_map_tail = kmalloc(sizeof(struct MMUNode),
					  GFP_KERNEL | MMU_NOWARN);

		simple_map_head->mtlb_start = -1;
		simple_map_head->stlb_start = -1;
		simple_map_head->mtlb_end = -1;
		simple_map_head->stlb_end = -1;
		simple_map_head->process_id = 0;
		simple_map_head->filp = NULL;
		simple_map_head->page_count = 0;
		simple_map_head->prev = NULL;
		simple_map_head->next = simple_map_tail;

		simple_map_tail->mtlb_start = -1;
		simple_map_tail->stlb_start = -1;
		simple_map_tail->mtlb_end = -1;
		simple_map_tail->stlb_end = -1;
		simple_map_tail->process_id = 0;
		simple_map_tail->filp = NULL;
		simple_map_tail->page_count = 0;
		simple_map_tail->prev = simple_map_head;
		simple_map_tail->next = NULL;

		mmu->region[i].simple_map_head = simple_map_head;
		mmu->region[i].simple_map_tail = simple_map_tail;
	}
	return MMU_STATUS_OK;
}

static enum MMUStatus CreateNode(struct MMU *mmu)
{
	struct MMUNode *free_map_head, *map_head, *p, **pp;
	struct MMUNode *free_map_tail, *map_tail;
	int i;
	unsigned int page_count;
	unsigned int prev_stlb = 0, prev_mtlb = 0;

	/* Init each region map node */
	for (i = 0; i < MMU_REGION_COUNT; i++) {
		free_map_head = kmalloc(sizeof(struct MMUNode),
					GFP_KERNEL | MMU_NOWARN);
		map_head = kmalloc(sizeof(struct MMUNode),
				   GFP_KERNEL | MMU_NOWARN);
		free_map_tail = kmalloc(sizeof(struct MMUNode),
					GFP_KERNEL | MMU_NOWARN);
		map_tail = kmalloc(sizeof(struct MMUNode),
				   GFP_KERNEL | MMU_NOWARN);

		free_map_head->mtlb_start = map_head->mtlb_start = -1;
		free_map_head->stlb_start = map_head->stlb_start = -1;
		free_map_head->mtlb_end = map_head->mtlb_end = -1;
		free_map_head->stlb_end = map_head->stlb_end = -1;
		free_map_head->process_id = map_head->process_id = 0;
		free_map_head->filp = map_head->filp = NULL;
		free_map_head->page_count = map_head->page_count = 0;
		free_map_head->prev = map_head->prev = NULL;
		free_map_head->next = free_map_tail;
		map_head->next = map_tail;

		free_map_tail->mtlb_start = map_tail->mtlb_start = -1;
		free_map_tail->stlb_start = map_tail->stlb_start = -1;
		free_map_tail->mtlb_end = map_tail->mtlb_end = -1;
		free_map_tail->stlb_end = map_tail->stlb_end = -1;
		free_map_tail->process_id = map_tail->process_id = 0;
		free_map_tail->filp = map_tail->filp = NULL;
		free_map_tail->page_count = map_tail->page_count = 0;
		free_map_tail->prev = free_map_head;
		map_tail->prev = map_head;
		free_map_tail->next = map_tail->next = NULL;

		mmu->region[i].free_map_head = free_map_head;
		mmu->region[i].map_head = map_head;
		mmu->region[i].free_map_tail = free_map_tail;
		mmu->region[i].map_tail = map_tail;

		p = kmalloc(sizeof(struct MMUNode), GFP_KERNEL | MMU_NOWARN);
		pp = &p;

		switch (i) {
		case MMU_REGION_IN:
			page_count = (REGION_IN_END - REGION_IN_START + 1) /
				     PAGE_SIZE;
			p->stlb_start = region_in_mmu_start >> 12 &
					0x3FF; //hold mmu addr: 0x0
			p->mtlb_start = region_in_mmu_start >> 22;
			//end point next region start: +1; for remainder: +1
			p->stlb_end = prev_stlb =
				region_in_mmu_end >> 12 & 0x3FF;
			p->mtlb_end = prev_mtlb = region_in_mmu_end >> 22;
			p->page_count = page_count - 1; //hold mmu addr: 0x0
			break;
		case MMU_REGION_OUT:
			page_count = (REGION_OUT_END - REGION_OUT_START + 1) /
				     PAGE_SIZE;
			p->stlb_start = region_out_mmu_start >> 12 & 0x3FF;
			p->mtlb_start = region_out_mmu_start >> 22;
			p->stlb_end = prev_stlb =
				region_out_mmu_end >> 12 & 0x3FF;
			p->mtlb_end = prev_mtlb = region_out_mmu_end >> 22;
			p->page_count = page_count;
			break;
		case MMU_REGION_PRIVATE:
			page_count = (REGION_PRIVATE_END -
				      REGION_PRIVATE_START + 1) /
				     PAGE_SIZE;
			p->stlb_start = region_private_mmu_start >> 12 & 0x3FF;
			p->mtlb_start = region_private_mmu_start >> 22;
			p->stlb_end = prev_stlb =
				region_private_mmu_end >> 12 & 0x3FF;
			p->mtlb_end = prev_mtlb = region_private_mmu_end >> 22;
			p->page_count = page_count;
			break;
		case MMU_REGION_PUB:
			p->stlb_start = prev_stlb;
			p->mtlb_start = prev_mtlb;
			p->stlb_end = prev_stlb = MMU_STLB_4K_ENTRY_NUM - 1;
			p->mtlb_end = prev_mtlb = MMU_MTLB_ENTRY_NUM - 1;
			p->page_count = (p->mtlb_end - p->mtlb_start) *
						MMU_STLB_4K_ENTRY_NUM +
					p->stlb_end - p->stlb_start + 1;
			break;
		default:
			XDX_DEV_ERROR(" *****MMU Region Error*****\n");
			break;
		}

		p->process_id = 0;
		p->filp = NULL;
		p->next = p->prev = NULL;

		InsertNode(i, pp, mmu, 1);
	}

	return MMU_STATUS_OK;
}

/* A simpile function to check if the map buffer is existed.
 *it needs more complex version
 */
#if 0
static enum MMUStatus SMCheckAddress(enum MMURegion e, struct MMU *mmu,
				     void *virtual_address)
{
	struct MMUNode *p;

	p = mmu->region[e].simple_map_head->next;
	while (p) {
		if (p->buf_virtual_address == virtual_address)
			return MMU_STATUS_FALSE;
		p = p->next;
	}
	return MMU_STATUS_OK;
}

static enum MMUStatus FindFreeNode(enum MMURegion e, struct MMUNode **node,
				   struct MMU *mmu, unsigned int page_count)
{
	struct MMUNode *p;

	p = mmu->region[e].free_map_head->next;
	while (p) {
		if (p->page_count >= page_count) {
			*node = p;
			return MMU_STATUS_OK;
		}
		p = p->next;
	}
	return MMU_STATUS_FALSE;
}

static enum MMUStatus SplitFreeNode(enum MMURegion e, struct MMUNode **node,
				    struct MMU *mmu, unsigned int page_count)
{
	struct MMUNode *p, **new;

	p = kmalloc(sizeof(struct MMUNode), GFP_KERNEL | MMU_NOWARN);
	new = &p;

	**new = **node;

	(*new)->mtlb_start = (*node)->mtlb_start;
	(*new)->stlb_start = (*node)->stlb_start;
	(*new)->mtlb_end =
		(page_count + (*node)->stlb_start) / MMU_STLB_4K_ENTRY_NUM +
		(*node)->mtlb_start;
	(*new)->stlb_end =
		(page_count + (*node)->stlb_start) % MMU_STLB_4K_ENTRY_NUM;
	(*new)->process_id = (*node)->process_id;
	(*new)->page_count = page_count;

	XDX_DEV_DEBUG(" *****new mtlb_start*****%d\n", (*new)->mtlb_start);
	XDX_DEV_DEBUG(" *****new stlb_start*****%d\n", (*new)->stlb_start);
	XDX_DEV_DEBUG(" *****new mtlb_end*****%d\n", (*new)->mtlb_end);
	XDX_DEV_DEBUG(" *****new stlb_end*****%d\n", (*new)->stlb_end);
	/* Insert a new node in map */
	InsertNode(e, new, mmu, 0);

	/* Update free node in free map*/
	(*node)->page_count -= page_count;
	if ((*node)->page_count == 0) {
		DeleteNode(node);
		XDX_DEV_DEBUG(" *****old node deleted*****\n");
	} else {
		(*node)->mtlb_start = (*new)->mtlb_end;
		(*node)->stlb_start = (*new)->stlb_end;

		XDX_DEV_DEBUG(" *****old mtlb_start*****%d\n", (*node)->mtlb_start);
		XDX_DEV_DEBUG(" *****old stlb_start*****%d\n", (*node)->stlb_start);
		XDX_DEV_DEBUG(" *****old mtlb_end*****%d\n", (*node)->mtlb_end);
		XDX_DEV_DEBUG(" *****old stlb_end*****%d\n", (*node)->stlb_end);
	}
	/* return a new node for map buffer */
	*node = *new;

	return MMU_STATUS_OK;
}
#endif

static enum MMUStatus SMRemoveNode(enum MMURegion e, struct MMU *mmu,
				   void *buf_virtual_address,
				   unsigned int process_id)
{
	struct MMUNode *p, **pp;

	p = mmu->region[e].simple_map_head->next;
	pp = &p;
	while (*pp) {
		if ((*pp)->buf_virtual_address == buf_virtual_address &&
		    (*pp)->process_id == process_id) {
			SMDeleteNode(pp);
			break;
		}
		*pp = (*pp)->next;
	}

	return MMU_STATUS_OK;
}

static enum MMUStatus RemoveNode(enum MMURegion e, void *buf_virtual_address,
				 struct MMU *mmu, unsigned int process_id)
{
	struct MMUNode *p, **pp;

	p = mmu->region[e].map_head->next;
	pp = &p;
	while (*pp) {
		if ((*pp)->buf_virtual_address == buf_virtual_address &&
		    (*pp)->process_id == process_id) {
			InsertNode(e, pp, mmu, 1);
			break;
		}
		*pp = (*pp)->next;
	}

	return MMU_STATUS_OK;
}

static enum MMUStatus Delay(unsigned int delay)
{
	if (delay > 0) {
		ktime_t dl = ktime_set((delay / MSEC_PER_SEC),
				       (delay % MSEC_PER_SEC) * NSEC_PER_MSEC);
		__set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_hrtimeout(&dl, HRTIMER_MODE_REL);
	}

	return MMU_STATUS_OK;
}

static enum MMUStatus CreateMutex(void **mtx)
{
	enum MMUStatus status;

	/* Allocate the mutex structure. */
	status = AllocateMemory(sizeof(struct mutex), mtx);
	if (MMU_IS_SUCCESS(status)) {
		/* Initialize the mutex. */
		mutex_init(*(struct mutex **)mtx);
	}

	return status;
}

static enum MMUStatus AcquireMutex(void *mtx, unsigned int timeout)
{
	if (timeout == MMU_INFINITE) {
		/* Lock the mutex. */
		mutex_lock(mtx);

		/* Success. */
		return MMU_STATUS_OK;
	}

	for (;;) {
		/* Try to acquire the mutex. */
		if (mutex_trylock(mtx))
			return MMU_STATUS_OK;

		if (timeout-- == 0)
			break;

		/* Wait for 1 millisecond. */
		Delay(1);
	}

	return MMU_STATUS_OK;
}

static enum MMUStatus ReleaseMutex(void *mtx)
{
	/* Release the mutex. */
	mutex_unlock(mtx);

	return MMU_STATUS_OK;
}

static inline enum MMUStatus QueryProcessPageTable(void *logical,
						   unsigned long long *address)
{
	unsigned long lg = (unsigned long)logical;
	unsigned long offset = lg & ~PAGE_MASK;
	struct vm_area_struct *vma;
	spinlock_t *ptl;
	pgd_t *pgd;
	pud_t *pud;
	pmd_t *pmd;
	pte_t *pte;

	if (is_vmalloc_addr(logical)) {
		/* vmalloc area. */
		*address = page_to_phys(vmalloc_to_page(logical)) | offset;
		return MMU_STATUS_OK;
	} else if (virt_addr_valid(logical)) {
		/* Kernel logical address. */
		*address = virt_to_phys(logical);
		return MMU_STATUS_OK;
	}
	/* Try user VM area. */
	if (!current->mm)
		return MMU_STATUS_NOT_FOUND;

#if KERNEL_VERSION(5, 8, 0) > LINUX_VERSION_CODE
	down_read(&current->mm->mmap_sem);
#else
	mmap_read_lock(current->mm);
#endif
	vma = find_vma(current->mm, lg);
#if KERNEL_VERSION(5, 8, 0) > LINUX_VERSION_CODE
	up_read(&current->mm->mmap_sem);
#else
	mmap_read_unlock(current->mm);
#endif

	/* To check if mapped to user. */
	if (!vma)
		return MMU_STATUS_NOT_FOUND;

	pgd = pgd_offset(current->mm, lg);
	if (pgd_none(*pgd) || pgd_bad(*pgd))
		return MMU_STATUS_NOT_FOUND;

#if (defined(CONFIG_CPU_CSKYV2) || defined(CONFIG_X86)) &&                     \
	KERNEL_VERSION(4, 12, 0) <= LINUX_VERSION_CODE
	pud = pud_offset((p4d_t *)pgd, lg);
#elif (defined(CONFIG_CPU_CSKYV2)) &&                                          \
	KERNEL_VERSION(4, 11, 0) <= LINUX_VERSION_CODE
	pud = pud_offset((p4d_t *)pgd, lg);
#else
#if (defined(XDXGPU_CENTOS79_SUPPORT))
        pud = pud_offset(pgd, lg);
#else
        pud = pud_offset((p4d_t *)pgd, lg);
#endif
#endif
	if (pud_none(*pud) || pud_bad(*pud))
		return MMU_STATUS_NOT_FOUND;

	pmd = pmd_offset(pud, lg);
	if (pmd_none(*pmd) || pmd_bad(*pmd))
		return MMU_STATUS_NOT_FOUND;

	pte = pte_offset_map_lock(current->mm, pmd, lg, &ptl);
	if (!pte) {
		spin_unlock(ptl);
		return MMU_STATUS_NOT_FOUND;
	}

	if (!pte_present(*pte)) {
		pte_unmap_unlock(pte, ptl);
		return MMU_STATUS_NOT_FOUND;
	}

	*address = (pte_pfn(*pte) << PAGE_SHIFT) | offset;
	pte_unmap_unlock(pte, ptl);

	*address -= gBaseDDRHw;

	return MMU_STATUS_OK;
}

static inline int GetProcessID(void)
{
	return current->tgid;
}

/*
static enum MMUStatus GetPhysicalAddress(void *logical,
					 unsigned long long *address)
{
	enum MMUStatus status;

	status = QueryProcessPageTable(logical, address);

	return status;
}
*/

static enum MMUStatus GetPageEntry(struct MMUNode *node,
				   unsigned int **page_table_entry,
				   struct MMU *mmu, unsigned int i)
{
	int num =
		node->mtlb_start * MMU_STLB_4K_ENTRY_NUM + node->stlb_start + i;
	*page_table_entry = (unsigned int *)mmu->stlb_virtual + num;
	return MMU_STATUS_OK;
}

static enum MMUStatus SetupDynamicSpace(struct device *dev, struct MMU *mmu,
					unsigned long BaseDDRAddr)
{
	unsigned int i;
	unsigned int stlb_entry;
	void *pointer = NULL;
	unsigned long long address;
	unsigned int num_entries = MMU_MTLB_ENTRY_NUM;
	unsigned int *mtlb_virtual = (unsigned int *)mmu->mtlb_virtual;

	AcquireMutex(mmu->page_table_mutex, MMU_INFINITE);
	if (pcie) {
		XDX_DEV_DEBUG(" *****stlb_virtual = %p**%d\n", pointer,
			 num_entries * MMU_STLB_4K_SIZE);

		address = mmu->stlb_physical;
		for (i = 0; i < num_entries; i++) {
			stlb_entry = address
				     /* 4KB page size */
				     | (0 << 2)
				     /* Ignore exception */
				     | (0 << 1)
				     /* Present */
				     | (1 << 0);
			WritePageEntry(mtlb_virtual++, stlb_entry);
			address += MMU_STLB_4K_SIZE;
		}
	} else {
		mmu->stlb_virtual =
			(void *)((u64)(mmu->mtlb_virtual) + MMU_MTLB_SIZE);
		mmu->stlb_physical = address =
			mmu->mtlb_physical + MMU_MTLB_SIZE;
		mmu->stlb_size = num_entries * MMU_STLB_4K_SIZE;

		for (i = 0; i < num_entries; i++) {
			stlb_entry = (address & 0xffffffe0)
				     /* 4KB page size */
				     | (0 << 2)
				     /* Ignore exception */
				     | (0 << 1)
				     /* Present */
				     | (1 << 0);
			WritePageEntry(mtlb_virtual++, stlb_entry);
			address += MMU_STLB_4K_SIZE;
		}
	}
	ReleaseMutex(mmu->page_table_mutex);

	/* Initial map info. */
	if (simple_map)
		SMCreateNodes(mmu);
	else
		CreateNode(mmu);

	return MMU_STATUS_OK;
}

static struct vpu_fpriv *get_vpu_fpriv(struct file *filp)
{
	struct drm_file *file_priv = filp->private_data;
	struct xdx_drm_fpriv *fpriv = file_priv->driver_priv;
	struct vpu_fpriv *vpu_fpriv_data = fpriv->vpu_fpriv;

	return vpu_fpriv_data;
}

static int add_file_bo(struct file *filp, struct xdx_bo *bo)
{
	struct vpu_fpriv *vpu_fpriv_data = get_vpu_fpriv(filp);
	struct list_binder *binder;

	binder = kmalloc(sizeof(struct list_binder), GFP_KERNEL);
	if (!binder)
		return -ENOMEM;

	binder->data = (void*)bo;
	mutex_lock(&vpu_fpriv_data->mm.mutex);
	list_add_tail(&binder->list, &vpu_fpriv_data->mm.bo_list);
	mutex_unlock(&vpu_fpriv_data->mm.mutex);
	return 0;
}

static int del_file_bo(struct file *filp, struct xdx_bo *bo)
{
	struct vpu_fpriv *vpu_fpriv_data = get_vpu_fpriv(filp);
	struct list_binder *binder;

	mutex_lock(&vpu_fpriv_data->mm.mutex);
	list_for_each_entry(binder, &vpu_fpriv_data->mm.bo_list, list) {
		if (binder->data == (void *)bo)
			break;
	}
	list_del(&binder->list);
	mutex_unlock(&vpu_fpriv_data->mm.mutex);

	kfree(binder);
	return 0;
}
/*
static int unmap_bo(struct file *filp, struct xdx_bo *bo)
{
	struct drm_file *file_priv = filp->private_data;
	struct drm_device *ddev = file_priv->minor->dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);

	struct xdx_bo_va *va;

	va = xdx_vm_bo_find(&xdev->vmm, &xdev->vmm.vpu, bo);
	if (!va) {
		return -EFAULT;
	}

	xdx_vm_bo_remove(&xdev->vmm, va);

	return 0;
}
*/
static int del_file_bo_all(struct file *filp)
{
	struct vpu_fpriv *vpu_fpriv_data = get_vpu_fpriv(filp);
	struct list_binder *binder, *tmp;
	u32 count = 0;

	if (!vpu_fpriv_data)
		return 0;

	mutex_lock(&vpu_fpriv_data->mm.mutex);

	list_for_each_entry_safe(binder, tmp, &vpu_fpriv_data->mm.bo_list, list) {
		count++;
		list_del(&binder->list);
		mutex_unlock(&vpu_fpriv_data->mm.mutex);
//		unmap_bo(filp, binder->data);
		kfree(binder);
		mutex_lock(&vpu_fpriv_data->mm.mutex);
	}

	mutex_unlock(&vpu_fpriv_data->mm.mutex);

	if (count > 0) {
		mm_probe = count;
		XDX_DEV_DEBUG("%d bo released.\n", mm_probe);
	}
	return 0;
}

void vpu_mmu_release(struct file *filp)
{
	del_file_bo_all(filp);
}


static int vpu_mmu_hw_init(struct mmu_t *mmucore, struct xdx_device *xdev, unsigned long ddr_base, unsigned long reg_base)
{
#if VPU_SHARE_PAGETABLE
	struct mmu_t *mcore;
#endif
	unsigned int address;
	unsigned int address_ext = 0;
	char *hwregs = NULL;

	mmucore->core_cfg.mmucorebase = mmucore->core_cfg.mmucorebase + reg_base;
	if (!request_mem_region(mmucore->core_cfg.mmucorebase, mmucore->core_cfg.iosize,
			"hantrommu")) {
		XDX_DEV_ERROR("hantrovcmd: failed to reserve mmu HW regs\n");
		return -EBUSY;
	}
	mmucore->hwregs = (u8 *)ioremap_wc(mmucore->core_cfg.mmucorebase,
					  mmucore->core_cfg.iosize);

	address = mmucore->g_mmu->page_table_array_physical-pci_resource_start(xdev->pdev, 2);
	address_ext = 0;
	hwregs = mmucore->hwregs;

#if VPU_SHARE_PAGETABLE
	mcore=get_mmunodebytype(xdev, 0, HANTRO_CORE_DEC);
	if (mcore != mmucore) {
		mmucore = mcore;
		pr_info("[%s]%d VPU use share pagetable.\n",__func__,__LINE__);
	}

#endif
	AcquireMutex(mmucore->g_mmu->hw_mutex, MMU_INFINITE);
	iowrite32(((0x100000000+mmucore->g_mmu->mtlb_physical)>>8) & 0xfffffff0, (void __iomem *)(hwregs +
		MMU_REG_PDENTRY0));
	iowrite32(0x21, (void __iomem *)(hwregs + MMU_REG_CONTROL));

#ifndef PD_MODE
	iowrite32(address, (void *)(hwregs + MMU_REG_ADDRESS));
	iowrite32(address_ext, (void *)(hwregs + MMU_REG_ADDRESS_MSB));
	iowrite32(0x10000, (void *)(hwregs + MMU_REG_PAGE_TABLE_ID));

	iowrite32(0x00000, (void *)(hwregs + MMU_REG_PAGE_TABLE_ID));
	iowrite32(0x1, (void *)(hwregs + MMU_REG_CONTROL));
#else
	iowrite32(((0x100000000+mmucore->g_mmu->mtlb_physical)>>8) & 0xfffffff0, (void __iomem *)(hwregs +
			   MMU_REG_PDENTRY0));
	iowrite32(0x21, (void __iomem *)(hwregs + MMU_REG_CONTROL));
#endif

	ReleaseMutex(mmucore->g_mmu->hw_mutex);

	return MMU_STATUS_OK;
}
static struct mmu_t * vpu_mmu_sw_init(struct xdx_device *xdev, struct mmu_core_cfg *core_cfg)
{
	struct mmu_t *mmucore;
	struct device *dev = xdev->dev;
	u64 paddr = core_cfg->mmucorebase + xdev->rmmio_base;
	u64 size = core_cfg->iosize;

	mmucore = kmalloc(sizeof(struct mmu_t), GFP_KERNEL);
	if (!mmucore) {
		XDX_DEV_DEBUG();
		return NULL;
	}

	mmucore->g_mmu = NULL;
	mmucore->mmu_enable = 0;
	mmucore->dev = dev;
	mmucore->core_cfg = *core_cfg;
	mmucore->hwregs = (u8 *)ioremap_wc(paddr, size);

	add_mmunode(xdev, mmucore->core_cfg.sliceidx, mmucore);

	return mmucore;
}

static void vpu_mmu_sw_deinit(struct mmu_t *mmucore)
{
	iounmap((void *)mmucore->hwregs);
	list_del(&mmucore->list);
}

int vpu_mmu_probe(dtbnode *pnode, int type, struct mmu_t *pmmu,
		   struct xdx_device *xdev, unsigned long ddr_base, unsigned long reg_base)
{
	enum MMUStatus status = 0;
	unsigned int i,k;
	void *pointer;
	struct mmu_t *mmucore;
	struct device *dev = xdev->dev;
	struct mmu_core_cfg mmu_cfg;

	for (i = 0; i < ARRAY_SIZE(mmu_core_array); i++) {
		if (mmu_core_array[i].parenttype == type)
			mmu_cfg = mmu_core_array[i];
	}

	if (mmu_cfg.parenttype == HANTRO_CORE_DEC && type == HANTRO_CORE_DEC)
		pr_info("HANTRO_CORE_DEC initialized.\n");
	else if (mmu_cfg.parenttype == HANTRO_CORE_ENC && type == HANTRO_CORE_ENC)
		pr_info("HANTRO_CORE_ENC initialized.\n");
	else {
		pr_err("Error MMU unknown.\n");
		return MMU_STATUS_NOT_FOUND;
	}

	pcie = 1;

	if (gBaseDDRHw == (unsigned long)-1) {
		gBaseDDRHw = ddr_base;
		simple_map = 0;
		map_shift = 0;
	}

	mmucore = vmalloc(sizeof(struct mmu_t));
	if (!mmucore) {
		XDX_DEV_DEBUG();
		return -ENOMEM;
	}
	pointer = kmalloc(sizeof(struct MMU), GFP_KERNEL | MMU_NOWARN);
	if (!pointer) {
		XDX_DEV_ERROR("can't alloc mmu node, no mem");
		vfree(mmucore);
		return -ENOMEM;
	}
	memset(pointer, 0, sizeof(struct MMU));
	mmucore->g_mmu = pointer;
	mmucore->mmu_enable = 0;
	mmucore->dev = dev;
	mmucore->core_cfg = mmu_cfg;
	XDX_DEV_DEBUG("mmu step 0: %llx:%d", mmucore->core_cfg.mmucorebase,
				   mmucore->core_cfg.iosize);

	MMU_ON_ERROR(CreateMutex(&mmucore->g_mmu->page_table_mutex));
	MMU_ON_ERROR(CreateMutex(&mmucore->g_mmu->hw_mutex));

	for (k = 0; k < MMU_REGION_COUNT; k++)
		MMU_ON_ERROR(CreateMutex(&mmucore->g_mmu->region[k].node_mutex));

	add_mmunode(xdev, mmucore->core_cfg.sliceidx, mmucore);

	MMUEnableStep0(xdev, ddr_base, mmucore);

	vpu_mmu_hw_init(mmucore, xdev, ddr_base, reg_base);

	return MMU_STATUS_OK;
onerror:
	XDX_DEV_ERROR(" *****MMU Init Error*****\n");
	vfree(mmucore);
	kfree(pointer);
	return status;

}

int hantroMMUHwInit(struct xdx_device *xdev, int type)
{
	struct mmu_t *mmu = NULL;
	u8 *hwregs = NULL;
	struct slice_info *slice = getslicenode(xdev, 0);
	u32 mtlb = 0, ext_mtlb = 0;
	u32 config = 0;
	struct dc8200_mmu *dcmmu = (struct dc8200_mmu*)xdev->vmm.vpu.context;

	mutex_lock(xdev->vmm.vpu.hw_mmu_mutex);
	list_for_each_entry(mmu, &slice->mmu_head, list) {
		if (mmu->core_cfg.parenttype == type) {
			hwregs = mmu->hwregs;

                        mtlb = (u32)(dcmmu->mtlb_physical & 0xFFFFFFFF);
                        ext_mtlb = (u32)((dcmmu->mtlb_physical >> 32) & 0xff);
                        if (ext_mtlb & 0xFFFFFF00) {
                                dev_err(xdev->dev, "Mtlb address out of range.\n");
	                        mutex_unlock(xdev->vmm.vpu.hw_mmu_mutex);
                                return MMU_STATUS_FALSE;
                        }

                        config = (((dcmmu->mtlb_physical >> 8) & 0xfffffff0) | dcmmu->mode);

                        iowrite32(0x20, (void __iomem *)(hwregs + MMU_REG_CONTROL));
                        iowrite32(config, (void __iomem *)(hwregs + MMU_REG_PDENTRY0));
                        iowrite32(0x21, (void __iomem *)(hwregs + MMU_REG_CONTROL));
		}
	}
	mutex_unlock(xdev->vmm.vpu.hw_mmu_mutex);

	return MMU_STATUS_OK;
}

int hantroMMUInit(struct xdx_device *xdev, int type)
{
	struct dc8200_mmu *mmu = (struct dc8200_mmu*)xdev->vmm.vpu.context;
	unsigned long reg_base;
	struct mmu_t * vpu_mmu = NULL;
	u32 mtlb = 0, ext_mtlb = 0;
	u32 config = 0;
	u8 *hwregs = NULL;
	int i = 0;

	if (CreateMutex(&xdev->vmm.vpu.hw_mmu_mutex) < 0) {
		dev_err(xdev->dev, "failed to create vpu mmu mutex.\n");
		return MMU_STATUS_FALSE;
	}

	reg_base = xdev->rmmio_base;

	for (i = 0; i < ARRAY_SIZE(mmu_core_array); i++) {
		if (type == mmu_core_array[i].parenttype)
			vpu_mmu = vpu_mmu_sw_init(xdev, &mmu_core_array[i]);
	}

	if (!vpu_mmu)
		return MMU_STATUS_FALSE;
	/*
	if (!request_mem_region(mmu_core_base, mmu_core_array[0].iosize, "hantrommu")) {
	XDX_DEV_ERROR("hantrovcmd: failed to reserve mmu HW regs\n");
	return -EBUSY;
	}
	*/
	mutex_lock(xdev->vmm.vpu.hw_mmu_mutex);
	hwregs = vpu_mmu->hwregs;

	mtlb = (u32)(mmu->mtlb_physical & 0xFFFFFFFF);
	ext_mtlb = (u32)((mmu->mtlb_physical >> 32) & 0xff);
	if (ext_mtlb & 0xFFFFFF00) {
		dev_err(xdev->dev, "Mtlb address out of range.\n");
	        mutex_unlock(xdev->vmm.vpu.hw_mmu_mutex);
		return MMU_STATUS_FALSE;
	}

	config = (((mmu->mtlb_physical >> 8) & 0xfffffff0) | mmu->mode);

	iowrite32(0x20, (void __iomem *)(hwregs + MMU_REG_CONTROL));
	iowrite32(config, (void __iomem *)(hwregs + MMU_REG_PDENTRY0));
	iowrite32(0x21, (void __iomem *)(hwregs + MMU_REG_CONTROL));
	mutex_unlock(xdev->vmm.vpu.hw_mmu_mutex);

	return MMU_STATUS_OK;
}

int hantroMMUprobe(dtbnode *pnode, int loop, struct mmu_t *pmmu,
		   struct xdx_device *xdev, unsigned long ddr_base, unsigned long reg_base)
{
#ifndef USE_DTB_PROBE
	unsigned int address;
	unsigned int address_ext = 0;
	char *hwregs = NULL;
#endif
	enum MMUStatus status = 0;
	unsigned int i, k, id;

	void *pointer;
	struct mmu_t *mmucore;
	struct device *dev = xdev->dev;

#ifdef VSI_FPGA
	pcie = 1;
#else
	pcie = 0;
#endif

	if (loop == 0 && gBaseDDRHw != (unsigned long)-1)
		return 0;

	if (gBaseDDRHw == (unsigned long)-1) {
		gBaseDDRHw = ddr_base;
		simple_map = 0;
		map_shift = 0;
	}
	XDX_DEV_DEBUG("%s %d ddr = %lx", __func__, pcie, ddr_base);
	if (loop == 0) {
#ifndef USE_DTB_PROBE /*simulate and compatible with old code*/
		for (i = 0; i < ARRAY_SIZE(mmu_core_array); i++) {
			mmucore = vmalloc(sizeof(struct mmu_t));
			if (!mmucore)
				break;
			pointer =
				kmalloc(sizeof(struct MMU),
					GFP_KERNEL | MMU_NOWARN);
			if (!pointer) {
				XDX_DEV_ERROR("can't alloc mmu node, no mem");
				vfree(mmucore);
				break;
			}
			memset(pointer, 0, sizeof(struct MMU));
			mmucore->g_mmu = pointer;
			mmucore->mmu_enable = 0;
			mmucore->dev = dev;
			mmucore->core_cfg = mmu_core_array[i];
			XDX_DEV_DEBUG("mmu step 0: %llx:%x",
				 mmucore->core_cfg.mmucorebase,
				 mmucore->core_cfg.iosize);

			MMU_ON_ERROR(
				CreateMutex(&mmucore->g_mmu->page_table_mutex));
			MMU_ON_ERROR(
				CreateMutex(&mmucore->g_mmu->hw_mutex));

			for (k = 0; k < MMU_REGION_COUNT; k++)
				MMU_ON_ERROR(CreateMutex(
					&mmucore->g_mmu->region[k].node_mutex));
			add_mmunode(xdev, mmucore->core_cfg.sliceidx, mmucore);
			MMUEnableStep0(xdev, ddr_base, mmucore);

			mmucore->core_cfg.mmucorebase = mmucore->core_cfg.mmucorebase + reg_base;
			if (!request_mem_region(mmucore->core_cfg.mmucorebase,
						mmucore->core_cfg.iosize,
						"hantrommu")) {
				XDX_DEV_ERROR("hantrovcmd: failed to reserve mmu HW regs\n");
				return -EBUSY;
			}
			mmucore->hwregs =
				(u8 *)ioremap_wc(mmucore->core_cfg.mmucorebase,
						mmucore->core_cfg.iosize);
			if (pcie) {
				address = mmucore->g_mmu->page_table_array_physical-pci_resource_start(xdev->pdev, 2);
				address_ext = 0;
			} else {
				address = mmucore->g_mmu->page_table_array_physical;
				address_ext = ((u32)(mmucore->g_mmu->page_table_array_physical >> 32))&0xff;
			}

			hwregs = mmucore->hwregs;
			AcquireMutex(mmucore->g_mmu->hw_mutex, MMU_INFINITE);
			iowrite32(((0x100000000+mmucore->g_mmu->mtlb_physical)>>8) & 0xfffffff0, (void __iomem *)(hwregs +
					MMU_REG_PDENTRY0));
			iowrite32(0x21, (void __iomem *)(hwregs + MMU_REG_CONTROL));
		#ifndef PD_MODE
			iowrite32(address, (void *)(hwregs + MMU_REG_ADDRESS));
			iowrite32(address_ext, (void *)(hwregs + MMU_REG_ADDRESS_MSB));
			iowrite32(0x10000, (void *)(hwregs + MMU_REG_PAGE_TABLE_ID));

			iowrite32(0x00000, (void *)(hwregs + MMU_REG_PAGE_TABLE_ID));
			iowrite32(0x1, (void *)(hwregs + MMU_REG_CONTROL));
		#else
			iowrite32(((0x100000000+mmucore->g_mmu->mtlb_physical)>>8) & 0xfffffff0, (void __iomem *)(hwregs +
					MMU_REG_PDENTRY0));
			iowrite32(0x21, (void __iomem *)(hwregs +
					MMU_REG_CONTROL));
		#endif
			ReleaseMutex(mmucore->g_mmu->hw_mutex);
		}
#else //USE_DTB_PROBE
		mmucore = vmalloc(sizeof(struct mmu_t));
		if (!mmucore)
			return -ENOMEM;
		pointer = kmalloc(sizeof(struct MMU),
				  GFP_KERNEL | MMU_NOWARN);
		if (!pointer) {
			vfree(mmucore);
			return -ENOMEM;
		}
		memset(pointer, 0, sizeof(struct MMU));
		mmucore->g_mmu = pointer;
		mmucore->core_cfg.mmucorebase = pnode->ioaddr;
		mmucore->core_cfg.iosize = pnode->iosize;
		mmucore->core_cfg.sliceidx = pnode->sliceidx;
		mmucore->core_cfg.parenttype = HANTRO_CORE_DEC;
		mmucore->mmu_enable = 0;
		mmucore->dev = dev;
		MMU_ON_ERROR(CreateMutex(&mmucore->g_mmu->page_table_mutex));
		MMU_ON_ERROR(CreateMutex(&mmucore->g_mmu->hw_mutex));

		for (i = 0; i < MMU_REGION_COUNT; i++)
			MMU_ON_ERROR(CreateMutex(
				&mmucore->g_mmu->region[i].node_mutex));
		add_mmunode(xdev, mmucore->core_cfg.sliceidx, mmucore);
		MMUEnableStep0(xdev, ddr_base, mmucore);
#endif //USE_DTB_PROBE
		return 0;
	}

	//loop == 1 part
	if (!pmmu)
		return -ENODEV;
	if (!request_mem_region(pmmu->core_cfg.mmucorebase,
				pmmu->core_cfg.iosize, "hantrommu")) {
		kfree(pmmu->g_mmu);
		return -ENODEV;
	}
	pmmu->hwregs = (u8 *)ioremap_wc(pmmu->core_cfg.mmucorebase,
				     pmmu->core_cfg.iosize);
	if (!pmmu->hwregs) {
		release_mem_region(pmmu->core_cfg.mmucorebase,
				   pmmu->core_cfg.iosize);
		kfree(pmmu->g_mmu);
		return -ENODEV;
	}
	XDX_DEV_DEBUG("mmu reg %llx map to %p", pmmu->core_cfg.mmucorebase,
		 pmmu->hwregs);
	id = ioread32((void *)(pmmu->hwregs + MMU_REG_HW_ID)) >> 16;
	if (id != 0x4D4D) {
		iounmap((void *)pmmu->hwregs);
		release_mem_region(pmmu->core_cfg.mmucorebase,
				   pmmu->core_cfg.iosize);
		kfree(pmmu->g_mmu);
		return -ENODEV;
	}
	MMUEnableStep1(pmmu->hwregs, pmmu->g_mmu);
	return MMU_STATUS_OK;

onerror:
	XDX_DEV_ERROR(" *****MMU Init Error*****\n");
	return status;
}

static int vpu_mmu_flush(struct xdx_device *xdev, int type)
{
	struct mmu_t *mmu;
	u8 *hwregs = NULL;

	struct slice_info * slice = getslicenode(xdev, 0);

	mutex_lock(xdev->vmm.vpu.hw_mmu_mutex);
	list_for_each_entry(mmu, &slice->mmu_head, list) {
		if (mmu->core_cfg.parenttype == type) {
			hwregs = mmu->hwregs;
			iowrite32(0x10, (void __iomem *)(hwregs + MMU_REG_FLUSH));
			iowrite32(0x00, (void __iomem *)(hwregs + MMU_REG_FLUSH));
		}
	}
	mutex_unlock(xdev->vmm.vpu.hw_mmu_mutex);

	return MMU_STATUS_OK;
}
#if 0
static long vpu_mmu_flushall(struct xdx_device *xdev)
{
	struct mmu_t *mmu;
	u8 *hwregs = NULL;

	struct slice_info * slice = getslicenode(0);

	mutex_lock(xdev->vmm.vpu.hw_mmu_mutex);
	list_for_each_entry(mmu, &slice->mmu_head, list) {
		hwregs = mmu->hwregs;
		iowrite32(0x10, (void __iomem *)(hwregs + MMU_REG_FLUSH));
		iowrite32(0x00, (void __iomem *)(hwregs + MMU_REG_FLUSH));
	}
	mutex_unlock(xdev->vmm.vpu.hw_mmu_mutex);

	return 0;
}
#endif

int hantroMMUMap(struct xdx_device *xdev, struct xdx_bo *bo, unsigned long* vaddr, int type)
{
	struct xdx_bo_va *va;

	va = xdx_vm_bo_find(&xdev->vmm, &xdev->vmm.vpu, bo);
	if (!va) {
		va = xdx_vm_bo_add(&xdev->vmm, &xdev->vmm.vpu, bo);
		vpu_mmu_flush(xdev, type);
	}

	WARN_ON(!va);
	*vaddr = va->addr.value;

	return 0;
}

static int hantroMMUBufferMapIoctl(struct file *f, struct addr_desc *desc, int type)
{
	struct drm_file *file_priv = f->private_data;
	struct drm_device *ddev = file_priv->minor->dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);
	struct drm_gem_object *obj;
	struct xdx_bo *bo;
	struct xdx_bo_va *va;
	int handle = desc->handle;

	obj = drm_gem_object_lookup(file_priv, handle);
	bo = to_xdx_gem(obj);

	va = xdx_vm_bo_find(&xdev->vmm, &xdev->vmm.vpu, bo);
	if (!va) {
		va = xdx_vm_bo_add(&xdev->vmm, &xdev->vmm.vpu, bo);
		add_file_bo(f, bo);
	}
	if (va){ 
		desc->bus_address = va->addr.value;
	} else {
		//WARN_ON(!va);
		desc->bus_address = 0;
	}
	xdx_gem_object_put(obj);

	return 0;
}


static int hantroMMUBufferUnMapIoctl(struct file *filp, int handle)
{
	struct drm_file *file_priv = filp->private_data;
	struct drm_device *ddev = file_priv->minor->dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);

	struct drm_gem_object *obj;
	struct xdx_bo *bo;
	struct xdx_bo_va *va;

	obj = drm_gem_object_lookup(file_priv, handle);
	bo = to_xdx_gem(obj);

	XDX_DEV_DEBUG("bo:0x%llx\n", (u64)bo);
	va = xdx_vm_bo_find(&xdev->vmm, &xdev->vmm.vpu, bo);
	if (!va) {
		return -EFAULT;
	}

	xdx_vm_bo_remove(&xdev->vmm, va);

	del_file_bo(filp, bo);
	xdx_gem_object_put(obj);
	return 0;
}

void hantroMMURelease(struct file *filp)
{
        struct drm_file *file_priv = filp->private_data;
        struct drm_device *ddev = file_priv->minor->dev;
        struct xdx_device *xdev = drm_to_xdev(ddev);
	struct mmu_t *mmucore;
	int i, slicen = get_slicenumber(xdev);
 

	for (i = 0; i < slicen; i++) {
		mmucore = get_mmunode(xdev, i, 0);
		while (mmucore) {
			if (simple_map) {
				if (!mmucore->g_mmu->region[0].simple_map_head)
					continue;
			} else {
				if (!mmucore->g_mmu->region[0].map_head)
					continue;
			}

			MMURelease(filp, mmucore->g_mmu);
			mmucore = mmucore->next;
		}
	}
}

static enum MMUStatus MMURelease(void *filp, struct MMU *mmu)
{
	unsigned int i, j;
	struct MMUNode *p, *tmp;
	unsigned long long address;
	unsigned int *page_table_entry;

	/* if mmu or TLB not enabled, return */
	XDX_DEV_DEBUG(" *****MMU Release*****\n");

	AcquireMutex(mmu->page_table_mutex, MMU_INFINITE);

	if (simple_map) {
		for (i = 0; i < MMU_REGION_COUNT; i++) {
			p = mmu->region[i].simple_map_head->next;

			while (p) {
				tmp = p->next;
				if (p->filp == (struct file *)filp) {
					for (j = 0; j < p->page_count; j++) {
						GetPageEntry(p,
							     &page_table_entry,
							     mmu, j);
						address = 0;
						WritePageEntry(page_table_entry,
							       address);
					}

					SMRemoveNode(i, mmu,
						     p->buf_virtual_address,
						     p->process_id);
				}
				p = tmp;
			}
		}
	} else {
		for (i = 0; i < MMU_REGION_COUNT; i++) {
			p = mmu->region[i].map_head->next;

			while (p) {
				tmp = p->next;
				if (p->filp == (struct file *)filp) {
					for (j = 0; j < p->page_count; j++) {
						GetPageEntry(p,
							     &page_table_entry,
							     mmu, j);
						address = 0;
						WritePageEntry(page_table_entry,
							       address);
					}
					RemoveNode(i, p->buf_virtual_address,
						   mmu, p->process_id);
				}
				p = tmp;
			}
		}
	}
	ReleaseMutex(mmu->page_table_mutex);

	return MMU_STATUS_OK;
}

void hantroMMUCleanup(struct xdx_device *xdev)
{
	int i, slicen = get_slicenumber(xdev);
	struct mmu_t *mmucore, *pnext;

	for (i = 0; i < slicen; i++) {
		mmucore = get_mmunode(xdev, i, 0);
		while (mmucore) {
			vpu_mmu_sw_deinit(mmucore);

			pnext = mmucore->next;
			kfree(mmucore);
			mmucore = pnext;
		}
	}
}

/*------------------------------------------------------------------------------
 *Function name: MMUEnable
 *Description:
 *Create TLB, set registers and enable MMU

 *For pcie, TLB buffers come from FPGA memory and The distribution is as follows
 *MTLB:              start from: 0x00100000, size: 4K bits
 *page table array:              0x00200000        64 bits
 *STLB:                          0x00300000        4M bits
 *------------------------------------------------------------------------------
 */
static enum MMUStatus MMUEnableStep0(struct xdx_device *xdev,
				     unsigned long BaseDDRAddr,
				     struct mmu_t *mmucore)
{
	enum MMUStatus status;
	dma_addr_t dma_handle;
	struct MMU *mmu;
	struct MMU *mmu0;
	struct slice_info *parentslice;
	int total_table_size;
	struct device *dev=xdev->dev;
	//void *cpu_vaddr;

	XDX_DEV_DEBUG(" *****MMU Enable...*****\n");

	parentslice = (struct slice_info *)(mmucore->parentslice);
	mmu = mmucore->g_mmu;

	if (parentslice->mmucore_num > 0) {
		AcquireMutex(mmu->page_table_mutex, MMU_INFINITE);
		if (pcie) {
			mmu->mtlb_size = MMU_MTLB_SIZE;
			mmu->mtlb_physical = vpu_alloc_gddr(xdev, &mmucore->mtlb_bo, mmu->mtlb_size);
			mmu->mtlb_virtual = ioremap_wc(mmu->mtlb_physical, mmu->mtlb_size);
			mmu->mtlb_physical = mmu->mtlb_physical - pci_resource_start(xdev->pdev, 2);

			mmu->page_table_array_physical = vpu_alloc_gddr(xdev, &mmucore->pagetbl_bo, PAGE_TABLE_ENTRY_SIZE);
			mmu->page_table_array = ioremap_wc(mmu->page_table_array_physical, PAGE_TABLE_ENTRY_SIZE);
			total_table_size = mmu->mtlb_size;

			mmu->stlb_size = MMU_MTLB_ENTRY_NUM*MMU_STLB_4K_SIZE;
			mmu->stlb_physical = vpu_alloc_gddr(xdev, &mmucore->stlb_bo, mmu->stlb_size);
			mmu->stlb_virtual = ioremap_wc(mmu->stlb_physical, mmu->stlb_size);
			mmu->stlb_physical = mmu->stlb_physical - pci_resource_start(xdev->pdev, 2);
		} else {
			/* Allocate the 4K mode MTLB table. */

			total_table_size =
				MMU_MTLB_SIZE +
				MMU_MTLB_ENTRY_NUM * MMU_STLB_4K_SIZE;
			mmu->mtlb_size = MMU_MTLB_SIZE;
			mmu->mtlb_virtual =
				dma_alloc_coherent(dev, total_table_size,
						   &dma_handle,
						   GFP_KERNEL | GFP_DMA);
			XDX_DEV_DEBUG(" *****g_mmu->mtlb_virtual = 0x%p\n",
				 mmu->mtlb_virtual);
			mmu->mtlb_physical = (unsigned long long)dma_handle;
			XDX_DEV_DEBUG(" *****mtlb_physical = 0x%08x\n",
				 (unsigned int)mmu->mtlb_physical);
			if (!mmu->mtlb_virtual) {
				XDX_DEV_ERROR("hantrodec alloc buffer fail\n");
				status = MMU_STATUS_FALSE;
				ReleaseMutex(mmu->page_table_mutex);
				goto onerror;
			}

			mmu->page_table_array_size = PAGE_TABLE_ENTRY_SIZE;
			mmu->page_table_array = dma_alloc_coherent(
				dev, mmu->page_table_array_size, &dma_handle,
				GFP_KERNEL | GFP_DMA);
			XDX_DEV_DEBUG(" *****g_mmu->page_table_array = 0x%p\n",
				 mmu->page_table_array);
			mmu->page_table_array_physical =
				(unsigned long long)dma_handle;
			pr_debug(" *****page_table_array_physical = 0x%llx\n",
				 mmu->page_table_array_physical);
			if (!mmu->page_table_array) {
				XDX_DEV_ERROR("hantrodec alloc buffer fail\n");
				status = MMU_STATUS_FALSE;
				ReleaseMutex(mmu->page_table_mutex);
				goto onerror;
			}
		}

		*((unsigned int *)mmu->page_table_array) =
			(mmu->mtlb_physical & 0xFFFFF000) | (0 << 0);
		*((unsigned int *)mmu->page_table_array + 1) =
			(u32)(mmu->mtlb_physical >> 32) & 0xff;
		/* a workaround for Fifo overflow */
		*((unsigned int *)mmu->page_table_array + 2) =
			(mmu->mtlb_physical & 0xFFFFF000) | (0 << 0);
		*((unsigned int *)mmu->page_table_array + 3) =
			(u32)(mmu->mtlb_physical >> 32) & 0xff;

		pr_debug(" Page table array[0]: lsb = 0x%08x\n",
			 ((int *)mmu->page_table_array)[0]);
		pr_debug("                      msb = 0x%08x\n",
			 ((int *)mmu->page_table_array)[1]);

		ZeroMemory(mmu->mtlb_virtual, total_table_size);
		ReleaseMutex(mmu->page_table_mutex);
		MMU_ON_ERROR(SetupDynamicSpace(dev, mmu, BaseDDRAddr));
	} else {
		mmu0 = parentslice->mmuhdr->g_mmu;

		AcquireMutex(mmu->page_table_mutex, MMU_INFINITE);
		/* Allocate the 4K mode MTLB table. */
		mmu->mtlb_size = mmu0->mtlb_size;
		mmu->mtlb_virtual = mmu0->mtlb_virtual;
		mmu->mtlb_physical = mmu0->mtlb_physical;
		XDX_DEV_DEBUG(" *****mtlb_physical = 0x%08x\n",
			 (unsigned int)mmu->mtlb_physical);

		mmu->page_table_array_size = mmu0->page_table_array_size;
		mmu->page_table_array = mmu0->page_table_array;
		XDX_DEV_DEBUG(" *****g_mmu->page_table_array = 0x%p\n",
			 mmu->page_table_array);
		mmu->page_table_array_physical =
			mmu0->page_table_array_physical;

		XDX_DEV_DEBUG(" Page table array[0]: lsb = 0x%08x\n",
			 ((int *)mmu->page_table_array)[0]);
		XDX_DEV_DEBUG("                      msb = 0x%08x\n",
			 ((int *)mmu->page_table_array)[1]);

		mmu->stlb_virtual = mmu0->stlb_virtual;
		mmu->stlb_physical = mmu0->stlb_physical;
		mmu->stlb_size = mmu0->stlb_size;
		if (simple_map)
			SMCreateNodes(mmu);
		else
			CreateNode(mmu);

		ReleaseMutex(mmu->page_table_mutex);
	}
	return MMU_STATUS_OK;
onerror:
	XDX_DEV_ERROR(" *****MMU Enable Error*****\n");
	return status;
}

void MMUEnableStep1(unsigned char *hwregs, struct MMU *mmu)
{
	/* set regs of all MMUs */
	unsigned int address;
	unsigned int address_ext = 0;

	if (pcie) {
		address = PAGE_PCIE_START_ADDRESS;
	} else {
		address = mmu->page_table_array_physical;
		address_ext =
			((u32)(mmu->page_table_array_physical >> 32)) & 0xff;
	}

	AcquireMutex(mmu->hw_mutex, MMU_INFINITE);
	iowrite32(address, (void *)(hwregs + MMU_REG_ADDRESS));
	iowrite32(address_ext, (void *)(hwregs + MMU_REG_ADDRESS_MSB));
	iowrite32(0x10000, (void *)(hwregs + MMU_REG_PAGE_TABLE_ID));
	iowrite32(0x00000, (void *)(hwregs + MMU_REG_PAGE_TABLE_ID));
	iowrite32(1, (void *)(hwregs + MMU_REG_CONTROL));
	ReleaseMutex(mmu->hw_mutex);
}

long hantroMMUIoctl(unsigned int cmd, struct file *filp, unsigned long arg)
{
	struct addr_desc addr;
	int ret = 0;
	int nodeType;
	struct drm_file *file_priv = filp->private_data;
	struct drm_device *ddev = file_priv->minor->dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);

	if (copy_from_user(&addr, (void *)arg, sizeof(struct addr_desc)))
		return -EFAULT;

	switch (cmd) {
	case HANTRO_IOCS_MMU_MEM_MAP:
		ret = hantroMMUBufferMapIoctl((struct file *)filp, &addr, nodeType);

		break;
	case HANTRO_IOCS_MMU_MEM_UNMAP:
		ret = hantroMMUBufferUnMapIoctl((struct file *)filp, addr.handle);
		return 0;

	case HANTRO_IOCS_MMU_FLUSH:
		ret = vpu_mmu_flush(xdev, addr.codec_type);
		return 0;
	default:
		return -EINVAL;
	}

	if (copy_to_user((void *)arg, &addr, sizeof(struct addr_desc))) {
		XDX_DEV_ERROR("copy_to_user failed\n");
		return -MMU_EFAULT;
	}
	return 0;
}
