/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#include <linux/rbtree.h>
#include <linux/debugfs.h>
#include <drm/drm_device.h>

#include "xdx_gem.h"
#include "hantro_priv.h"
#include "hantro_vcmd.h"
#include "vpu_xmem.h"

#ifdef VPU_HW_STAT
struct _vpu_stat vpu_stat = {0};
struct _vpu_irq_stat vpu_irq_stat = {0};
#endif
u32 xvcmd_debug_start=0;
struct _xvcmd xvcmd;

u32 mm_probe = 0;
u32 x_probe2 = 0;

struct xmem_node {
	struct rb_node node;
	void *bo;	//	struct xdx_bo *bo;
	void *p_gddr;
	u64	size;

	pid_t	pid;
	char	comm[TASK_COMM_LEN];
};


static struct rb_root xmem_root = RB_ROOT;
static DEFINE_SPINLOCK(xmem_node_lock);

static struct xmem_node *xmem_search(struct rb_root *root, void *key)
{
	struct rb_node *node = root->rb_node;

	spin_lock(&xmem_node_lock);
	while (node) {
		struct xmem_node * data = container_of(node,struct xmem_node,node);
		if (key < data->bo)
			node = node->rb_left;
		else if (key > data->bo)
			node = node->rb_right;
		else {
			spin_unlock(&xmem_node_lock);
			return data;
		}
	}
	spin_unlock(&xmem_node_lock);
	return NULL;
}
static bool xmem_insert(struct rb_root *root, struct xmem_node * data)
{
	struct rb_node **new = &(root->rb_node), *parent = NULL;

	spin_lock(&xmem_node_lock);
	while(*new) {
		struct xmem_node *this = container_of(*new, struct xmem_node, node);

		parent = *new;
		if (data->bo < this->bo)
			new = &((*new)->rb_left);
		else if (data->bo > this->bo)
			new = &((*new)->rb_right);
		else {
			spin_unlock(&xmem_node_lock);
			return false;
		}

	}

	rb_link_node(&data->node, parent, new);
	rb_insert_color(&data->node, root);

	spin_unlock(&xmem_node_lock);
	return true;
}
void xmem_remove(struct rb_root *root, void *key)
{
	struct xmem_node *dst_data = xmem_search(root, key);

	spin_lock(&xmem_node_lock);
	if (dst_data) {
		rb_erase(&dst_data->node, root);
	}
	spin_unlock(&xmem_node_lock);
	if (dst_data)
		vfree(dst_data);
}

void vpumem_remove(struct rb_root *root, void *key)
{
	if (root == NULL)
		root = &xmem_root;

	xmem_remove(root, key);
}

int xmem_create(struct rb_root *root, void *bo, u64 size)
{
	struct xmem_node *data;

	data = vzalloc(sizeof(struct xmem_node));
	if (!data)
	{
		XDX_DEV_DEBUG();
		return -ENOMEM;
	}
	data->bo = bo;
	data->size = size;

	data->pid = current->pid;
	strncpy(data->comm,current->comm,TASK_COMM_LEN);

	xmem_insert(root, data);

	return 0;
}

int vpumem_create(struct rb_root *root, void *bo, u64 size)
{
	if (root == NULL)
		root = &xmem_root;

	return xmem_create(root,bo,size);
}

# if 0
static void xmem_foreach(struct rb_root *root)
{
	struct rb_node *node;
	XDX_DEV_DEBUG();

	if (root == NULL)
		root = &xmem_root;

	spin_lock(&xmem_node_lock);
	for(node = rb_first(root); node; node=rb_next(node))
		XDX_DEV_DEBUG("x_addr:0x%llx x_size:0x%llx \n",((struct xdx_bo *)(rb_entry(node,struct xmem_node, node)->bo))->pages[0], //xdx_bo_gpu_offset
					   rb_entry(node,struct xmem_node, node)->size);
	spin_unlock(&xmem_node_lock);
}
#endif

static void xmem_cleanup(struct rb_root *root)
{
	struct rb_node *node;

	if (root == NULL)
		root = &xmem_root;

	XDX_DEV_DEBUG();
	for(node = rb_first(root); node; node=rb_next(node)) {
		spin_lock(&xmem_node_lock);
		rb_erase(node, root);
		spin_unlock(&xmem_node_lock);
		vfree(rb_entry(node,struct xmem_node, node));
	}
}

#define DEBUG_GDDR_ALLOC_SIZE	0x400000
#define DEBUG_GDDR_BO_MAX	1024
struct dentry *vpu_debugfs_dir=NULL;
struct xdx_bo *debug_gddr_bo_ptr[DEBUG_GDDR_BO_MAX];

static u32 debug_gddr_bo_cnt = 0;

#if 0
static ssize_t xmem_read(struct file *file, char __user *user_buf, size_t count, loff_t *ppos)
{

	XDX_DEV_DEBUG();
	xmem_foreach(&xmem_root);

	return 0;
}
#endif

static ssize_t xmem_write(struct file *file, const char __user *user_buf,size_t count, loff_t *ppos)
{
	char buf[32] = {0};
	int size;
	struct xdx_device *xdev;
	int ret;
	u64 *paddr = NULL;
	void *vaddr;
	unsigned int i;

	XDX_DEV_DEBUG();

	xdev = file_inode(file)->i_private;
	size = min(sizeof(buf)-1, count);
	if (copy_from_user(buf, user_buf, size))
		return -EFAULT;

	if(!strncmp(buf, "alloc", size)) {

		if (debug_gddr_bo_cnt >= DEBUG_GDDR_BO_MAX) {
			XDX_DEV_ERROR("ERR!\n");
			return -EFAULT;
		}
		ret = xdx_bo_create_reserved(xdev,PAGE_ALIGN(DEBUG_GDDR_ALLOC_SIZE),PAGE_SIZE,XDXGPU_PL_VRAM,
			&debug_gddr_bo_ptr[debug_gddr_bo_cnt], paddr,
			&vaddr);
		if (ret) {
			XDX_DEV_ERROR("Failed to alloc memory\n");
			ret = -ENOMEM;
			return ret;
		}

		xmem_create(&xmem_root,&debug_gddr_bo_ptr[debug_gddr_bo_cnt],PAGE_ALIGN(DEBUG_GDDR_ALLOC_SIZE));
		debug_gddr_bo_cnt++;

	}
	else if (!strncmp(buf, "xvcmd_debug_start", size)) {
		XDX_DEV_ERROR("xvcmd debug beginnig\n");
		if (xvcmd_debug_start == 0) {
			ret = xdx_bo_create_reserved(xdev,PAGE_ALIGN(CMDBUF_POOL_TOTAL_SIZE),PAGE_SIZE,XDXGPU_PL_VRAM,
				&xvcmd.buf_bo, &xvcmd.p_buf,
				&xvcmd.buf);
			if (ret) {
				XDX_DEV_ERROR("Failed to alloc memory\n");
				ret = -ENOMEM;
				return ret;
			}
			ret = xdx_bo_create_reserved(xdev,PAGE_ALIGN(CMDBUF_POOL_TOTAL_SIZE),PAGE_SIZE,XDXGPU_PL_VRAM,
				&xvcmd.status_buf_bo, &xvcmd.p_status_buf,
				&xvcmd.status_buf);
			if (ret) {
				XDX_DEV_ERROR("Failed to alloc memory\n");
				xdx_gem_object_put(&xvcmd.buf_bo->base);
				ret = -ENOMEM;
				return ret;
			}
			ret = xdx_bo_create_reserved(xdev,PAGE_ALIGN(CMDBUF_VCMD_REGISTER_TOTAL_SIZE),PAGE_SIZE,XDXGPU_PL_VRAM,
				&xvcmd.registers_bo, &xvcmd.p_registers,
				&xvcmd.registers);
			if (ret) {
				XDX_DEV_ERROR("Failed to alloc memory\n");
				xdx_gem_object_put(&xvcmd.buf_bo->base);
				xdx_gem_object_put(&xvcmd.status_buf_bo->base);
				ret = -ENOMEM;
				return ret;
			}
			xvcmd_debug_start = 1;
		}else {
			XDX_DEV_ERROR("xvcmd debug already start\n");
		}
	}
	else if (!strncmp(buf, "xvcmd_debug_end", size)) {
		if (xvcmd_debug_start == 1) {
			xdx_gem_object_put(&xvcmd.buf_bo->base);
			xdx_gem_object_put(&xvcmd.status_buf_bo->base);
			xdx_gem_object_put(&xvcmd.registers_bo->base);

			xvcmd_debug_start = 0;
		} else {
			XDX_DEV_ERROR("xvcmd debug do not start\n");
		}
	}
	else if (!strncmp(buf, "freeall", size)) {
//		xmem_remove(&xmem_root, xdx_bo_gpu_offset(debug_gddr_bo_ptr[i]));

		for(i=0;i<debug_gddr_bo_cnt;i++) {
			xmem_remove(&xmem_root, &debug_gddr_bo_ptr[i]);
			xdx_gem_object_put(&debug_gddr_bo_ptr[i]->base);
		}
		debug_gddr_bo_cnt=0;

	}
	else if (!strncmp(buf, "cleanstat", size)) {
		xmem_cleanup(&xmem_root);
	}
	else {
		XDX_DEV_ERROR("input error.\n");
		return -EFAULT;
	}

	return count;
}
static void vcmd_show(struct seq_file *m, void *p)
{
	struct xdx_device *xdev = m->private;
	vcmd_slice_str *cur_slice;

	list_for_each_entry(cur_slice, &xdev->vpu.vcmd_slice, list) {
		seq_printf(m, "dec:%d enc:%d \n", cur_slice->dec_vcmd.vcmd_timer.count/(4096*2160),
			cur_slice->enc_vcmd.vcmd_timer.count/(1920*1080));
	}

	return;
}

static void stats_show(struct seq_file *m, void *p)
{
	struct xdx_device *xdev = m->private;
	vcmd_slice_str *cur_slice;
	struct _vpu_irq_stat *vpu_Decirq_stat = NULL;
	struct _vpu_irq_stat *vpu_Encirq_stat = NULL;
	struct _vpu_stat *vpu_Decstat = NULL;
	struct _vpu_stat *vpu_Encstat = NULL;

	list_for_each_entry(cur_slice, &xdev->vpu.vcmd_slice, list) {
		vpu_Decirq_stat = &cur_slice->dec_vcmd.vcmd_core->vpu_irq_stat;
		vpu_Encirq_stat = &cur_slice->enc_vcmd.vcmd_core->vpu_irq_stat;
		vpu_Decstat = &cur_slice->dec_vcmd.vcmd_core->vpu_stat;
		vpu_Encstat = &cur_slice->enc_vcmd.vcmd_core->vpu_stat;

		seq_printf(m, "enc_stat:%llu dec_stat:%llu vcmd_enc_stat:%llu vcmd_dec_stat:%llu \n",
			   vpu_Encstat->enc_stat, vpu_Decstat->dec_stat, vpu_Encstat->vcmd_enc_stat, vpu_Decstat->vcmd_dec_stat);

		seq_printf(m, "DEC_IRQ none:%llu noirq:%llu reset:%llu abort:%llu buserr:%llu timeout:%llu cmderr:%llu endcmd:%llu cmdid:%llu \n",
			   vpu_Decirq_stat->irq_none, vpu_Decirq_stat->irq_noirq,
			   vpu_Decirq_stat->irq_reset, vpu_Decirq_stat->irq_abort, vpu_Decirq_stat->irq_buserr,
			   vpu_Decirq_stat->irq_timeout, vpu_Decirq_stat->irq_cmderr, vpu_Decirq_stat->irq_endcmd,
			   vpu_Decirq_stat->irq_cmdid);
		seq_printf(m, "ENC_IRQ none:%llu noirq:%llu reset:%llu abort:%llu buserr:%llu timeout:%llu cmderr:%llu endcmd:%llu cmdid:%llu \n",
			   vpu_Encirq_stat->irq_none, vpu_Encirq_stat->irq_noirq,
			   vpu_Encirq_stat->irq_reset, vpu_Encirq_stat->irq_abort, vpu_Encirq_stat->irq_buserr,
			   vpu_Encirq_stat->irq_timeout, vpu_Encirq_stat->irq_cmderr, vpu_Encirq_stat->irq_endcmd,
			   vpu_Encirq_stat->irq_cmdid);

	}

	return;
}

static int xmem_show(struct seq_file *m, void *p)
{
//	struct rb_node *node;
//	struct xmem_node *xmem;
//	struct xdx_bo * xbo;
	struct xdx_device *xdev = m->private;

	seq_printf(m, "device: 0x%x\n", xdev->pdev->device);
//	seq_printf(m,"mm_probe:%d, x_probe2:%d\n", mm_probe, x_probe2);

	vcmd_show(m, p);
#ifdef VPU_HW_STAT
	stats_show(m, p);

	if (xvcmd_debug_start == 1)
	{
		seq_printf(m, "current cmdbuf_id:%u buf:%llx %p \n", xvcmd.cmdbuf_id,
				   xvcmd.p_buf - PANGU_VRAM_DEV_PHYADDR_BASE, xvcmd.buf);
		{
			int i;
			for(i=0;i<CMDBUF_MAX_SIZE/4;i++)
			{
				seq_printf(m, "0x%x ", *((u32 *)xvcmd.buf + xvcmd.cmdbuf_id*CMDBUF_MAX_SIZE/4 + i));
				if (i%8 == 0 && i != 0)
					seq_printf(m, "\n");
			}
		}
	}
#endif

#if 0
	for(node = rb_first(&xmem_root); node; node=rb_next(node)) {
		xmem = rb_entry(node,struct xmem_node, node);
		if (xmem && xmem->bo && ((struct xdx_bo *)xmem->bo)->pages) {
			xbo = xmem->bo;
			seq_printf(m,"x_addr:0x%llx x_size:0x%llx pid:%u comm:%s \n", xbo->pages[0], xmem->size, xmem->pid,xmem->comm);
		}
		else
			seq_printf(m,"0\n");
	}
#endif
	return 0;
}

static int xmem_open(struct inode *inode, struct file *file)
{
	return single_open(file, xmem_show, inode->i_private);
}

static const struct file_operations vpu_xmem_fops = {
	.owner = THIS_MODULE,
	.open = xmem_open,
	.read = seq_read,
	.write = xmem_write,
	.llseek = seq_lseek,
	.release = single_release,

};

static const struct file_operations vpu_status_fops = {
	.owner = THIS_MODULE,
	.open = xmem_open,
	.read = seq_read,
	.write = xmem_write,
	.llseek = seq_lseek,
	.release = single_release,

};

void vpu_debugfs_init(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
	char filename[16] = {0};

	if (!vpu_debugfs_dir)
		vpu_debugfs_dir = debugfs_create_dir("vpu", NULL);

	snprintf(filename, sizeof(filename), "status-%d", ddev->primary->index);
	debugfs_create_file(filename, 0444, vpu_debugfs_dir, xdev, &vpu_status_fops);

#if 0
	snprintf(filename, sizeof(filename), "xmem-%d", ddev->primary->index);
	debugfs_create_file(filename, 0444, vpu_debugfs_dir, xdev, &vpu_xmem_fops);
#endif

	return ;
}
void vpu_debugfs_exit(void)
{
	if (!vpu_debugfs_dir) {
		XDX_DEV_INFO("vpu_debugfs_dir is not exist.\n");
		return;
	}

	debugfs_remove_recursive(vpu_debugfs_dir);
	vpu_debugfs_dir = NULL;
}

