/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/kthread.h>
#include <linux/pci.h>
#include <linux/uaccess.h>
#include <linux/pm_runtime.h>

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0)) || (defined(REDHAT_83)) || (defined(XDXGPU_CENTOS85_SUPPORT))
#include <drm/drm_device.h>
#include <drm/drm_file.h>
#else
#include <drm/drmP.h>
#endif

#include "xdx_device.h"
#include "xdx_debugfs.h"

#ifdef CONFIG_DEBUG_FS
/// for gpu info debugfs
static unsigned short Device_ID;
static struct gpu_info {
    unsigned int DID;
    const char *gpu_type_name;
    const char *gpu_chip_name;
    unsigned int GDDR_width;
} gpu_info_list[] = {
     {0x1140, "XDX120LP",    "TJ01-630-A",   32 },
     {0x1320, "XDX150LP",    "TJ01-750-A",   64 },
     {0x13C0, "XDX160",      "TJ01-760-A",   64 },
     {0x13C1, "XDX161",      "TJ01-761-A",   64 },
     {0x1500, "XDX180",      "TJ01-880-A",   128},
     {0x15A0, "XDX190",      "TJ01-990-A",   128},
 };
/*
static struct gpu_info gpu_info_list[] = {
    {0x1140, "XDX120LP",    "TJ01-630-A",   32 },
    {0x1320, "XDX150LP",    "TJ01-750-A",   64 },
    {0x13C0, "XDX160",      "TJ01-760-A",   64 },
    {0x13C1, "XDX161",      "TJ01-761-A",   64 },
    {0x1500, "XDX180",      "TJ01-880-A",   128},
    {0x15A0, "XDX190",      "TJ01-990-A",   128},
};*/

static int info_open(struct inode *inode, struct file *filp)
{
    filp->private_data = inode->i_private;
    return 0;
}

inline ssize_t debugfs_attr_read(struct file *file, char __user *buf,
                                       size_t len, loff_t *ppos)
{
       return -ENODEV;
}

inline ssize_t debugfs_attr_write(struct file *file,
                                       const char __user *buf,
                                       size_t len, loff_t *ppos)
{
       return -ENODEV;
}

#ifndef DEFINE_DEBUGFS_ATTRIBUTE
#define DEFINE_DEBUGFS_ATTRIBUTE(__fops, __get, __set, __fmt)          \
static int __fops ## _open(struct inode *inode, struct file *file)     \
{                                                                      \
       __simple_attr_check_format(__fmt, 0ull);                        \
       return simple_attr_open(inode, file, __get, __set, __fmt);      \
}                                                                      \
static const struct file_operations __fops = {                         \
       .owner   = THIS_MODULE,                                         \
       .open    = __fops ## _open,                                     \
       .release = simple_attr_release,                                 \
       .read    = debugfs_attr_read,                                   \
       .write   = debugfs_attr_write,                                  \
       .llseek  = no_llseek,                                           \
}
#endif

static ssize_t debugfs_readGPU(struct file *file, char __user *buf, size_t length, loff_t *off)
{
    u32 i, len, size = 4096;
    char *str;
    size_t ret;
    
    str = kzalloc(size, GFP_KERNEL);
    if(str == NULL)
	    return -ENOMEM;
    for(i = 0; i < sizeof(gpu_info_list)/sizeof(struct gpu_info); i++)
    {
        if(Device_ID == gpu_info_list[i].DID )
        {
	    len = scnprintf(str, size,  "----------------- GPU info -----------------\n");
	    len += scnprintf(str+len, size-len,  "GPU type : %-10s\n", gpu_info_list[i].gpu_type_name);
            len += scnprintf(str+len, size-len, "GPU chip : %-10s\n", gpu_info_list[i].gpu_chip_name);
            len += scnprintf(str+len, size-len, "GPU GDDR width : %-10d bit\n", gpu_info_list[i].GDDR_width);
            len += scnprintf(str+len, size-len, "--------------------------------------------\n\n");
	    break;
        }
        else
        {
            continue;
        }
    }

    if(i >= sizeof(gpu_info_list)/sizeof(struct gpu_info))
    {
        printk(KERN_EMERG "----------------- NO GPU info -----------------\n");
	return -EFAULT;
    }

    ret = simple_read_from_buffer(buf, length ,off, str, len);

    kfree(str);
    return ret;
}

static ssize_t debugfs_writeGPU(struct file *file, const char __user *buf, size_t len, loff_t *off)
{
    printk(KERN_ERR "%s %d\n", __func__, __LINE__);
    return 0;
}

struct file_operations gpuinfo_fop = {
    .owner = THIS_MODULE,
    .open = info_open,
    .read = debugfs_readGPU,
    .write = debugfs_writeGPU,
};

/**
 * xdx_debugfs_regs_read - Read from a MMIO register
 *
 * @f: open file handle
 * @buf: User buffer to store read data in
 * @size: Number of bytes to read
 * @pos:  Offset to seek to
 *
 * The lower bits are the BYTE offset of the register to read.  This
 * allows reading multiple registers in a single call and having
 * the returned size reflect that.
 */
static ssize_t xdx_debugfs_regs_read(struct file *f, char __user *buf,
					size_t size, loff_t *pos)
{
	struct xdx_device *xdev = file_inode(f)->i_private;
	ssize_t result = 0;
	int r;

	if (size & 0x3 || *pos & 0x3)
		return -EINVAL;

	r = pm_runtime_get_sync(xdev_to_drm(xdev)->dev);
	if (r < 0) {
		pm_runtime_put_autosuspend(xdev_to_drm(xdev)->dev);
		return r;
	}

	while (size) {
		uint32_t value;

		value = xdx_mmio_readl(xdev, *pos);
		r = put_user(value, (uint32_t *)buf);
		if (r) {
			pm_runtime_mark_last_busy(xdev_to_drm(xdev)->dev);
			pm_runtime_put_autosuspend(xdev_to_drm(xdev)->dev);
			return r;
		}

		result += 4;
		buf += 4;
		*pos += 4;
		size -= 4;
	}

	pm_runtime_mark_last_busy(xdev_to_drm(xdev)->dev);
	pm_runtime_put_autosuspend(xdev_to_drm(xdev)->dev);

	return result;
}

/**
 * xdx_debugfs_regs_write - Write to a MMIO register
 *
 * @f: open file handle
 * @buf: User buffer to write data from
 * @size: Number of bytes to write
 * @pos:  Offset to seek to
 *
 * The lower bits are the BYTE offset of the register to write.  This
 * allows writing multiple registers in a single call and having
 * the returned size reflect that.
 */
static ssize_t xdx_debugfs_regs_write(struct file *f, const char __user *buf,
					 size_t size, loff_t *pos)
{
	struct xdx_device *xdev = file_inode(f)->i_private;
	ssize_t result = 0;
	int r;

	if (size & 0x3 || *pos & 0x3)
		return -EINVAL;

	r = pm_runtime_get_sync(xdev_to_drm(xdev)->dev);
	if (r < 0) {
		pm_runtime_put_autosuspend(xdev_to_drm(xdev)->dev);
		return r;
	}

	while (size) {
		uint32_t value;

		r = get_user(value, (uint32_t *)buf);
		if (r) {
			pm_runtime_mark_last_busy(xdev_to_drm(xdev)->dev);
			pm_runtime_put_autosuspend(xdev_to_drm(xdev)->dev);
			return r;
		}

		xdx_mmio_writel(xdev, *pos, value);

		result += 4;
		buf += 4;
		*pos += 4;
		size -= 4;
	}

	pm_runtime_mark_last_busy(xdev_to_drm(xdev)->dev);
	pm_runtime_put_autosuspend(xdev_to_drm(xdev)->dev);

	return result;
}

static const struct file_operations xdx_debugfs_regs_fops = {
	.owner = THIS_MODULE,
	.read = xdx_debugfs_regs_read,
	.write = xdx_debugfs_regs_write,
	.llseek = default_llseek
};

static int xdx_debugfs_sclk_set(void *data, u64 val)
{
	int ret = 0;
	struct xdx_device *xdev = (struct xdx_device *)data;

	ret = pm_runtime_get_sync(xdev_to_drm(xdev)->dev);
	if (ret < 0) {
		pm_runtime_put_autosuspend(xdev_to_drm(xdev)->dev);
		return ret;
	}

    // TODO: set GPU clk

	pm_runtime_mark_last_busy(xdev_to_drm(xdev)->dev);
	pm_runtime_put_autosuspend(xdev_to_drm(xdev)->dev);

	if (ret)
		return -EINVAL;

	return 0;
}

DEFINE_DEBUGFS_ATTRIBUTE(fops_sclk_set, NULL,
        xdx_debugfs_sclk_set, "%llu\n");

int xdx_debugfs_init(struct xdx_device *xdev)
{
	struct dentry *root = xdev_to_drm(xdev)->primary->debugfs_root;
	struct dentry *ent, *gpuinfo_ent, *regs_ent;

	if (!debugfs_initialized())
		return 0;

	ent = debugfs_create_file("xdxgpu_force_sclk", 0200, root, xdev,
				  &fops_sclk_set);
	if (IS_ERR(ent)) {
		dev_err(xdev->dev, "unable to create xdxgpu_force_sclk debugsfs file\n");
		return PTR_ERR(ent);
	}

	gpuinfo_ent = debugfs_create_file("GPUinfo", 0666, root, xdev, &gpuinfo_fop);
	if (IS_ERR(gpuinfo_ent)) {
		dev_err(xdev->dev, "unable to create gpu info debugsfs file\n");
		return PTR_ERR(ent);
	}
	Device_ID = xdev->pdev->device;

	regs_ent = debugfs_create_file("xdxgpu_regs", 0666, root, xdev,
				       &xdx_debugfs_regs_fops);
	if (IS_ERR(regs_ent)) {
		dev_err(xdev->dev,
			"unable to create regs access debugsfs file\n");
		return PTR_ERR(ent);
	}

	return 0;
}
#else
int xdx_debugfs_init(struct xdx_device *xdev)
{
    return 0;
}
#endif /* CONFIG_DEBUG_FS */
