/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/irq.h>
#include <linux/pci.h>

#include <drm/drm_crtc_helper.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0))
#include <drm/drm_vblank.h>
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0))
#include <drm/drm_irq.h>
#else
#include <drm/drmP.h>
#endif
#include <drm/xdxgpu_drm.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
#include <drm/drm_drv.h>
#endif
#include <linux/pm_runtime.h>

#include "xdx_device.h"
#include "xdx_irq.h"
#include "xdx_drv.h"

static irqreturn_t xdx_irq_handler(int irq, void *arg)
{
	struct xdx_irq_entry *irq_entry = (struct xdx_irq_entry *)arg;
	struct xdx_device *xdev = irq_entry->xdev;
	unsigned long flags;

	if (xdev->shutdown)
		return IRQ_NONE;

	switch (xdev->irq.irq_type) {
		case XDX_IRQ_MSIX: {
			uint32_t clientid = XDX_IRQ_CLIENTID_INVALID;

			if (xdev->irq.ih_funcs->get_client_id)
				clientid = xdev->irq.ih_funcs->get_client_id(irq_entry);

			if (clientid == XDX_IRQ_CLIENTID_INVALID) {
				dev_err(xdev->dev, "unsupported msix vector %d\n",
					irq_entry->vec_idx);
				return IRQ_NONE;
			}

			spin_lock_irqsave(&xdev->irq.lock, flags);

			if (xdev->irq.ih_funcs->set)
				xdev->irq.ih_funcs->set(xdev, irq_entry, XDX_IRQ_STATE_DISABLE);

			if (xdev->irq.clients[clientid])
				xdev->irq.clients[clientid]->funcs->process(
					xdev, xdev->irq.clients[clientid], irq_entry);

			if (xdev->irq.ih_funcs->set)
				xdev->irq.ih_funcs->set(xdev, irq_entry, XDX_IRQ_STATE_ENABLE);

			spin_unlock_irqrestore(&xdev->irq.lock, flags);

			break;
		}
		case XDX_IRQ_INTX: {
			if (xdev->irq.ih_funcs->intx_handler)
				xdev->irq.ih_funcs->intx_handler(xdev, irq_entry);
			break;
		}
		case XDX_IRQ_MSI: {
			if (xdev->irq.ih_funcs->msi_handler)
				xdev->irq.ih_funcs->msi_handler(xdev, irq_entry);
			break;
		}
		default: {
			dev_err(xdev->dev, "unrecognized irq type\n");
			break;
		}
	}

	pm_runtime_mark_last_busy(xdev_to_drm(xdev)->dev);
	return IRQ_HANDLED;
}

int xdx_irq_register_src(struct xdx_device *xdev, struct xdx_irq_src *irq_src)
{
	unsigned long flags;

	if (xdev->irq.clients[irq_src->id]) {
		dev_err(xdev->dev,
			"irq handle has been registered for source id %d\n",
			irq_src->id);
		return -EINVAL;
	}

	spin_lock_irqsave(&xdev->irq.lock, flags);
	xdev->irq.clients[irq_src->id] = irq_src;
	spin_unlock_irqrestore(&xdev->irq.lock, flags);

	dev_info(xdev->dev, "register irq handle for source %d\n", irq_src->id);

	return 0;
}

int xdx_irq_unregister_src(struct xdx_device *xdev, struct xdx_irq_src *irq_src)
{
	unsigned long flags;

	if (xdev->irq.clients[irq_src->id] != irq_src)
		return -EINVAL;

	dev_info(xdev->dev, "unregister irq handle for source %d\n", irq_src->id);

	spin_lock_irqsave(&xdev->irq.lock, flags);
	xdev->irq.clients[irq_src->id] = NULL;
	spin_unlock_irqrestore(&xdev->irq.lock, flags);

	return 0;
}

int xdx_irq_init(struct xdx_device *xdev)
{
	int r = 0;
	unsigned int i;

	if (xdev->flags.no_hardware)
		return 0;

	spin_lock_init(&xdev->irq.lock);

	if (xdev->soc_funcs.set_interrupt_funcs) {
		xdev->soc_funcs.set_interrupt_funcs(xdev);
	} else {
		dev_err(xdev->dev, "set_interrupt_funcs is NULL\n");
		return -EINVAL;
	}

	xdev->irq.msi_enabled = false;
	xdev->irq.irq_type = XDX_IRQ_UNKNOW;

	if (likely(!xdev->flags.run_in_emu && xdx_msi_enable)) {
		int nvec = pci_msix_vec_count(xdev->pdev);

		dev_info(xdev->dev, "nvec = %d\n", nvec);

		/* firstly we try to alloc MSIX */
		r = pci_alloc_irq_vectors(xdev->pdev, nvec,
					  nvec, PCI_IRQ_MSIX);
		/* fallback to MSI when MSIX is not available */
		if (r != nvec) {
			/* we only need one vector in this case */
			r = pci_alloc_irq_vectors(xdev->pdev, 1,
						1, PCI_IRQ_MSI);
		}
		if (r < 1) {
			/* fallback to INTx when MSI is not available */
			xdev->irq.irq_type = XDX_IRQ_INTX;
			dev_warn(xdev->dev,
				"failed to alloc irq vectors, ret(%d), fallback to INTx\n", r);
		} else {
			uint32_t clientid = XDX_IRQ_CLIENTID_INVALID;
			if (r == 1) {
				/* fallback to MSI and use the software to dispatch IRQ */
				nvec = 1;
				xdev->irq.irq_type = XDX_IRQ_MSI;
			} else {
				xdev->irq.irq_type = XDX_IRQ_MSIX;
			}

			xdev->irq.irqs =
				kzalloc(nvec * sizeof(struct xdx_irq_entry), GFP_KERNEL);
			if (!xdev->irq.irqs) {
				pci_free_irq_vectors(xdev->pdev);
				r = -ENOMEM;
				goto err_end;
			}

			xdev->irq.num_irq = nvec;
			xdev->irq.msi_enabled = true;
			for (i = 0; i < xdev->irq.num_irq; i++) {
				xdev->irq.irqs[i].xdev = xdev;
				xdev->irq.irqs[i].irq = pci_irq_vector(xdev->pdev, i);
				xdev->irq.irqs[i].vec_idx = i;

				if (xdev->irq.ih_funcs->get_client_id)
					clientid = xdev->irq.ih_funcs->get_client_id(&xdev->irq.irqs[i]);

				if (clientid != XDX_IRQ_CLIENTID_INVALID)
					xdev->irq.irqs[i].interested = true;
			}
		}
	}

	if (!xdev->irq.msi_enabled) {
		xdev->irq.irqs =
			kcalloc(1, sizeof(struct xdx_irq_entry), GFP_KERNEL);
		if (!xdev->irq.irqs) {
			r = -ENOMEM;
			goto err_end;
		}

		/* hardcode irq number */
		xdev->irq.num_irq = 1;
		xdev->irq.irqs[0].irq = 17;
		xdev->irq.irqs[0].xdev = xdev;
		xdev->irq.irqs[0].vec_idx = 0;
		xdev->irq.irqs[0].interested = true;
	}

	for (i = 0; i < xdev->irq.num_irq; i++) {
		/* Skip uninterested interrupt */
		if (!xdev->irq.irqs[i].interested)
			continue;

		/* PCI devices require shared interrupts. */
		r = request_irq(xdev->irq.irqs[i].irq, xdx_irq_handler, IRQF_SHARED,
				xdev->irq.ih_funcs->get_irq_vector_name(&xdev->irq.irqs[i]),
				&xdev->irq.irqs[i]);
		if (r) {
			dev_err(xdev->dev, "failed to request_irq %d, %d\n", i,
				xdev->irq.irqs[i].irq);
			goto err_end;
		}
	}

	xdev->irq.installed = true;

	dev_info(xdev->dev, "irq initialized, msix interrupt %s\n",
		xdev->irq.msi_enabled ? "enabled" : "disabled");
	dev_info(xdev->dev, "irq initialized, msi software dispatch interrupt %s\n",
		(xdev->irq.irq_type == XDX_IRQ_MSI) ? "enabled" : "disabled");

	return 0;

err_end:
	xdx_irq_fini(xdev);

	return r;
}

void xdx_irq_fini(struct xdx_device *xdev)
{
	uint32_t i;

	if (xdev->flags.no_hardware)
		return;

	if (xdev->irq.installed) {
		for (i = 0; i < xdev->irq.num_irq; i++) {
			/* Skip uninterested interrupt */
			if (!xdev->irq.irqs[i].interested)
				continue;

			free_irq(xdev->irq.irqs[i].irq, &xdev->irq.irqs[i]);
		}
		xdev->irq.installed = false;
	}

	if (xdev->irq.msi_enabled) {
		pci_free_irq_vectors(xdev->pdev);
		xdev->irq.msi_enabled = false;
	}

	if (xdev->irq.irqs)
		kfree(xdev->irq.irqs);
}

/**
 * xdxgpu_irq_set - update hardware interrupt state
 *
 * @xdev: xdxgpu device pointer
 * @src: interrupt source pointer
 * @state: enable or disable
 *
 * Updates interrupt state for the specific source (all ASICs).
 */
int xdxgpu_irq_set(struct xdx_device *xdev,
		      struct xdx_irq_src *src,
		      unsigned type,
		      enum xdx_interrupt_state state)
{
	unsigned long irqflags;
	int r;

	spin_lock_irqsave(&xdev->irq.lock, irqflags);
	r = src->funcs->set(xdev, src, type, state);
	spin_unlock_irqrestore(&xdev->irq.lock, irqflags);
	return r;
}
