/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/device.h>
#include <linux/errno.h>
#include <linux/kernel.h>

#include "xdx_vmm.h"
#include "xdx_device.h"
#include "xdx_gem.h"

int xdx_vmm_init(struct xdx_vmm *vmm)
{
	struct xdx_device *xdev = vmm_to_xdev(vmm);

	if (vmm->dc.funcs) {
		vmm->dc.context = vmm->dc.funcs->create(xdev);
		if (IS_ERR(vmm->dc.context)) {
			dev_err(xdev->dev, "failed to create dc mmu context\n");
			return IS_ERR(vmm->dc.context);
		}
		strcpy(vmm->dc.name, "DC");
	}

#ifdef SUPPORT_VPU
	if (vmm->dc.context) {
		vmm->vpu.context = vmm->dc.context;
		strcpy(vmm->vpu.name, "VPU");
	} else if (vmm->vpu.funcs) {
		vmm->vpu.context = vmm->vpu.funcs->create(xdev);
		strcpy(vmm->vpu.name, "VPU");
	} else {
		dev_err(xdev->dev, "failed to create vpu mmu context\n");
		return -1;
	}
#endif
	return 0;
}

int xdx_vmm_fini(struct xdx_vmm *vmm)
{
	if (vmm->dc.funcs) {
		vmm->dc.funcs->destrory(vmm->dc.context);
		vmm->dc.context = NULL;
	}

	return 0;
}

struct xdx_bo_va *xdx_vm_bo_find(struct xdx_vmm *vmm, struct xdx_vm *vm,
				 struct xdx_bo *bo)
{
	struct xdx_bo_va *va, *tmp, *ret = NULL;

	xdx_bo_reserve(bo, true);
	list_for_each_entry_safe (va, tmp, &bo->va_list, node) {
		if (va->vm == vm) {
			ret = va;
			break;
		}
	}
	xdx_bo_unreserve(bo);
	return ret;
}

struct xdx_bo_va *xdx_vm_bo_add(struct xdx_vmm *vmm, struct xdx_vm *vm,
				struct xdx_bo *bo)
{
	struct xdx_device *xdev = vmm_to_xdev(vmm);
	struct xdx_bo_va *va;
	int ret;

	va = kzalloc(sizeof(struct xdx_bo_va), GFP_KERNEL);
	if (!va)
		return NULL;

	va->vm = vm;
	va->bo = bo;

	ret = vm->funcs->bo_map(vm->context, bo, &va->addr);
	if (ret) {
		dev_err(xdev->dev, "%s VM: failed to map bo %p (%d)\n",
			vm->name, bo, ret);
		kfree(va);
		return NULL;
	}

	xdx_bo_reserve(bo, true);
	list_add(&va->node, &bo->va_list);
	xdx_bo_unreserve(bo);

	return va;
}

void xdx_vm_bo_remove(struct xdx_vmm *vmm, struct xdx_bo_va *va)
{
	struct xdx_device *xdev = vmm_to_xdev(vmm);
	struct xdx_vm *vm = va->vm;
	struct xdx_bo *bo = va->bo;
	int ret;

	ret = vm->funcs->bo_unmap(vm->context, bo, va->addr);
	if (ret)
		dev_warn(xdev->dev, "%s VM: failed to unmap bo %p, va %llx\n",
			 vm->name, bo, va->addr.value);

	xdx_bo_reserve(bo, true);
	list_del(&va->node);
	xdx_bo_unreserve(bo);

	kfree(va);
}
